/* BootMain.c
 * Main code for managing boot selection
 *
 * --------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ------------------------------------------------------------------------
 *
 *
 * system: SA1200
 * subsystem: BootMgr
 * author: jdg 05/03/00
 * revisions:
 * $History: bootMain.c $
 * 
 * *****************  Version 11  *****************
 * User: Jdguilfo     Date: 11/14/00   Time: 5:21p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Spectacle Island Port, includes fixes bugs 1226 & 1227, more checks
 * before updating flash util block
 * 
 * *****************  Version 10  *****************
 * User: Jdguilfo     Date: 10/27/00   Time: 2:46p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Fix boot flags
 * 
 * *****************  Version 9  *****************
 * User: Jdguilfo     Date: 10/27/00   Time: 2:17p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Fix spacing on text
 * 
 * *****************  Version 8  *****************
 * User: Jdguilfo     Date: 10/26/00   Time: 3:38p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Merge serial and HPC versions into one version, see bootMgr.txt for
 * details.
 * 
 * *****************  Version 7  *****************
 * User: Jdguilfo     Date: 10/20/00   Time: 2:27p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Merge Boot Manager & Flash Utility, add HPC
 * 
 * *****************  Version 4  *****************
 * User: Tclancyx     Date: 6/13/00    Time: 10:51a
 * Updated in $/dev_1200/BoardSupport/BootMgr
 * change version to 1.1.NoBldNum
 * 
 * *****************  Version 3  *****************
 * User: Jdguilfo     Date: 5/11/00    Time: 10:56a
 * Updated in $/dev_1200/BoardSupport/BootMgr
 * Change default countdown to 10 seconds
 * 
 * *****************  Version 2  *****************
 * User: Jdguilfo     Date: 5/11/00    Time: 10:15a
 * Updated in $/dev_1200/BoardSupport/BootMgr
 * Flush output before booting OS
 * 
 * *****************  Version 1  *****************
 * User: Jdguilfo     Date: 5/10/00    Time: 4:46p
 * Created in $/dev_1200/BoardSupport/BootMgr
 * Add BootMgr feature
 * 
 * ------------------------------------------------------------------------
 */

#include "chip.h"
#include "os.h"

#include "genericIO.h"
#include "flashUart.h"
#include "utils.h"
#include "hpc.h"
#include "smalloc.h"
#include "flashUtil.h"

#ifndef NULL
#define NULL ((void*)0)
#endif

typedef unsigned int UINT;

#define COUNT 29000

#define NVRAM_ADDR 0x00008000
#define NVRAM_SIZE 0x00002000
#define NVRAM_OFFSET   0x0200 /* VxWorks has what's below this */

#define NVRAM_VALID 0x30746f42
#define DEFAULT_COUNT 10
#define DEFAULT_OS     3
#define DEFAULT_SILENT 0
#define DEFAULT_IO_TYPE GIT_SERIAL

int sysCpuRev = 0;
unsigned int sysCpuId = 0;
int bootFlags = 0;
int flash32BitMode = 1;
extern UINT flashBase;
extern unsigned char flashType;

typedef struct {
    UINT valid;
    int count;
    int os;
    int silent;
    enum GENERIC_IO_TYPE io_type;
} NVRAM_STRUCT;

typedef struct {
    char *name;
    UINT addr;
    UINT data;
    UINT boot_flags_serial;
    UINT boot_flags_hpc;
} OS;

enum POLL_RESULT {
    POLL_TIMEOUT,
    POLL_SERIAL,
    POLL_HPC
};

void flash_main(unsigned int bootflags);

OS os_table[] = OS_TABLE;
#define NUM_OS_ENTRIES ((sizeof(os_table)) / (sizeof(OS)))

static int marquee[] = {0xB0, 0x0B};

typedef void (*VOIDFUNCPTR) ();  

void copyLongs (register UINT *source, register UINT *destination, UINT nlongs);

void
drainUart()
{
    char c;
    while (uartPollInput(&c) == 0);
}

enum POLL_RESULT
pollOneSec()
{
    int i, invalid;
    char c, prev_c_s = ' ', prev_c_h = ' ';
    for (i=0; i<COUNT; i++) {
        invalid = uartPollInput(&c);
        if (0 == invalid) {
            if ((c == ' ') && (prev_c_s == ' '))
                return POLL_SERIAL;
            prev_c_s = c;
        }
        
        invalid = hpcGetMaybe(&c);
        if (0 == invalid) {
            if ((c == ' ') && (prev_c_h == ' '))
                return POLL_HPC;
            prev_c_h = c;
        }
    }
    return POLL_TIMEOUT;
}

char *
itoa(unsigned int x, int len)
{
    static char buf[12];
    char *c = buf+11;
    unsigned int d, new_x;
    
    if (x == 0) {
        *--c = '0';
    } else {
        while (x > 0) {
            new_x = x / 10;
            d = x - new_x * 10;
            *--c = '0' + d;
            x = new_x;
        }
    }

    while (c > buf+11-len)
        *--c = ' ';

    return c;
}

void
getBootParams(int *count, int *os, int *silent, enum GENERIC_IO_TYPE *io_type)
{
    NVRAM_STRUCT* params = (NVRAM_STRUCT*) (NVRAM_ADDR + NVRAM_OFFSET);

    if (params->valid == NVRAM_VALID) {
        *count = params->count;
        *os = params->os;
        *silent = params->silent;
        *io_type = params->io_type;
    } else {
        *count = DEFAULT_COUNT;
        *os = DEFAULT_OS;
        *silent = DEFAULT_SILENT;
        *io_type = DEFAULT_IO_TYPE;
    }
}

enum POLL_RESULT
checkAutoBoot(int count, int silent)
{
    char greeting[]  = "Press space bar to stop auto-boot...\n\r### ";
    char updateStr[] = "\r### ";
    char *p;
    enum POLL_RESULT presult = POLL_TIMEOUT;

    if (! silent) {
        p = itoa(count,3);
        greeting[sizeof(greeting)-5] = p[0];
        greeting[sizeof(greeting)-4] = p[1];
        greeting[sizeof(greeting)-3] = p[2];
        outStr(greeting);
        hpcSend(greeting);
    }
    IXP1200_REG_WRITE(IXP1200_LED_ADDR, count);

    while (count > 0) {
        presult = pollOneSec();
        if (presult != POLL_TIMEOUT)
            break;
        count--;
        IXP1200_REG_WRITE(IXP1200_LED_ADDR, count);
        if (! silent) {
            p = itoa(count,3);
            updateStr[sizeof(updateStr)-5] = p[0];
            updateStr[sizeof(updateStr)-4] = p[1];
            updateStr[sizeof(updateStr)-3] = p[2];
            outStr(updateStr);
            hpcSendMaybe(updateStr);
        }
    }
    if (! silent) {
        outStr("\n\r");
        hpcSendMaybe("\n\r");
    }
    return presult;
}

#define BSIZE 8

char *
getStr()
{
    static char buf[BSIZE];
    char c;
    int i;

    i = 0;
    while (1) {
        c = genericInChar();
        if ((c == '\n') || (c == '\r'))
            break;
        if ((c == 0x08) || (c == 0x7F)) {
            if (i > 0) {
                genericOutStr("\010 \010");
                i--;
            }
            continue;
        }
        if (i < BSIZE-1) {
            buf[i] = c;
        }
        genericOutChar(c);
        i++;
    }

    buf[i] = 0;
        
    genericOutStr("\n\r");
    return buf;
}

void
usage()
{
    genericOutStr("BootMgr commands:\n\r");
    genericOutStr("p : Print boot parameters\n\r");
    genericOutStr("c : Change boot parameters\n\r");
    genericOutStr("b : Boot with current parameters\n\r");
    genericOutStr("b <os> : Boot given os without changing parameters\n\r");
    genericOutStr("h : Print this help message\n\r");
}

void
print(int count, int os, int silent, enum GENERIC_IO_TYPE io_type)
{
    unsigned int i;

    genericOutStr("BootMgr Version " VERSION "\n\r");
    genericOutStr("CPU Revision ");
    genericOutStr(hex(sysCpuId,8));
    genericOutStr("\n\rOS list:\n\r");
    for (i=0; i<NUM_OS_ENTRIES; i++) {
        genericOutStr(itoa(i, 3));
        genericOutStr(" ");
        genericOutStr(os_table[i].name);
        genericOutStr("\n\r");
    }
    genericOutStr("IO Type list:\n\r  0 Serial\n\r  1 HPC\n\r");
    genericOutStr("Default OS: ");
    genericOutStr(itoa(os, 0));
    genericOutStr("\n\r");
    genericOutStr("Countdown value: ");
    genericOutStr(itoa(count, 0));
    genericOutStr("\n\r");
    genericOutStr("Disable initial display: ");
    genericOutStr(itoa(silent, 0));
    genericOutStr("\n\r");
    genericOutStr("Default IO type: ");
    genericOutStr(itoa(io_type, 0));
    genericOutStr("\n\r");
}

#define BUFSIZ ((NVRAM_SIZE) / sizeof(UINT))

void
update(int count, int os, int silent, enum GENERIC_IO_TYPE io_type)
{
    UINT buf[BUFSIZ];
    NVRAM_STRUCT* params = (NVRAM_STRUCT*) (((UINT)buf) + NVRAM_OFFSET);

    copyLongs((UINT*)NVRAM_ADDR, buf, NVRAM_SIZE / sizeof(UINT));
    params->valid = NVRAM_VALID;
    params->count = count;
    params->os    = os;
    params->silent = silent;
    params->io_type = io_type;

    flashErase(NVRAM_ADDR);
    flashWrite(buf, NVRAM_SIZE, NVRAM_ADDR, 0);
}

/* returns 0 if a valid number is not found */
int 
atoi(char *c, int *val)
{
    int v = 0;
    int rv = 0;

    while (*c == ' ')
        c++;
    while ((*c >= '0') && (*c <= '9')) {
        v = 10 * v + (*c - '0');
        rv++;
        c++;
    }
    *val = v;
    return rv;
}

void
change(int *count, int *os, int *silent, enum GENERIC_IO_TYPE *io_type)
{
    int new_count, new_os, new_silent;
    enum GENERIC_IO_TYPE new_io_type;
    char *c;

    print(*count, *os, *silent, *io_type);
    genericOutStr("Enter blank line to leave value unchanged\n\r");
    while (1) {
        genericOutStr("Default OS: ");
        c = getStr();
        if (! atoi(c, &new_os)) {
            new_os = *os;
            break;
        }
        if ((new_os >= 0) && (new_os < (int)NUM_OS_ENTRIES))
            break;
        genericOutStr("Invalid OS value\n\r");
    }
    while (1) {
        genericOutStr("Countdown value: ");
        c = getStr();
        if (! atoi(c, &new_count)) {
            new_count = *count;
            break;
        }
        if (new_count < 1) {
            genericOutStr("Minimum countdown value is 1\n\r");
            continue;
        }
        if (new_count > 999) {
            genericOutStr("Maximum countdown value is 999\n\r");
            continue;
        }
        break;
    }
    while (1) {
        genericOutStr("Disable initial display: ");
        c = getStr();
        if (! atoi(c, &new_silent)) {
            new_silent = *silent;
            break;
        }
        if ((new_silent >= 0) && (new_silent < 2))
            break;
        genericOutStr("Invalid value\n\r");
    }
    while (1) {
        genericOutStr("Default IO type: ");
        c = getStr();
        if (! atoi(c, (int*)&new_io_type)) {
            new_io_type = *io_type;
            break;
        }
        if ((new_io_type >= GIT_SERIAL) && (new_io_type <= GIT_HPC))
            break;
        genericOutStr("Invalid value\n\r");
    }
    if ((new_count != *count) ||
        (new_os != *os) ||
        (new_silent != *silent) ||
        (new_io_type != *io_type) )
        update(new_count, new_os, new_silent, new_io_type);
    *os = new_os;
    *count = new_count;
    *silent = new_silent;
    *io_type = new_io_type;
}

void
interactive(int count, int *os, int silent, enum GENERIC_IO_TYPE io_type)
{
    char *c;
    int new_os;

    while (1) {
        genericOutStr("\n\r[BootMgr]: ");
        c = getStr();
        while (*c == ' ')
            c++;
        if ((*c >= 'A') && (*c <= 'Z'))
            *c += 'a' - 'A';
        switch(*c) {
        case 'b':
            if (atoi(c+1, &new_os)) {
                if ((new_os >= 0) && (new_os < (int)NUM_OS_ENTRIES)) {
                    *os = new_os;
                } else {
                    genericOutStr("Invalid OS value\n\r");
                    break;
                }
            } /* end if atoi */
#ifdef TESTING
            genericOutStr("** booting to ");
            genericOutStr(itoa(*os,3));
            genericOutStr("\n\r");
            break;
#endif
            return;
        case 'c':
            change(&count, os, &silent, &io_type);
            break;
        case 'p':
            print(count, *os, silent, io_type);
            break;
#ifdef LOSS
        case 'd': {
            int i;
            char *c = (char*) NVRAM_ADDR;
            for (i=0; i<0x220; i++) {
                genericOutStr(" ");
                genericOutStr(hex(c[i],2));
            }
            genericOutStr("\n\r");
            break;
        }
#endif
        case 'h':
            usage();
            break;
        default:
            genericOutStr("Unknown command: ");
            genericOutChar(*c);
            genericOutStr("\n\r");
            usage();
        }
    }
}

void
_main(int startType)
{
    int count, os, silent;
    OS *os_entry;
    VOIDFUNCPTR os_addr;
    UINT gpio_en, val;
    enum POLL_RESULT presult;
    enum GENERIC_IO_TYPE new_type;
    UINT data;

    bootFlags = startType;

    sysCpuId = sysCpuIDGet();
    sysCpuRev = sysCpuId & 0xF;

    gpio_en = IXP1200_REG_READ1(IXP1200_GPIO_EN);
    IXP1200_REG_WRITE(IXP1200_GPIO_EN, GPIO_OWN_CCCC | GPIO_OUTPUT(0));
    val = IXP1200_REG_READ1(IXP1200_GPIO_DATA);
    IXP1200_REG_WRITE(IXP1200_GPIO_EN, gpio_en);

    if ((val & (1<<3)) && ((sysCpuRev & 0xF) > 0)) {
        flash32BitMode = 0;
    } else {
        flash32BitMode = 1;
    }

    
    set_marquee(marquee);

    uartInit(9600);
    hpcInit();

    getBootParams(&count, &os, &silent, &genericIoType);
    presult = checkAutoBoot(count, silent);
    if (POLL_TIMEOUT != presult) {
        switch (presult) {
        case POLL_HPC:
            new_type = GIT_HPC;
            break;
        default:
            new_type = GIT_SERIAL;
        }

        /* See if IO type has changed */
        if (genericIoType != new_type) {
            genericIoType = new_type;
            update(count, os, silent, genericIoType);
        }

        /* prompt for changes */
        interactive(count, &os, silent, genericIoType);
    }

    /* wait for output to flush */
    switch (genericIoType) {
    case GIT_SERIAL:
        uartFlush();
        {int delay; for (delay=0; delay<10000; delay++);}
        break;
    case GIT_HPC:
        hpcSend("\n");
        break;
    }

    if (os < 0)
        os = 0;
    if (os >= (int)NUM_OS_ENTRIES)
        os = NUM_OS_ENTRIES-1;
    os_entry = os_table + os;
    os_addr = (VOIDFUNCPTR)(os_entry->addr);
    data = os_entry->data;
    switch (genericIoType) {
    case GIT_SERIAL:
        data |= os_entry->boot_flags_serial;
        break;
    case GIT_HPC:
        data |= os_entry->boot_flags_hpc;
        break;
    }
    os_addr(data);
}
