/* flash.cpp
 * Classes to deal with flash information
 *
 * --------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ------------------------------------------------------------------------
 *
 *
 * system: SA1200
 * subsystem: FUtil
 * author: jdg 05/03/00
 * revisions:
 * $History: flash.cpp $
 * 
 * *****************  Version 4  *****************
 * User: Jdguilfo     Date: 11/16/00   Time: 3:53p
 * Updated in $/dev_1200/BoardSupport/BootFlash/host
 * Add support for 28F160 parts (untested)
 * 
 * *****************  Version 3  *****************
 * User: Jdguilfo     Date: 11/14/00   Time: 5:21p
 * Updated in $/dev_1200/BoardSupport/BootFlash/host
 * Spectacle Island Port, includes fixes bugs 1226 & 1227, more checks
 * before updating flash util block
 * 
 * *****************  Version 1  *****************
 * User: Jdguilfo     Date: 5/03/00    Time: 4:39p
 * Created in $/dev_1200/BoardSupport/FlashUtil/host
 * Add new Flash Utility (FUtil)
 * 
 * ------------------------------------------------------------------------
 */

#include <windows.h>

#include "flash.h"

bool
FLASH_INFO::isValid(int type)
{
    if ((type < 0) || (type >= 4))
        return false;
    switch (m_type[type]) {
    case FLASH_28F800T:
    case FLASH_28F800B:
    case FLASH_28F320B:
    case FLASH_28F320T:
    case FLASH_28F160B:
    case FLASH_28F160T:
        return true;
    }
    return false;
}

DWORD
FLASH_INFO::validVector()
{
    DWORD res = 0;
    int i;

    for (i=0; i<4; i++) {
        if (isValid(i))
            res |= (1<<i);
    }
    return res;
}

int
FLASH_INFO::lowestValid()
{
    int i;

    for (i=0; i<4; i++) {
        if (isValid(i))
            return i;
    }
    return -1;
}

DWORD
FLASH_INFO::flashSize(int bank)
{
    DWORD size;

    switch (m_type[bank]) {
    case FLASH_28F800T:
    case FLASH_28F800B:
        size = m_width * (1 * 1024 * 1024 / 16);
        break;
    case FLASH_28F160B:
    case FLASH_28F160T:
        size = m_width * (2 * 1024 * 1024 / 16);
        break;
    case FLASH_28F320B:
    case FLASH_28F320T:
        size = m_width * (4 * 1024 * 1024 / 16);
        break;
    default:
        size = m_bank_size;
    }
    return size;
}

bool 
FLASH_INFO::startOfBlock(int bank, DWORD addr)
{
    DWORD a1, a2;
    a1 = nextBlock(bank, addr-1);
    a2 = nextBlock(bank, addr);
    return ((a2 != MAXDWORD) && (a1 != a2));
}

DWORD
FLASH_INFO::nextBlock(int bank, DWORD addr)
{
    DWORD blockSize;

    if ((bank < 0) || (bank > 3))
        return MAXDWORD;

    switch (m_type[bank]) {
    case FLASH_28F800B:
    case FLASH_28F160B:
    case FLASH_28F320B:
        blockSize = 0x1000 * m_width; // large block size
        if ((addr / blockSize) == 0)  // first large block = 8 small blocks
            blockSize /= 8;
        return (addr + blockSize) & ~(blockSize-1);
    case FLASH_28F800T:
        blockSize = 0x1000 * m_width; // large block size
        if ((addr / blockSize) == 15)  // last large block = 8 small blocks
            blockSize /= 8;
        return (addr + blockSize) & ~(blockSize-1);
    case FLASH_28F160T:
        blockSize = 0x1000 * m_width; // large block size
        if ((addr / blockSize) == 31)  // last large block = 8 small blocks
            blockSize /= 8;
        return (addr + blockSize) & ~(blockSize-1);
    case FLASH_28F320T:
        blockSize = 0x1000 * m_width; // large block size
        if ((addr / blockSize) == 63)  // last large block = 8 small blocks
            blockSize /= 8;
        return (addr + blockSize) & ~(blockSize-1);
    }
    return MAXDWORD;
}

void 
FLASH_INFO::printSummary(FILE *fp)
{
    int i;
    char *c;

    fprintf(fp,"CPU Revision: %08X\n",m_cpu_rev);
    fprintf(fp,"Flash Width: %d\n",m_width);
    fprintf(fp,"Flash Bank Size: %d KBytes\n", 8*m_width*m_size);

    for (i=0; i<4; i++) {
        switch (m_type[i]) {
        case FLASH_28F800T:
            c = "28F800B3-T";
            break;
        case FLASH_28F800B:
            c = "28F800B3-B";
            break;
        case FLASH_28F160B:
            c = "28F160B3-B";
            break;
        case FLASH_28F160T:
            c = "28F160B3-T";
            break;
        case FLASH_28F320B:
            c = "28F320B3-B";
            break;
        case FLASH_28F320T:
            c = "28F320B3-T";
            break;
        default:
            c = "Unknown";
        }
        fprintf(fp,"Flash Type Bank[%d]: %s\n",i,c);
    }
}
