/* span.cpp
 * Class for dealing with regions of addresses
 *
 * --------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ------------------------------------------------------------------------
 *
 *
 * system: SA1200
 * subsystem: FUtil
 * author: jdg 05/03/00
 * revisions:
 * $History: span.cpp $
 * 
 * *****************  Version 1  *****************
 * User: Jdguilfo     Date: 5/03/00    Time: 4:39p
 * Created in $/dev_1200/BoardSupport/FlashUtil/host
 * Add new Flash Utility (FUtil)
 * 
 * ------------------------------------------------------------------------
 * This code defines a class that represents ranges of addresses. The ranges
 * can have holes.
 * ------------------------------------------------------------------------
 */

#include <windows.h>
#include <assert.h>

#include "span.h"

span_item::span_item(unsigned char *buf, DWORD bsize)
{
    m_buf = buf;
    m_bsize = bsize;
    m_start = 0;
    m_end = bsize;
    m_zero = false;
    m_next = m_prev = NULL;
}

span_item::span_item(span_item *sp, DWORD start, DWORD end, bool zero,
                     span_item *next, span_item *prev)
{
    m_buf = sp->m_buf;
    m_bsize = sp->m_bsize;
    m_start = start;
    m_end = end;
    m_zero = zero;
    m_next = next;
    m_prev = prev;
}

span_item::~span_item()
{
    if (m_prev) {
        m_prev->m_next = m_next;
    }
    if (m_next) {
        m_next->m_prev = m_prev;
    }
}

void
span_item::makeZeros(DWORD threash)
{
    int start, end;
    span_item *new_sp;

    end = m_end - 1;
    while (1) {
        for (; end >= (int) m_start; end--) {
            if (m_buf[end] == 0)
                break;
        }
        if (end < (int)(m_start + threash))
            return;
        for (start = end-1; start >= (int) m_start; start--) {
            if (m_buf[start] != 0)
                break;
        }
        if ((end - start) < (int) threash) {
            end = start-1;
            continue;
        }
        // end points to the last zero byte and start points to the byte
        // before the first zero byte, so we need to increment them
        end++;
        start++;

	// round them to word boundaries
	start = (start + 3) & ~3;
	end = end & ~3;

        // make new block from start to end
        splitAfter(end);
        new_sp = splitAfter(start);
        new_sp->m_zero = true;
        end = start-1;
    } // end while 1
}

span_item*
span_item::splitBefore(DWORD ptr)
{
    span_item *sp;

    if (ptr <= m_start)
        return NULL;
    if (ptr >= m_end)
        return this;
    sp = new span_item(this, m_start, ptr, false, this, m_prev);
    m_start = ptr;
    if (m_prev) {
        m_prev->m_next = sp;
    }
    m_prev = sp;
    return sp;
}

span_item*
span_item::splitAfter(DWORD ptr)
{
    span_item *sp;

    if (ptr <= m_start)
        return this;
    if (ptr >= m_end)
        return NULL;
    sp = new span_item(this, ptr, m_end, false, m_next, this);
    m_end = ptr;
    if (m_next) {
        m_next->m_prev = sp;
    }
    m_next = sp;
    return sp;
}

//////////////////////////////////////////////////////////

span::span(unsigned char *buf, DWORD bsize)
{
    m_head = new span_item(buf, bsize);
}

span::~span()
{
    span_item *span, *next;

    for (span = m_head; span; span = next) {
        next = span->m_next;
        delete span;
    }
}

void 
span::limit(DWORD start, DWORD end)
{
    insertHole(0, start);
    insertHole(end, MAXDWORD);
}

void
span::insertHole(DWORD start, DWORD end)
{
    span_item *sp, *next;

    sp = m_head;
    while (1) {
        if (start >= sp->m_end) {
            if (sp->m_next == NULL) {
                break;
            }
            sp = sp->m_next;
            continue;
        }
        // start <= m_end
        if (start <= sp->m_start) {
            if (end >= sp->m_end) {
                // case 1. delete entire span
                if (sp == m_head)
                    m_head = sp->m_next;
                next = sp->m_next;
                delete sp;
                if (next == NULL)
                    break;
                sp = next;
                continue;
            } else {
                // case 2. delete start of span
                sp->m_start = end;
                break;
            }
        } else {
            // start > sp->m_start
            if (end >= sp->m_end) {
                // case 3. delete end of span
                sp->m_end = start;
                if (sp->m_next == NULL)
                    break;
                sp = sp->m_next;
                continue;
            } else {
                // case 4. delete middle of span
                // new span covers second half
                next = new span_item(sp, end, sp->m_end, false, sp->m_next, sp);
                sp->m_end = start;
                sp->m_next = next;
                break;
            }
        }
    } // end while 1
}

void 
span::makeZeros(DWORD threash)
{
    span_item *item, *next;

    for (item = m_head; item; item = next) {
        next = item->m_next;
        item->makeZeros(threash);
    }
}

bool
span::get_span(DWORD& start, DWORD& end, bool& zero, void* &state)
{
    span_item *item;

    if (state == NULL)
        item = m_head;
    else
        item = ((span_item*) state)->m_next;

    if (item == NULL)
        return false;

    start = item->m_start;
    end   = item->m_end;
    zero  = item->m_zero;
    state = (void*) item;
    return true;
}

bool 
span::get_span_joined(DWORD& start, DWORD& end, void* &state)
{
    span_item *item, *next;

    if (state == NULL)
        item = m_head;
    else
        item = ((span_item*) state)->m_next;

    if (item == NULL)
        return false;

    start = item->m_start;
    for (next = item->m_next;
         next;
         item = next, next = next->m_next) {
        if (item->m_end != next->m_start)
            break;
    }
    end = item->m_end;
    state = (void*) item;
    return true;
}

bool
span::overlaps(DWORD start, DWORD end)
{
    span_item *item, *next;

    for (item = m_head; item; item = next) {
        next = item->m_next;
        if ((start < item->m_end) && (end > item->m_start))
            return true;
    }
    return false;
}
