/* smalloc.c
 * Simple malloc utility functions for OS-independent code
 *
 * --------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ------------------------------------------------------------------------
 *
 *
 * system: SA1200
 * subsystem: FUtil
 * author: jdg 05/03/00
 * revisions:
 * $History: smalloc.c $
 * 
 * *****************  Version 5  *****************
 * User: Jdguilfo     Date: 11/14/00   Time: 5:21p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Spectacle Island Port, includes fixes bugs 1226 & 1227, more checks
 * before updating flash util block
 * 
 * *****************  Version 4  *****************
 * User: Jdguilfo     Date: 10/26/00   Time: 3:38p
 * Updated in $/dev_1200/BoardSupport/BootFlash
 * Merge serial and HPC versions into one version, see bootMgr.txt for
 * details.
 * 
 * *****************  Version 1  *****************
 * User: Jdguilfo     Date: 5/03/00    Time: 4:37p
 * Created in $/dev_1200/BoardSupport/FlashUtil
 * Add new Flash Utility (FUtil)
 * 
 * ------------------------------------------------------------------------
 * This code defines a simple malloc interface using a fixed pool. It is
 * intended for OS-independent code where is no native malloc.
 * ------------------------------------------------------------------------
 */

typedef unsigned int UINT;

#include "smalloc.h"

#define START_ADDR 0xC0020000
#define POOL_SIZE  (20*1024*1024)

#define NULL ((void*)0)

typedef struct header {
    struct header *ptr;
    unsigned int size;
} HEADER;

static HEADER *base = NULL;
static HEADER *allocp;

/**************************************************/

void*
memcpy(void* dest, void* src, int size)
{
    if ((((UINT)dest & 3) == 0) && (((UINT)src & 3) == 0)) {
        UINT *d, *s, *e;
        d = (UINT*)dest;
        s = (UINT*)src;
        e = (UINT*)((char*)src + size);
        while(s < e)
            *(d++) = *(s++);
    } else {
        char *d, *s, *e;
        d = (char*)dest;
        s = (char*)src;
        e = (char*)((char*)src + size);
        while(s < e)
            *(d++) = *(s++);
    }

    return dest;
}

int 
bcmp(unsigned char *buf1, unsigned char *buf2, int nBytes)
{
    if ((((int)buf1 & 3) == 0) &&
        (((int)buf2 & 3) == 0) &&
        ((nBytes & 3) == 0)) {
        unsigned int *b1, *b2;
        b1 = (unsigned int*) buf1;
        b2 = (unsigned int*) buf2;
        nBytes >>= 2;
        while (nBytes-- > 0) {
            if (*b1++ != *b2++)
                return 1;
        }
        return 0;
    }
    while (nBytes-- > 0) {
        if (*buf1++ != *buf2++)
            return 1;
    }
    return 0;
}

void*
calloc( int num_of_elts, int size)
{
    return malloc(num_of_elts * size);
}

/**************************************************/

static UINT*
alloc(int size)
{
    unsigned int nunits;
    HEADER *p, *q;

    if (base == NULL) {
        base = (HEADER*) START_ADDR;
        base->ptr = base;
        base->size = POOL_SIZE / sizeof(HEADER);
        allocp = base;
    }

    nunits = 1 + (size + sizeof(HEADER) - 1) / sizeof(HEADER);
    for (q = allocp, p=q->ptr; ; q=p, p=p->ptr) {
        if (p->size >= nunits) { /* big enough */
            if (p->size == nunits) { /* exactly */
                q->ptr = p->ptr;
            } else { /* allocate tail end */
                p->size -= nunits;
                p += p->size;
                p->size = nunits;
            }
            allocp = q;
            return (UINT*)(p+1);
        }
        if (p == allocp)
            return NULL;
    }
}

void*
malloc(int size)
{
    UINT *p, *q;
    p = alloc(size);
    if (p != 0) {
        for (q = p + (size / sizeof(UINT))-1; q >= p; q--)
            *q = 0;
    }
    return p;
}

void
free(void *ptr)
{
    HEADER *p, *q;

    p = (HEADER*)ptr - 1;
    for (q = allocp; !((p > q) && (p < q->ptr)); q=q->ptr)
        if ((q >= q->ptr) && ((p > q) || (p < q->ptr)))
            break;
    if (p+p->size == q->ptr) { /* join to upper nbr */
        p->size += q->ptr->size;
        p->ptr = q->ptr->ptr;
    } else {
        p->ptr = q->ptr;
    }
    if (q + q->size == p) { /* join to lower nbr */
        q->size += p->size;
        q->ptr = p->ptr;
    } else {
        q->ptr = p;
    }
    allocp = q;
}

char*
hex(unsigned int val, int size)
{
    static char buf[20];
    static char h[16] = {'0','1','2','3','4','5','6','7',
                         '8','9','A','B','C','D','E','F'};
    char *b = &buf[0];
    while (--size >= 0) {
        *(b++) = h[(val >> (4*size)) & 0xF];
    }
    *b = 0;
    return &buf[0];
}
