/* mmuArmLib.h - ARM MMU library header file */

/* Copyright 1996-1998 Wind River Systems, Inc. */

/*
modification history
--------------------
01g,31may00,jdg  added TOR201 hack for Tornado 2.0.1 support
01f,17may99,jdg  added MMU_ID_SA1200
01e,09mar98,jpd  added MMUCTR_ definitions for ARM810 and extra cache states.
01d,20jan98,jpd  added _func_armPhysToVirt() etc. function pointers.
01c,17sep97,jpd  added MMU_INIT_VALUE and mmuIntLock/Unlock definitions.
01b,24jan97,jpd  modified mmuAEnable definition.
01a,09may96,cdp  created.
*/

#ifdef TOR201
#include "../h/arch/arm/mmuArmLib.h"
#else

#ifndef	__INCmmuArmLibh
#define	__INCmmuArmLibh

/*
 * N.B. although this library contains code written for the ARM810 CPU,
 * at the time of writing, this code has not been fully tested on
 * that CPU. YOU HAVE BEEN WARNED.
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifndef	_ASMLANGUAGE

#include "memLib.h"

#endif /* _ASMLANGUAGE */

/*
 * MMUs are not fitted to all processors, and so the definitions below need not
 * be included for all processors. In any case, only the following are
 * supported:
 */

#if ((CPU == ARM710A) || (CPU == ARM810) || (CPU == ARMSA110))

/* MMU identifiers as read from ID register */

#define MMU_ID_CPU_MASK	0xFFF0	/* Mask for CPU types */

#define MMU_ID_ARM710A	0x7100
#define MMU_ID_ARM810	0x8100
#define MMU_ID_SA110	0xA100
#define MMU_ID_SA1100	0xA110
#define MMU_ID_SA1200	0xC120


/* The page size we will use. Ignore the sub-page concept */

#define PAGE_SIZE 4096


/*
 * The amount described by a Level 1 Descriptor, which equals the smallest
 * amount of VM allocatable in VxWorks.
 */

#define PAGE_BLOCK_SIZE 0x100000	/* 1 MB */


#define PTE_INDEX_MASK 0x000FF000	/* extract PTE index from Virt Addr */
#define PTE_INDEX_SHIFT 12		/* shift to make that into PTE index */

#define L1D_TO_BASE_SHIFT 10		/* to turn L1D to PT base addr */

#define ADDR_PI_MASK 0x00000FFF		/* extract page index from Virt Addr */

#define ADDR_TO_PAGE 12	/* shift phys address to PTE page base address */


/* Level 1 Descriptor types */

#define DESC_TYPE_PAGE 1

#define DEF_L1_PAGE 0x00000011		/* domain zero, Page descriptor */

/*
 * Section descriptors, such as might be used to set up an intermediate
 * set of page tables on SA-1100 from BSP initialisation, before
 * vm(Base)Lib sets up a proper set of page tables.
 */

#define RAM_SECT_DESC   0xC1E   /* R/W cacheable bufferable domain 0 */
#define OTHER_SECT_DESC 0xC12   /* R/W domain 0 neither cache- nor bufferable */


/* Level 2 Descriptor or Page Table Entry (PTE) types */

#define PTE_TYPE_FAULT	 0		/* any access will cause a fault */
#define PTE_TYPE_SM_PAGE 2		/* Small page descriptor */


/*
 * The Translation Table Base register (TTBR) points to a table of Level 1
 * Descriptors. these are either Invalid Section descriptors, Section
 * Descriptors, or Page Table Descriptors. If Page Table Descriptors, they
 * each point to a table of Level 2 Page Descriptors, or Page Table Entries
 * (PTEs).
 * The 32-bit (virtual) address space allows for 4096 M. Each Level 1
 * Descriptor describes a 1 M area of memory. There are therefore 4096 Level
 * 1 Descriptors, and each table of 256 Level 2 Page Descriptors (PTEs)
 * describes 256 4 kbyte pages.
 */

#define NUM_L1_DESCS 4096
#define NUM_L2_DESCS 256
#define L1_TABLE_SIZE (NUM_L1_DESCS * sizeof(LEVEL_1_DESC))


/* No. of pages a Level 1 Descriptor table takes up */

#define L1_DESC_PAGES (L1_TABLE_SIZE / PAGE_SIZE)


/* Size of a Page Table */

#define PAGE_TABLE_SIZE (NUM_L2_DESCS * sizeof(PTE))


/*
 * Architecture-dependent MMU states. These are states settable for pages and
 * here they correspond to settings in the Page Table Entries (PTEs).
 *
 * We set Valid/Invalid by setting a Fault second-level descriptor rather
 * than by using the Access Permissions within a small page second-level
 * descriptor. This is because we will want to use the Access Permissions to
 * allow read/write from supervisor mode and we cannot then use the AP bits
 * to disallow access as the SR bits in theMMU Control register must be set
 * to 01 in order to control read/write access from Supervisor mode.
 */

#define MMU_STATE_MASK_VALID		0x03		/* 2nd level desc type*/
#define MMU_STATE_MASK_WRITABLE		0x00000FF0	/* All APn bits */
#define MMU_STATE_MASK_CACHEABLE	0x0000000C	/* CB bits */

#define MMU_STATE_VALID			PTE_TYPE_SM_PAGE /* set to page type */
#define MMU_STATE_VALID_NOT		PTE_TYPE_FAULT	/* set to type fault */
#define MMU_STATE_WRITABLE		0x00000FF0	/* APn bits to 11 */
#define MMU_STATE_WRITABLE_NOT		0x0		/* APn bits to 00 */


/* Write-through mode is only available on 710A */

#if (CPU == ARM710A)
#define MMU_STATE_CACHEABLE_WRITETHROUGH	0x8
#endif
#if (CPU == ARMSA110)
/* actually only available on SA-1100 */

#define MMU_STATE_CACHEABLE_MINICACHE		0x8  /* allocate in minicache */
#endif

#define MMU_STATE_CACHEABLE_COPYBACK		0xC

/*
 * Set the default state to be copyback. CACHEABLE_WRITETHROUGH can also be
 * selected on 710A.
 */

#define MMU_STATE_CACHEABLE		MMU_STATE_CACHEABLE_COPYBACK

#define MMU_STATE_CACHEABLE_NOT		0x0

#define MMU_STATE_BUFFERABLE		0x4	/* bufferable, not cacheable */
#define MMU_STATE_BUFFERABLE_NOT	0x0	/* will also set not cacheable */


/* The coprocessor number of the MMU System Control Processor */

#define CP_MMU 15


/* MMU Control Register bit allocations */

#define MMUCR_M_ENABLE	(1)	/* MMU enable */
#define MMUCR_A_ENABLE	(1<<1)	/* Address alignment fault enable */
#define MMUCR_C_ENABLE	(1<<2)	/* (data) cache enable */
#define MMUCR_W_ENABLE	(1<<3)	/* write buffer enable */
#define MMUCR_PROG32	(1<<4)	/* PROG32 */
#define MMUCR_DATA32	(1<<5)	/* DATA32 */
#define MMUCR_L_ENABLE	(1<<6)	/* Late abort on earlier CPUs */
#define MMUCR_BIGEND	(1<<7)	/* Big-endian (=1), little-endian (=0) */
#define MMUCR_SYSTEM	(1<<8)	/* System bit, modifies MMU protections */
#define MMUCR_ROM	(1<<9)	/* ROM bit, modifies MMU protections */
#define MMUCR_F		(1<<10)	/* Should Be Zero */
#define MMUCR_Z_ENABLE	(1<<11)	/* Branch prediction enable on 810 */
#define MMUCR_I_ENABLE	(1<<12)	/* Instruction cache enable on SA110 */


/*
 * Values to be used when mmuEnable() is called. This will be after the MMU has
 * been initialised by sysInit()/romInit() and after cacheLib has set whatever
 * cache enable settings have been chosen.
 *
 * M 1 Enable MMU
 * A 1 Enable address alignment fault
 * C X ((D-)Cache Enable) Controlled by cacheLib
 * W X (Write Buffer) Controlled by cacheLib
 * P X (PROG32) should be set before this
 * D X (DATA32) should be set before this
 * L X (Late abort on earlier CPUs) ignore
 * B X (Big/Little-endian) should be set before this
 * S 0 (System)
 * R 1 (ROM)
 * F 0 Should be Zero
 * Z X (Branch prediction enable on 810) Controlled by cacheLib
 * I X (I-cache enable) Controlled by cacheLib
 */

/*
 * For time being, do not enable the address alignment fault, as GCC currently
 * generates incorrect code, and switching this on will cause immediate
 * faults.
 */

#define MMU_ENABLE_MASK  (0xFFFFE000 | \
			MMUCR_M_ENABLE | \
			MMUCR_SYSTEM | MMUCR_ROM | MMUCR_F | MMUCR_Z_ENABLE)

#if (CPU == ARM710A)
#define MMU_ENABLE_VALUE (MMUCR_M_ENABLE | MMUCR_A_ENABLE | MMUCR_ROM)
#endif

#if (CPU == ARM810)
#define MMU_ENABLE_VALUE (MMUCR_M_ENABLE | MMUCR_A_ENABLE | MMUCR_ROM )
#endif

#if (CPU == ARMSA110)
#define MMU_ENABLE_VALUE (MMUCR_M_ENABLE | MMUCR_A_ENABLE | MMUCR_ROM)
#endif


/*
 * Values to be used when the MMU Control Register is initialised from within
 * romInit.s or sysALib.s (before the MMU is enabled).
 */

#if (CPU == ARM710A)
#define MMU_INIT_VALUE (MMUCR_PROG32 | MMUCR_DATA32 | MMUCR_L_ENABLE | MMUCR_ROM)
#endif

#if (CPU == ARM810)
#define MMU_INIT_VALUE (MMUCR_PROG32 | MMUCR_DATA32 | MMUCR_L_ENABLE | MMUCR_ROM)
#endif

#if (CPU == ARMSA110)
#define MMU_INIT_VALUE (MMUCR_PROG32 | MMUCR_DATA32 | MMUCR_L_ENABLE | MMUCR_ROM)
#endif

#if (CPU == ARM810)

/*
 * Definitions for register 15 of the MMU coprocessor (clock and test
 * configuration).
 */

#define MMUCTR_D 	(1<<0)	/* Enable Dynamic Clock switching */
#define MMUCTR_S 	(1<<1)	/* Enable Synchronous clock switching */
#define MMUCTR_F_BUS	(0<<2)	/* Fast clock source: ... bus clock */
#define MMUCTR_F_REF	(1<<2)	/* ... REFCLK */
#define MMUCTR_F_PLL	(3<<2)	/* ... PLL output clock */

#endif /* (CPU == ARM810) */

#ifndef	_ASMLANGUAGE

/* Translation Table Base Register (TTBR) */

typedef struct
    {
    UINT pad	: 14;
    UINT addr	: 18;		/* address of Translation Table */
				/* (must reside on 16 kbyte boundary) */
    } TTB_REG;

typedef struct
    {
    UINT type	: 2;		/* descriptor type, 1 => page */
    UINT pad1	: 2;		/* SBZ */
    UINT pad2	: 1;		/* SBO */
    UINT domain	: 4;		/* domain number */
    UINT pad3	: 1;		/* SBZ */
    UINT addr	: 22;		/* base address of page table */
    } PAGE_DESC_FIELD;


typedef union
    {
    PAGE_DESC_FIELD fields;
    UINT32 bits;
    } LEVEL_1_DESC;


/* Layout of Page Table Entries (PTEs), actually small page descriptors */

typedef struct
    {
    UINT type	: 2;		/* page type, 2 => small page */
    UINT cb	: 2;		/* cacheable/bufferable bits */
    UINT ap0	: 2;		/* access permission, sub-page 0 */
    UINT ap1	: 2;		/* access permission, sub-page 1 */
    UINT ap2	: 2;		/* access permission, sub-page 2 */
    UINT ap3	: 2;		/* access permission, sub-page 3 */
    UINT addr	: 20;		/* page base address */
    } PTE_FIELD;

typedef union
    {
    PTE_FIELD fields;
    UINT32 bits;
    } PTE;

typedef struct mmuTransTblStruct
    {
    LEVEL_1_DESC *pLevel1Table;
    } MMU_TRANS_TBL;


/* Externally visible mmuLib and mmuALib routines */

IMPORT STATUS mmuLibInit (int pageSize);
IMPORT UINT32 mmuReadId (void);
IMPORT UINT32 mmuCrGet (void);
IMPORT UINT32 mmuModifyCr (UINT32 value, UINT32 mask);
IMPORT void   mmuTtbrSet (LEVEL_1_DESC *);
IMPORT void   mmuDacrSet (UINT32 dacRVal);
IMPORT void   mmuAEnable (UINT32 cacheState);
IMPORT void   mmuADisable (void);
IMPORT void   mmuTLBIDFlushEntry (void *addr);
IMPORT void   mmuTLBIDFlushAll (void);
IMPORT int    mmuIntLock (void);
IMPORT void   mmuIntUnlock (int lockKey);

/*
 * Pointers to functions that can be filled in by the BSP to point to
 * functions that convert between virtual and physical addresses. This is
 * particularly necessary on those BSPs where it is not possible to provide
 * a memory map where virtual and physical addresses are identical.
 */

IMPORT void * (* _func_armVirtToPhys) (void *);
IMPORT void * (* _func_armPhysToVirt) (void *);

/*
 * Pointer to a function that can be filled in by the BSP to point to a
 * function that returns a memory partition id for an area of memory to store
 * the Level 1 and Level 2 page tables. This area must be big enough for all
 * use. No provision is made to use that memory and then continue using
 * system memory once that has been filled.
 * N.B. at the time of writing, this feature has NEVER been tested at all.
 */

IMPORT PART_ID (* _func_armPageSource) (void);


#endif	/* _ASMLANGUAGE */

#endif /* ((CPU == ARM710A) || (CPU == ARM810) || (CPU == ARMSA110)) */

#ifdef __cplusplus
}
#endif

#endif	/* __INCmmuArmLibh */

#endif /* ifdef TOR201 */
