/* flashUtil.c - flash memory utility */

/* Copyright 1997-1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */

/*
modification history
--------------------
01a,08jul99,jdg  Created from version 01i of flashMem.c
*/

/*
DESCRIPTION
This provides a series of routines for writing flash parts.
It is aimed primarily at the Intel 28F800 parts, but it might
also work on other Intel 28F and 29F parts.
*/

#include "flashUtil.h"

#define FLASH28CMD_READ_MEM             0x00000000
#define FLASH28CMD_READ_ID              0x90909090
#define FLASH28CMD_ERASE_SETUP          0x20202020
#define FLASH28CMD_ERASE                0x20202020
#define FLASH28CMD_ERASE_CONFIRM        0xd0d0d0d0
#define FLASH28CMD_ERASE_VERIFY         0xa0a0a0a0
#define FLASH28CMD_PROG_SETUP           0x40404040
#define FLASH28CMD_PROG_VERIFY          0xc0c0c0c0
#define FLASH28CMD_RESET                0xffffffff

#define FTYPE_28F800T                   0x92          /* device code 28F800-T */
#define FTYPE_28F800B                   0x93          /* devide code 28F800-B */
#define FTYPE_28F160T                   0x90          /* device code 28F160-T */
#define FTYPE_28F160B                   0x91          /* devide code 28F160-B */
#define FTYPE_28F320T                   0x96          /* device code 28F320-T */
#define FTYPE_28F320B                   0x97          /* device code 28F320-B */

#define CAST_FLASH (FLASH_TYPE*)

#ifndef SYS_FLASH_WRITE_ENABLE_RTN
#define SYS_FLASH_WRITE_ENABLE_RTN()
#endif

#ifndef SYS_FLASH_WRITE_DISABLE_RTN
#define SYS_FLASH_WRITE_DISABLE_RTN()
#endif

#ifndef FLASH_MEM_TYPE
#define FLASH_MEM_TYPE FTYPE_NONE
#endif

#define BIT(i) (1 << (i))

UINT8 flashType = FLASH_MEM_TYPE;
UINT32 flashBase = FLASH_BASE;

/*
 * flash32BitMode is a boolean indicating whether we should program the
 * flash in 32-bit mode or 16-bit mode. It is normally set to whatever
 * mode the HW is in. If it is set to the mode the HW is not in, i.e. the
 * HW is in 16-bit mode and we want to program in 32-bit mode, or vice versa,
 * then flash32BitMode should be set to the new mode and flashModeSwap should
 * be set to 1. That is: flash32BitMode should be set to the mode we want
 * to program in, and the actual mode is (flash32BitMode ^ flashModeSwap).
 */
extern int flash32BitMode;
int flashModeSwap = 0;

#define WRITE_ADDR(offset) (flash32BitMode ? (offset) : ((offset)*2))
#define READ_ADDR(offset) \
  (flashModeSwap ? (flash32BitMode ? ((offset)/2) : ((offset)*2)) : (offset))

/******************************************************************************
*
* flashTypeGet - determine the device type of on-board flash memory 
*
* This routine uses the `autoselect' command to determine the device type of
* on-board flash memory for flash 28F devices.
*
* RETURNS: An integer indicating the device type of on-board flash memory.
*/

UINT8 flashTypeGet (UINT32 flash_offset)
    {
    volatile FLASH_TYPE* pFR = CAST_FLASH (flashBase + READ_ADDR(flash_offset));
    volatile FLASH_TYPE* pFW = CAST_FLASH (flashBase + WRITE_ADDR(flash_offset));
    UINT8 retVal;

    SYS_FLASH_WRITE_ENABLE_RTN ();                      /* enable writes */

#ifdef ENABLE_29
    *(CAST_FLASH FLASH29_REG_FIRST_CYCLE)  = FLASH29_CMD_FIRST;
    *(CAST_FLASH FLASH29_REG_SECOND_CYCLE) = FLASH29_CMD_SECOND;
    *(CAST_FLASH FLASH29_REG_FIRST_CYCLE)  = FLASH29_CMD_AUTOSELECT;
#else
    *(pFW) = FLASH28CMD_READ_ID;
#endif
    retVal = (UINT8) ((pFR[1]) >> 16);

#ifdef ENABLE_29
    *(CAST_FLASH FLASH29_REG_FIRST_CYCLE)  = FLASH29_CMD_FIRST;
    *(CAST_FLASH FLASH29_REG_SECOND_CYCLE) = FLASH29_CMD_SECOND;
    *(CAST_FLASH FLASH29_REG_FIRST_CYCLE)  = FLASH29_CMD_READ_RESET;
#else
    *(pFW) = FLASH28CMD_RESET;
#endif

    SYS_FLASH_WRITE_DISABLE_RTN ();                     /* disable writes */

    return (retVal);
    }

/******************************************************************************
*
* flashErase - erase the contents of flash memory
*
* This routine clears the contents of flash memory.
*
* Flash 28F\f2xxx\f1 devices are erased by writing a flash erase command to
* the device and verifying that each flash location is set to a high value
* (0xFF).
* 
* Flash 29F\f2xxx\f1 devices are erased by writing the six-byte erase code
* into specific address locations, which sets all byte locations to a high
* value (0xFF).
*
* RETURNS: OK, or ERROR if the contents of flash memory cannot be erased.
*/

STATUS flashErase
    (
    UINT32 flash_offset
    )
    {
    volatile FLASH_TYPE* pFR = CAST_FLASH (flashBase + READ_ADDR(flash_offset));
    volatile FLASH_TYPE* pFW = CAST_FLASH (flashBase + WRITE_ADDR(flash_offset));
    STATUS retVal = OK;
    int ix, status;

    if (flashType == FTYPE_NONE)
        flashType = flashTypeGet(flash_offset);

    switch (flashType)
        {
        case (FTYPE_28F800B):
        case (FTYPE_28F800T):
        case (FTYPE_28F160B):
        case (FTYPE_28F160T):
        case (FTYPE_28F320T):
        case (FTYPE_28F320B):
            {
            SYS_FLASH_WRITE_ENABLE_RTN ();              /* raise Vpp */

            *pFW = FLASH28CMD_ERASE_SETUP;              /* setup */
            *pFW = FLASH28CMD_ERASE_CONFIRM;            /* erase */

            /* wait for completion */
            for (ix = (FLASH_WIDTH/2)-1; ix >= 0; ix--) 
                {
                for (;;)
                    {
                    status = (*pFR) >> (ix * 16);
                    if (status & BIT(7))
                        break;
                    }
                if (status & (BIT(5) | BIT(4) | BIT(3) | BIT(1)))
                    {
                    retVal = ERROR;
                    break;
                    }
                }

            *pFW = FLASH28CMD_RESET;                    /* read array */
            
            SYS_FLASH_WRITE_DISABLE_RTN ();             /* disable enable */
            break;
            }
        default:
            retVal = ERROR;
        }

    return (retVal);
    }


/******************************************************************************
*
* flashWrite - write data to flash memory
*
* This routine copies specified data of a specified length, <size>, into a 
* specified offset, <offset>, in the flash memory.  Data is passed as a string,
* <pFB>, if not NULL.  If NULL, data is taken as a repeated sequence of
* <value>.
* 
* RETURNS: OK, or ERROR if the write operation fails.
* 
* SEE ALSO: sysFlashSet()
*/

STATUS flashWrite
    (
    FLASH_TYPE * pFB,           /* string to be copied; use <value> if NULL */
    int size,                   /* size to program in bytes */
    int flash_offset,           /* byte offset into flash memory */
    FLASH_TYPE value            /* value to program */
    )
    {
    volatile FLASH_TYPE * pFW;          /* flash address */
    volatile FLASH_TYPE * pFR;          /* flash address */
    STATUS retVal = OK;
    int i, ix, status, pass;
    FLASH_TYPE valtmp;

    if (flashType == FTYPE_NONE)
        flashType = flashTypeGet(flash_offset);

    switch (flashType)
        {
        case (FTYPE_28F800B):
        case (FTYPE_28F800T):
        case (FTYPE_28F160B):
        case (FTYPE_28F160T):
        case (FTYPE_28F320T):
        case (FTYPE_28F320B):
            {
            SYS_FLASH_WRITE_ENABLE_RTN ();              /* raise Vpp */

            pFR = CAST_FLASH(flashBase + READ_ADDR(flash_offset));
            pFW = CAST_FLASH(flashBase + WRITE_ADDR(flash_offset));
            for (i = 0; (i < size) && (retVal == OK); i+=4)
                {
                if (pFB != NULL)
                    valtmp = *pFB++;
                else
                    valtmp = value;

                for (pass = 0; pass < 2-flash32BitMode; pass++) {

                    *pFW = FLASH28CMD_PROG_SETUP;           /* write setup */
                    *pFW = valtmp;
                    pFW++;
                    valtmp >>= 16;

                    /* wait for completion */
                    for (ix = (FLASH_WIDTH/2)-1; ix >= 0; ix--) 
                        {
                            for (;;)
                                {
                                    status = (*pFR) >> (ix * 16);
                                    if (status & BIT(7))
                                        break;
                                }
                            if (status & (BIT(5) | BIT(4) | BIT(3) | BIT(1)))
                                {
                                    retVal = ERROR;
                                    break;
                                }
                        } /* end for ix */
                    } /* end for pass */
                } /* end for i */
            
            *--pFW = FLASH28CMD_RESET;                  /* read array */
            
            SYS_FLASH_WRITE_DISABLE_RTN ();             /* disable enable */
            break;
            }

        default:
            retVal = ERROR;
        }
 
    return (retVal);
    }


/******************************************************************************
*
* flashNextBlock - Advance a pointer to the next flash block
*
* This routine advances a pointer to the start of the next flash block
* for a multi-block device.
* 
* RETURNS: OK, or ERROR if the device type is unknown
*/

STATUS flashNextBlock
    (
    UINT32 *flash_offset      /* Offset into flash memory */
    )
    {
    STATUS retVal = OK;
    UINT32 blockSize;
    UINT32 flashSize;

    if (flashType == FTYPE_NONE)
        flashType = flashTypeGet(*flash_offset);

    switch (flashType)
        {
        case (FTYPE_28F800B):
        case (FTYPE_28F800T):
            flashSize = 1024*1024;
            break;
        case (FTYPE_28F160B):
        case (FTYPE_28F160T):
            flashSize = 2 * 1024*1024;
            break;
        case (FTYPE_28F320B):
        case (FTYPE_28F320T):
            flashSize = 4 * 1024*1024;
            break;
        default:
            return ERROR;
        }
    if (flash32BitMode)
        flashSize *= 2;

    switch (flashType)
        {
        case (FTYPE_28F800B):
        case (FTYPE_28F160B):
        case (FTYPE_28F320B):
            if (flash32BitMode) {
                blockSize = 0x8000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x8000 * FLASH_WIDTH))
                    == 0)
                    blockSize /= 8;
            } else {
                blockSize = 0x4000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x4000 * FLASH_WIDTH))
                    == 0)
                    blockSize /= 8;
            }
            *flash_offset = (*flash_offset + blockSize) & ~(blockSize-1);
            break;
        case (FTYPE_28F800T):
            if (flash32BitMode) {
                blockSize = 0x8000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x8000 * FLASH_WIDTH))
                    == 15)
                    blockSize /= 8;
            } else {
                blockSize = 0x4000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x4000 * FLASH_WIDTH))
                    == 15)
                    blockSize /= 8;
            }
            *flash_offset = (*flash_offset + blockSize) & ~(blockSize-1);
            break;
        case (FTYPE_28F160T):
            if (flash32BitMode) {
                blockSize = 0x8000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x8000 * FLASH_WIDTH))
                    == 31)
                    blockSize /= 8;
            } else {
                blockSize = 0x4000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x4000 * FLASH_WIDTH))
                    == 31)
                    blockSize /= 8;
            }
            *flash_offset = (*flash_offset + blockSize) & ~(blockSize-1);
            break;
        case (FTYPE_28F320T):
            if (flash32BitMode) {
                blockSize = 0x8000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x8000 * FLASH_WIDTH))
                    == 63)
                    blockSize /= 8;
            } else {
                blockSize = 0x4000 * FLASH_WIDTH; /* Large block size */
                if ((((*flash_offset) & (flashSize-1)) / (0x4000 * FLASH_WIDTH))
                    == 63)
                    blockSize /= 8;
            }
            *flash_offset = (*flash_offset + blockSize) & ~(blockSize-1);
            break;
        default:
            retVal = ERROR;
        }

    return (retVal);
    }

#ifdef FLASH_SIZE

/******************************************************************************
*
* sysFlashGet - get the contents of flash memory
*
* This routine copies the contents of flash memory into a specified
* string.  The string is terminated with an EOS.
*
* RETURNS: OK, or ERROR if access is outside the flash memory range.
*
* SEE ALSO: sysFlashSet()
*
* INTERNAL
* If multiple tasks are calling sysFlashSet() and sysFlashGet(),
* they should use a semaphore to ensure mutually exclusive access.
*/

STATUS sysFlashGet
    (
    char *string,    /* where to copy flash memory      */
    int strLen,      /* maximum number of bytes to copy */
    int offset       /* byte offset into flash memory   */
    )
    {
    if ((offset < 0) || (strLen < 0) || ((offset + strLen) > FLASH_SIZE))
        return (ERROR);

    bcopyBytes ((char *) (FLASH_ADRS + offset), string, strLen);
    string [strLen] = EOS;

    return (OK);
    }

/******************************************************************************
*
* sysFlashSet - write to flash memory
*
* This routine copies a specified string into flash memory after calling 
* sysFlashErase() and clearing flash memory.
*
* If the specified string must be overlaid on the contents of flash memory, 
* undefine FLASH_NO_OVERLAY.
*
* RETURNS: OK, or ERROR if the write fails or the input parameters are 
* out of range.
*
* SEE ALSO: sysFlashErase(), sysFlashGet(), sysFlashTypeGet(), sysFlashWrite()
*
* INTERNAL
* If multiple tasks are calling sysFlashSet() and sysFlashGet(),
* they should use a semaphore to ensure mutually exclusive access to flash
* memory.
*/

STATUS sysFlashSet
    (
    char *string,     /* string to be copied into flash memory */
    int strLen,       /* maximum number of bytes to copy       */
    int offset        /* byte offset into flash memory         */
    )
    {
    int flashSize = FLASH_SIZE;

    static char tempBuffer [FLASH_SIZE];

    if ((offset < 0) || (strLen < 0) || ((offset + strLen) > flashSize))
        return (ERROR);

    /* see if contents are actually changing */

    if (bcmp ((char *) (FLASH_ADRS + offset), string, strLen) == 0)
        return (OK);

    /* first read existing data */

#ifndef FLASH_NO_OVERLAY
    bcopyBytes ((char *) FLASH_ADRS, tempBuffer, flashSize);
    bcopyBytes (string, (tempBuffer + offset), strLen);
#else
    bcopyBytes (string, tempBuffer, strLen);
#endif  /* FLASH_NO_OVERLAY */

    if (flashErase (FLASH_ADRS - FLASH_BASE) == ERROR) /* erase device */
        return (ERROR);

#ifndef FLASH_NO_OVERLAY                                /* program device */
    if (flashWrite ((FLASH_TYPE*)tempBuffer, flashSize,
                    (FLASH_ADRS - FLASH_BASE), 0) == ERROR)
#else   /* FLASH_NO_OVERLAY */
    if (flashWrite ((FLASH_TYPE*)tempBuffer, strLen,
                    (FLASH_ADRS - FLASH_BASE) + offset, 0) == ERROR)
#endif  /* FLASH_NO_OVERLAY */
        return (ERROR);

    return (OK);
    }

#endif /* ifdef FLASH_SIZE */
