/* FlashUtilMgr.c - User Interface for utility for reprogramming flash */

/* Copyright 1999 Intel Corp. */

/*
modification history
--------------------
01a,08Jul99,jdg  created
*/

/*
DESCRIPTION
This file contains a semi-generic user interface for reprogramming flash.
It depends on flashUtil.c to provide the actual routines for manipulating
the flash. It must be #included in a preamble file that defines certain
types and constants. In particular, it needs:

#define MAKE_PAGES_WRITABLE(start_addr, end_addr) ...
    Routine to make the pages between start_addr and (end_addr-1) writable.
    If these pages are already writable, then it can be made empty.

#define SERPORTNAME "string"
    Name to be passed to fopen to get a file pointer to the serial port used
    to interact with the user.

#define FLASH_TOTAL_SIZE const
    Size in bytes of Flash part.

#define FLASH_WIDTH const
    Width in bytes of Flash part. Valid values are 1, 2, or 4.

typedef UINT8, UINT16, UINT32
    Types evaluating to 8-, 16-, and 32-bit unsigned ints

typedef STATUS
#define OK, ERROR
    Type defining return type of functions. It can return OK or ERROR. This 
    may be defined as an enumeration.

#define FLASH_BASE addr
    Address of start of Flash

#define FLASH_ADRS addr
    Optional address of block containing NVRAM parameters that should
    not be overwritten.

#define ALT_FLASH_BASE addr
    Address of alternate start of Flash, when the flash is connected to
    CE1 rather than CE0.

*/


#include <stdio.h>
#include "flashUtil.h"
#include "flashUtil.c"



UINT8 *fuBuf = NULL;

#ifdef FLASH_ADRS
#define NVRAM_OFFSET (FLASH_ADRS - FLASH_BASE)
#endif

#define PAGESIZE (4 * 1024)

FILE* fuFp = NULL;

extern UINT32 flashBase;

#define FILE_GRAN 0x10000

/*******************************************************************************
*
* fuFRead - Flash Utility: Read file into buffer
*
* This routine reads a file into a buffer while printing a status
* report on fuFp.
*
* RETURNS: N/A
*
* NOMANUAL
*/

static UINT32 fuFRead(UINT8 *buf, UINT32 size, FILE *fp)
    {
    UINT8 *lbuf;
    UINT32 read, count = 0;

    fprintf(fuFp,"Reading file:\n");
    for (lbuf = buf; lbuf < buf + size; )
        {
        fprintf(fuFp,"\r%08X bytes",count);
        fflush(fuFp);
        read = fread(lbuf, sizeof(char), FILE_GRAN, fp);
        count += read;
        if (feof(fp) || ferror(fp)) 
            break;
        lbuf = buf + count;
        }
    fprintf(fuFp,"\r%08X bytes ... done\n",count);
    return count;
    }

/*******************************************************************************
*
* fuReadFile - Flash Utility: Read a file into a buffer
*
* This routine allocates a buffer and reads a file into it.
*
* RETURNS: malloc'ed buffer or NULL if error
*
* NOMANUAL
*/

static UINT8* fuReadFile(char *name, UINT32 *size)
    {
    UINT8 *buf;
    FILE *fp;
    int read;

    buf = malloc(FLASH_TOTAL_SIZE);
    if (buf == NULL)
        {
        fprintf(fuFp,"Can't malloc buffer.\n");
        return NULL;
        }
    fprintf(fuFp,"Opening file '%s' ... ", name);
    fflush(fuFp);
    fp = fopen(name,"rb");
    if (fp == NULL)
        {
        fprintf(fuFp,"Can't open file: '%s'\n",name);
        free(buf);
        return NULL;
        }
    else
        {
        fprintf(fuFp,"done\n");
        }
    bfill(buf, FLASH_TOTAL_SIZE, 0);
    
    read = fuFRead(buf, FLASH_TOTAL_SIZE, fp);
    if (ferror(fp))
        {
        fprintf(fuFp,"Error reading file\n");
        free(buf);
        return NULL;
        }
    if (EOF != getc(fp))
        {
        fprintf(fuFp,"Not at EOF?!!!!\n");
        free(buf);
        return NULL;
        }
    fclose(fp);
    if (size != NULL)
        *size = read;
    return buf;
    }

/*******************************************************************************
*
* fuSanityCheck - Flash Utility: Try to sanity check a buffer
*
* This routine tries to sanity check a buffer to see if it contains
* a viable ARM raw binary image. 
*
* RETURNS: OK or ERROR
*
* NOMANUAL
*/

static STATUS fuSanityCheck(UINT8* buf)
    {
    /* Not too many sanity checks we can make */

    if (((*((UINT32*)buf) & 0xFF000000) != 0xEA000000) &&
        ((*((UINT32*)buf) & 0xFF00F000) != 0xE500F000))
        {
        fprintf(fuFp,"First instruction is not a branch instruction.\n");
        return ERROR;
        }

    return OK;
}

/*******************************************************************************
*
* fuIsWhite - Flash Utility: Check a buffer to see if it contains white space
*
* This routine checks a buffer to see if it contains only spaces or tabs
* up to the end of the buffer or the first line-feed.
*
* RETURNS: TRUE if white, FALSE otherwise
*
* NOMANUAL
*/

static BOOL fuIsWhite(char *buf)
    {
    while(1)
        {
        if ((*buf == '\0') || (*buf == '\n'))
            return TRUE;
        if ((*buf != ' ') && (*buf != '\t'))
            return FALSE;
        buf++;
        }
    }

/*******************************************************************************
*
* fuPromptHex - Flash Utility: prompt for a hex value
*
* This routine displays a prompt with a default value, and then reads a
* response. If the response is not white (blank), it parses the response
* as a hex value and updates the default.
*
* If a period is entered, then the value is not changed, and an "exit"
* flag is returned to the caller.
*
* RETURNS: ERROR if exit value specified (period), OK otherwise
*
* NOMANUAL
*/

static STATUS fuPromptHex(char *prompt, UINT32 *val)
    {
    char buf[160];
    UINT32 newval;

    while (1)
        {
        fprintf(fuFp,"%s (%06X): ",prompt,*val);
        fgets(buf,160,fuFp);
        if ((buf[0] == '.') && (buf[1] == '\n'))
            {
            return ERROR;
            }
        else if (fuIsWhite(buf))
            {
            return OK;
            }
        else if (1 == sscanf(buf, "%X", &newval))
            {
            *val = newval;
            return OK;
            }
        else
            {
            fprintf(fuFp,"Invalid hex value entered: '%s'.\n",buf);
            }
        }
    }

/*******************************************************************************
*
* fuPromptString - Flash Utility: prompt for a string value
*
* This routine displays a prompt with a default value, and then reads a
* response. If the response is not white (blank), it parses the response
* as a string value and updates the default.
*
* If a period is entered, then the value is not changed, and an "exit"
* flag is returned to the caller.
*
* RETURNS: ERROR if exit value specified (period), OK otherwise
*
* NOMANUAL
*/

static STATUS fuPromptString(char *prompt, char* *val)
    {
    static char buf[160];
    char buf2[160];

    fprintf(fuFp,"%s (%s): ",prompt,*val);
    fgets(buf2,160,fuFp);
    if ((buf2[0] == '.') && (buf2[1] == '\n'))
        {
        return ERROR;
        }
    else if (fuIsWhite(buf2))
        {
        return OK;
        }
    else
        {
        sscanf(buf2,"%s",buf);
        *val = buf;
        return OK;
        }
    }

/*******************************************************************************
*
* freeFlashBuf - Free the buffer allocated by updateFlash in case we crash
*
* This routine will free fuBuf if necessary. This can be used in case
* updateFlash gets an error and leaves the buffer allocated.
* If the buffer was alread freed, then nothing is done.
*/

void freeFlashBuf()
    {
    if (NULL != fuBuf)
        {
        free(fuBuf);
        fuBuf = NULL;
        }
    }       

/*******************************************************************************
*
* programFlash - Program the flash from a buffer
*
* This routine updates the flash with the contents of a buffer, optionally
* skipping the block containing the NVRam value. It displays a status
* update during the programming process.
*
* RETURNS: ERROR or OK
*/

STATUS programFlash(UINT8* buf,
                           UINT32 flashStart, UINT32 flashEnd,
                           BOOL skipNVRam, BOOL skipFUtil)
    {
    UINT32 NVOffset;
    UINT32 FUtil_start, FUtil_end;
    UINT32 addr, next_addr;

#ifdef FLASH_ADRS
    if (skipNVRam)
        NVOffset = NVRAM_OFFSET;
    else
#endif
        NVOffset = (UINT32)(-1);

#ifdef FUTIL_ADRS
    if (skipFUtil) {
        FUtil_start = FUTIL_ADRS;
        FUtil_end   = FUTIL_ADRS + FUTIL_LEN;
    }
    else
#endif
        FUtil_start = FUtil_end = (UINT32)(-1);

    fprintf(fuFp,"Progress:\n");
    for (addr = flashStart; addr < flashEnd; addr = next_addr)
        {
        fprintf(fuFp,"\r%3d%%  ",(100 * addr) / (flashEnd - flashStart));
        fflush(fuFp);
        next_addr = addr;
        flashNextBlock(&next_addr);
        if (addr == NVOffset)
            continue;
        if ((addr >= FUtil_start) && (addr < FUtil_end))
            continue;
        if (OK != flashErase(addr))
            {
            fprintf(fuFp,"Error erasing flash block %X\n",addr);
            return ERROR;
            }

        if (OK != flashWrite(CAST_FLASH (buf + addr - flashStart),
                             next_addr - addr,
                             addr,
                             0))
            {
            fprintf(fuFp,"Error writing flash block %X\n",addr);
            return ERROR;
            }
        if (0 != bcmp((char*)(buf + addr - flashStart), 
                      (char*)(flashBase + addr), 
                      next_addr - addr))
            {
            fprintf(fuFp,"Error: data miscompares after write, block = %X.\n",addr);
            /* return ERROR; */
            }
        } /* end for addr */
    fprintf(fuFp,"\r100%% ... done\n");

    return OK;
}

/*******************************************************************************
*
* updateFlash - Enter a dialog with the user to update the flash
*
* This routine enters into a dialog with the user over the serial port
* to get enough information to update the flash image. It is assumed that
* the routine is executing out of RAM.
*
* It will mark the flash pages as writable, prompt the user (over the serial
* port) for necessary information, read the specified file, sanity check
* it, and then update the flash image.
*
* RETURNS: ERROR or OK
*/

STATUS updateFlash()
    {
    UINT32 size;
    UINT32 flashStart, flashEnd, fileOffset;
    UINT32 nextBlock1, nextBlock2;
    BOOL skipNVRam, skipFUtil;
    char *fileName = "";
    char *answer;
    STATUS status = OK;

    flashStart = 0;
    flashEnd = FLASH_TOTAL_SIZE;
    fileOffset = 0;

    fuFp = fopen(SERPORTNAME, "r+");
    if (fuFp == NULL)
        {
        fprintf(stderr,"Can't open serial port: '%s'\n",SERPORTNAME);
        return ERROR;
        }

    fprintf(stderr,
            "Input and Output are being done on the serial port.\n"
            "Switch to the serial port to continue operation.\n");

    /* Make sure all of flash is writable and valid */
    MAKE_PAGES_WRITABLE(flashBase, flashBase + FLASH_TOTAL_SIZE);
    
    /* check for errors in flash utils */
    nextBlock1 = 0;
    if (OK != flashNextBlock(&nextBlock1))
        {
        fprintf(fuFp,"Error in flashNextBlock\n");
        return ERROR;
        }

    fprintf(fuFp,
            "Note that reprogramming the flash is a risky endeavor.\n"
            "If the machine crashes halfway through, or if you load an\n"
            "invalid image, then the machine will not be bootable.\n"
            "In that event, you will need to remove the flash parts and\n"
            "reprogram them externally.\n\n"
            "Note also that all of the numeric parameters are considered\n"
            "hexadecimal, *without* the leading '0x'.\n\n"
            "Enter a period ('.') to abort this function.\n"
            "Enter a newline to accept the default.\n\n"
        );

    while(1)
        {
        if (OK != fuPromptString("Filename", &fileName))
            goto abort;
        fuBuf = fuReadFile(fileName, &size);
        if (fuBuf == NULL)
            continue;
        if (fuSanityCheck(fuBuf) != OK)
            {
            answer = "No";
            if (OK != fuPromptString("This does not appear to be a valid ARM "
                                     "file, do you want to continue",
                                     &answer))
                goto abort;;
            if ((*answer != 'y') && (*answer != 'Y'))
                {
                free(fuBuf);
                continue;
                }
            }
        while (1)
            {
            if (OK != fuPromptHex("Starting offset in flash", &flashStart))
                goto abort;
            if (flashStart != 0)
                {
                nextBlock1 = flashStart;
                nextBlock2 = flashStart-1;
                flashNextBlock(&nextBlock1);
                flashNextBlock(&nextBlock2);
                if (nextBlock1 == nextBlock2)
                    {
                    fprintf(fuFp,
                            "Starting offset %X does not begin on a flash block boundary.\n",
                            flashStart);
                    continue;
                    }
                }
            if (OK != fuPromptHex("Ending offset in flash  ", &flashEnd))
                goto abort;
            nextBlock1 = flashEnd;
            nextBlock2 = flashEnd-1;
            flashNextBlock(&nextBlock1);
            flashNextBlock(&nextBlock2);
            if (nextBlock1 == nextBlock2)
                {
                fprintf(fuFp,
                        "Ending offset %X does not begin on a flash block boundary.\n",
                        flashEnd);
                continue;
                }
            if (OK != fuPromptHex("Starting offset in file ", &fileOffset))
                goto abort;
            if (fileOffset & 3)
                {
                fprintf(fuFp,"File offset %X is not longword aligned.\n",
                        fileOffset);
                continue;
                }

            if (fileOffset + (flashEnd - flashStart) > FLASH_TOTAL_SIZE)
                {
                fprintf(fuFp,"Area specified extends beyond end of buffer.\n");
                continue;
                }

#ifdef FUTIL_ADRS
            if ((flashStart < FUTIL_ADRS + FUTIL_LEN) &&
                (flashEnd > FUTIL_ADRS))
                {
                answer = "Yes";
                if (OK != fuPromptString("Do you wish to skip the FUtil block",
                                         &answer))
                    goto abort;
                skipFUtil = ((*answer == 'y') || (*answer == 'Y'));
                }
            else
#endif
                skipFUtil = FALSE;

#ifdef FLASH_ADRS
            if ((flashStart <= NVRAM_OFFSET) &&
                (flashEnd > NVRAM_OFFSET))
                {
                answer = "Yes";
                if (OK != fuPromptString("Do you wish to skip the NVRAM block",
                                         &answer))
                    goto abort;
                skipNVRam = ((*answer == 'y') || (*answer == 'Y'));
                }
            else
#endif
                skipNVRam = FALSE;

            if (programFlash(fuBuf + fileOffset, flashStart, flashEnd, skipNVRam,
                             skipFUtil)
                == ERROR)
                {
                status = ERROR;
                break;
                }

            if ((flashStart == 0) && (flashEnd == FLASH_TOTAL_SIZE))
                break;

            answer = "No";
            if (OK != fuPromptString("Do you want to program other offsets",
                                     &answer))
                goto abort;
            if ((*answer == 'y') || (*answer == 'Y'))
                continue;

            break;
            } /* while (1): loop for programming offsets */
        break;
        } /* while (1): main loop */
abort:
    if (fuBuf != NULL)
        {
        free(fuBuf);
        fuBuf = NULL;
        }
    if (fuFp != NULL)
        {
        fclose(fuFp);
        fuFp = NULL;
        }
    return status;
    }
