/* ixp1200IntrCtl.c - Intel IXP1200 interrupt controller driver */

/* Copyright 1984-1997, Wind River Systems, Inc.; Copyright 1999 Intel Corp. */

/*
modification history
--------------------
01c,07Mar00,jdg  Added support for B0 revision ixp1200
01b,13aug99,jdg  changed name from vbsa2100 to ixp1200eb
01a,28apr99,jdg  created from 01a of sa1100IntrCtl.c.
*/

/*
This module implements the Intel IXP1200 interrupt controller driver.

The IXP1200 does not have a true interrupt controller, so the interrupt
processing must be in non-preemptive mode. Effectively there is only one 
interrupt level, but there are 32 interrupt vectors. These are managed
by the intArchLib.

The number of interrupts supported by the software is IXP1200_INT_NUM_VECTORS.

The actual registers referenced are:
  IXP1200_IRQ
  IXP1200_IRQSTATUS
We assume that config.h or <bsp>.h has defined the above addresses.

This driver assumes that the chip is memory-mapped and does direct
memory accesses to the registers which are assumed to be 32 bits wide.
If a different access method is needed, the BSP can redefine the macros
IXP1200_INT_REG_READ(addr,result) and IXP1200_INT_REG_WRITE(addr,data).

The BSP will initialize this driver in sysHwInit2(), after initializing
the main interrupt library, usually intLibInit().  The initialization
routine, ixp1200IntDevInit() will setup the interrupt service.

All of the functions in this library are global.  This allows them to
be used by the BSP if it is necessary to create wrapper routines or to
incorporate several drivers together as one.
*/

#include "vxWorks.h"
#include "config.h"
#include "intLib.h"


/* hardware access methods */

#ifndef IXP1200_REG_READ
#define IXP1200_REG_READ(a,val) ((val) = *(volatile UINT32 *)(a))
#endif	/*IXP1200_REG_READ*/

#ifndef IXP1200_REG_WRITE
#define IXP1200_REG_WRITE(a,val) (*(volatile UINT32 *)(a) = (val))
#endif /*IXP1200_REG_WRITE*/


/* driver constants */

/* lsbTab[256] returns the index of the lsb which is 1 */
LOCAL const
unsigned char lsbTab[0x100] =
    {
    255,0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    5,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    6,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    5,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    7,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    5,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    6,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    5,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0,
    4,  0,  1,  0,  2,  0,  1,  0,
    3,  0,  1,  0,  2,  0,  1,  0
    };    

/* Local data */

/* forward declarations */

STATUS	ixp1200IntLvlVecChk (int*, int*);
STATUS  ixp1200IntLvlVecAck (int, int);
int	ixp1200IntLvlChg (int);
STATUS	ixp1200IntLvlEnable (int);
STATUS	ixp1200IntLvlDisable (int);


/*******************************************************************************
*
* ixp1200IntDevInit - initialize the interrupt controller
*
* This routine will initialize the interrupt controller device,
* disabling all interrupt sources.  It will also connect the device
* driver specific routines into the architecture level hooks.  If the BSP
* needs to create a wrapper routine around any of the arhitecture level
* routines, it should install the pointer to the wrapper routine after
* calling this routine.  
* 
* RETURNS: N/A
*/

void ixp1200IntDevInit (void)
    {
    /* install the driver routines in the architecture hooks */

    sysIntLvlVecChkRtn	= ixp1200IntLvlVecChk;
    sysIntLvlVecAckRtn	= ixp1200IntLvlVecAck;
    sysIntLvlChgRtn	= ixp1200IntLvlChg;
    sysIntLvlEnableRtn	= ixp1200IntLvlEnable;
    sysIntLvlDisableRtn	= ixp1200IntLvlDisable;

    }


/*******************************************************************************
*
* ixp1200IntLvlVecChk - check for and return any pending interrupts
*
* This routine interrogates the hardware to determine the highest priority
* interrupt pending.  It returns the vector associated with that interrupt, and
* also the interrupt priority level prior to the interrupt (not the
* level of the interrupt).  Since there is no notion of interrupt level,
* this is a constant.
*
* The return value ERROR indicates that no pending interrupt was found and
* that the level and vector values were not returned.
*
* RETURNS: OK or ERROR if no interrupt is pending.
*/

STATUS  ixp1200IntLvlVecChk
    (
    int* pLevel,  /* ptr to receive old interrupt level */
    int* pVector  /* ptr to receive current interrupt vector */
    )
    {
    UINT32 isr, isr2;
    int vector;

    /* Read IXP1200_IRQ register */

    IXP1200_REG_READ(IXP1200_IRQ, isr);

    /* If no interrupt is pending, return ERROR */

    if (isr == 0)
	return ERROR;

    if (isr & (1 << 2))
        {

        /* if PCI bit is set */

        IXP1200_REG_READ(IXP1200_IRQSTATUS, isr2);
        isr2 |= (isr & ~(1<<2)) << 6;

        if ((vector = (isr2 & 0xFF))) 
            {
            vector = lsbTab[vector];
            }
        else if ((vector = ((isr2 >> 8) & 0xFF)))
            {
            vector = lsbTab[vector] + 8;
            }
        else if ((vector = ((isr2 >> 16) & 0xFF)))
            {
            vector = lsbTab[vector] + 16;
            }
        else if ((vector = ((isr2 >> 24) & 0xFF)))
            {
            vector = lsbTab[vector] + 24;
            }
        else
            {
            return ERROR;
            }
        }
    else
        {

        /* PCI bit is not set */

        vector = lsbTab[isr>>2] + 8;
        }

    /* fix up RTC/UART for Rev A chips */
    if (sysCpuRev < 1) {
        if (vector == INT_VEC_RTC)
            vector = INT_VEC_UART;
    }

    *pVector = vector;
    *pLevel = 0;

    return OK;
    }

/*******************************************************************************
*
* ixp1200IntLvlVecAck - acknowledge the current interrupt
*
* This is a no-op
*
* Since ixp1200IntLvlVecChk doesn't change the state of the system,
* there is nothing here that needs to be undone.
*
* It is assumed that the interrupt service routine has cleared the 
* interrupt source itself.
*
* RETURNS: OK or ERROR if a hardware fault is detected.
* ARGSUSED
*/

STATUS  ixp1200IntLvlVecAck
    (
    int level,	/* old interrupt level to be restored */
    int vector	/* current interrupt vector, if needed */
    )
    {
    return OK;
    }


/*******************************************************************************
*
* ixp1200IntLvlChg - change the interrupt level value
*
* This is a no-op since we effectively have one level.
*
* RETURNS: Previous interrupt level.
*/

int  ixp1200IntLvlChg
    (
    int level	/* new interrupt level */
    )
    {
    return 0;
    }

/*******************************************************************************
*
* ixp1200IntLvlEnable - enable a single interrupt level
*
* Enable a single interrupt level. This only works on the PCI interrupts.
*
* RETURNS: OK or ERROR if the specified level cannot be enabled.
*/

STATUS  ixp1200IntLvlEnable
    (
    int level  /* level to be enabled */
    )
    {
    if (level < 0 ||
	level > IXP1200_INT_NUM_VECTORS)
	return ERROR;

    level = 1 << level;
    if ((level & IXP1200_INT_CTRL_MASK) == 0)
        return ERROR;

    IXP1200_REG_WRITE(IXP1200_IRQENABLE_SET, level);

    return OK;
    }

/*******************************************************************************
*
* ixp1200IntLvlDisable - disable a single interrupt level
*
* Disable a single interrupt level. This only works on the PCI interrupts
*
* RETURNS: OK or ERROR, if the specified interrupt level cannot be disabled.
*/

STATUS  ixp1200IntLvlDisable
    (
    int level  /* level to be disabled */
    )
    {
    if (level < 0 ||
	level > IXP1200_INT_NUM_VECTORS)
	return ERROR;

    level = 1 << level;
    if ((level & IXP1200_INT_CTRL_MASK) == 0)
        return ERROR;

    IXP1200_REG_WRITE(IXP1200_IRQENABLECLEAR, level);

    return OK;
    }
