/* ixp1200Pci.c - Intel IXP1200 PCI Unit driver */

/* Copyright 1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01d,06Oct00,jdg  Added Big-Endian support
01c,14dec00,jdg  added automatic pci address configuration
01b,13aug99,jdg  changed name from vbsa1200 to ixp1200eb
01a,25may99,jdg  created from ebsa sysLib.c, version 01j
*/

/*
DESCRIPTION
This device driver code for the Intel IXP1200 PCI unit.

It the function sysPciInit(), which should be called from
sysHwInit() to initialize the PCI unit on the IXP1200.

If we are the central function, it sets up PCI space as:
 I/O Space:
  0000 0000: Ethernet card
  0000 F000: IXP1200 CSRs (128 bytes)
 Mem Space:
  0000 0000: IXP1200 SDRAM (16 Meg)
  4000 0000: IXP1200 CSRs  (128 bytes)

It also contains the function sysPciAssignAddrs() which will 
(if we are central function) examine PCI space and program ourself
and all other PCI devices dynamically, based on how much addresses
space each device needs.

*/

#include "vxWorks.h"
#include "config.h"
#include "ixp1200Pci.h"
#ifdef AUTO_PCI_CONFIG
#ifdef INCLUDE_AUTO_PCI_CONFIG_SHOW_ROUTINE
#include <stdio.h>
#endif
#endif
#include "taskLib.h"

/* local defines  */

#ifdef AUTO_PCI_CONFIG
#define MAX_BARS 30
#define MAX_DEVICES 6

#ifndef INITIAL_PCI_CMD
#define INITIAL_PCI_CMD PCI_CMD_IO_ENABLE
#endif
#endif /* ifdef AUTO_PCI_CONFIG */

/* locals */

#ifdef AUTO_PCI_CONFIG

static PciDevice devices[MAX_DEVICES];
static int nDevices;

#endif /* ifdef AUTO_PCI_CONFIG */


/* function prototypes */

void sysPciAssignAddrs(void);

/* driver functions */

/******************************************************************************
*
* sysPciIsCentralFunc - returns a boolean indicating whether we are central func
*
* This function returns whether we are central function or not
*
* RETURNS: true if central function
*/

LOCAL int sysPciIsCentralFunc()
{
    int central_func;
#ifdef SPECTACLE_ISLAND
    central_func = 1;
#else
    central_func =
        IXP1200_SA_CONTROL_PCF & *(UINT32 *)IXP1200_SA_CONTROL;
#endif /* SPECTACLE_ISLAND */
    return central_func;
}


/*******************************************************************************
*
* sysPciInit - initialise PCI bridge controller
*
* This function initialises the PCI controlling parts.
*
* In particular, it will:
*   Reset the PCI bus
*   Reset the Doorbell registers
*   Set the PCI address Extension register to 0x00000000
*   Unreset the PCI bus
*   Set the Base Address Mask Register based on LOCAL_MEM_SIZE_ACTUAL
*   Set the CSR Memory BAR to 0x40000000 (IXP1200 CSRs in PCI Mem space)
*   Set the CSR BAR        to 0x0000F000 (IXP1200 CSRs in PCI I/O space)
*   Set the SDRAM BAR      to 0x00000000 (IXP1200 SDRAM in PCI Mem space)
*   Allow the PCI unit to be bus master, respond to I/O and Mem space,
*     and to generate Master Write and Invalidate transactions.
*   Allow the PCI unit to respond to PCI configuration cycles
* 
* RETURNS: N/A
*/

void sysPciInit (void)
    {
    volatile int i, j;
    int central_func;

    /* assert PCI reset */

    *(UINT32 *)(IXP1200_SA_CONTROL) =
	    (*(UINT32 *)(IXP1200_SA_CONTROL)) & ~IXP1200_SA_CONTROL_PNR;

    /* wait a few usecs to settle the device and PCI bus (no timers yet) */

    for (i = 0; i < 50 ; i++)
        j = i + 1;


    /* Disable Outbound & Doorbell Ints in Outbound Interrupt Mask Register*/

    *(UINT32 *)(IXP1200_OUTB_INT_MASK) = 0xC;


    /* Disable doorbell interrupts to PCI in Doorbell PCI Mask Register */

    *(UINT32 *)(IXP1200_DBELL_PCI_MASK) = 0;


    /* Disable doorbell interrupts to SA-110 in Doorbell SA-110 Mask Reg */

    *(UINT32 *)(IXP1200_DBELL_SA_MASK) = 0;


    /* Set PCI Address Extension Register to known state */

    *(UINT32 *)(IXP1200_PCI_ADDR_EXT) = 0x00000000;


    /* Set Int pin */
    *(UINT32 *)(IXP1200_PCI_INT_LAT) = 1 << 8;

    /* Negate PCI reset (has no effect if we are not central function) */

    *(UINT32 *)(IXP1200_SA_CONTROL) =
	    (*(UINT32 *)(IXP1200_SA_CONTROL)) | IXP1200_SA_CONTROL_PNR;

    /* wait a few usecs to settle the device and PCI bus (no timers yet) */

    for (i = 0; i < 50 ; i++)
        j = i + 1;


    /* Setup SDRAM Base Address Mask Reg: LOCAL_MEM_SIZE_ACTUAL window (32Meg)
       from PCI into SDRAM */

    *(UINT32 *)(IXP1200_DRAM_BASE_ADDR_MASK) = LOCAL_MEM_SIZE_ACTUAL - 0x00040000; 


    /* Set up CSR Base Address Mask Reg. 
     * Due to a Pass-A HW bug, we need to ask for at least 1 Meg. */

    *(UINT32 *)(IXP1200_CSR_BASE_ADDR_MASK) = 0x000c0000;

    /* Stop IXP1200 from responding to any PCI transactions */

    sysWrite16 ((volatile UINT16 *)(IXP1200_PCI_COMMAND), 0);


    /* Check to see if we are central function, if not then do not
     * set up BARs */

#ifdef SPECTACLE_ISLAND
    central_func = 1;
#else
    central_func =
        IXP1200_SA_CONTROL_PCF & *(UINT32 *)IXP1200_SA_CONTROL;
#endif /* SPECTACLE_ISLAND */

    if (central_func)
        {

        /* Set CSR Memory Base Address Register */

        *(UINT32 *)(IXP1200_PCI_MEM_BAR) = 0x40000000;


        /*
         * Set CSR I/O Base Address Register: set base address of CSRs from PCI,
         * CSR Base Address Mask Register will setup 128 byte window.
         */

        *(UINT32 *)(IXP1200_PCI_IO_BAR) = 0x0000F000;


        /* Setup SDRAM Base Address Register: */

        *(UINT32 *)(IXP1200_PCI_DRAM_BAR) = 0x00000000;


        /*
         * Enable IXP1200 as Bus master and respond to I/O space and memory space
         * transactions as target
         */

        sysWrite16 ((volatile UINT16 *)(IXP1200_PCI_COMMAND),
                    IXP1200_PCI_COMMAND_IOSE |  /* I/O Space Enable */
                    IXP1200_PCI_COMMAND_MSE  |  /* Mem Space Enable */
                    IXP1200_PCI_COMMAND_ME   |  /* Master Enable */
                    IXP1200_PCI_COMMAND_MWI     /* Master Wr & Invalidate En */
            );

        }

    /*
     * Compute Big Endian PCI setting
     */

#if (_BYTE_ORDER == _BIG_ENDIAN)
#define BE_PCI (IXP1200_SA_CONTROL_BEO | \
                IXP1200_SA_CONTROL_DEO | \
                IXP1200_SA_CONTROL_BEI | \
                IXP1200_SA_CONTROL_DEI )
#else
#define BE_PCI 0
#endif

    /*
     * Set Initialise Complete in PCI Control Register: allow IXP1200 to
     * respond to PCI configuration cycles.
     */

    *(UINT32 *)(IXP1200_SA_CONTROL) =
	    (*(UINT32 *)(IXP1200_SA_CONTROL)) | (0x00000001 | BE_PCI);

#if defined(INCLUDE_PCI) && defined(INCLUDE_HPC)
    /*
     * We need to do this here because some BIOS's are too quick for
     * the spawned thread to correct things. So we'll do it for a short
     * while here, and then spawn a task later to fix later ones.
     */
    if ((sysCpuRev == 1) && (!central_func)) {
        int i,j;

        /* Fix B0 bug where some BIOS will clobber the INT_LAT reg */
        DEBUG_LED(0x77);
        for (i=0; i<5000000; i++) {
            j = *(volatile UINT32*)(IXP1200_PCI_INT_LAT);
            if ((j & ~0xFF) != (1<<8)) {
                *(volatile UINT32*)(IXP1200_PCI_INT_LAT) =
                    (1<<8) | (j & 0xFF);
            }
        }
        DEBUG_LED(0x88);
    }
#endif

    return;
    }


/*******************************************************************************
*
* sysPciFix1 - Fix PCI config registers due to B0 chip bug
*
* Some BIOS's will clobber parts of the PCI INT_LAT register that should
* have been read only. So we poll for a while, and if it gets clobbered,
* we fix it.
*
* RETURNS: N/A
*/

void sysPciFix1 (void)
{
    int i,j;

    /* Fix B0 bug where some BIOS will clobber the INT_LAT reg */
    for (i=0; i<60*60; i++) {
        j = *(volatile UINT32*)(IXP1200_PCI_INT_LAT);
        if ((j & ~0xFF) != (1<<8)) {
            *(volatile UINT32*)(IXP1200_PCI_INT_LAT) =
                (1<<8) | (j & 0xFF);
        }
        taskDelay(1);
    }
}

/*******************************************************************************
*
* sysPciFix - spawn a task to fix PCI config registers due to B0 chip bug
*
* Some BIOS's will clobber parts of the PCI INT_LAT register that should
* have been read only. So we poll for a while, and if it gets clobbered,
* we fix it. This is done in a separate task so that we don't hold up the boot.
*
* RETURNS: N/A
*/

void sysPciFix (void)
{
#ifndef SPECTACLE_ISLAND
    int central_func;

    central_func =
        IXP1200_SA_CONTROL_PCF & *(UINT32 *)IXP1200_SA_CONTROL;

    if (!central_func)
        {
        taskSpawn ("tPciFix",
                   5,    /* priority, 255==lowest */
                   0,    /* options */
                   1024, /* stack size (bytes) */
                   (FUNCPTR)sysPciFix1,
                   0,0,0,0,0,0,0,0,0,0);
        }
#endif
}

/******************************************************************************
*
* sysPciToPhys - translate a Pci address to a physical address
*
* This function converts a Pci address to a physical address. This only works
* for PCI Memory space.
*
* RETURNS: the physical adddress
*/

void * sysPciToPhys
    (
    void *	PciAddr	/* PCI address */
    )
    {
    if (((UINT32) PciAddr) < LOCAL_MEM_SIZE_ACTUAL) 
        { 
        /* SDRAM address */
        return (void*) (((UINT32) PciAddr) + SDRAM_PHYS_BASE);
        } else {
        /* Other Memspace address */
        return (void*) (((UINT32) PciAddr) + PCI_MEM_BASE);
        }
    }

/******************************************************************************
*
* sysPhysToPci - translate a physical address to a Pci address
*
* This function converts a physical address to a Pci address. This only works
* for PCI Memory space.
*
* RETURNS: the Pci adddress
*/

void * sysPhysToPci
    (
    void *	PhysAddr	/* Physical address */
    )
    {
    if ((((UINT32) PhysAddr) >= SDRAM_VIRT_OFFSET) &&
        (((UINT32) PhysAddr) <  SDRAM_PF_BASE))
        {
        /* SDRAM address */
        return (void*) (((UINT32) PhysAddr) - SDRAM_PHYS_BASE);
        } else {
        /* Other Memspace address */
        return (void*) (((UINT32) PhysAddr) - PCI_MEM_BASE);
        }
    }

#ifdef AUTO_PCI_CONFIG

/******************************************************************************
*
* sortBars - Sort Bars largest to smallest and assign addresses
*
* This function takes in an array of pointers to BARs and a starting
* PCI address, then it will sort the BARS from largest to smallest. It will
* then align the address appropriately for the first (largest) BAR, and then
* assign increasing addresses with no gaps.
*
* RETURNS: N/A
*/

static void
sortBars
    (
    PciBar *Bars[],    /* Array of pointers to BARs */
    const int nBars,   /* Number of BARs in array */
    UINT32 startAddr   /* PCI address at which to start */
    )
{
    int i, j;
    PciBar *tmp;

    /* Sort biggest to smallest */
    for (i = 0; i < nBars-1; i++) {
        for (j = i+1; j < nBars; j++) {
            if (Bars[j]->size > Bars[i]->size) {
                /* swap them */
                tmp = Bars[i];
                Bars[i] = Bars[j];
                Bars[j] = tmp;
            }
        } /* end for j */
    } /* end for i */

    /* Assign Addresses; first align startAddr */

    startAddr = (startAddr + Bars[0]->size - 1) & ~(Bars[0]->size - 1);

    for (i=0; i < nBars; i++) {
        Bars[i]->address |= startAddr;
        startAddr += Bars[i]->size;
    } /* end for i */
}

/******************************************************************************
*
* addPageEntries - Call sysMmuMapAdd to add page entries for specified range
*
* This function takes a range of addresses containing a gap, and then calls
* sysMmuMapAdd to add page table entries for the one or two memory ranges
* specified.
*
* RETURNS: N/A
*/

static void
addPageEntries
    (
    UINT32 start,      /* Starting PCI address of range*/
    UINT32 end,        /* Ending PCI address of range */
    UINT32 startSkip,  /* Starting PCI address of hole */
    UINT32 skipLen,    /* Ending PCI address of hole */
    UINT32 offset      /* Offset from PCI address to virtual address */
    )
{
    UINT32 start1, end1;

    start1 = start;
    end1 = startSkip;
    if (start1 != end1) {
        start1 = ROUND_DOWN(start1, PAGE_SIZE);
        end1 = ROUND_UP(end1, PAGE_SIZE);
        sysMmuMapAdd((void*)(offset + start1), end1-start1,
                     VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
                     VM_STATE_VALID | VM_STATE_WRITABLE     | VM_STATE_CACHEABLE_NOT);
    }
    start1 = startSkip + skipLen;
    end1 = end;
    if (start1 != end1) {
        start1 = ROUND_DOWN(start1, PAGE_SIZE);
        end1 = ROUND_UP(end1, PAGE_SIZE);
        sysMmuMapAdd((void*)(offset + start1), end1-start1,
                     VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE,
                     VM_STATE_VALID | VM_STATE_WRITABLE     | VM_STATE_CACHEABLE_NOT);
    }
}


/******************************************************************************
*
* sysPciAssignAddrs - Scan PCI bus and automatically assign PCI addresses
*
* This function scans the PCI bus to see what other devices are there. It then
* sorts the BARs that it finds and assigns them addresses. It also adds page
* entries to cover the range of PCI space used. It enables PCI transactions
* as defined by INITIAL_PCI_CMD, which defaults to enabling IO transactions.
*
* As a special case, it assigns its own SDRAM space to PCI address zero.
*
* This funciton only has an effect if this processor is the PCI central function.
*
* RETURNS: N/A
*/

void
sysPciAssignAddrs(void)
{
    PciBar   *memBars[MAX_BARS];
    PciBar   *ioBars[MAX_BARS];

    int nMBars, nIOBars;
    int bus, dev, func, i, j;
    UINT16 data;
    INT32 data32, tmp, size;

#ifndef SPECTACLE_ISLAND
    int central_func;

    central_func =
        IXP1200_SA_CONTROL_PCF & *(UINT32 *)IXP1200_SA_CONTROL;
    if (! central_func)
        return;
#endif

    /* Assign first device to ourselves */
    devices[0].bus = 0;
    devices[0].device = 0;
    devices[0].func = 0;
    devices[0].vendor_id = sysRead16((void*)IXP1200_PCI_VENDOR_ID);
    devices[0].device_id = sysRead16((void*)IXP1200_PCI_DEVICE_ID);
    devices[0].error = FALSE;
    devices[0].bar[CSR_BAR].size = 0x00100000; /* 1 Meg: CSR Mem Base */
    devices[0].bar[CSR_BAR].address = 0;       /* Mem addr */
    devices[0].bar[IO_BAR].size = 0x00000080;  /* 128: CSR I/O Base */
    devices[0].bar[IO_BAR].address = 1;        /* I/O addr */
    devices[0].bar[SD_BAR].size = LOCAL_MEM_SIZE_ACTUAL; /* SDRAM Mem Base */
    devices[0].bar[SD_BAR].address = 0;                  /* Mem addr */
    devices[0].bar[NO_BAR].size = 0;
    nDevices = 1;
    nMBars = nIOBars = 0;
    /* But SDRAM base first */
    memBars[nMBars++] = &devices[0].bar[SD_BAR];
    memBars[nMBars++] = &devices[0].bar[CSR_BAR];
    ioBars[nIOBars++] = &devices[0].bar[IO_BAR];

    func = 0;
    for ( bus = 0; bus < PCI_MAX_BUS; bus++ ) {
        for (dev = 0; dev < PCI_MAX_DEV; dev++ ) {
            pciConfigInWord(bus, dev, func, PCI_CFG_VENDOR_ID, &data);
            if (data == 0xFFFF)
                continue;
            if (nDevices == MAX_DEVICES)
                break;
            devices[nDevices].bus = bus;
            devices[nDevices].device = dev;
            devices[nDevices].func = func;
            devices[nDevices].vendor_id = data;
            pciConfigInWord(bus, dev, func, PCI_CFG_DEVICE_ID, &data);
            devices[nDevices].device_id = data;
            devices[nDevices].error = FALSE;
            for (i=0; i<6; i++) {
                pciConfigOutLong(bus, dev, func,
                                 PCI_CFG_BASE_ADDRESS_0 + (4*i),
                                 0xFFFFFFFF);
                pciConfigInLong(bus, dev, func,
                                PCI_CFG_BASE_ADDRESS_0 + (4*i),
                                &data32);
                if (data32 == 0)
                    break;
                devices[nDevices].bar[i].address = (data32 & 1);
                if (data32 & 1) {
                    /* IO space */
                    tmp = data32 & ~0x00000003;
                    size = ~(tmp-1);
                    devices[nDevices].bar[i].size = size;
                    if (tmp != ~(size-1))
                        devices[nDevices].error = TRUE;
                    else {
                        if (nIOBars < MAX_BARS)
                            ioBars[nIOBars++] = &devices[nDevices].bar[i];
                    }
                } else {
                    /* Mem space */
                    tmp = data32 & ~0x0000000F;
                    size = ~(tmp-1);
                    devices[nDevices].bar[i].size = size;
                    if (tmp != ~(size-1))
                        devices[nDevices].error = TRUE;
                    else {
                        if (nMBars < MAX_BARS)
                            memBars[nMBars++] = &devices[nDevices].bar[i];
                    }
                }
            } /* end for i */
            devices[nDevices].bar[i].size = 0;
            nDevices++;
        } /* end for dev */
    } /* end for bus */

    /* Sort bars, Leave our SDRAM at Mem zero */

    sortBars(memBars+1, nMBars-1, LOCAL_MEM_SIZE_ACTUAL);
    sortBars(ioBars, nIOBars, 0);


    data = PCI_CMD_IO_ENABLE;
    for (i=1; i<nDevices; i++) {
        if (devices[i].error)
            continue;

        for (j=0; devices[i].bar[j].size; j++) {
            pciConfigOutLong(devices[i].bus,
                             devices[i].device,
                             devices[i].func,
                             PCI_CFG_BASE_ADDRESS_0 + (4*j),
                             devices[i].bar[j].address);
        } 
        pciConfigOutLong(devices[i].bus,
                         devices[i].device,
                         devices[i].func,
                         PCI_CFG_DEV_INT_LINE,
                         INT_VEC_PIL);
        pciConfigOutWord(devices[i].bus,
                         devices[i].device,
                         devices[i].func,
                         PCI_CFG_COMMAND, 
                         data);

    } /* end for i */

    /* Add page table entries, skip our SDRAM */
    addPageEntries(memBars[1]->address,        /* start */
                   memBars[nMBars-1]->address + memBars[nMBars-1]->size, /*end*/
                   devices[0].bar[CSR_BAR].address, /* skip start */
                   devices[0].bar[CSR_BAR].size,    /* skip len */
                   PCI_MEM_BASE);
    addPageEntries(ioBars[0]->address,         
                   ioBars[nIOBars-1]->address + ioBars[nIOBars-1]->size,
                   devices[0].bar[IO_BAR].address,
                   devices[0].bar[IO_BAR].size,
                   PCI_IO_BASE);
}

#ifdef INCLUDE_AUTO_PCI_CONFIG_SHOW_ROUTINE

/******************************************************************************
*
* pciAddrShow - Shows PCI address assignments
*
* This function prints the PCI address assignments automatically generated. 
* It prints a table giving for each device the bus number, device number,
* a flag indicating whether there was an error for that device, and the
* vendor/device ID. Then it lists the BARs for that device, giving the
* size of the region and its address. If the low-order bit of the address
* is "1", then the address corresponds to I/O space, otherwise it is in
* memory space.
*
* RETURNS: N/A
*/

void pciAddrShow(void)
{
    int i, j;
    for (i=0; i<nDevices; i++) {
        printf("bus %d  dev %d  error %d  id %04X%04X\n",
               devices[i].bus,
               devices[i].device,
               devices[i].error,
               devices[i].vendor_id,
               devices[i].device_id);
        for (j=0; devices[i].bar[j].size; j++) {
            printf("  size %08X  addr %08X\n",
                   devices[i].bar[j].size,
                   devices[i].bar[j].address);
        }
    } /* end for i */
}


/******************************************************************************
*
* pciInfoGet - Retrieves PCI device information
*
* This function retrieves the PCI assignments of the specified device. 
* It returns a pointer to a PciDevice structure containing the bus number,
* device number, a flag indicating whether there was an error for that
* device, the vendor/device ID and a lists the BARs indicating the size of
* the region and its address for that device. A NULL pointer will be
* returned if the specified device is not within range of the configured
* devices.
*
* RETURNS: Pointer to PCI device information, or NULL
*/
PciDevice *pciInfoGet
	(
	UINT16 device
	)
{
    if(device >= nDevices) return(NULL);
	return(&devices[device]);
}

#endif /* ifdef #ifdef INCLUDE_AUTO_PCI_CONFIG_SHOW_ROUTINE */
#endif /* ifdef AUTO_PCI_CONFIG */
