/* ixp1200Sio.c - Intel IXP1200 UART tty driver */

/* Copyright 1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */

#include "copyright_wrs.h"

/*
modification history
--------------------
01d,06Oct00,jdg  Added Big-Endian support
01c,07Mar00,jdg  Added support for B0 revision ixp1200
01b,13aug99,jdg  changed name from vbsa1200 to ixp1200eb
01a,07may99,jdg  created from sa1100Sio.c version 01a.
*/

/*
DESCRIPTION
This is the device driver for the Intel IXP1200 UART.

The UART contains a universal
asynchronous receiver/transmitter, and a baud-rate generator, The UART
contains an 8-entry, 8-bit FIFO to buffer outgoing data and an 8-entry
11-bit FIFO to buffer incoming data.  If a framing, overrun or parity
error occurs during reception, the appropriate error bits are stored in
the receive FIFO along with the received data.  The only mode of
operation supported is with the FIFOs enabled.

The UART design does not support modem control input or output signals
e.g. DTR, RI, RTS, DCD, CTS and DSR.

An interrupt is generated when the receive FIFO is non-empty and when 
the transmit FIFO is empty.

Only asynchronous serial operation is supported by the UARTs which
supports 7 or 8 bit word lengths with or without parity and with one or
two stop bits. The only serial word format supported by the driver is 8
data bits, 1 stop bit, no parity,  The default baud rate is determined
by the BSP by filling in the IXP1200_CHAN structure before calling
ixp1200UartDevInit().

The UART supports baud rates from 56.24 to 230.4 kbps.

.SH DATA STRUCTURES
An IXP1200_CHAN data structure is used to describe each channel, this
structure is described in ixp1200Sio.h.

.SH CALLBACKS
Servicing a "transmitter ready" interrupt involves making a callback to
a higher level library in order to get a character to transmit.  By
default, this driver installs dummy callback routines which do nothing.
A higher layer library that wants to use this driver (e.g. ttyDrv)
will install its own callback routine using the SIO_INSTALL_CALLBACK
ioctl command.  Likewise, a receiver interrupt handler makes a callback
to pass the character to the higher layer library.
 
.SH MODES
This driver supports both polled and interrupt modes.

.SH USAGE
The driver is typically only called by the BSP. The directly callable
routines in this modules are ixp1200UartDevInit(), and ixp1200UartInt().

The BSP's sysHwInit() routine typically calls sysSerialHwInit(), which
initialises the hardware-specific fields in the IXP1200_CHAN structure
before calling ixp1200UartDevInit() which
resets the device and installs the driver function pointers.  After
this the UART will be enabled but it will not have interrupts enabled.

The BSP's sysHwInit2() routine typically calls ixp1200UartDevInit2(),
which enables the interrupts. This is done because the IXP1200 does not
have a true interrupt controller and hence the only way to enable/disable
the UART interrupts is by modifying the UART control registers directly.

Likewise, there is a ixp1200UartDevReset() routine which disables
the UART interrupts. This is typically used in sysSerialReset to quiet
the UART interrupts within sysToMonitor.

The following example shows the first parts of the initialisation:

.CS
#include "ixp1200Sio.h"

LOCAL IXP1200_CHAN ixp1200Chan[N_IXP1200_UART_CHANS];

void sysSerialHwInit (void)
    {
    int i;

    for (i = 0; i < N_IXP1200_UART_CHANNELS; i++)
        {
        ixp1200Chan[i].baudRate = CONSOLE_BAUD_RATE;
        ixp1200Chan[i].xtal = UART_XTAL_FREQ; 

        ...


        /@
         * Initialise driver functions, getTxChar, putRcvChar
         * and channelMode and initialise UART
         @/

        ixp1200UartDevInit(&ixp1200Chan[i]);
        }
    }
.CE

The BSP's sysHwInit2() routine typically calls sysSerialHwInit2(),
which connects the chips interrupts via intConnect() and enables those
interrupts by calling ixp1200UartDevInit2(), as shown in the following example:

.CS

void sysSerialHwInit2 (void)
    {
    int i;

    for (i = 0; i < N_IXP1200_UART_CHANNELS; i++)
        {
        /@ connect and enable interrupts @/

        (void) intConnect (INUM_TO_IVEC(INT_VEC_UART),
                           ixp1200UartInt, (int) &ixp1200Chan[i]);
        ixp1200UartDevInit2(&ixp1200Chan[i]);
        }
    }
.CE

.SH BSP
By convention all the BSP-specific serial initialisation is performed
in a file called sysSerial.c, which is #include'ed by sysLib.c.
sysSerial.c implements at least four functions, sysSerialHwInit()
sysSerialHwInit2(), sysSerialChanGet(), and sysSerialReset(). The first
two have been described above, the others work as follows:

sysSerialChanGet is called by usrRoot to get the serial channel
descriptor associated with a serial channel number. The routine takes a
single parameter which is a channel number ranging between zero and
NUM_TTY. It returns a pointer to the corresponding channel descriptor,
SIO_CHAN *, which is just the address of the IXP1200_CHAN structure.
 
sysSerialReset is called from sysToMonitor() and should reset the
serial devices to an inactive state (prevent them from generating any
interrupts).

.SH INCLUDE FILES:
ixp1200Sio.h sioLib.h

.SH SEE ALSO:
*/

/* includes */

#include "vxWorks.h"
#include "intLib.h"
#include "errnoLib.h"
#include "errno.h"
#include "sioLib.h"
#include "ixp1200Sio.h"
#include "ixp1200eb.h"

/* local defines  */

/* locals */

/* function prototypes */

LOCAL STATUS ixp1200UartDummyCallback (void);
LOCAL void ixp1200UartInitChannel (IXP1200_CHAN * pChan);
LOCAL STATUS ixp1200UartIoctl (SIO_CHAN * pSioChan, int request, int arg);
LOCAL int ixp1200UartTxStartup (SIO_CHAN * pSioChan);
LOCAL int ixp1200UartCallbackInstall (SIO_CHAN * pSioChan, int callbackType,
                                      STATUS (*callback)(), void * callbackArg);
LOCAL int ixp1200UartPollInput (SIO_CHAN * pSioChan, char *);
LOCAL int ixp1200UartPollOutput (SIO_CHAN * pSioChan, char);

/* driver functions */

LOCAL SIO_DRV_FUNCS ixp1200SioDrvFuncs =
    {
    (int (*)())ixp1200UartIoctl,
    ixp1200UartTxStartup,
    ixp1200UartCallbackInstall,
    ixp1200UartPollInput,
    ixp1200UartPollOutput
    };

/*******************************************************************************
*
* ixp1200UartDummyCallback - dummy callback routine.
*
* RETURNS: ERROR, always.
*/

LOCAL STATUS ixp1200UartDummyCallback (void)
    {
    return ERROR;
    }

/*******************************************************************************
*
* ixp1200UartDevInit - initialise an IXP1200 channel
*
* This routine initialises some SIO_CHAN function pointers and then resets
* the chip to a quiescent state.  Before this routine is called, the BSP
* must already have initialised all the device addresses, etc. in the
* IXP1200_CHAN structure.
*
* RETURNS: N/A
*/

void ixp1200UartDevInit
    (
    IXP1200_CHAN *       pChan   /* ptr to IXP1200_CHAN describing this channel */
    )
    {
    int oldlevel = intLock();

    /* initialise the driver function pointers in the SIO_CHAN */

    pChan->sio.pDrvFuncs = &ixp1200SioDrvFuncs;


    /* set the non BSP-specific constants */

    pChan->getTxChar = ixp1200UartDummyCallback;
    pChan->putRcvChar = ixp1200UartDummyCallback;

    pChan->channelMode = 0;    /* am undefined mode */


    /* initialise the chip */

    ixp1200UartInitChannel (pChan);


    intUnlock (oldlevel);

    return;
    }

/*******************************************************************************
*
* ixp1200UartDevInit2 - initialise UART interupts
*
* This routine performs hardware initialisation of the UART channel.
* It will enable receive interrupts. 
* Transmit interrupts are enabled, but the ISR will probably find that
* there is nothing to do and disable them.
*
* RETURNS: N/A
*/

void ixp1200UartDevInit2
    (
    IXP1200_CHAN *       pChan   /* ptr to IXP1200_CHAN describing this channel */
    )
    {
    int oldlevel = intLock();

    pChan->uartCR |= UART_RX_INT_EN | UART_TX_INT_EN;
    IXP1200_REG_WRITE( UART_CR, pChan->uartCR );

    intUnlock (oldlevel);
    }

/*******************************************************************************
*
* ixp1200UartInitChannel - initialise UART
*
* This routine performs hardware initialisation of the UART channel.
*
* This does not enable the interrupts because we cannot mask them out
* due to lack of interrupt controller.
*
* RETURNS: N/A
*/

LOCAL void ixp1200UartInitChannel
    (
    IXP1200_CHAN *       pChan   /* ptr to IXP1200_CHAN describing this channel */
    )
    {
    UINT32      cr;

    if (sysCpuRev < 1) {
        /* Clear any outstanding interrupts and disable the uart */
        IXP1200_REG_WRITE( UART_CR, UART_INT_CLR);
    } else {
        /* Disable uart */
        IXP1200_REG_WRITE( UART_CR, 0);
    }

    /* Compute baud rate divisor */
    cr = (((pChan->xtal / (16 * pChan->baudRate)) - 1) & 0x3FF) << 16;

    /* Set word format: 8 bits, 1 stop bit, no parity */
    /* Leave interrupts disabled */
    cr |=
         UART_DSS_8BIT |
         UART_SBS_1BIT |
         UART_EN;


    pChan->uartCR = cr;
    IXP1200_REG_WRITE( UART_CR, 0);
    IXP1200_REG_WRITE( UART_CR, cr );
    }

/*******************************************************************************
*
* ixp1200UartIoctl - special device control
*
* This routine handles the IOCTL messages from the user.
*
* RETURNS: OK on success, ENOSYS on unsupported request, EIO on failed
* request.
*/

LOCAL STATUS ixp1200UartIoctl
    (
    SIO_CHAN *  pSioChan,       /* ptr to SIO_CHAN describing this channel */
    int         request,        /* request code */
    int         arg             /* some argument */
    )
    {
    int         oldlevel;       /* current interrupt level mask */
    STATUS      status;         /* status to return */
    UINT32      brd;            /* baud rate divisor */
    IXP1200_CHAN * pChan = (IXP1200_CHAN *)pSioChan;

    status = OK;        /* preset to return OK */

    switch (request)
        {
        case SIO_BAUD_SET:
            /*
             * Set the baud rate. Return EIO for an invalid baud rate, or
             * OK on success.
             *
             * Calculate the baud rate divisor: it must be non-zero and must
             * fit in a 10-bit register.
             */

            brd = (pChan->xtal / (16 * arg)) - 1;

            if (brd & ~0x3FF)
                {
                status = EIO;           /* baud rate out of range */
                break;
                }

            /* disable interrupts during chip access */

            oldlevel = intLock ();
            pChan->uartCR &= ~UART_BDR_MASK;
            pChan->uartCR |= brd << UART_BDR_OFFSET;

            IXP1200_REG_WRITE( UART_CR, pChan->uartCR );
        
            pChan->baudRate = arg;
 
            intUnlock (oldlevel);
            break;


        case SIO_BAUD_GET:

            /* Get the baud rate and return OK */

            *(int *)arg = pChan->baudRate;
            break; 


        case SIO_MODE_SET:
            /*
             * Set the mode (e.g., to interrupt or polled). Return OK
             * or EIO for an unknown or unsupported mode.
             */

            if ((arg != SIO_MODE_POLL) && (arg != SIO_MODE_INT))
                {
                status = EIO;
                break;
                }
           
            oldlevel = intLock ();
            
            if (arg == SIO_MODE_INT)
                {
                /* Enable appropriate interrupts */
                
                pChan->uartCR |= UART_RX_INT_EN | UART_TX_INT_EN;
                IXP1200_REG_WRITE( UART_CR, pChan->uartCR );
                }
            else
                {
                /* Disable the interrupt */ 

                pChan->uartCR &= ~(UART_RX_INT_EN | UART_TX_INT_EN);
                IXP1200_REG_WRITE( UART_CR, pChan->uartCR | UART_INT_CLR);
                }

            pChan->channelMode = arg;

            intUnlock (oldlevel);
            break;          


        case SIO_MODE_GET:

            /* Get the current mode and return OK */

            *(int *)arg = pChan->channelMode;
            break;


        case SIO_AVAIL_MODES_GET:

            /* Get the available modes and return OK */

            *(int *)arg = SIO_MODE_INT | SIO_MODE_POLL;
            break;


        case SIO_HW_OPTS_SET:

            /*
             * Optional command to set the hardware options (as defined
             * in sioLib.h).
             * Return OK, or ENOSYS if this command is not implemented.
             * Note: several hardware options are specified at once.
             * This routine should set as many as it can and then return
             * OK. The SIO_HW_OPTS_GET is used to find out which options
             * were actually set.
             */

        case SIO_HW_OPTS_GET:

            /*
             * Optional command to get the hardware options (as defined
             * in sioLib.h). Return OK or ENOSYS if this command is not
             * implemented.  Note: if this command is unimplemented, it
             * will be assumed that the driver options are CREAD | CS8
             * (e.g., eight data bits, one stop bit, no parity, ints enabled).
             */

        default:
            status = ENOSYS;
        }

    return status;

    }

/*******************************************************************************
*
* ixp1200UartInt - handle an interrupt
*
* This routine handles interrupts from the UART.
*
* RETURNS: N/A
*/
 
void ixp1200UartInt
    (
    IXP1200_CHAN *       pChan   /* ptr to IXP1200_CHAN describing this channel */
    )
    {
    UINT32      ch;             /* Possible char to be in/output */
    UINT32      statusReg;      /* Status from UART */
    int         gotOne;
    char        c;

    /* read status registers */

    IXP1200_REG_READ( UART_SR, statusReg );
    

    /*
     * loop, removing characters from the rx FIFO (other Sio drivers do this
     * so it must be OK)
     */

    while (statusReg & UART_RX_RDY)                        /* data? */
        {
        IXP1200_REG_READ ( UART_DR, ch);            /* read character */
        (*pChan->putRcvChar) (pChan->putRcvArg, (char)ch); /* pass it on */
        IXP1200_REG_READ( UART_SR, statusReg );   /* update status */
        }


    /*
     * if tx FIFO is interrupting, loop, writing characters to the FIFO
     * until the FIFO is full or there are no more characters to transmit
     */

    if (statusReg & UART_TX_EMPTY)              /* tx FIFO needs filling? */
        {
        do
            {
            gotOne = (*pChan->getTxChar) (pChan->getTxArg, &c) != ERROR;
            if (gotOne)
                IXP1200_REG_WRITE(UART_DR, c);          /* tx char */

            IXP1200_REG_READ( UART_SR, statusReg );     /* update status */
            }
        while (gotOne && (statusReg & UART_TX_RDY));    /* while not full */

        if (!gotOne)
            {
            /* no more chars to send - disable transmitter interrupts */
            
            pChan->uartCR &= ~UART_TX_INT_EN;
            }
        }

    if (sysCpuRev < 1) {
        /* Clear interrupts */
        /* Note that if an interrupt comes through during the above processing,
           then since the UART interrupts are level-sensitive, not edge-sensitive,
           we won't lose an interrupt */
        IXP1200_REG_WRITE( UART_CR, pChan->uartCR | UART_INT_CLR);
    } else {
        IXP1200_REG_WRITE( UART_CR, pChan->uartCR);
    }

    }

/*******************************************************************************
*
* ixp1200UartTxStartup - transmitter startup routine
*
* Call interrupt level char output routine and enable interrupt
*
* RETURNS: OK on success, ENOSYS if the device is polled-only, or
* EIO on hardware error.
*/

LOCAL int ixp1200UartTxStartup
    (
    SIO_CHAN *  pSioChan        /* ptr to SIO_CHAN describing this channel */
    )
    {
    IXP1200_CHAN * pChan = (IXP1200_CHAN *)pSioChan;
    UINT oldlevel;

    if (pChan->channelMode == SIO_MODE_INT)
        {
        /* Enable Transmitter interrupts */
        
        oldlevel = intLock ();

        pChan->uartCR |= UART_TX_INT_EN;
        IXP1200_REG_WRITE (UART_CR, pChan->uartCR);
        
        intUnlock (oldlevel);
        return OK;
        }
    else
        {
        return ENOSYS;
        }
    }

/******************************************************************************
*
* ixp1200UartPollOutput - output a character in polled mode.
*
* RETURNS: OK if a character arrived, EIO on device error, EAGAIN
* if the output buffer is full, ENOSYS if the device is interrupt-only.
*/

LOCAL int ixp1200UartPollOutput
    (
    SIO_CHAN *  pSioChan,       /* ptr to SIO_CHAN describing this channel */
    char        outChar         /* char to output */
    )
    {

    /* IXP1200_CHAN * pChan = (IXP1200_CHAN *)pSioChan; */
    UINT32 pollStatus;

    IXP1200_REG_READ (UART_SR, pollStatus);

    /* is the transmitter ready to accept a character? */

    if (pollStatus & UART_TX_FULL)
        return EAGAIN;


    /* write out the character */

    IXP1200_REG_WRITE (UART_DR, outChar);        /* transmit character */

    return OK;
    }

/******************************************************************************
*
* ixp1200UartPollInput - poll the device for input.
*
* RETURNS: OK if a character arrived, EIO on device error, EAGAIN
* if the input buffer is empty, ENOSYS if the device is interrupt-only.
*/

LOCAL int ixp1200UartPollInput
    (
    SIO_CHAN *  pSioChan,       /* ptr to SIO_CHAN describing this channel */
    char *      thisChar        /* pointer to where to return character */
    )
    {

    /* IXP1200_CHAN * pChan = (IXP1200_CHAN *)pSioChan; */
    UINT32 pollStatus;

    IXP1200_REG_READ (UART_SR, pollStatus);

    if (!(pollStatus & UART_RX_RDY))
        return EAGAIN;


    /* got a character */

    IXP1200_REG_READ (UART_DR, *thisChar);
#ifdef IXP1200_A0
    IXP1200_REG_READ (UART_DR, *thisChar); /* pass-1 bug */
#endif

    return OK;
    }

/******************************************************************************
*
* ixp1200UartCallbackInstall - install ISR callbacks to get/put chars.
*
* This routine installs interrupt callbacks for transmitting characters
* and receiving characters.
*
* RETURNS: OK on success, or ENOSYS for an unsupported callback type.
*
*/

LOCAL int ixp1200UartCallbackInstall
    (
    SIO_CHAN *  pSioChan,       /* ptr to SIO_CHAN describing this channel */
    int         callbackType,   /* type of callback */
    STATUS      (*callback)(),  /* callback */
    void *      callbackArg     /* parameter to callback */
                 
    )
    {
    IXP1200_CHAN * pChan = (IXP1200_CHAN *)pSioChan;

    switch (callbackType)
        {
        case SIO_CALLBACK_GET_TX_CHAR:
            pChan->getTxChar    = callback;
            pChan->getTxArg     = callbackArg;
            return OK;

        case SIO_CALLBACK_PUT_RCV_CHAR:
            pChan->putRcvChar   = callback;
            pChan->putRcvArg    = callbackArg;
            return OK;

        default:
            return ENOSYS;
        }
    }

/*******************************************************************************
*
* ixp1200UartDevReset - Reset an IXP1200 channel
*
* This routine resets the UART to a quiet state. We do this here because
* there is no real interrupt controller to do it in a generic way
*
* RETURNS: N/A
*/

void ixp1200UartDevReset
    (
    IXP1200_CHAN *       pChan   /* ptr to IXP1200_CHAN describing this channel */
    )
    {
    int oldlevel = intLock();

    pChan->uartCR &= ~( UART_TX_INT_EN | UART_EN | UART_RX_INT_EN );
    IXP1200_REG_WRITE( UART_CR, pChan->uartCR | UART_INT_CLR );

    intUnlock (oldlevel);
    }
