/* sysALib.s - Intel ixp1200eb system-dependent routines */

/* Copyright 1997-1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */
	.data
	.globl	_copyright_wind_river
	.long	_copyright_wind_river

/*
modification history
--------------------
01d,06Oct00,jdg  Added Big-Endian support
01c,07Mar00,jdg  Added support for B0 revision ixp1200
01b,13aug99,jdg  changed name from vbsa1200 to ixp1200eb
01a,19Apr99,jdg  created from 01c of brutus.
*/

/*
DESCRIPTION
This module contains system-dependent routines written in assembly
language. It contains the entry code, sysInit(), for VxWorks images
that start running from RAM, such as 'vxWorks'. These images are loaded
into memory by some external program (e.g., a boot ROM) and then
started.  The routine sysInit() must come first in the text segment.
Its job is to perform the minimal setup needed to call the generic C
routine usrInit().

sysInit() masks interrupts in the processor, initialises the MMU, sets
the initial stack pointer and intialises system hardware.  Other
hardware and device initialisation is performed later in the sysHwInit
routine in sysLib.c

NOTE
Many routines in this module don't use the "C" frame pointer %r11@ ! or
establish a stack frame. This is either for efficiency, or because no stack
has yet been established.

SEE ALSO:
.I "ARM Architecture Reference Manual,"
*/

#define _ASMLANGUAGE
#include "vxWorks.h"
#include "regs.h"
#include "arch/arm/arm.h"
#include "arch/arm/mmuArmLib.h"
#include "sysLib.h"
#include "config.h"


/* Exported internal functions */

	.globl	_sysInit		/* start of system code */
	.globl	_sysIntStackSplit	/* routine to split interrupt stack */
	.globl	_sysMmuTtbrGet		/* Read MMU TTBR */
	.globl  _sysToRomInit           /* Turn off MMU and branch to romInit */

	.globl	_sysWrite16		/* routine to do STRH */
	.globl	_sysRead16		/* routine to do LDRH */
	.globl  _swap32                 /* routine to swap bytes in 32 bit int */
	.globl  _swap16                 /* routine to swap bytes in 16 bit int */

	.globl  _sysCpuVerGet           /* Get CPU version revision */
	
/* externals */

	.extern	_usrInit		/* system initialization routine */

	.extern	_vxSvcIntStackBase	/* base of SVC-mode interrupt stack */
	.extern	_vxSvcIntStackEnd	/* end of SVC-mode interrupt stack */
	.extern	_vxIrqIntStackBase	/* base of IRQ-mode interrupt stack */
	.extern	_vxIrqIntStackEnd	/* end of IRQ-mode interrupt stack */


/* absolutes */

	/* absolute address of cache and mini-cache flush areas */
	.globl _sysCacheFlushReadArea
	.equ   _sysCacheFlushReadArea, CACHE_FLUSH_BASE
	.globl _sysMinicacheFlushReadArea
	.equ   _sysMinicacheFlushReadArea, MINICACHE_FLUSH_BASE


	.text
	.balign 4

/*******************************************************************************
*
* sysInit - start after boot
*
* This routine is the system start-up entry point for VxWorks in RAM, the
* first code executed after booting.  It disables interrupts, sets up
* the stack, and jumps to the C routine usrInit() in usrConfig.c.
*
* The initial stack is set to grow down from the address of sysInit().  This
* stack is used only by usrInit() and is never used again.  Memory for the
* stack must be accounted for when determining the system load address.
*
* NOTE: This routine should not be called by the user.
*
* RETURNS: N/A (never returns)

* sysInit ()		  /@ THIS IS NOT A CALLABLE ROUTINE @/

*/

_sysInit:

	/* disable interrupts and force SVC32 mode, just like reset */

	MRS	r1, cpsr
	BIC	r1, r1, #MASK_MODE
	ORR	r1, r1, #MODE_SVC32 | I_BIT | F_BIT
	MSR	cpsr, r1

#ifdef USE_LEDS_FOR_DEBUG
	/* Initialize LED's */
	LDR	r1, L$_IXP1200_LED_ADDR
#ifndef LED_MEM_MAPPED
	/* Set GPIO's to core ownership, outputs, and output a zero */
	/* IXP1200_LED_ADDR is GPIO_DATA */
	MOV	r2, #(GPIO_OWN_CCCC | GPIO_OUTPUT(0xF))
	STR	r2, [r1,#-(IXP1200_LED_ADDR-IXP1200_GPIO_EN)]
#endif /* LED_MEM_MAPPED */
	MOV	r2, #2
	STR	r2, [r1]
#endif /* USE_LEDS_FOR_DEBUG */

	/*
	 * Check for Chip revision: read chip ID into r8
	 * 0:	A rev
	 * 1:	B rev
	 */

	MRC	CP_MMU, 0, r8, c0, c0, 0
	AND	r8,r8,#0xF
	

	/*
	 * CPU INTERRUPTS DISABLED
	 *
	 * In order to remap RAM down to zero upwards, we must have
	 * been entered by some process that has switched the MMU on
	 * and remapped the machine. So leave the MMU switched on, and
	 * pointing to its set of page tables, but otherwise initialise
	 * MMU.
	 */

	MOV	r1, #MMU_INIT_VALUE		/* normal init value */
	ORR	r1, r1, #MMUCR_M_ENABLE		/* leave MMU switched on */
	ORR	r1, r1, #MMUCR_ROM		/* and set protection bit */
	MCR	CP_MMU, 0, r1, c1, c0, 0	/* Write to MMU CR */

	/*
	 * Set Process ID Register to zero, this effectively disables
	 * the process ID remapping feature.
	 */

	MOV	r1, #0
	MCR	CP_MMU, 0, r1, c13, c0, 0

#if 0
	/*
	 * Set core clock speed in Power Manager PLL Configuration Register.
	 * This must be done before enabling clock switching - there are
	 * strong suspicions that it will have no effect otherwise.
	 */

	MOV	r2, #CCCR_INIT_VALUE	/* coded core clock speed value */
	LDR	r1, L$_IXP1200_PPL_CFG	/* PLL Configuration Reg*/
	STR	r2, [r1]
#endif

	/*
	 * Put the CPU in its normal (fast) clock switching mode
	 * Note that this may prevent any kernel that a bootrom loads from
	 * changing the core clock.
	 */

	MCR	CP_MMU, 0, r14, c15, c1, 2

	/* Clear Debug Support features: VxWorks makes no use of these */

	MOV	r1, #0
	MCR	CP_MMU, 0, r1, c14, c3, 0	/* Clear DBCR */
	MCR	CP_MMU, 0, r1, c14, c8, 0	/* Clear IBCR */

	/*
	 * We would like to discard the contents of the Write-Buffer
	 * altogether, but there is no facility to do this. Failing that,
	 * we do not want any pending writes to happen at a later stage,
	 * so drain the Write-Buffer, i.e. force any pending writes to
	 * happen now.
	 */

	MCR	CP_MMU, 0, r1, c7, c10, 4	/* Drain write-buffer */

	/* Flush, (i.e. invalidate) all entries in all caches */

	MCR	CP_MMU, 0, r1, c7, c6, 0	/* Flush D-cache & mini-cache */
	MCR	CP_MMU, 0, r1, c7, c5, 0	/* Flush (inval.) all I-cache */

	/*
	 * On StrongARM, the next four instructions could still come
	 * from (what was in) the Instruction cache.
	 *
	 * The BSP makes no explicit use of Read Buffer, but make sure
	 * that nothing already in it affects us.
	 */

	MCR	CP_MMU, 0, r1, c9, c0, 0	/* Flush Read Buffer */

	/* Bring other functional units out of reset */

	LDR	r1, L$_IXP1200_RESET
	MOV	r2, #RESET_MICROENGINE
	STR	r2, [r1]

	/* 
	 * We don't want to reprogram the SDRAM controller, but
	 * we should make sure that the interrupt is not set 
	 */
	
	LDR	r1, L$_IXP1200_SDRAM_CSR
	LDR	r2, [r1]
	BIC	r2, r2, #(1<<SDRAM_EN_INT_BIT)
	STR	r2, [r1]

#ifndef USE_LEDS_FOR_DEBUG
	
	/* Initialize LED's */
	
	LDR	r1, L$_IXP1200_LED_ADDR
#ifndef LED_MEM_MAPPED
	
	/* 
	 * Set GPIO's to core ownership, outputs, and output a zero
	 * IXP1200_LED_ADDR is GPIO_DATA 
	 */
	
	MOV	r2, #(GPIO_OWN_CCCC | GPIO_OUTPUT(0xF))
	STR	r2, [r1,#-(IXP1200_LED_ADDR-IXP1200_GPIO_EN)]
#endif /* LED_MEM_MAPPED */
	MOV	r2, #2
	STR	r2, [r1]
#endif /* USE_LEDS_FOR_DEBUG */
#ifdef USE_LEDS_FOR_DEBUG
	LDR	r1, L$_IXP1200_LED_ADDR
	MOV	r2, #3
	STR	r2, [r1]
#endif /* USE_LEDS_FOR_DEBUG */

	/*
	 * We don't reload the MMU page tables here. Since we are executing
	 * out of SDRAM (which has been remapped), we have to assume that
	 * we have valid page tables. If we don't and we try to remap them
	 * here, then we will suddenly change the virtual to physical
	 * mapping where we are executing.
	 */

	/* set initial stack pointer so stack grows down from start of code */

	ADR	sp, _sysInit		/* initialise stack pointer */


	/* now call usrInit */

	MOV	fp, #0			/* initialise frame pointer */

#ifdef INCLUDE_HPC
	MOV	r0, #(BOOT_WARM_AUTOBOOT | BOOT_DONT_INIT_HPC | BOOT_DONT_INIT_PCI)	/* pass startType */
#else
	MOV	r0, #BOOT_WARM_AUTOBOOT	/* pass startType */
#endif
	B	_usrInit

/*******************************************************************************
*
* sysIntStackSplit - split interrupt stack and set interrupt stack pointers
*
* This routine is called, via a function pointer, during kernel
* initialisation.  It splits the allocated interrupt stack into IRQ and
* SVC-mode stacks and sets the processor's IRQ stack pointer. Note that
* the pointer passed points to the bottom of the stack allocated i.e.
* highest address+1.
*
* IRQ stack needs 6 words per nested interrupt;
* SVC-mode will need a good deal more for the C interrupt handlers.
* For now, use ratio 1:7 with any excess allocated to the SVC-mode stack
* at the lowest address.
*
* Note that FIQ is not handled by VxWorks so no stack is allocated for it.
*
* The stacks and the variables that describe them look like this.
* .CS
*
*	  - HIGH MEMORY -
*     ------------------------ <--- vxIrqIntStackBase (r0 on entry)
*     | 		     |	
*     |       IRQ-mode	     |	
*     |    interrupt stack   |
*     | 		     |	
*     ------------------------ <--{ vxIrqIntStackEnd
*     | 		     |	  { vxSvcIntStackBase
*     |       SVC-mode	     |	
*     |    interrupt stack   |
*     | 		     |	
*     ------------------------ <--- vxSvcIntStackEnd
*	  - LOW  MEMORY -
* .CE
*
* NOTE: This routine should not be called by the user.
*
* void sysIntStackSplit
*     (
*     char *pBotStack	/@ pointer to bottom of interrupt stack @/
*     long size		/@ size of stack @/
*     )

*/

_sysIntStackSplit:

	/*
	 * r0 = base of space allocated for stacks (i.e. highest address)
	 * r1 = size of space
	 */

	SUB	r2, r0, r1			/* r2->lowest usable address */
	LDR	r3, L$_vxSvcIntStackEnd
	STR	r2, [r3]			/*  == end of SVC-mode stack */
	SUB	r2, r0, r1, ASR #3		/* leave 1/8 for IRQ */
	LDR	r3, L$_vxSvcIntStackBase
	STR	r2, [r3]

	/* now allocate IRQ stack, setting irq_sp */

	LDR	r3, L$_vxIrqIntStackEnd
	STR	r2, [r3]
	LDR	r3, L$_vxIrqIntStackBase
	STR	r0, [r3]

	MRS	r2, cpsr
	BIC	r3, r2, #MASK_MODE
	ORR	r3, r3, #MODE_IRQ32 | I_BIT	/* set irq_sp */
	MSR	cpsr, r3
	MOV	sp, r0

	/* switch back to original mode and return */

	MSR	cpsr, r2
	MOV	pc, lr

/******************************************************************************
*
* sysMmuTtbrGet - get the Translation Table Base Register
*
* This routine reads the Translation Table Base Register of the MMU.
*
* RETURNS: the value in the TTBR
*
* NOMANUAL
*
*  LEVEL_1_DESC *  sysMmuTtbrGet(void)
*/

_sysMmuTtbrGet:

	MRC	CP_MMU, 0, r0, c2, c0, 0	/* Read the TTBR in MMU */
	MOV	pc, lr				/* exit */


/******************************************************************************
*
* sysCpuVerGet - get the CPU version
*
* This routine reads chip revision field from the coprocessor
*
* RETURNS: the CPU version
*
*/

_sysCpuVerGet:	

	MRC	CP_MMU, 0, r0, c0, c0, 0
	MOV	pc, lr				/* exit */


/******************************************************************************
*
* sysToRomInit - turn off MMU and branch to romInit
*
* This copies a small bit of code to SDRAM
*
* RETURNS: This never returns
*
*
* void sysToRomInit
*     (
*     int startType
*     )
*/
_sysToRomInit:	
	/* disable interrupts in CPU and switch to SVC32 mode */

	MRS	r1, cpsr
	BIC	r1, r1, #MASK_MODE
	ORR	r1, r1, #MODE_SVC32 | I_BIT | F_BIT
	MSR	cpsr, r1
	
#ifdef USE_LEDS_FOR_DEBUG
	LDR	r1, L$_IXP1200_LED_ADDR
	MOV	r2, #5
	STR	r2, [r1]
#endif /* USE_LEDS_FOR_DEBUG */
	
	/* 
	 * Copy three instruction from copy_start2 to SRAM, and then
	 * execute from there during the MMU transition 
	 */
	
	ADR	r1, copy_start2
	LDR	r4, L$_SRAM_COPY_ADDR
	LDMIA	r1, {r5,r6,r7}
	STMIA	r4, {r5,r6,r7}

	/* Load register constants first, so we have less to copy */
	
	MOV	r1, #MMU_INIT_VALUE
	MOV	r2, #0
	LDR	r3, L$_ROM_INIT_PLUS4

	/* Jump to SRAM to finish */
	
	MOV	pc, r4
copy_start2:	
	
	/* Disable the MMU */
	
	MCR	CP_MMU, 0, r1, c1, c0, 0	/* Write to MMU CR */

	/* Flush I-cache */
	
	MCR	CP_MMU, 0, r2, c7, c7, 0	/* Flush (inval) all I,D-cache */

	/* jump to romInit + 4 */
	
	MOV	pc, r3
	
/*******************************************************************************
*
* sysWrite16 - explicitly perform a 16-bit write
*
* This routine simply performs a 16-bit write, i.e. an STRH instruction. This
* would be used for I/O peripherals where it is significant whether a
* store is an STR, STRB or STRH. This routine may be particularly
* useful when the behaviour of the C compiler is not appropriate.
*
* NOTE
* For reasons of efficiency, this routine does not establish a stack frame.
*
* RETURNS: N/A

* void sysWrite16
*	(
*	volatile UINT16 *	address,	/@ address to write to @/
*	UINT16			data		/@ data to write @/
*	)

*/

_sysWrite16:

	STRH	r1, [r0]	/* Store the 16-bit quantity */
	MOV	pc, lr		/* Return */

/*******************************************************************************
*
* sysRead16 - explicitly perform a 16-bit read
*
* This routine simply performs a 16-bit read, i.e. an LDRH instruction. This
* would be used for I/O peripherals where it is significant whether a store
* is an LDR, LDRB or LDRH.
*
* NOTE
* For reasons of efficiency, this routine does not establish a stack frame.
*
* RETURNS: the value read

* UINT16 sysRead16
*	(
*	volatile UINT16 *	address,	/@ address to read @/
*	)

*/

_sysRead16:

	LDRH	r0, [r0]	/* Load the 16-bit quantity */
	MOV	pc, lr		/* Return */

	
/*******************************************************************************
*
* swap32 - swap the bytes in a 32-bit word
*
* This routine swaps the bytes in a 32-bit word;  i.e. it would convert
* 0x12345678 to 0x78563412.
*
* NOTE
* For reasons of efficiency, this routine does not establish a stack frame.
*
* RETURNS: the swapped data
*
* UINT32 swap32
*	(
*	UINT32			data		/@ data to be swapped @/
*	)
*/

_swap32:

	/* Assume r0 initially contains 11223344 */
	AND	r1, r0, #0x0000FF00 /* r1 = 00003300 */
	MOV	r1, r1, lsl #8      /* r1 = 00330000 */
	ORR     r1, r1, r0, lsl #24 /* r1 = 44330000 */
	AND     r2, r0, #0x00FF0000 /* r2 = 00220000 */
	ORR	r1, r1, r2, lsr #8  /* r1 = 44332200 */
	ORR     r0, r1, r0, lsr #24 /* r0 = 44332211 */
	MOV	pc, lr		/* Return */

	
/*******************************************************************************
*
* swap16 - swap the bytes in a 16-bit word
*
* This routine swaps the bytes in a 16-bit word;  i.e. it would convert
* 0x1234 to 0x3412.
*
* NOTE
* For reasons of efficiency, this routine does not establish a stack frame.
*
* RETURNS: the swapped data
*
* UINT16 swap16
*	(
*	UINT16			data		/@ data to be swapped @/
*	)
*/

_swap16:

	/* Assume r0 initially contains xxxx1122 */
	MOV     r0, r0, lsl #16     /* r0 = 11220000 */
	MOV     r1, r0, lsl #8      /* r1 = 22000000 */
	ORR	r1, r1, r0, lsr #8  /* r1 = 22112200 */
	MOV	r0, r1, lsr #16     /* r0 = 00002211 */
	MOV	pc, lr		/* Return */


/*******************************************************************************
*
* getFP - Return the frame pointer of the caller of this routine
*
* This routine returns the frame pointer from the caller of this routine.
*
* RETURNS: the frame pointer

* UINT32 getFP(void)

*/
	.globl _getFP
_getFP:
	MOV r0, fp
	LDR r0, [r0,#-12]
	MOV pc, lr

/******************************************************************************/

/*
 * PC-relative-addressable pointers - LDR Rn,=sym was (is?) broken
 * note "_" after "$" to stop preprocessor performing substitution
 */

	.balign	4

L$_vxSvcIntStackBase:
	.long	_vxSvcIntStackBase

L$_vxSvcIntStackEnd:
	.long	_vxSvcIntStackEnd

L$_vxIrqIntStackBase:
	.long	_vxIrqIntStackBase

L$_vxIrqIntStackEnd:
	.long	_vxIrqIntStackEnd

L$_IXP1200_PPL_CFG:	
	.long IXP1200_PLL_CFG

L$_IXP1200_RESET:	
	.long IXP1200_RESET
	
L$_IXP1200_SDRAM_CSR:
	.long IXP1200_SDRAM_CSR

	/* IXP1200_LED_ADDR is either in memory or is the same as GPIO_DATA */
	
L$_IXP1200_LED_ADDR:	
	.long IXP1200_LED_ADDR

L$_SRAM_COPY_ADDR:
	.long SRAM_COPY_ADDR

	/* 
	 * ROM_INIT_PLUS4 needs to be a physical address as we are turning
	 * the MMU off before we use it; so we must remove the
	 * ROM_VIRT_OFFSET 
	 */
	
L$_ROM_INIT_PLUS4:	
	.long (ROM_TEXT_ADRS + 4 - ROM_VIRT_OFFSET)

