/* sysEnd.c - System enhanced network interface support library */

/* Copyright 1997-1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */

#include "copyright_wrs.h"

/*
modification history
--------------------
01h,05Jan00,jdg  Added support for auto PCI address assignment
01g,13aug99,jdg  changed name from vbsa1200 to ixp1200eb
01f,08jul99,jdg  change pciMemBase and usrFlags for SA1200
01e,04mar98,jpd  comments/doc'n changes.
01d,29jan98.rlp  renamed macro names containing lower case characters.
01c,09dec97,tam  changed init of endLoadString to prevent tgtsvr checksum error
01b,06dec97,tam  modified to allow auto detection/initialization of DEC chip;
                 added sysDecPciIntAck() routine.
01a,27sep97,rlp  written.
*/

/*
DESCRIPTION
This file contains the board-specific routines for Ethernet adapter
initialisation. Currently, this module only supports the Digital
semiconductor EB143 LAN Adapter, Digital 21140 or 21040 based LAN
Adapters.  If support for more 21143-based adapters is desired, this
module must be modified.


NOTE
At the time of writing, this module has only been tested with the following
Ethernet cards:
        Digital 21143 controller: DEC EB143 evaluation card.
        Digital 21140 controller: D-LINK DFE-500TX and NETGEAR FA310TX.
        Digital 21040 controller: Digital DE435.
Refer to the BSP man page for any eventual limitations or problems related
to the BSP.


SEE ALSO: ifLib,
.I "Digital Semiconductor 21143 PCI/CardBus Fast Ethernet LAN Controller."
.I "Digital Semiconductor 21143 10/100Base-TX Evaluation Board User's Guide."
*/

#include "vxWorks.h"
#include "config.h"

#include "vmLib.h"
#include "stdio.h"

#ifdef INCLUDE_DEC      /*  Digital Semiconductor 21X4X Support */

#include "drv/end/dec21x4xEnd.h"


/* defines */

#define CSR_BASE_MSK    0x7f            /* Mask Base Address Register */
#define MMU_MSK         0x0fff          /* Mask MMU page */
#define END_LD_STR_SIZE 80

/* PCI memory base address register configuration mode */

#define FORCE           0x00    /* overwrite membase address register */
#define AUTO            0x01    /* read membase address register */

#if defined(AUTO_PCI_CONFIG)
#define PCI_REG_MOD     AUTO    /* define the register configuration mode */
#else
#define PCI_REG_MOD     FORCE   /* define the register configuration mode */
#endif


/* DEC 21X4X 10/100Base-TX Board type */

#define EB143           1       /* DEC 21143 10/100Base-TX Evaluation Board */
#define DC140           2       /* DEC 21140 10/100Base-TX Based Board */
#define DC040           3       /* DEC 21040 10/100Base-TX Based Board */

#define BOARD_TYPE_NB   3


/* DEC driver user flags */

#define DEC_USR_FLAGS_143 (DEC_USR_21143)
#define DEC_USR_FLAGS_140 (DEC_USR_BAR_RX | DEC_USR_RML | DEC_USR_CAL_08 | \
                           DEC_USR_PBL_04 | DEC_USR_21140)
#define DEC_USR_FLAGS_040 (DEC_USR_BAR_RX | DEC_USR_CAL_08 | DEC_USR_PBL_04)


/* DEC 21X4X PCI/CardBus specific definitions */

#define DEC21X4X_PCI_VENDOR_ID  0x1011  /* PCI vendor ID */
#define DEC21143_PCI_DEVICE_ID  0x0019  /* PCI device ID */
#define DEC21140_PCI_DEVICE_ID  0x0009  /* PCI device ID */
#define DEC21040_PCI_DEVICE_ID  0x0002  /* PCI device ID */


/* forward declarations */

STATUS sysDec21143MediaSelect (DEC21X4X_DRV_CTRL *, UINT *);
STATUS sysDec21140MediaSelect (DEC21X4X_DRV_CTRL *, UINT *);
void sysDecPciIntAck ();

/* typedefs */
 
typedef struct pciResource              /* PCI_RESOURCES */
    {
    UINT32      iobaseCsr;              /* Base Address Register 0 */
    UINT32      membaseCsr;             /* Base Address Register 1 */
    char        irq;                    /* Interrupt Request Level */
    UINT32      irqvec;                 /* Interrupt Request vector */
    UINT32      configType;             /* type of configuration */
    UINT32      boardType;              /* type of LAN board this unit is */
    UINT32      pciBus;                 /* PCI Bus number */
    UINT32      pciDevice;              /* PCI Device number */
    UINT32      pciFunc;                /* PCI Function number */
    } PCI_RESOURCES;

typedef struct boardResource            /* BOARD_RESOURCES */
    {
    UINT32      type;           /* type of the board */
    UINT32      vendorId;               /* Vendor ID */
    UINT32      deviceId;               /* Device ID */
    UINT32      decUsrFlags;            /* DEC driver user flags */
    FUNCPTR     mediaSelectFunc;                /* media select routine */
    } BOARD_RESOURCES;


/* locals */

/*
 * This array defines the board-specific PCI resources, the base address
 * register configuration mode and the Ethernet adapter type. It's indexed
 * using the device number returned from pciFindDevice().
 *
 * Originally, the four different slots had different parameters, but for us
 * they are the same. So we can just duplicate the last set for the rest of the
 * table.
 */
 
LOCAL PCI_RESOURCES pciResources [PCI_MAX_DEV] =
    {
    {PCI_IO_ADR0, PCI_MEM_ADR0, PCI_INT_LVL0, PCI_INT_VEC0, PCI_REG_MOD, EB143},
    {PCI_IO_ADR1, PCI_MEM_ADR1, PCI_INT_LVL1, PCI_INT_VEC1, PCI_REG_MOD, EB143},
    {PCI_IO_ADR2, PCI_MEM_ADR2, PCI_INT_LVL2, PCI_INT_VEC2, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    {PCI_IO_ADR3, PCI_MEM_ADR3, PCI_INT_LVL3, PCI_INT_VEC3, PCI_REG_MOD, EB143},
    };

/*
 * This array defines board-specific vendor and device ids, flags to pass to
 * the drive load routine and the function used to select the media.
 */

LOCAL BOARD_RESOURCES boardResources [BOARD_TYPE_NB] =
    {
    {EB143, DEC21X4X_PCI_VENDOR_ID, DEC21143_PCI_DEVICE_ID, DEC_USR_FLAGS_143,\
     sysDec21143MediaSelect},
    {DC140, DEC21X4X_PCI_VENDOR_ID, DEC21140_PCI_DEVICE_ID, DEC_USR_FLAGS_140,\
     sysDec21140MediaSelect},
    {DC040, DEC21X4X_PCI_VENDOR_ID, DEC21040_PCI_DEVICE_ID, DEC_USR_FLAGS_040,\
     NULL},
    };

LOCAL UINT32    pciDeviceNo = 0;                /* PCI device populated */
LOCAL char      endLoadStr[END_LD_STR_SIZE];    /* END load string */


/* imports */

IMPORT END_TBL_ENTRY endDevTbl[];


/*******************************************************************************
*
* sysDec21x4xPciInit - prepare LAN adapter for DEC21X4X initialization
*
* This routine find out the PCI device, and map its memory and IO address.
* It must be done prior to initializing the DEC21X4X, sysDec21x4xInit().  Also
* must be done prior to MMU initialization, usrMmuInit().
*
* RETURNS: N/A
*/

STATUS sysDec21x4xPciInit (void)
    {
    PCI_RESOURCES *     pRsrc;          /* dec resource */
#if !defined(AUTO_PCI_CONFIG)
    PHYS_MEM_DESC *     pMmu;           /* MMU table */
#endif
    UINT32              pciBus;         /* PCI Bus number */
    UINT32              pciDevice;      /* PCI Device number */
    UINT32              pciFunc;        /* PCI Function number */
    int                 ix;             /* counter */
    UINT32              membaseCsr;     /* Base Address Register 1 */
    UINT32              iobaseCsr;      /* Base Address Register 0 */
    char                irq;            /* Interrupt Request Level */
    UINT32              boardType = NONE; /* board type detected */
    BOOL                found = FALSE;

    /*
     * The following code tries to automatically detect the first instance
     * of a DEC21143, DEC21140 or DEC21040 Ethernet based PCI board.
     */

    for (ix = 0; ix < BOARD_TYPE_NB; ix++)
        {
        if (pciFindDevice (boardResources[ix].vendorId,
                           boardResources[ix].deviceId,
                           0, &pciBus, &pciDevice, &pciFunc) == OK)
            {
            /* board detected */

            found = TRUE;

            boardType = boardResources[ix].type;

            /* select the media function */

            _func_dec2114xMediaSelect = boardResources[ix].mediaSelectFunc;

            /* update the END device table load string */

            sprintf (endLoadStr, "0x00:0x00:0x00:0x00:-1:0x00:0x%x:2",
                     boardResources[ix].decUsrFlags);
            endDevTbl[0].endLoadString = endLoadStr;

            break;
            }
        }

    if (!found || (pciDevice > PCI_MAX_DEV))
        return (ERROR);

    /* we found the right one */

#if (defined(TARGET_CMA220) || defined(TARGET_CMA221) || defined(TARGET_CMA225))
    /* attach board specific interrup ack. routine if required by the h/w */

    _func_dec2114xIntAck = sysDecPciIntAck;
#endif  /* TARGET_CMA220 || TARGET_CMA221 || TARGET_CMA225 */

    pciDeviceNo = pciDevice;
    pciResources[pciDevice].boardType = boardType;

    pRsrc = &(pciResources[pciDevice]);
    pRsrc->pciBus    = pciBus;
    pRsrc->pciDevice = pciDevice;
    pRsrc->pciFunc   = pciFunc;

    if (pRsrc->configType == AUTO)
        {
        /* get memory base address and IO base address */

        pciConfigInLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_BASE_ADDRESS_0, &iobaseCsr);
        pciConfigInLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_BASE_ADDRESS_1, &membaseCsr);
        pciConfigInByte (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                         PCI_CFG_DEV_INT_LINE, &irq);

        membaseCsr &= ~CSR_BASE_MSK;
        iobaseCsr &= ~CSR_BASE_MSK;

        /* modify the sysPhysMemDesc[] */
#if !defined(AUTO_PCI_CONFIG)
        for (ix = 0; ix < sysPhysMemDescNumEnt; ix++)
            {
            pMmu = &sysPhysMemDesc[ix];
            if (pMmu->physicalAddr == (void *)pRsrc->membaseCsr)
                {
                pMmu->physicalAddr = (void *) (membaseCsr & ~MMU_MSK);
                pMmu->virtualAddr  = (void *) (membaseCsr & ~MMU_MSK);
                }
            }
#endif
        /* overwrite the resource table with read value */
 
        pRsrc->membaseCsr   = membaseCsr;
        pRsrc->iobaseCsr    = iobaseCsr;
        pRsrc->irq          = irq;
        pRsrc->irqvec       = IVEC_TO_INUM(irq);

        }
    else
        {
        /* set memory base address and IO base address */

        pciConfigOutLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                          PCI_CFG_BASE_ADDRESS_0, pRsrc->iobaseCsr |
                          PCI_BASE_IO);
        pciConfigOutLong (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                          PCI_CFG_BASE_ADDRESS_1, pRsrc->membaseCsr);
        pciConfigOutByte (pRsrc->pciBus, pRsrc->pciDevice, pRsrc->pciFunc,
                          PCI_CFG_DEV_INT_LINE, pRsrc->irq);
        }
    
    /* enable mapped memory and IO addresses */

    pciConfigOutWord (pciBus, pciDevice, pciFunc, PCI_CFG_COMMAND,
                      PCI_CMD_IO_ENABLE | PCI_CMD_MASTER_ENABLE);

    /* disable sleep mode */

    pciConfigOutWord (pciBus, pciDevice, pciFunc, PCI_CFG_MODE,
                      SLEEP_MODE_DIS);

    return (OK);
    }

/*******************************************************************************
*
* sysDec21x4xInit - prepare LAN adapter for dec21X4X initialization
*
* This routine is expected to perform any adapter-specific or target-specific
* initialization that must be done prior to initializing the dec21X4X.
*
* The dec21X4X driver calls this routine from the driver endLoad routine before
* any other routines in this library.
*
* This routine returns the base address, the main memory address over the PCI
* bus, the interrupt level and the interrupt vector parameter.
*
* RETURNS: OK or ERROR if the adapter could not be prepared for initialization.
*/
 
STATUS sysDec21x4xInit
    (
    DEC21X4X_DRV_CTRL * pDrvCtrl
    )
    {
    PCI_RESOURCES * pRsrc = &(pciResources [pciDeviceNo]);
 
    /* locate the dec21X4X based adapter. */

    switch (pRsrc->boardType)
        {

        case EB143 :
        case DC140 :
        case DC040 :
            /* set the device control structure */
            
            pDrvCtrl->devAdrs = pRsrc->iobaseCsr + CPU_PCI_IO_ADRS;

            /*
             * pciMemBase is the PCI address that corresponds to physical
             * address 0. Since we are mapping SDRAM address 0 to be the
             * first PCI visible address, we need to subtract out
             * the physical address of the start of SDRAM.
             * Another way to look at this is that even though it falls
             * outside of the PCI window, physical address zero corresponds
             * to SDRAM_PHYS_BASE bytes before the BASE_ADRS.
             */

            pDrvCtrl->pciMemBase = PCI2DRAM_BASE_ADRS - SDRAM_PHYS_BASE;
            pDrvCtrl->ilevel = pRsrc->irq;
            pDrvCtrl->ivec = pRsrc->irqvec;
            break;
            
        default :
            return (ERROR);
 
        }

#if 0
#define	DEC_USR_SF	0x00008000	/* Enable store&forward CSR6:21 */
#define DEC_USR_THR_072	0x00000000      /* Use TxTheshold 72/128 bytes */
#define	DEC_USR_THR_096 0x00010000      /* Use TxTheshold 96/256 bytes */
#define	DEC_USR_THR_128 0x00020000      /* Use TxTheshold 128/512 bytes */
#define	DEC_USR_THR_160 0x00030000      /* Use TxTheshold 160/1024 bytes */
#endif

#ifdef IXP1200_A0
	pDrvCtrl->usrFlags |= DEC_USR_SF;
#else
	pDrvCtrl->usrFlags |= DEC_USR_THR_128;
#endif
    
    return (OK);
    }
 
/*******************************************************************************
*
* sysLanIntEnable - enable dec21X4X interrupts
*
* This routine enables dec21X4X interrupts.  This may involve operations on
* interrupt control hardware.
*
* RETURNS: OK or ERROR for invalid arguments.
*/

STATUS sysLanIntEnable
    (
    int level           /* level number */
    )
    {
    return (intEnable(level));
    }

/*******************************************************************************
*
* sysLanIntDisable - disable dec21X4X interrupts
*
* This routine disables dec21X4X interrupts.  This may involve operations on
* interrupt control hardware.
*
* RETURNS: OK or ERROR for invalid arguments.
*/

STATUS sysLanIntDisable
    (
    int level           /* level number */
    )
    {
    return (intDisable(level));
    }

/*******************************************************************************
*
* sysDec21x4xEnetAddrGet - get Ethernet address
*
* This routine provides a target-specific interface for accessing a
* device Ethernet address.
*
* RETURNS: OK or ERROR if could not be obtained.
*/

STATUS  sysDec21x4xEnetAddrGet
    (
    int         unit,
    char *      enetAdrs
    )
    {
    /*
     * There isn't a target-specific interface for accessing a
     * device Ethernet address.
     */

    return (ERROR);
    }

/*******************************************************************************
*
* sysDec21143MediaSelect - hook routine for dec21x4xStart
*
* This is hook routine for dec21x4xStart(). This routine selects a media for
* dec21143 board.
*
* RETURNS: OK, always.
*/
 
STATUS sysDec21143MediaSelect
    (
    DEC21X4X_DRV_CTRL *      pDrvCtrl,   /* Driver control */
    UINT *                   pCsr6Val    /* CSR6 return value */
    )
    {
    ULONG *     csrReg;
    ULONG       csrData;

    /* set CSR6 value */
 
    *pCsr6Val |= CSR6_2114X_MB1;

    /* Led/Control selection */
 
    csrReg   = (ULONG *)(pDrvCtrl->devAdrs + (CSR15 * DECPCI_REG_OFFSET));
    csrData  = (CSR15_21143_CWE | CSR15_21143_LG3 | CSR15_21143_LG1
                | CSR15_MD_MSK);
    *csrReg  = PCISWAP (csrData);

    csrData = (CSR15_21143_LG3 | CSR15_21143_LG1 | CSR15_MODE_10);
    *csrReg  = PCISWAP (csrData);

    /* force 10Base-T halh duplex configuration */
 
    /* set SIA registers */
 
    csrReg   = (ULONG *)(pDrvCtrl->devAdrs + (CSR14 * DECPCI_REG_OFFSET));
    csrData  = PCISWAP (*csrReg);
    csrData &= ~(CSR14_21143_T4 | CSR14_21143_TXF| CSR14_21143_TXH |
                 CSR14_21143_TAS | CSR14_21143_TH | CSR14_21143_ANE);
    *csrReg = PCISWAP(csrData);
 
    csrReg   = (ULONG *)(pDrvCtrl->devAdrs + (CSR13 * DECPCI_REG_OFFSET));
    csrData  = PCISWAP (*csrReg);
    csrData |= CSR13_SRL_SIA;
    *csrReg  = PCISWAP (csrData);
 
    return (OK);
    }

/*******************************************************************************
*
* sysDec21140MediaSelect - hook routine for dec21x4xStart
*
* This is hook routine for dec21x4xStart(). This routine selects a media for
* dec21140 board.
*
* RETURNS: OK, always.
*/

STATUS sysDec21140MediaSelect
    (
    DEC21X4X_DRV_CTRL *      pDrvCtrl,   /* Driver control */
    UINT *                   pCsr6Val    /* CSR6 return value */
    )
    {

    /* set CSR6 value */

    *pCsr6Val = CSR6_2114X_PS;

    return (OK);
    }

#if (defined(TARGET_CMA220) || defined(TARGET_CMA221) || defined(TARGET_CMA225))
/*******************************************************************************
*
* sysDecPciIntAck - hook routine to acknowledge LAN interrupt
*
* This is a hook routine for the END driver. It's purpose is to perform any
* interrupt acknowledgement that cannot be done in the END driver but which is
* required by the hardware.
*
* RETURNS: N/A.
*/

void sysDecPciIntAck ()
    {
    /*
     * The V3 spec implies that acknowledging this interrupt should
     * not be necessary  and that the IRQ (from the V3) will be deasserted
     * when the PCI interrupt pin becomes deasserted (when the Ethernet
     * controller has its interrupt cleared above). However, without
     * this, the interrupt is never cleared on the Cogent board.
     */

    *(UINT8 *)(V3_BASE + 0x76) = 0;
    }
#endif  /* TARGET_CMA220 || TARGET_CMA221 || TARGET_CMA225 */
#endif /* INCLUDE_DEC */
