/* sysLib.c - Intel ixp1200eb system-dependent routines */

/* Copyright 1997-1998 Wind River Systems, Inc.; Copyright 1999 Intel Corp. */
#include "copyright_wrs.h"

/*
modification history
--------------------
01d,07Mar00,jdg  Added support for B0 revision ixp1200
01c,05Jan00,jdg  Added support for auto PCI address assignment & EEV ethernet
01b,13aug99,jdg  changed name from vbsa1200 to ixp1200eb
01a,19apr99,jdg  created from 01e of brutus.
*/

/*
DESCRIPTION
This library provides board-specific routines for the Intel ixp1200eb board.

It #includes the following chip drivers:
    ixp1200Timer.c - timer drivers
    ixp1200IntrCtl.c -	interrupt controller driver
    ixp1200Sio.c -       Serial I/O driver
    ixp1200Pci.c -       PCI unit driver
if INCLUDE_FLASH is defined, it #includes
    nvRamToFlash.c -	driver to use some Flash like NVRAM
else it #includes
    nullNvRam.c -	dummy NVRAM routines

It #includes the following BSP files:
    sysSerial.c -	serial device initialisation routines
    sysEnd.c -          END network driver support routines


INCLUDE FILES: sysLib.h string.h intLib.h vxLib.h cacheLib.h
muxLib.h config.h private/vmLibP.h arch/arm/mmuArmLib.h dllLib.h
pciIomapLib.h

SEE ALSO:
*/

/* includes */

#include "vxWorks.h"
#include "config.h"

#include "sysLib.h"
#include "intLib.h"
#include "string.h"
#include "vxLib.h"
#include "cacheLib.h"
#include "private/vmLibP.h"
#include "arch/arm/mmuArmLib.h"
#include "muxLib.h"
#include "dllLib.h"
#include "pciIomapLib.h"
#include "drv/mem/flash29.h"
#include "hpcIf.h"
#ifdef INCLUDE_EARLY_SERIAL_DEBUG
#include <stdio.h>
#endif

/* defines */

#ifdef USE_LEDS_FOR_DEBUG
#define DEBUG_LED(val) IXP1200_REG_WRITE(IXP1200_LED_ADDR, (val))
#else
#define DEBUG_LED(val)
#endif

#define MEM_DESC(vaddr,paddr,size,attr) \
    {(void *) (vaddr), (void *) (paddr), ROUND_UP((size), PAGE_SIZE), \
    VM_STATE_MASK_VALID | VM_STATE_MASK_WRITABLE | VM_STATE_MASK_CACHEABLE, \
         (attr)}
#define ATTR_WC (VM_STATE_VALID | VM_STATE_WRITABLE     | VM_STATE_CACHEABLE    )
#define ATTR_RC (VM_STATE_VALID | VM_STATE_WRITABLE_NOT | VM_STATE_CACHEABLE    )
#define ATTR_RM (VM_STATE_VALID | VM_STATE_WRITABLE_NOT | VM_STATE_CACHEABLE_MINICACHE)
#define ATTR_WN (VM_STATE_VALID | VM_STATE_WRITABLE     | VM_STATE_CACHEABLE_NOT)
#define ATTR_RN (VM_STATE_VALID | VM_STATE_WRITABLE_NOT | VM_STATE_CACHEABLE_NOT)
#define ATTR_INV (VM_STATE_VALID_NOT | VM_STATE_WRITABLE_NOT | VM_STATE_CACHEABLE_NOT)

/* imports */

IMPORT char end;		  	   /* end of system, created by ld */
IMPORT VOIDFUNCPTR _func_armIntStackSplit; /* ptr to fn to split stack */
IMPORT UINT32 sysMmuTtbrGet (void);	   /* read MMU TTBR register */
IMPORT void sysToRomInit(int bootType);    /* Branch to romInit after turning off
                                              the MMU */
IMPORT UINT16	sysRead16  (volatile UINT16 * ptr);
IMPORT void	    sysWrite16 (volatile UINT16 * ptr, UINT16 val);
IMPORT UINT32   swap32(UINT32 data);
IMPORT UINT16   swap16(UINT16 data);
IMPORT void     sysIntStackSplit (char *, long);
IMPORT char*    creationDate;
IMPORT int      sysStartType;

IMPORT int      sysCpuVerGet(void);    /* Read CPU HW revision */

/* globals */

#if !defined(INCLUDE_MMU) && \
    (defined(INCLUDE_CACHE_SUPPORT) || defined(INCLUDE_MMU_BASIC) || \
     defined(INCLUDE_MMU_FULL))
#define INCLUDE_MMU		/* should always be defined */
#endif

/*
 * The following variable describes which hardware CPU version we are
 * running on. Currently, the values are:
 *  0: A rev
 *  1: B rev
 *  3: C rev
 */

int sysCpuRev = 0;
int sysCpuId  = 0; /* Same as above but has all 32 bits */

/*
 * The following variable describes what flash mode we are in:
 * 0: 16-bit mode
 * 1: 32-bit mode
 *
 * In 16-bit mode, each flash read reads two 16-values and returns a 32-bit
 * result, whereas each flash write writes one 32-bit value. Note that the
 * flash offsets for reads are doubled by the hardware.
 */

#ifdef SPECTACLE_ISLAND
int flash32BitMode = 0;
#else
int flash32BitMode = 1;
#endif

/*
 * The following structure describes the various different parts of the
 * memory map, to be used during initialisation by vm(Base)GlobalMapInit()
 * when INCLUDE_MMU_BASIC or INCLUDE_MMU_FULL are defined.
 *
 * The following are not the smallest areas that could be allocated for a
 * working system. If the amount of memory used by the page tables is
 * critical, they could be reduced.
 */

PHYS_MEM_DESC sysPhysMemDesc [] =
    {
    /*
     * Address and length parameters must be page-aligned (multiples of 0x1000)
     * or else VxWorks will fail to boot, possibly silently.
     *
     * SDRAM: setup contiguously from 0 upwards.
     *   Start after RESERVED_LOW_MEM bytes, since the area below this is special
     */
    MEM_DESC(LOCAL_MEM_LOCAL_ADRS, SDRAM_VIRT_OFFSET, LOCAL_MEM_SIZE, ATTR_WC),
    /*
     * ROM / Flash: remapped to make room for SDRAM
     *
     * ROM is normally marked as uncacheable by VxWorks. We leave it
     * like that for the time being, even though this has a severe impact
     * on execution speed from ROM.
     *
     * In order to use the Flash as NVRAM, the Flash area must be
     * writable.  Strictly speaking, only the first and last pages need
     * be marked as writable.
     */

#ifdef INCLUDE_FLASH_UTILS

    /*
     * Map early pages, but make them invalid, this is needed to allow
     * for rewriting the entire flash later on.
     */

    MEM_DESC(FLASH_BASE,
             (FLASH_BASE - ROM_VIRT_OFFSET),
             ROM_BASE_ADRS-FLASH_BASE,
             ATTR_INV),
#endif

     MEM_DESC(ROM_BASE_ADRS, (ROM_BASE_ADRS-ROM_VIRT_OFFSET), ROM_SIZE, ATTR_RN),
#ifdef INCLUDE_FLASH

    /* The NVRAM page needs to be writable */

    MEM_DESC(FLASH_ADRS, (FLASH_ADRS-ROM_VIRT_OFFSET), FLASH_SIZE, ATTR_WN),

    /* The following pages need to be writable so that the FLASH29 tests work */

    MEM_DESC(ROUND_DOWN(FLASH29_REG_FIRST_CYCLE, PAGE_SIZE),
             ROUND_DOWN(FLASH29_REG_FIRST_CYCLE, PAGE_SIZE)-ROM_VIRT_OFFSET,
             PAGE_SIZE,
             ATTR_WN),
    MEM_DESC(ROUND_DOWN(FLASH29_REG_SECOND_CYCLE, PAGE_SIZE),
             ROUND_DOWN(FLASH29_REG_SECOND_CYCLE, PAGE_SIZE)-ROM_VIRT_OFFSET,
             PAGE_SIZE,
             ATTR_WN),

#endif

#ifdef INCLUDE_ALT_FLASH

    /*
     * Map address of alternate flash (using CE1) so we can boot from
     * NetROM and rewrite the flash on the board.
     */

    MEM_DESC(ALT_FLASH_BASE, (ALT_FLASH_BASE-ROM_VIRT_OFFSET), ALT_FLASH_SIZE,
             ATTR_RN),
#endif

    /* The rest of memory is flat, but do not map all of it, otherwise
       we will use up too much memory for page tables */
    /* SRAM */
    /* Details: SRAM_POP area is only one page (one word) long.
                The Read-Lock, CAM-Unlock, and Pop areas are ReadOnly */
    MEM_DESC(SRAM_BASE,         SRAM_BASE,         SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_RL_BASE,      SRAM_RL_BASE,      SRAM_SIZE, ATTR_RN),
    MEM_DESC(SRAM_WU_BASE,      SRAM_WU_BASE,      SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_CU_BASE,      SRAM_CU_BASE,      SRAM_SIZE, ATTR_RN),
    MEM_DESC(SRAM_BWC_BASE,     SRAM_BWC_BASE,     SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_BWS_BASE,     SRAM_BWS_BASE,     SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_BTC_BASE,     SRAM_BTC_BASE,     SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_BTS_BASE,     SRAM_BTS_BASE,     SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q0, SRAM_PUSH_BASE_Q0, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q1, SRAM_PUSH_BASE_Q1, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q2, SRAM_PUSH_BASE_Q2, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q3, SRAM_PUSH_BASE_Q3, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q4, SRAM_PUSH_BASE_Q4, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q5, SRAM_PUSH_BASE_Q5, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q6, SRAM_PUSH_BASE_Q6, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_PUSH_BASE_Q7, SRAM_PUSH_BASE_Q7, SRAM_SIZE, ATTR_WN),
    MEM_DESC(SRAM_POP_BASE_Q0,  SRAM_POP_BASE_Q0,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q1,  SRAM_POP_BASE_Q1,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q2,  SRAM_POP_BASE_Q2,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q3,  SRAM_POP_BASE_Q3,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q4,  SRAM_POP_BASE_Q4,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q5,  SRAM_POP_BASE_Q5,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q6,  SRAM_POP_BASE_Q6,  PAGE_SIZE, ATTR_RN),
    MEM_DESC(SRAM_POP_BASE_Q7,  SRAM_POP_BASE_Q7,  PAGE_SIZE, ATTR_RN),

    MEM_DESC(SRAM_CSR_BASE,     SRAM_CSR_BASE,     PAGE_SIZE, ATTR_WN),
    MEM_DESC(SRAM_IO_BASE,      SRAM_IO_BASE,      IO_SIZE,   ATTR_WN),
    
#ifdef LED_MEM_MAPPED
    /* LED */
    MEM_DESC(ROUND_DOWN(IXP1200_LED_ADDR,PAGE_SIZE),
             ROUND_DOWN(IXP1200_LED_ADDR,PAGE_SIZE),
             PAGE_SIZE,
             ATTR_WN),
#endif
    
    /* System Registers */
    MEM_DESC(IXP1200_SYS_REG_BASE, IXP1200_SYS_REG_BASE, IXP1200_SYS_REG_SIZE, ATTR_WN),

    /* ATU registers */
    MEM_DESC(UENG_CSR_BASE,    UENG_CSR_BASE,    UENG_CSR_SIZE,    ATTR_WN),
    MEM_DESC(UENG_XFER_BASE,   UENG_XFER_BASE,   UENG_XFER_SIZE,   ATTR_WN),
    MEM_DESC(FBI_CSR_BASE,     FBI_CSR_BASE,     PAGE_SIZE,        ATTR_WN),
    MEM_DESC(FBI_SCRATCH_BASE, FBI_SCRATCH_BASE, FBI_SCRATCH_SIZE, ATTR_WN),

    /* SDRAM */
    /* SDRAM non-prefetch already done up above, except for low reserved mem */
    MEM_DESC(SDRAM_PHYS_BASE,  SDRAM_PHYS_BASE,  RESERVED_LOW_MEM, ATTR_WN),
    MEM_DESC(SDRAM_PF_BASE,    SDRAM_PF_BASE,    LOCAL_MEM_SIZE,   ATTR_WN),
    MEM_DESC(CACHE_FLUSH_BASE, CACHE_FLUSH_BASE, D_CACHE_SIZE,     ATTR_WC),
    MEM_DESC(MINICACHE_FLUSH_BASE, MINICACHE_FLUSH_BASE, PAGE_SIZE,ATTR_RM),
    MEM_DESC(SDRAM_CSR_BASE,   SDRAM_CSR_BASE,   PAGE_SIZE,        ATTR_WN),

    /* PCI Unit */
    MEM_DESC(PCI_CSR_BASE, PCI_CSR_BASE, PAGE_SIZE, ATTR_WN),

#ifdef INCLUDE_PCI
    MEM_DESC(CPU_PCI_CNFG_ADRS, CPU_PCI_CNFG_ADRS, CPU_PCI_CNFG_SIZE, ATTR_WN),
#if (PCI_MAX_BUS > 1)
    MEM_DESC(CPU_PCI_CNFG_1_ADRS, CPU_PCI_CNFG_1_ADRS, CPU_PCI_CNFG_1_SIZE, ATTR_WN),
#endif

#ifndef AUTO_PCI_CONFIG
    /* 1 page in PCI I/O Space for ethernet card CSRs */
    MEM_DESC(CPU_PCI_IO_ADRS, CPU_PCI_IO_ADRS, CPU_PCI_IO_SIZE, ATTR_WN),
#endif

#ifdef INCLUDE_EXTRA_PCI_MAPPING
	/* For PCI device driver -- drj: Oct 11,99 */
    MEM_DESC(CPU_PCI_MEM_ADRS, CPU_PCI_MEM_ADRS, CPU_PCI_MEM_SIZE, ATTR_WN),
#endif

#endif /* ifdef INCLUDE_PCI */

    /* Allocate space for sysMmuMapAdd */
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    MEM_DESC(NULL, NULL, 0, 0),
    };

int sysPhysMemDescNumEntMax = NELEMENTS (sysPhysMemDesc);
int sysPhysMemDescNumEnt = NELEMENTS (sysPhysMemDesc);

int	sysBus	    = BUS;		/* system bus type (VME_BUS, etc) */
int	sysCpu	    = CPU;		/* system cpu type (ARMSA110) */
char *	sysBootLine = BOOT_LINE_ADRS; 	/* address of boot line */
char *	sysExcMsg   = EXC_MSG_ADRS;	/* catastrophic message area */
int	sysProcNum;			/* processor number of this cpu */
int	sysFlags;			/* boot flags */
char	sysBootHost [BOOT_FIELD_LEN];	/* name of host from which we booted */
char	sysBootFile [BOOT_FIELD_LEN];	/* name of file from which we booted */


/* locals */

static int sys_timer_clk = SYS_TIMER_CLK_DEFAULT;

/* forward declarations */

char * sysPhysMemTop (void);
int sysLedsReadWrite (int and, int eor);
#ifdef INCLUDE_EEV
void syseeVPciInit (void);
#endif

STATUS sysMmuMapAdd(void *addr, UINT len, UINT mask, UINT state);
void * sysVirtToPhys(void *virtAddr);
void * sysPhysToVirt(void *physAddr);

/* included source files */

#ifdef INCLUDE_EARLY_SERIAL_DEBUG
#include "dbgSerUtils.c"
#endif

#include "sysSerial.c"
#include "ixp1200Timer.c"  /* was "timer/ambaTimer.c" */
#include "ixp1200IntrCtl.c"
#include "ixp1200Sio.c"
#include "ixp1200Pci.c"
#ifdef INCLUDE_FLASH
#include "mem/nvRamToFlash.c"
#else
#include "mem/nullNvRam.c"
#endif

#include "pciIomapLib.c"
#include "pciIomapShow.c"

#ifdef INCLUDE_END
#include "sysEnd.c"
#endif
#ifdef INCLUDE_EEV
#include "sysNetIf.c"
#endif

#include "ixp1200HPC.c"
#ifdef INCLUDE_HPC_SERIAL
#include "HPCSerial.c"
#endif

/*******************************************************************************
*
* sysModel - return the model name of the CPU board
*
* This routine returns the model name of the CPU board.
*
* RETURNS: A pointer to a string identifying the board and CPU.
*/

char * sysModel (void)
    {
#ifdef SPECTACLE_ISLAND
    return "Intel Spectacle Island - ARM IXP1200";
#else
    return "Intel ixp1200eb - ARM IXP1200";
#endif
    }

/*******************************************************************************
*
* sysBspRev - return the BSP version with the revision eg 1.1/<x>
*
* This function returns a pointer to a BSP version with the revision.
* for eg. 1.1/<x>. BSP_REV is concatenated to BSP_VERSION to form the
* BSP identification string.
*
* RETURNS: A pointer to the BSP version/revision string.
*/

char * sysBspRev (void)
    {
    return (BSP_VERSION BSP_REV);
    }

/*******************************************************************************
*
* ixpVersion - return the Intel IXP version
*
* This function returns a pointer to a BSP version with the revision.
* This is the Intel IXP version in the form "major.minor.build_num".
*
* RETURNS: A pointer to the BSP version/revision string.
*/

char * ixpVersion (void)
    {
    return (IXP_VERSION);
    }

/*******************************************************************************
*
* sysCpuIDGet - return the CPU ID
*
* This function returns the CPU ID. The low order 4-bits gives the revision
* number.
*
* RETURNS: CPU ID
*/

int sysCpuIDGet (void)
    {
    return (sysCpuId);
    }

/*******************************************************************************
*
* sysBoardIDGet - return the Board ID
*
* This function returns the Board ID. The low order 4-bits gives the revision
* number.
*
* RETURNS: Board ID
*/

int sysBoardIDGet (void)
    {
#ifdef SPECTACLE_ISLAND
    return (0x8086FFFF);
#else
    return (0x80860001);
#endif
    }

/*******************************************************************************
*
* sysTorCommMethod - return method tornado is using for communications
*
* This function returns the method that tornado will use for communications.
*
* RETURNS: comm method:
*          0: Unknown
*          1: Serial Port
*          2: ethernet
*          3: HPC
*/

int sysTorCommMethod (void)
    {
#if (WDB_COMM_TYPE == WDB_COMM_SERIAL)
# if (WDB_TTY_CHANNEL == 0)
    return 1;
# else
    return 3;
# endif
#elif (WDB_COMM_TYPE == WDB_COMM_NETWORK)
    return 2;
#else
    return 0;
#endif
    }

/******************************************************************************
*
* sysVirtToPhys - translate a virtual address to a physical address (ARM)
*
* This function converts a virtual address (of part of a page table in
* DRAM) to a physical address. This routine is used both by the BSP MMU
* initialisation and by the vm(Base)Lib code later (via function
* pointer). The mapping set up in the initial BSP MMU initialisation is
* the same as that done later and that done previously by Angel.
*
* RETURNS: the physical adddress
*/

void * sysVirtToPhys
    (
    void *	virtAddr	/* virtual address in DRAM */
    )
    {
    if (((UINT32) virtAddr) < ROM_VIRT_OFFSET)
        {

        /* SDRAM */

        return (void *) (((UINT32) virtAddr) + SDRAM_VIRT_OFFSET);
        }
    else if (((UINT32) virtAddr) < SRAM_BASE)
        {

        /* Flash */

        return (void *) (((UINT32) virtAddr) - ROM_VIRT_OFFSET);
        }
    else
        {
        return virtAddr;
        }
    }

/******************************************************************************
*
* sysPhysToVirt - translate a physical address to a virtual address (ARM)
*
* This function converts a physical address (of part of a page table in
* DRAM) to a virtual address. This routine is used by the vm(Base)Lib
* code later (via function pointer).
*
* RETURNS: the virtual adddress
*/

void * sysPhysToVirt
    (
    void *	physAddr	/* physical address in DRAM */
    )
    {
    if (((UINT32) physAddr) < SRAM_BASE)
        {

        /* Flash */

        return (void *) (((UINT32) physAddr) + ROM_VIRT_OFFSET);
        }
    else if ((((UINT32) physAddr) >= SDRAM_VIRT_OFFSET) &&
             (((UINT32) physAddr) <  SDRAM_PF_BASE))
        {

        /* SDRAM */

        return (void *) (((UINT32) physAddr) - SDRAM_VIRT_OFFSET);
        }
    else
        {
        return physAddr;
        }
    }


/******************************************************************************
*
* sysHwInit0 - Intialize _func_armVirtToPhys before sysHwInit is called
*
* This function initialized _func_armVirtToPhys and _func_armPhysToVirt
* before sysHwInit is called. It is called from usrInit in bootConfig.c
* before cacheLibInit is called.
*
* RETURNS: N/A
*/

void sysHwInit0()
    {
#ifndef TOR201
    _func_armVirtToPhys = sysVirtToPhys;
    _func_armPhysToVirt = sysPhysToVirt;
#else
#if defined(INCLUDE_CACHE_SUPPORT) || defined(INCLUDE_MMU)
    /*
     * Install the appropriate cache libary, using the our
     * address translation routines.
     */

    cacheArmSA1100LibInstall(sysPhysToVirt, sysVirtToPhys);
#endif
#ifdef INCLUDE_MMU
    /* Install the appropriate MMU library and translation routines */

    mmuArmSA1100LibInstall (sysPhysToVirt, sysVirtToPhys);
#endif /* INCLUDE_MMU */
#endif
    }

/*******************************************************************************
*
* sysHwInit - initialise the CPU board hardware
*
* This routine initialises various features of the hardware.
* Normally, it is called from usrInit() in usrConfig.c.
*
* NOTE: This routine should not be called directly by the user.
*
* RETURNS: N/A
*/

void sysHwInit (void)
    {
    UINT32 gpio_val, gpio_en;

    /* install the IRQ/SVC interrupt stack splitting routine */

    _func_armIntStackSplit = sysIntStackSplit;

    DEBUG_LED(6);


    /* Initialize sysCpuRev */

    sysCpuId = sysCpuVerGet();
    sysCpuRev = sysCpuId & 0xF;


    /* Initialize flash32BitMode (equals ~ GPIO[3]) */
    IXP1200_REG_READ(IXP1200_GPIO_EN, gpio_en);
    IXP1200_REG_WRITE(IXP1200_GPIO_EN, GPIO_OWN_CCCC | GPIO_OUTPUT(0));
    IXP1200_REG_READ(IXP1200_GPIO_DATA, gpio_val);
    IXP1200_REG_WRITE(IXP1200_GPIO_EN, gpio_en);
    if ((gpio_val & (1<<3)) && (sysCpuRev > 0))
        flash32BitMode = 0;
    else
        flash32BitMode = 1;


    /* Initialize sysPhysMemDescNumEnt */

    while (sysPhysMemDesc[sysPhysMemDescNumEnt-1].len == 0)
        sysPhysMemDescNumEnt--;


    /* initialize HPC */
#ifdef INCLUDE_HPC
    if (!(sysStartType & BOOT_DONT_INIT_HPC)) {
        hpcMemInit();
    }
#else
    hpcMemDummyInit();
#endif


    /* initialize PCI unit */

#ifndef SPECTACLE_ISLAND
    if (!(sysStartType & BOOT_DONT_INIT_PCI))
#endif
        {
        sysPciInit();
        }

    DEBUG_LED(0xB);

#if defined (INCLUDE_SERIAL)
    /* initialise the appropriate serial devices */

    sysSerialHwInit ();      /* initialise serial data structure */
#else

    /* Quiet UART Interrupts */

    IXP1200_REG_WRITE(IXP1200_UART_CR, 0);
#endif


    /* Set timer speed based on Chip Rev */
    if (sysCpuRev < 1) {
        sys_timer_clk = SYS_TIMER_CLK_A;
#if 0
    } else if (sysCpuRev > 1) {
        sys_timer_clk = SYS_TIMER_CLK_C;
#endif
    }
    

#ifdef INCLUDE_EARLY_SERIAL_DEBUG
    dbgSerInit();

	dbgSerOutstr("BSP IXP Version: ");
	dbgSerOutstr(ixpVersion());
	dbgSerOutstr(", built on ");
	dbgSerOutstr(creationDate);
	dbgSerOutstr("\r\n");
    {
        char buf[16];
        sprintf(buf,"%08X",sysCpuId);
        dbgSerOutstr("CPU ID: ");
        dbgSerOutstr(buf);
        dbgSerOutstr("\r\n");
        dbgSerOutstr("CPU Speed: ");
        if (sysCpuRev < 1) {
            dbgSerOutstr(CLOCK_SPEED_STRING_A);
#if 0
        } else if (sysCpuRev > 1) {
            dbgSerOutstr(CLOCK_SPEED_STRING_C);
#endif
        } else {
            dbgSerOutstr(CLOCK_SPEED_STRING);
        }
        dbgSerOutstr(" MHz\r\n");
    }
    dbgSerOutstr("Early serial debug initialized\r\n");
#endif


#ifndef TOR201
#if defined(INCLUDE_MMU_BASIC) || defined(INCLUDE_MMU_FULL)

    /* initialise function pointers used by mmuLib */

    _func_armVirtToPhys = sysVirtToPhys;
    _func_armPhysToVirt = sysPhysToVirt;

    cacheLib.dmaPhysToVirtRtn	= (FUNCPTR) _func_armPhysToVirt;
    cacheLib.dmaVirtToPhysRtn	= (FUNCPTR) _func_armVirtToPhys;
#endif
#endif

#if defined (INCLUDE_PCI)

#if (_BYTE_ORDER == _BIG_ENDIAN)
    if (sysCpuRev < 1) {
        dbgSerOutstr("Big Endian support requires CPU revision 1 or higher.\n\r");
        sysToMonitor(BOOT_NO_AUTOBOOT);
    }
#endif

    /* Initialize PCI driver library */

    if (pciIomapLibInit (PCI_MECHANISM_3,
                         PCI_CONFIG0_BASE,
                         PCI_CONFIG1_BASE,
                         0) != OK)
        sysToMonitor(BOOT_NO_AUTOBOOT);


#if defined (AUTO_PCI_CONFIG)
    sysPciAssignAddrs();
#endif 

#if defined (INCLUDE_DEC)

    /* Map/configure DEC21X4X */

    sysDec21x4xPciInit();

#endif  /* INCLUDE_DEC */

#if defined (INCLUDE_EEV)
    /* We must assign PCI addresses before this */
    syseeVPciInit ();
#endif
#endif  /* INCLUDE_PCI */

    /* Quiet PCI Interrupts */

    IXP1200_REG_WRITE(IXP1200_IRQENABLECLEAR, 0xFFFFFFFF);
    IXP1200_REG_WRITE(IXP1200_FIQENABLECLEAR, 0xFFFFFFFF);
    
    /* Quiet FBI Interrupts */

    IXP1200_REG_WRITE(IXP1200_IREG, 0);

    /* Quiet RTC Interrupts */

    IXP1200_REG_WRITE(IXP1200_RTC_DIV, RTCDIV_RESET_VAL);

    DEBUG_LED(7);

    return;
    }

/*******************************************************************************
*
* sysHwInit2 - additional system configuration and initialisation
*
* This routine connects system interrupts and does any additional
* configuration necessary.
*
* RETURNS: N/A
*
* NOMANUAL
*/

void sysHwInit2 (void)
    {
    /* initialise the interrupt library and interrupt driver */

    DEBUG_LED(8);

    intLibInit (IXP1200_INT_NUM_LEVELS, IXP1200_INT_NUM_VECTORS, INT_MODE);
    ixp1200IntDevInit();

    DEBUG_LED(9);


    /* connect sys clock interrupt and auxiliary clock interrupt */

    intConnect (INUM_TO_IVEC (SYS_TIMER_INT_LVL), sysClkInt, 0);
    intConnect (INUM_TO_IVEC (AUX_TIMER_INT_LVL), sysAuxClkInt, 0);


    /* Fix B0 PCI problem */
#if defined(INCLUDE_PCI) && defined(INCLUDE_HPC)
    if (sysCpuRev == 1)
        sysPciFix();
#endif


    /* 
     * Initialize HPC driver
     * Note that this must occur before sysSerialHwInit2 if HPC serial
     * ports are being used.
     */
#ifdef INCLUDE_HPC
    if (!(sysStartType & BOOT_DONT_INIT_HPC)) {
        IXP1200_REG_WRITE(IXP1200_DBELL_PCI_MASK, 0);
        IXP1200_REG_WRITE(IXP1200_DBELL_SA_MASK,  0);
    
        intConnect((void*)INT_VEC_DFH, pciInt, INT_VEC_DFH);
        intEnable(INT_VEC_DFH);

        hpcSysInit(0);
    } else {
        intConnect((void*)INT_VEC_DFH, pciInt, INT_VEC_DFH);
        intEnable(INT_VEC_DFH);

        hpcSysInit(1);
    }
#endif
    
#if defined (INCLUDE_SERIAL)
    /* connect serial interrupt */

    sysSerialHwInit2();
#endif

#if defined(INCLUDE_MMU_BASIC) || defined(INCLUDE_MMU_FULL)

    /* These need to be set up for the case of a bootROM image
     * where the standard libraries do not set these values, but
     * the ethernet controller (dec21x4x) needs them.
     */

    cacheDmaFuncs.virtToPhysRtn = (FUNCPTR) sysVirtToPhys;
    cacheDmaFuncs.physToVirtRtn = (FUNCPTR) sysPhysToVirt;
#endif

    /* Initialize other drivers here */
#if defined (INCLUDE_PCI)
    if (sysPciIsCentralFunc()) {
        intConnect((void*)INT_VEC_PIL, pciInt, INT_VEC_PIL);
        intEnable(INT_VEC_PIL);
    }
#endif

    /* DEBUG_LED(0xA); */
    sysLedsReadWrite(0x0,0xA); /* Set LED to A */

    return;
    }

/*******************************************************************************
*
* sysPhysMemTop - get the address of the top of physical memory
*
* This routine returns the address of the first missing byte of memory,
* which indicates the top of memory.
*
* Normally, the user specifies the amount of physical memory with the
* macro LOCAL_MEM_SIZE in config.h.  BSPs that support run-time
* memory sizing do so only if the macro LOCAL_MEM_AUTOSIZE is defined.
* If not defined, then LOCAL_MEM_SIZE is assumed to be, and must be, the
* true size of physical memory.
*
* NOTE: Do not adjust LOCAL_MEM_SIZE to reserve memory for application
* use.  See sysMemTop() for more information on reserving memory.
*
* RETURNS: The address of the top of physical memory.
*
* SEE ALSO: sysMemTop()
*/

char * sysPhysMemTop (void)
    {
    static char * physTop = NULL;

    if (physTop == NULL)
        {
#ifdef LOCAL_MEM_AUTOSIZE
        
	/* If auto-sizing is possible, this would be the spot.  */

#	error   "Dynamic memory sizing not supported"

#else
	/* Don't do autosizing, if size is given */

	physTop = (char *)(LOCAL_MEM_LOCAL_ADRS + LOCAL_MEM_SIZE);

#endif /* LOCAL_MEM_AUTOSIZE */
	}

    return physTop;
    }

/*******************************************************************************
*
* sysMemTop - get the address of the top of VxWorks memory
*
* This routine returns a pointer to the first byte of memory not
* controlled or used by VxWorks.
*
* The user can reserve memory space by defining the macro USER_RESERVED_MEM
* in config.h.  This routine returns the address of the reserved memory
* area.  The value of USER_RESERVED_MEM is in bytes.
*
* RETURNS: The address of the top of VxWorks memory.
*/

char * sysMemTop (void)
    {
    static char * memTop = NULL;

    if (memTop == NULL)
	{
	memTop = sysPhysMemTop () - USER_RESERVED_MEM;
	}

    return memTop;
    }

/*******************************************************************************
*
* sysToMonitor - transfer control to the ROM monitor
*
* This routine transfers control to the ROM monitor.  It is usually called
* only by reboot() -- which services ^X -- and bus errors at interrupt
* level.  However, in some circumstances, the user may wish to introduce a
* new <startType> to enable special boot ROM facilities.
*
* RETURNS: Does not return.
*/

STATUS sysToMonitor
    (
    int startType   /* passed to ROM to tell it how to boot */
    )
    {
    UINT32 *    p = (UINT32 *)ROM_TEXT_ADRS;
#if defined(INCLUDE_DEC) || defined(INCLUDE_EEV)
    END_OBJ * pEnd;
#endif

    DEBUG_LED(4);

#if defined(INCLUDE_SERIAL)
    sysSerialReset ();	/* put serial devices into quiet state */
#endif

    /* Reset microengines and macs */

    IXP1200_REG_WRITE(IXP1200_RESET, RESET_MICROENGINE | RESET_EXTERNAL);

#if defined(INCLUDE_DEC)
    /*
     * If the Ethernet driver were left alone, the controller chip might
     * alter memory, so we stop the controller before jumping to the bootrom.
     */

    pEnd = endFindByName ("dc", 0);

    if (pEnd != NULL)
	pEnd->pFuncTable->stop(pEnd->devObject.pDevice);
#endif /* INCLUDE_DEC */
#if defined(INCLUDE_EEV)
    /*
     * If the Ethernet driver were left alone, the controller chip might
     * alter memory, so we stop the controller before jumping to the bootrom.
     */

    pEnd = endFindByName ("eeE", 0);

    if (pEnd != NULL)
	pEnd->pFuncTable->stop(pEnd->devObject.pDevice);
#endif /* INCLUDE_EEV */

    /*
     * Examine ROM - if it's a VxWorks boot ROM, jump to the warm boot
     * entry point; otherwise hang (we don't attempt to remap the machine
     * to find the Primary Boot Loader).
     * A VxWorks boot ROM begins
     *    MOV   R0,#BOOT_COLD
     *    B     ...
     *    DCB   "Copyright"
     * We check the first and third words only. This could be tightened up
     * if required (see romInit.s).
     *
     * Note that the MMU must stay on.
     */

    /*
     * Note that only p[2] is possibly swapped because it is byte data and
     * not longword data.
     */
    if (p[0] != 0xE3A00002 || p[2] != SWAP32PCI(0x79706F43))
	{
	while (TRUE)
	    /* do nothing */ ;
	}

    /* Actually jump to ROM, first we have to turn off the MMU */

    /* transition routine address */

    sysToRomInit(startType);			/* jump to bootrom */

    return OK;		/* in case we ever continue from ROM monitor */
    }

/****************************************************************************
*
* sysProcNumGet - get the processor number
*
* This routine returns the processor number for the CPU board, which is
* set with sysProcNumSet().
*
* RETURNS: The processor number for the CPU board.
*
* SEE ALSO: sysProcNumSet()
*/

int sysProcNumGet (void)
    {
    return 0;
    }

/****************************************************************************
*
* sysProcNumSet - set the processor number
*
* Set the processor number for the CPU board.  Processor numbers should be
* unique on a single backplane.
*
* NOTE
* By convention, only processor 0 should dual-port its memory.
*
* RETURNS: N/A
*
* SEE ALSO: sysProcNumGet()
*/

void sysProcNumSet
    (
    int procNum		/* processor number */
    )
    {
    sysProcNum = procNum;
    }

/******************************************************************************
*
* sysLedsReadWrite - read/write the state of the LEDs on the IXP1200EB board
*
* This routine can be used to read and write the state of the debug LEDs
* on the IXP1200EB board. The current state is ANDed and EORed with the
* supplied parameters. Bits 0..3 control the LEDs. The current state appears
* as a  hexadecimal digit on the LED.
*
* RETURNS: previous state of LEDs.
*/

int sysLedsReadWrite
    (
    int and,
    int eor
    )
    {
    static UINT8 current = 0;
    int oldLevel, previous;

    oldLevel = intLock ();	 /* no interrupts during read-modify-write */

    /*
     * Read current state, remove old bit and insert new one.
     * Note the extra '^', necessary because a 1 => OFF, 0 => ON
     */

    previous = current;
    current = ((current & and) ^ eor) & 0xF;

    IXP1200_REG_WRITE(IXP1200_LED_ADDR, current);

    intUnlock (oldLevel);

    return previous;
    }

/*******************************************************************************
*
* sysMmuMapAdd - Add entries to sysPhysMemDesc
*
* This routine adds mapping entries to sysPhysMemDesc.
*
* Add the entry {addr, addr, len, mask, state} to sysPhysMemDesc.
*
* RETURNS: 
*   OK or ERROR 
*/

STATUS sysMmuMapAdd
    (
    void *addr,      /* Virtual and physical address of region */
    UINT len,        /* Length of region */
    UINT mask,       /* Initial State mask */
    UINT state       /* Initial State */
    )
{
    if (sysPhysMemDescNumEnt < sysPhysMemDescNumEntMax) {
        sysPhysMemDesc[sysPhysMemDescNumEnt].virtualAddr = addr;
        sysPhysMemDesc[sysPhysMemDescNumEnt].physicalAddr = addr;
        sysPhysMemDesc[sysPhysMemDescNumEnt].len = ROUND_UP(len, PAGE_SIZE);
        sysPhysMemDesc[sysPhysMemDescNumEnt].initialStateMask = mask;
        sysPhysMemDesc[sysPhysMemDescNumEnt].initialState = state;
        sysPhysMemDescNumEnt++;
        return OK;
    } else {
        return ERROR;
    }
}

/*******************************************************************************
*
* sysMemVirtSize - get sizes of virtual memory regions
*
* This routine returns sizes of virtual memory regions.
*
* SYS_ALL is the actual amount of memory mapped to virtual address
* zero from the SDRAM. Some amount of the top end of this may be
* reserved by users, so the amount actually available to VxWorks
* is given by SYS. The size SDRAM which is available to the HW and 
* not used by VxWorks (mapped with virt addr == phys addr) is given
* by SDRAM.
*
* RETURNS: N/A
*/

void sysMemVirtSize
    (
    unsigned int* sys,       /* Size of memory mapped at zero for VxWorks */
    unsigned int* sys_all,   /* Size of memory mapped at zero and usable by VxWorks */
    unsigned int* sdram,     /* Size of memory mapped at C0000000 and usable by HW */
    unsigned int* sram,      /* Size of sram */
    unsigned int* flash      /* Size of Flash */
    )
{
    /* Size of memory mapped at zero for VxWorks */
    *sys_all = LOCAL_MEM_SIZE; 

    /* Size of memory mapped at zero and usable by VxWorks */
    *sys = LOCAL_MEM_SIZE - USER_RESERVED_MEM;

    /* Size of memory mapped at C0000000 and usable by HW */
    *sdram = RESERVED_LOW_MEM;

    /* Size of sram */
    *sram = SRAM_SIZE;

    /* Size of Flash */
    *flash = (ROM_BASE_ADRS + ROM_SIZE_TOTAL - ROM_VIRT_OFFSET);
}
