/* devconf.h
 *
 *------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------
 * system: SA1200
 * subsystem: Angel Debug Monitor
 * author: Henry Qian 01/29/99
 * revisions:
 *
 */

#ifndef angel_brutus_devconf_h
#define angel_brutus_devconf_h

#include "stacks.h"

#define CACHE_SUPPORTED    1
#define PROFILE_SUPPORTED  0
#define SYSTEM_COPROCESSOR_SUPPORTED	1

#define ENABLE_MMU         1
#define ENABLE_CLOCK_SWITCHING 1

/*
 * ring buffer size
 *
 * NOTE - the ring buffer macros are more efficient
 * when ring buffer size is declared as a power of 2
 */

#define RING_SIZE             256

#define RING_LOW_WATER_MARK   (RING_SIZE / 4)   /* when to fill tx buffer */
#define RING_HIGH_WATER_MARK  (RING_SIZE / 2)   /* when to empty rx buffer */

/*
 * Choose the method for debug output here.  Options supported for PIE are:
 *      panicblk        panic string written to RAM
 *      pidulate        via ARMulator (during emualtion only)
 *      logadp          via ADP to host
 */
#ifdef DEBUG
# define DEBUG_METHOD logterm
#endif

/*************************************************************************/
/**                                                                     **/
/**     These defines determine whether Angel uses FIQs or IRQs         **/
/**                                                                     **/
/*************************************************************************/
#define HANDLE_INTERRUPTS_ON_IRQ 1
#define HANDLE_INTERRUPTS_ON_FIQ 0

/* See serlock.h for the possible settings */
#define FIQ_SAFETYLEVEL FIQ_CannotBeOptimised

/* control whether the serial device is raw or Angel/shared */
#if !defined(MINIMAL_ANGEL) || MINIMAL_ANGEL == 0
#  define RAW_EBS_SERIAL 0      /* change to 1 if should always be raw */
#else
#  define RAW_EBS_SERIAL 1
#endif

/* the default active device */
#define DE_DEFAULT_ACTIVE DI_SERIAL

/* The SA1200 memory map is structured as follows. */
#include "platform.h"

/*
 * The ARM 7 PIE and EBSA110 allow unrestricted access to the entire memory space.
 * To stop the debugger acessing areas it should these two macros are used
 * to determine if an access should occur.

 * The ARM 6 PIE card prohibits user mode code from reading the I/O space.
 * So these macros are superfluous in that case.
*/
#define READ_PERMITTED(__addr__) (1)
#define WRITE_PERMITTED(__addr__) (1)
/*
#define WRITE_PERMITTED(__addr__) (((__addr__) < IOBase) || ((__addr__) > (IOBase + IOSize)))
*/

/* This can be overridden by the memory sizer - if present */
#define Angel_DefaultTopOfMemory (DRAM_BASE + DRAM_SIZE)

/* This should be set to 1 for systems with RAM slots */
#define MEMORY_SIZE_MAY_CHANGE_DYNAMICALLY 0

/* The following are the sizes of the various Angel stacks */

#define Angel_UNDStackSize      0x0100
#define Angel_ABTStackSize      0x0100
#define Angel_AngelStackSize    0x4000
#define Angel_SVCStackSize      0x1000

# if HANDLE_INTERRUPTS_ON_IRQ
#  define Angel_IRQStackSize     0x8000
# else
#  define Angel_IRQStackSize     0x0100
# endif

# if HANDLE_INTERRUPTS_ON_FIQ
#  define Angel_FIQStackSize     0x0800
# else
#  define Angel_FIQStackSize     0x0100
# endif

/* for the fatal error handler (only!) */
#define Angel_FatalStackSize    0x0400

/* This is set to 1 only if the system does not remap ROM to 0
 * on a reset
 */
#define BRANCH_TO_HIGH_ROM_ON_RESET 0

#define Angel_StacksAreRelativeToTopOfMemory 1

/*
** Allocate space for MMU page tables - put them at the top of memory
*/
#define Angel_PageTableAreaBaseOffsetFromTopOfMemory (-(TTB_SIZE))

/*
** Angel stacks will be right below MMU page tables
*/
# define Angel_StackBaseOffsetFromTopMemory  \
        Angel_PageTableAreaBaseOffsetFromTopOfMemory + \
        (-(Angel_CombinedAngelStackSize))

/*
** The application stack will sit below the Angel stacks.
*/
#define Angel_ApplStackOffset Angel_StackBaseOffsetFromTopMemory
#define Angel_ApplStackSize        0x2000
#define Angel_ApplStackLimitOffset (Angel_ApplStackOffset - Angel_ApplStackSize)

/* Profiling is not currently supported, so no need to allocate some
** space for profiling data
*/

/*
** Need room for ethernet frames and data
** Each frame is allowed 1600 bytes, then number
** of frames in the device layer being defined in the
** (driver).c files, eg dec21040.c - double this at least
** Then there are the table in socket.c for tracking the 
** frames and sockets
*/

/*
 * By default the heap starts at the end of the ROM image and the heap
 * limit is the bottom of the application stack.  These are the
 * defaults unless these two #defines are turned on in which case the
 * addresses specified are used by the C Library.
 */
#if 0
# define Angel_ApplHeapLimit  ???
# define Angel_ApplHeap       ???
#endif

/*
 * where to download a new agent to
 */
#define Angel_DownloadAgentArea        0x8000

/*
 * This constant should be changed only with great care.  It is the
 * amount of empty stack space left between two callback's stacks on
 * the AngelStack.
 */
#define Angel_AngelStackFreeSpace 0x400

/*
        Keep this value low enough to fit in one packet
        Which file should this be in - Not here for sure
*/
/*---------------------------------------------------------------------------*/
/*-- I/O memory map ---------------------------------------------------------*/
/*---------------------------------------------------------------------------*/

/*
 * the devices available
 * ORDER AND NUMBER MUST MATCH angel_Device[] TABLE IN devices.c
 */
typedef enum DeviceIdent
{
  DI_SERIAL = 0,
  DI_NUM_DEVICES
} DeviceIdent;

/* the device which controls the LED's */
#define DI_LED_DEVICE DI_SERIAL

/* 
 * the Interrupt handlers available 
 * ORDER AND NUMBER MUST MATCH angel_IntHandler[] in devices.c
 */
typedef enum IntHandlerID
{
    IH_SERIAL = 0
} IntHandlerID;

#define DE_NUM_INT_HANDLERS 1   /* must be no. of entries in IntHandlerID */


/* 
 * the POLL handlers available 
 * ORDER AND NUMBER MUST MATCH angel_PollHandler[] in devices.c
 */
#define DE_NUM_POLL_HANDLERS 0

#endif

 /* EOF devconf_h */

