/* serial.c
 *
 *------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------
 * system: SA1200
 * subsystem: Angel Debug Monitor
 * author: Henry Qian 01/29/99
 * revisions:
 *
 * Dan White 03/21/00 add B0 changes
 */

#include "serial.h"             /* header for this file */

#include "devdriv.h"            /* device driver support */
#include "platform.h"

#include "params.h"             /* parameter structures and utilities */
#include "rxtx.h"               /* rx/tx packet engines */
#include "support.h"            /* __rt_[en/dis]ableIRQ() */
#include "logging.h"            /* ERROR(), etc. */
#include "serlock.h"            /* serialisation, etc. */
#include "serring.h"            /* interface to high-level driver */

/* General purpose constants, macros, enums, typedefs */

#define UART_DR_BASE ((volatile unsigned int*)UART_DR)
#define UART_CR_BASE ((volatile unsigned int*)UART_CR)
#define UART_SR_BASE ((volatile unsigned int*)UART_SR)

/* Forward declarations */

static void serial_ControlTx(DeviceID devid);
static void serial_ControlRx(DeviceID devid);
DevError serial_Control(DeviceID devid, 
                        DeviceControl op, void *arg);
static void serial_KickStartFn(DeviceID devid);

static DevError serial_set_params(DeviceID devid, const ParameterConfig *conf);

/*
 * The set of parameter options supported by the device
 */
static unsigned int baud_options[] = { 9600, 19200, 38400, 57600, 115200 };

static ParameterList param_list[] = {
    {
        AP_BAUD_RATE,
        sizeof(baud_options) / sizeof(baud_options[0]),
        baud_options
    }
};

/* 
 *The default parameter config for the device
 */
static Parameter param_default[] = {
    { AP_BAUD_RATE, 9600 }
};

/* the state shared between the interrupt handler and the deferred processor */
static struct RingBuffer serial_rx_ring;
static struct RingBuffer serial_tx_ring;

#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0) /* Angel/shared */

/* the configuration needed by the TX and RX engines */

#define SERIAL_FC_SET  ((1<<serial_XON)|(1<<serial_XOFF))
#define SERIAL_CTL_SET ((1<<serial_STX)|(1<<serial_ETX)|(1<<serial_ESC))
#define SERIAL_ESC_SET (SERIAL_FC_SET|SERIAL_CTL_SET)

static const struct re_config engine_config = {
    serial_STX,
    serial_ETX,
    serial_ESC,         /* self-explanatory? */
    SERIAL_FC_SET,      /* set of flow-control characters  */
    SERIAL_ESC_SET,     /* set of characters to be escaped */
    serpkt_flow_control,
    (void *)DI_SERIAL,  /* what to do with FC chars */
    angel_DD_RxEng_BufferAlloc,
    (void *)DI_SERIAL   /* how to get a buffer */
};

/* the state of the rx engine */
static struct re_state rx_engine_state;

/* the state of the tx engine */
static struct te_state tx_engine_state;

/* packet for actual rx in progress */
static struct data_packet  rx_packet;

/* the current write packet */
static struct data_packet  tx_packet;

/* collected TX and RX engine state */
static RxTxState serial_rx_tx_state = 
{
    &engine_config,
    &rx_engine_state,
    &tx_engine_state,
    &rx_packet,
    &tx_packet,
    1 /* rx_flow */
};

#else  /* raw */

static RawState serial_raw_state;

#endif /* ... else raw ... */

/*
 * The control functions and interface
 */
static const SerialControl serial_ctrl = 
{
#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0)
    &serial_rx_tx_state,
    NULL,
    serpkt_int_tx_processing,
    serpkt_int_rx_processing,
#else
    NULL,
    &serial_raw_state,
    serraw_int_tx_processing,
    serraw_int_rx_processing,
#endif
    SER_DEV_IDENT_0,
    &serial_tx_ring,
    &serial_rx_ring,
    serial_ControlTx,
    serial_ControlRx,
    serial_Control,
    serial_KickStartFn
};

/* Publically-accessible globals */

/*
 * This is the device table entry for this device
 */
const struct angel_DeviceEntry angel_SerialDevice = {
#if !defined(RAW_EBS_SERIAL) || (RAW_EBS_SERIAL == 0)
    DT_ANGEL,
    {
        serpkt_AsyncWrite,
        serpkt_RegisterRead
    },
    serpkt_Control, 
#else
    DT_RAW,
    {
        /* nasty but necessary casts, as can only statically initialise
         * the first member of a union
         */
        (angel_DeviceWriteFn)serraw_Write,
        (angel_DeviceRegisterReadFn)serraw_Read
    },
    serraw_Control,
#endif
    &serial_ctrl,
    { sizeof(param_list) / sizeof(param_list[0]), param_list },
    { sizeof(param_default) / sizeof(param_default[0]), param_default }
};

unsigned int receivedCharErrorStatus = 0;

#define DEBUG_SA1200 1
#if DEBUG_SA1200

void putchar(unsigned char ch)
{
    static int debugCOM_Init = 0;
    
    if (debugCOM_Init == 0)
    {
        /* init UART port */
        *UART_CR_BASE = 0;
        *UART_CR_BASE = BAUD_9600 | \
                        UART_DATA_BITS_8 | \
						UART_ENABLE_BIT;
        /* only do this once */
        debugCOM_Init = 1;
    }
    
    while((*UART_SR_BASE & UART_TX_READY) == 0)
        ;   /* Wait until transmitter ready */
    
    *UART_DR_BASE = (unsigned int)ch;
    
    if (ch == '\n')
    {
        while((*UART_SR_BASE & UART_TX_READY) == 0)
            ;   /* Wait until transmitter ready */

        *UART_DR_BASE = (unsigned int)'\r';
    }
} /* putchar */

void DEBUG_putstring(char *string) 
{
    while (*string != '\0')
    {
        putchar(*string) ;
        string++ ;
    }
}

#else /* DEBUG_SA1200 */

static void DEBUG_putstring(char *string)
{
}

#endif /* DEBUG_SA1200 */


/* Private globals */

/* macros/routines to get and put serial characters */
#define _GetChar() (*UART_DR_BASE)
#define _GetStatus() (*UART_SR_BASE)


#pragma no_check_stack
static void _PutChar(const char ch) 
{
    while((*UART_SR_BASE & UART_TX_READY) == 0)
        ;   /* Wait until transmitter ready */
    
    *UART_DR_BASE = (unsigned int)ch;
}
#pragma check_stack


#pragma no_check_stack
/*
 *  Function: serial_KickStartFn
 *   Purpose: Kick-start tx by sending first character
 *
 *    Params:
 *       Input: devid   device ID of the driver
 *
 *   Returns: Nothing
 */
static void serial_KickStartFn(DeviceID devid)
{
    char c;
    IGNORE(devid);
    /* chuck out a character */
    c = ringBufGetChar( &serial_tx_ring );
    _PutChar(c) ;
}
#pragma check_stack


/*
 * enable async transmission if not already enabled
 * NB caller should ensure protection against interrupt
 */
#pragma no_check_stack

static void serial_ControlTx(DeviceID devid)
{
    unsigned tx_status = angel_DeviceStatus[DI_SERIAL] & SER_TX_MASK;
    
    IGNORE(devid);
    /*
    * We enable tx whenever we want to send a flow control char,
    * OR when we are sending a packet and we have not been XOFF'ed
    */
    
    if ((tx_status & SER_TX_FLOW_CONTROL) || (tx_status == SER_TX_DATA))
    {
        /* tx interrupt should be enabled */
        angel_DeviceStatus[DI_SERIAL] |= SER_TX_IRQ_EN;
        *UART_CR_BASE  |= UART_TX_INT_ENABLE ;
    }
    else
    {
        /* tx interrupt should be disabled */
        angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_IRQ_EN;
        *UART_CR_BASE &= ~UART_TX_INT_ENABLE ;
    }
}

#pragma check_stack


/*
 * control rx interrupt mask according to flow control and read interest
 * NB caller should ensure protection against interrupt
 */
static void serial_ControlRx( DeviceID devid )
{
    IGNORE(devid);
    
    if (!(angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED)
        && (angel_DeviceStatus[DI_SERIAL] & DEV_READ_BUSY_MASK))
    {
        /* rx interrupt should be enabled */
        angel_DeviceStatus[DI_SERIAL] |= SER_RX_IRQ_EN;
        *UART_CR_BASE |= UART_RX_INT_ENABLE ;
    }
    else
    {
        /* rx interrupt should be disabled */
        angel_DeviceStatus[DI_SERIAL] &= ~SER_RX_IRQ_EN;
        *UART_CR_BASE &= ~UART_RX_INT_ENABLE ;
    }
}


/*
 * wait for tx to drain completely
 */
static void serial_tx_drain( void )
{
    while (angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD)
        ;   /* Wait for message to move over to UART */
    while((*UART_SR_BASE & UART_TX_FIFO_EMPTY) == 0)
        ;   /* Wait for UART to drain */
}

/*
 * Harware reset of UART
 */
static DevError ResetDriver( DeviceID devid )
{
    DevError status ;
    
    serial_tx_drain();
    status = serial_set_params(devid, &angel_SerialDevice.default_config) ;
    return status ;
}


/*
 * Initialisation control operation
 */
static DevError serial_init(DeviceID devid)
{

    DevError err;
    /*
    * do one-time start-up initialisation
    * (for this device, just a device reset)
    */
    Angel_ExitToUSR();      /* because DeviceControl requires it */
    
#if defined(MINIMAL_ANGEL) && MINIMAL_ANGEL != 0
    err = Angel_RawDeviceControl(devid, DC_RESET, NULL);
#else
    err = angel_DeviceControl(devid, DC_RESET, NULL);
#endif
    
    Angel_EnterSVC();
    
    return err;
}

/*
 * Reset control operation
 */
static DevError serial_reset(DeviceID devid)
{
    (void)ResetDriver(devid);       /* low_level reset */
    
    /* reset private flags (LEAVE lowest 8 bits alone) */
    angel_DeviceStatus[devid] &= 0xFF;
    
    /* turn interrupts on/off for rx and tx */
    serial_ControlRx(devid);
    serial_ControlTx(devid);
    
    return DE_OKAY;
}


/*
 * Receive Mode control operation
 */
static DevError serial_recv_mode(DeviceID devid, DevRecvMode mode)
{
    DevError ret_code = DE_OKAY;
    
    if ( mode == DR_DISABLE )
    {
        if ( ! (angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED) )
        {
            /* disable reception and schedule send of XOFF */
            angel_DeviceStatus[DI_SERIAL] |= (SER_RX_DISABLED | SER_TX_FLOW_CONTROL);
            serial_ControlRx(devid);
            serial_ControlTx(devid);
        }
    }
    else if ( mode == DR_ENABLE )
    {
        if ( angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED )
        {
            /* enable reception and schedule send of XON */
            angel_DeviceStatus[DI_SERIAL] &= ~SER_RX_DISABLED;
            angel_DeviceStatus[DI_SERIAL] |= SER_TX_FLOW_CONTROL;
            serial_ControlRx(devid);
            serial_ControlTx(devid);
        }
    }
    else
    {
        ret_code = DE_INVAL;
    }
    
    return ret_code;
}


/*
 * Set Speed control operation
 */
static DevError serial_set_params(DeviceID devid, const ParameterConfig *conf)
{
    word    speed;
    word    baud_value;
    IGNORE(devid);
    
    if ( ! Angel_FindParam( AP_BAUD_RATE, conf, &speed ) ) {
        return DE_BAD_OP ;
    }
    
    switch (speed) {
    case 9600:
        baud_value = BAUD_9600;
        break;
    case 19200:
        baud_value = BAUD_19200;
        break;
    case 38400:
        baud_value = BAUD_38400;
        break;
    case 57600:
        baud_value = BAUD_57600;
        break;
    case 115200:
        baud_value = BAUD_115200;
        break;
    default:     
        return DE_BAD_OP;
    }
    
    serial_tx_drain();
    
    /* init UART port */
    *UART_CR_BASE = 0;
    *UART_CR_BASE = baud_value | \
                    UART_DATA_BITS_8 | \
                    UART_ENABLE_BIT | \
                    UART_TX_INT_ENABLE | \
                    UART_RX_INT_ENABLE;
    return DE_OKAY;
}


/* 
 * implementation of device control for serial driver 
 *
 * devid is guaranteed to be in range but op is not checked as
 * individual devices can extend the range of operations
 */
DevError serial_Control(DeviceID devid, 
                        DeviceControl op,
                        void *arg)
{
    DevError ret_code;
    
    /* we only implement the standard ones here */
    switch( op )
    {
    case DC_INIT:
        ret_code = serial_init(devid);
        break;
        
    case DC_RESET:
        ret_code = serial_reset(devid);
        break;
        
    case DC_RECEIVE_MODE:  
        ret_code = serial_recv_mode(devid, (DevRecvMode)((int)arg));
        break;
        
    case DC_SET_PARAMS:     
        ret_code = serial_set_params(devid, (const ParameterConfig *)arg);
        break;
        
    default:
        ret_code = DE_BAD_OP;
        break;
    }
    
    return ret_code;
}


/* 
 * interrupt handler for serial driver
 */
#pragma no_check_stack

void angel_SerialIntHandler(    unsigned ident,
                            unsigned data,
                            unsigned empty_stack )
{
    unsigned int receivedCharWithStatus;
    volatile unsigned int UART_Status;
    char rx_ch;
    
    IGNORE( ident );
    IGNORE( data );
#if defined(MINIMAL_ANGEL) && MINIMAL_ANGEL != 0
    IGNORE( empty_stack );
#endif
    
    UART_Status = _GetStatus();
    
    /* Receive interrupt */
    
    if (UART_Status & UART_RX_READY)
    {
        /*
         * Data Ready interrupt
         */
        bool queue_rx_deferred = FALSE;
        
        do {
            /*
             * For an rx interrupt, place the character into a ring-buffer
             * and queue processing for after we have re-enabled interrpts.
             */
            receivedCharWithStatus = _GetChar(); /* get a char from UART */
#ifdef SA1200_A0
            receivedCharWithStatus = _GetChar(); /* get a char from UART */
#endif
            if (receivedCharWithStatus & (UART_DATA_FRMING_ERR | UART_DATA_PARITY_ERR))
            {
                receivedCharErrorStatus++;
                UART_Status =  _GetStatus(); /* get status for while() */
                continue; /* ignor the erroreous char */
            }
            
            rx_ch = receivedCharWithStatus & 0xFF;
            
            if (!ringBufFull(&serial_rx_ring))
            {
                RawState *const raw_state = serial_ctrl.raw_state;
                
                ringBufPutChar(&serial_rx_ring, rx_ch);
                
                /* In case we use raw serial, do counting */
                if ((raw_state != NULL) && (raw_state->rx_to_go > 0))
                    --(raw_state->rx_to_go);
                
                if ((raw_state == NULL && rx_ch == serial_ETX) ||
                    ((raw_state != NULL) && (raw_state->rx_to_go == 0) && (raw_state->rx_data != NULL)) ||
                    ((ringBufCount(&serial_rx_ring) >= RING_HIGH_WATER_MARK)))
                {
                    /* note that we need to do deferred rx processing */
                    queue_rx_deferred = TRUE;
                }
            }
            else
            {
                queue_rx_deferred = TRUE ;
                break ;
            }
            
            UART_Status = _GetStatus();
        } while ((UART_Status & UART_RX_READY) && (!ringBufFull(&serial_rx_ring)));
        
        /* Clear interrupt pending bit, if there is no outstanding interrupts */
        /* if ((UART_Status & UART_TX_FIFO_EMPTY) == 0) -- For this board */
		if (IXP1200_SysCpuRev < 1)
            *UART_CR_BASE |= UART_INT_CLEAR;
        
        /* see if we need to do deferred rx processing */
        if (queue_rx_deferred && !(angel_DeviceStatus[DI_SERIAL] & SER_RX_QUEUED))
        {
            angel_DeviceStatus[DI_SERIAL] |= SER_RX_QUEUED;
#if !defined(MINIMAL_ANGEL) || MINIMAL_ANGEL == 0
            Angel_SerialiseTask(    0,
                serial_ctrl.rx_processing,
                (void *)DI_SERIAL,
                empty_stack,
                &Angel_GlobalRegBlock[RB_Interrupted]);
#else
            serial_ctrl.rx_processing((void *)DI_SERIAL);
#endif
        }
    } /* UART_Status & UART_RX_READY */
    
    
    /* Transmit interrupt */
    
    if (UART_Status & UART_TX_FIFO_EMPTY)
    {
        bool queue_tx_deferred = FALSE;
        
        do {
            /*
             * For a tx interrupt, get the next character to transmit and send
             * it to the device.
             */
            
            /* do we need to send a flow control character? */
            if (angel_DeviceStatus[DI_SERIAL] & SER_TX_FLOW_CONTROL)
            {
                bool rx_disabled = angel_DeviceStatus[DI_SERIAL] & SER_RX_DISABLED;
                
                /* send a flow control char to UART */
                _PutChar(rx_disabled ? serial_XOFF : serial_XON);
                
                angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_FLOW_CONTROL;
            }
            else if (ringBufNotEmpty(&serial_tx_ring))
            {
                /* send a packet char to UART */
                _PutChar(ringBufGetChar(&serial_tx_ring));
            }
            
            /*
             * if we have reached to low-water mark in the
             * ring buffer then it is time to ask for a refill,
             * as long as there is more of the packet to come
             */
            if ((ringBufCount(&serial_tx_ring) < RING_LOW_WATER_MARK)
                && !(angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD))
            {
                queue_tx_deferred = TRUE;
            }
            
            UART_Status =  _GetStatus();
        } while ((UART_Status & UART_TX_FIFO_EMPTY) && (!ringBufEmpty(&serial_tx_ring)));
        
        /* Clear interrupt pending bit, if there is no outstanding interrupts */
        /* if ((UART_Status & UART_RX_READY) == 0) -- For this board */
	 	if (IXP1200_SysCpuRev < 1)
            *UART_CR_BASE |= UART_INT_CLEAR;
        
        if (ringBufEmpty(&serial_tx_ring))
        {
            /*
             * if the end of the packet has been sent,
             * then queue deferred processing of this fact
             */
            if (angel_DeviceStatus[DI_SERIAL] & SER_TX_EOD)
            {
                queue_tx_deferred = TRUE;
                
                /* disable Tx if necessary */
                angel_DeviceStatus[DI_SERIAL] &= ~SER_TX_DATA;
                serial_ControlTx(DI_SERIAL);
            }
            else
            {
                /*
                 * We need to set a flag to tell fill_tx_ring
                 * that it needs to restart character Tx.
                 */
                angel_DeviceStatus[DI_SERIAL] |= SER_TX_KICKSTART;
                
                /* Need to disable the transmit interrupt, otherwise it would
                 * come back immediately after we return.
                 */
                *UART_CR_BASE &= ~UART_TX_INT_ENABLE;
            }
        }
        
        /* If we got here then we may need to queue deferred tx processing */
        if ( queue_tx_deferred && !(angel_DeviceStatus[DI_SERIAL] & SER_TX_QUEUED))
        {
            angel_DeviceStatus[DI_SERIAL] |= SER_TX_QUEUED;
            
#if !defined(MINIMAL_ANGEL) || MINIMAL_ANGEL == 0
            Angel_SerialiseTask(    0,
                serial_ctrl.tx_processing,
                DI_SERIAL,
                empty_stack,
                &Angel_GlobalRegBlock[RB_Interrupted]);
#else
            serial_ctrl.tx_processing((void *)DI_SERIAL);
#endif
            
        }
    } /* UART_Status & UART_TX_READY */

 	if (IXP1200_SysCpuRev < 1)
		*UART_CR_BASE |= UART_INT_CLEAR;
}

#pragma check_stack

/* Routines for logging debug information to a seperate serial port */

void logserial_Reset(int port, int baudvalue)
{
}


/* EOF serial.c */
