/* heap.c
 *
 *
 * --------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------
 * system: SA1200
 * subsystem: uHAL 
 * author: Henry Qian 4/9/99
 * revisions:
 *
 */

/*
 * This is a simple heap allocator.  It allocates memory in fixed "chunk" 
 * sizes.  The chunk size is a power of 2, for example 128.  It allocates
 * as many chunks as it takes to hold the memory size requested and a block
 * header which describes the block of chunks.  All blocks of chunks are kept
 * in the memory list pointed at by "first".  The header describes the number 
 * of chunks in this block, whether or not it is free and how much memory
 * the caller asked for.  It also contains a magic pattern that allows the
 * deallocation code to check for memory corruption.  The smaller the chunk
 * size the less wasted memory, however a chunk must be bigger than the block
 * header.
 *
 * The memory allocation algorithm is to find the first block of chunks big
 * enough.  If this block of chunks is too large, the block is broken into 
 * two blocks.  The second block of chunks is used for the memory and first
 * is made free.  This is a "first found" algorithm, it could be more efficient
 * but the search would take longer.
 *
 * When freeing memory we could check to see if the block of chunks being
 * freed could be merged with either the previous or next block of chunks (or
 * both) if they are already free.  This collects the memory back into larger
 * blocks and helps prevent excessive fragmentation.  Rather than do this 
 * every time memory is deallocated, we wait until either we cannot find a 
 * free block of chunks big enough or we are concerned that there is too much
 * fragmentation.
 *
 * The macros uHAL_HEAP_BASE and uHAL_HEAP_SIZE must be declared in the system
 * specific header files.
 *
 */

#include <stdio.h>
#include "heap.h"

/* memory chunk descriptor */
typedef struct BlockHeader_S {
    unsigned int Magic;         /* for sanity check */
    unsigned int chunks;        /* size of chunk in chunks */
    unsigned int allocatedMemory; /* allocated size in bytes */
    struct BlockHeader_S* next; /* next chunk in the list */
} BlockHeader;

typedef struct {
    unsigned int totalMemory;
    unsigned int maxChunks;
    unsigned int highWaterMarkOfAllocatedChunks;
    unsigned int allocatedMemory;
    unsigned int allocatedChunks;
    unsigned int totalBlocks;
} MemStat;

MemStat memStat;

unsigned int uHAL_HEAP_BASE;
unsigned int uHAL_HEAP_SIZE;

#define HEADER_SIZE sizeof(BlockHeader) 
#define CHUNK_MAGIC 0xDEADBEEF

/* minimum size of memory to be allocated */
#define CHUNK_SIZE_IN_BITS 7
#define CHUNK_SIZE (1 << CHUNK_SIZE_IN_BITS) 

/* first memory block pointers (held in ascending memory order) */
BlockHeader* firstBlock = 0;

/* local (static) routines */
int easyMergeBlocks(BlockHeader* pBlock);
void mergeBlocks(void);
BlockHeader* firstFit(unsigned int chunks);

/* The following variables established at boot time in boot.s */
extern unsigned int SAiv_BaseOfFreeMemory;
extern unsigned int SAiv_TopOfFreeMemory;

/*
**  void mergeBlocks(void)
**
**  DESCRIPTION
**  
**  RETURNS
**      None.
*/
void mergeBlocks(void)
{
    BlockHeader* pBlock = firstBlock;
    
    while (pBlock)
    {
        if (easyMergeBlocks(pBlock) == 0)
            pBlock = pBlock->next;
    }
}

/*
**  int easyMergeBlocks(BlockHeader*)
**
**  DESCRIPTION
**      merge this chunk and the next, if they are both free
**  
**  RETURNS
**      None.
*/
int easyMergeBlocks(BlockHeader* pBlock)
{
    BlockHeader* nextBlock = pBlock->next;
    
    if (nextBlock)
    {
        if ((pBlock->allocatedMemory == 0) && (nextBlock->allocatedMemory == 0))
        {
            /* Merge */
            pBlock->next = nextBlock->next;
            pBlock->chunks += nextBlock->chunks;
            nextBlock->Magic = ~CHUNK_MAGIC;
            
            /* account for the death of a block of chunks */
            memStat.totalBlocks--;
            return 1; /* merge successful */
        }
    }

    return 0; /* no merge */
}


/*
**  int SAr_InitHeap(void)
**
**  DESCRIPTION
**      initialize the heap allocation scheme
**  
**  RETURNS
**      None.
*/
void SAr_InitHeap(void)
{
    /* Round up to CHUNK_SIZE boundary */
    uHAL_HEAP_BASE = (SAiv_BaseOfFreeMemory + CHUNK_SIZE-1) & ~(CHUNK_SIZE - 1);
    uHAL_HEAP_SIZE = SAiv_TopOfFreeMemory - uHAL_HEAP_BASE;

    /* set up the free block (one big one to start with) */
    firstBlock = (BlockHeader*)uHAL_HEAP_BASE;
    firstBlock->Magic = CHUNK_MAGIC;
    firstBlock->allocatedMemory = 0;             
    firstBlock->chunks = uHAL_HEAP_SIZE >> CHUNK_SIZE_IN_BITS;
    firstBlock->next = 0;
    
    /* set up the memory statistics */
    memStat.totalMemory = uHAL_HEAP_SIZE;
    memStat.maxChunks = uHAL_HEAP_SIZE >> CHUNK_SIZE_IN_BITS;
    memStat.allocatedMemory = 0;
    memStat.allocatedChunks = 0;
    memStat.totalBlocks = 1;
    memStat.highWaterMarkOfAllocatedChunks = 0;
}


/*
**  BlockHeader* firstFit(unsigned int)
**
**  DESCRIPTION
**  
**  RETURNS
**      None.
*/
BlockHeader* firstFit(unsigned int chunks) 
{
    BlockHeader* pBlock = firstBlock;
    
    /* very, very simple algorithm, find the first block big enough */
    while (pBlock)
    {
        if ((pBlock->chunks >= chunks) && (pBlock->allocatedMemory == 0))
            return pBlock;
        pBlock = pBlock->next;
    }

    return 0;
}


/*
**  void* malloc(unsigned int)
**
**  DESCRIPTION
**      allocate memory, make this as quick as possible
**  
**  RETURNS
**      None.
*/
void* malloc(unsigned int size)
{
    unsigned int chunks;
    BlockHeader* pBlock;
    BlockHeader* newBlock;
    
    if (size == 0)
        return(0); /* return as an error */

    /* Figure out the number of chunks we need */
    chunks = (sizeof(BlockHeader) + size + CHUNK_SIZE-1) >> CHUNK_SIZE_IN_BITS;
    
    /* Find the first block that fits */
    if ((pBlock = firstFit(chunks)) == 0)
    {
        /* No large enough block. Try merge */
        mergeBlocks();
        pBlock = firstFit(chunks);
    }

    if (pBlock)
    {
        /* Big block needs to be broken into two smaller blocks */
        if (pBlock->chunks > chunks)
        {
            /*
             * make a new block and thread it into the list, make
             * sure that the allocator gets the block at the end,
             * this tends to keep the free blocks of chunks at the
             * start of the memory list.
             */
            newBlock = (BlockHeader*)((char*)pBlock +
                            ((pBlock->chunks - chunks) * CHUNK_SIZE));
        
            newBlock->Magic = CHUNK_MAGIC;
            newBlock->next = pBlock->next;
            newBlock->allocatedMemory = size;
            newBlock->chunks = chunks;
        
            pBlock->next = newBlock;
            pBlock->chunks -= chunks;
        
            /* account for the birth of a block */
            memStat.totalBlocks++;
        
            /* account for this memory */
            memStat.allocatedMemory += size;
            memStat.allocatedChunks += chunks;

            if (memStat.allocatedChunks > memStat.highWaterMarkOfAllocatedChunks)
                memStat.highWaterMarkOfAllocatedChunks = memStat.allocatedChunks;
        
            /* return the caller a pointer to its memory */
            return (void *)((unsigned char *)newBlock + sizeof(BlockHeader));
        }
        else if (pBlock->chunks == chunks)
        {
            /* New found block has the exactly number of chunks we need */
            pBlock->allocatedMemory = size;
            memStat.allocatedMemory += size;
            memStat.allocatedChunks += chunks;

            if (memStat.allocatedChunks > memStat.highWaterMarkOfAllocatedChunks)
                memStat.highWaterMarkOfAllocatedChunks = memStat.allocatedChunks;
        
            return (void *)((unsigned char *)pBlock + sizeof(BlockHeader));
        }

    }

    return(0); /* No memory could be allocated */
}

/*
**  void free(void*)
**
**  DESCRIPTION
**  
**  RETURNS
**      None.
*/
void free(void* memoryAddress)
{
    BlockHeader* pBlock = (BlockHeader*)((char*)memoryAddress - sizeof(BlockHeader));
    
    /* check the magic header number */
    if (pBlock->Magic != CHUNK_MAGIC)
        return;
    
    /* check that this block is indeed allocated */
    if (pBlock->allocatedMemory == 0)
        return;
    
    /* account for this memory */
    memStat.allocatedMemory -= pBlock->allocatedMemory;
    memStat.allocatedChunks -= pBlock->chunks;
    
    /* mark this block of chunks as free, don't attempt to merge (yet) */
    pBlock->allocatedMemory = 0;
}

/*
**  void SAr_ShowMalloc(void)
**
**  DESCRIPTION
**  
**  RETURNS
**      None.
*/
void SAr_ShowMalloc(void)
{
    unsigned int totalBlocks = 0;
    unsigned int allocatedMemory = 0;
    unsigned int allocatedChunks = 0;

    BlockHeader* pBlock = firstBlock;

    if (firstBlock == 0)
    {
        printf(" - Heap not initialized yet.\n");
        return;
    }

    printf(" ----------------------------------------------\n");
    printf("  - B_addr   B_next   M_addr   M_size   chunks \n");
    printf(" ----------------------------------------------\n");

    while(pBlock != 0)
    {
        if (pBlock->Magic != CHUNK_MAGIC)
        {
            printf(" - error: Magic: %08X\n", pBlock->Magic);
            return;
        }

        printf(" - %08x %08x %08x %08x %08x\n",
                (unsigned int)pBlock,
                (unsigned int)(pBlock->next),
                ((unsigned int)pBlock + sizeof(BlockHeader)),
                (unsigned int)(pBlock->allocatedMemory),
                (unsigned int)(pBlock->chunks) );

        totalBlocks++;

        if (pBlock->allocatedMemory != 0)
        {
            allocatedMemory += pBlock->allocatedMemory;
            allocatedChunks += pBlock->chunks;
        }

        pBlock = pBlock->next;
    }

    if ((allocatedMemory != memStat.allocatedMemory) ||
        (allocatedChunks != memStat.allocatedChunks) ||
        (totalBlocks != memStat.totalBlocks) ||
        (memStat.highWaterMarkOfAllocatedChunks > memStat.maxChunks))
    {
        printf(" - error:\n");
        if (allocatedMemory != memStat.allocatedMemory)
            printf(" - allocatedMemory: %08x vs. %08x\n", allocatedMemory, memStat.allocatedMemory);
        if (allocatedChunks != memStat.allocatedChunks)
            printf(" - allocatedChunks: %08x vs. %08x\n", allocatedChunks, memStat.allocatedChunks);
        if (totalBlocks != memStat.totalBlocks)
            printf(" - totalBlocks:     %08x vs. %08x\n", totalBlocks, memStat.totalBlocks);
        if (memStat.highWaterMarkOfAllocatedChunks > memStat.maxChunks)
            printf(" - highWaterMark:   %08x vs. %08x\n",
                    memStat.highWaterMarkOfAllocatedChunks, memStat.maxChunks);
    }

    printf(" - Memory Allocation Statistics: --------------\n");
    printf(" - Total Blocks: %d  %08x:%08x (base:end)\n",
                    memStat.totalBlocks, uHAL_HEAP_BASE, uHAL_HEAP_BASE+uHAL_HEAP_SIZE-1);
    printf(" - Allocated Memory: %08x  Total Memory: %08x\n",
                    memStat.allocatedMemory, memStat.totalMemory);
    printf(" - Allocated Chunks: %08x  Chunk size:   %08x\n",
                    memStat.allocatedChunks, CHUNK_SIZE);
    printf(" - High Water Mark:  %08x  Total Chunks: %08x\n",
                    memStat.highWaterMarkOfAllocatedChunks,
                    memStat.maxChunks);
}


/* end of file */
