/* interrupts.c
 *
 *------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------
 * system: SA1200
 * subsystem: SA1200 ARM Core
 * author: Henry Qian 02/22/99
 * revisions:
 *
 */

#include "platform.h"
#include "interrupts.h"

// -----------------------
// Regular Interrupt (IRQ)
// -----------------------

IntHandler IRQ_Handlers[MAX_NUM_OF_SOURCES];
unsigned int orignal_IRQ_Vector;

#define INT_NOT_USED 0

/*
**  void SA_InitIRQ()
**
**  DESCRIPTION
**      Disable IRQ.
**      Initialize IRQ stack and install a IRQ
**      interrupt vector.
** 
**  RETURNS
**      None.
*/
void SA_InitIRQ(void)
{
    int ii;

    unsigned int origIntMask = SAr_DisableInt();

    /* Initialize internal vector table */
    for (ii = 0; ii < MAX_NUM_OF_SOURCES; ii++)
    {
        IRQ_Handlers[ii].func = 0;
        IRQ_Handlers[ii].data = 0;
    }

    /* SAr_DisableAllIRQs(); */

    /* Save the orignal in case we need to chain interrupt */
    orignal_IRQ_Vector = *(unsigned int*)IRQ_VECTOR;

    /* install IRQ Interrupt Service Routine */
    *(unsigned int*)IRQ_VECTOR = (unsigned int)SAir_IRQTrapWrap;

    SAr_EnableInt(origIntMask);
}


/*
**  void SA_IRQ_Connect()
**
**  DESCRIPTION
**      Register ISR for IRQ interrupts.
** 
**  RETURNS
**      success: original interrupt handler;
**      error: IntHandler.func == -1.
*/
IntHandler SA_IRQ_Connect(  IntSource intSource,
                            PFVI intServiceRoutine,
                            int userParam)
{
    IntHandler retval;

    unsigned int origIntMask = SAr_DisableInt();

    if ((intSource >= 0) &&
        (intSource < MAX_NUM_OF_SOURCES) &&
        (intSource != REAL_TIME_CLOCK))
    {
        /* Return old interrupt handler */
        retval = IRQ_Handlers[intSource];
        /* install the new handler */
        IRQ_Handlers[intSource].func = intServiceRoutine;
        IRQ_Handlers[intSource].data = userParam;
    }
    else
        retval.func = (PFVI)-1;

    SAr_EnableInt(origIntMask);

    return(retval);
}


/*
**  int SA_IRQ_Dispatcher()
**
**  DESCRIPTION
**      Sort out different IRQ interrupts.
**      Read different interrupt sourse status registers
**      to determine what interrupt(s) we have and call
**      their ISR respectively.
** 
**  RETURNS
**      determine pre-emption task reschedule
**      1: pre-emption; 0: non pre-emption.
*/
int SA_IRQ_Dispatcher(void)
{
	int ii;
    unsigned int IRQ_Status;
    unsigned int PCI_IRQ_Status;

    /* To determine if this interrupt would cause task re-scheduling */
    IRQ_Status = *(unsigned int*)(IRQ);

    // UART interrupt
    //
    if (*(unsigned int*)UART_CR & (UART_TX_FIFO_EMPTY | UART_RX_READY))
    {
        if (IRQ_Handlers[SERIAL].func)
            (IRQ_Handlers[SERIAL].func)(IRQ_Handlers[SERIAL].data);
#if 0
        else
            *(unsigned int*)UART_CR |= UART_INT_CLEAR;
#endif
    }
    // External C pin interrupt (MAC device interrupt)
    //
    if (IRQ_Status & I_CINT)
    {
        if (IRQ_Handlers[EXTERNAL_C_PIN].func)
            (IRQ_Handlers[EXTERNAL_C_PIN].func)(IRQ_Handlers[EXTERNAL_C_PIN].data);
        else
        {
            // Clear CINT interrupt, if driver is not there.
            *(unsigned int*)IREG |= CINT_BIT;
        }
    }

    // Micro engin/thread interrupt
    //
    if (IRQ_Status & I_UENG)
    {
        if (IRQ_Handlers[MICRO_THREAD].func)
            (IRQ_Handlers[MICRO_THREAD].func)(IRQ_Handlers[MICRO_THREAD].data);
        else
        {
            // Clear UENG interrupt, if driver is not there.
            *(unsigned int*)IREG &= 0xF8000000;
        }
    }

    // SRAM Interrupt
    //
    if (IRQ_Status & I_SRAM)
    {
        if (IRQ_Handlers[SRAM_MEMORY].func)
            (IRQ_Handlers[SRAM_MEMORY].func)(IRQ_Handlers[SRAM_MEMORY].data);
        else
            *(unsigned int*)SRAM_CSR &= ~(SRAM_RLRS | SRAM_RLS);
    }

    // SRAM Interrupt
    //
    if (IRQ_Status & I_SDRAM)
    {
        if (IRQ_Handlers[SDRAM_MEMORY].func)
            (IRQ_Handlers[SDRAM_MEMORY].func)(IRQ_Handlers[SDRAM_MEMORY].data);
    }

    // Interrupts generated by PCI unit
    //
    if (IRQ_Status & I_PCI)
    {
        PCI_IRQ_Status = *(unsigned int*)(CSR_BASE + IRQ_STATUS);

        if (PCI_IRQ_Status & DOORBELL_INT)
        {
            if (IRQ_Handlers[PCI_DOORBELL].func)
                (IRQ_Handlers[PCI_DOORBELL].func)(IRQ_Handlers[PCI_DOORBELL].data);
        }

        if (PCI_IRQ_Status & SOFT_INT)
        {
            if (IRQ_Handlers[SOFT].func)
                (IRQ_Handlers[SOFT].func)(IRQ_Handlers[SOFT].data);
			else
				*(unsigned int*)(CSR_BASE + IRQ_SOFT) = 0;
        }

        if (PCI_IRQ_Status & TIMER1_INT)
        {
            if (IRQ_Handlers[TIMER1].func)
                (IRQ_Handlers[TIMER1].func)(IRQ_Handlers[TIMER1].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_1_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_IRQ_Status & TIMER2_INT)
        {
            if (IRQ_Handlers[TIMER2].func)
                (IRQ_Handlers[TIMER2].func)(IRQ_Handlers[TIMER2].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_2_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_IRQ_Status & TIMER3_INT)
        {
            if (IRQ_Handlers[TIMER3].func)
                (IRQ_Handlers[TIMER3].func)(IRQ_Handlers[TIMER3].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_3_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_IRQ_Status & TIMER4_INT)
        {
            if (IRQ_Handlers[TIMER4].func)
                (IRQ_Handlers[TIMER4].func)(IRQ_Handlers[TIMER4].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_4_CLEAR) = 0xFFFFFFFF;
        }

#if INT_NOT_USED
        if (PCI_IRQ_Status & DMA1_INT)
        {
            if (IRQ_Handlers[DMA1].func)
                (IRQ_Handlers[DMA1].func)(IRQ_Handlers[DMA1].data);
        }

        if (PCI_IRQ_Status & DMA2_INT)
        {
            if (IRQ_Handlers[DMA2].func)
                (IRQ_Handlers[DMA2].func)(IRQ_Handlers[DMA2].data);
        }

        if (PCI_IRQ_Status & PCI_IRQ_L_INT)
        {
            if (IRQ_Handlers[PCI_IRQ_L].func)
                (IRQ_Handlers[PCI_IRQ_L].func)(IRQ_Handlers[PCI_IRQ_L].data);
        }

        if (PCI_IRQ_Status & DMA1_NOT_BUSY_INT)
        {
            if (IRQ_Handlers[DMA1_NOT_BUSY].func)
                (IRQ_Handlers[DMA1_NOT_BUSY].func)(IRQ_Handlers[DMA1_NOT_BUSY].data);
        }

        if (PCI_IRQ_Status & DMA2_NOT_BUSY_INT)
        {
            if (IRQ_Handlers[DMA2_NOT_BUSY].func)
                (IRQ_Handlers[DMA2_NOT_BUSY].func)(IRQ_Handlers[DMA2_NOT_BUSY].data);
        }

        if (PCI_IRQ_Status & START_BIST_INT)
        {
            if (IRQ_Handlers[START_BIST].func)
                (IRQ_Handlers[START_BIST].func)(IRQ_Handlers[START_BIST].data);
        }

        if (PCI_IRQ_Status & SERR_INT)
        {
            if (IRQ_Handlers[SERR].func)
                (IRQ_Handlers[SERR].func)(IRQ_Handlers[SERR].data);
        }

        if (PCI_IRQ_Status & SDRAM_PARITY_INT)
        {
            if (IRQ_Handlers[SDRAM_PARITY].func)
                (IRQ_Handlers[SDRAM_PARITY].func)(IRQ_Handlers[SDRAM_PARITY].data);
        }

        if (PCI_IRQ_Status & I2O_INPOST_INT)
        {
            if (IRQ_Handlers[I2O_INPOST].func)
                (IRQ_Handlers[I2O_INPOST].func)(IRQ_Handlers[I2O_INPOST].data);
        }

        if (PCI_IRQ_Status & DISCARD_TIMER_INT)
        {
            if (IRQ_Handlers[DISCARD_TIMER].func)
                (IRQ_Handlers[DISCARD_TIMER].func)(IRQ_Handlers[DISCARD_TIMER].data);
        }

        if (PCI_IRQ_Status & DATA_PARITY_INT)
        {
            if (IRQ_Handlers[DATA_PARITY].func)
                (IRQ_Handlers[DATA_PARITY].func)(IRQ_Handlers[DATA_PARITY].data);
        }

        if (PCI_IRQ_Status & PCI_MASTER_ABORT_INT)
        {
            if (IRQ_Handlers[PCI_MASTER_ABORT].func)
                (IRQ_Handlers[PCI_MASTER_ABORT].func)(IRQ_Handlers[PCI_MASTER_ABORT].data);
        }

        if (PCI_IRQ_Status & PCI_TARGET_ABORT_INT)
        {
            if (IRQ_Handlers[PCI_TARGET_ABORT].func)
                (IRQ_Handlers[PCI_TARGET_ABORT].func)(IRQ_Handlers[PCI_TARGET_ABORT].data);
        }

        if (PCI_IRQ_Status & PCI_PARITY_INT)
        {
            if (IRQ_Handlers[PCI_PARITY].func)
                (IRQ_Handlers[PCI_PARITY].func)(IRQ_Handlers[PCI_PARITY].data);
        }
#endif // #if INT_NOT_USED
    }
}


/*
**  int SA_IRQ_Enable(IntSource intSource)
**  int SA_IRQ_Disable(IntSource intSource)
**
**  DESCRIPTION
**      Enable/disable a IRQ interrupt.
** 
**  RETURNS
**      success:    SA_SUCCESS
**      error:      SA_FAIL
*/
int SA_IRQ_Enable(IntSource intSource)
{
    int retval = SA_SUCCESS;

    unsigned int origIntMask = SAr_DisableInt();

    switch (intSource)
    {
        case MICRO_THREAD:
            *(unsigned int*)IREG = (*(unsigned int*)IREG | UEIRQ_ENABLE) & 0xF0000000;
            break;

        case EXTERNAL_C_PIN:
            *(unsigned int*)IREG = (*(unsigned int*)IREG | CINTIRQ_ENABLE) & 0xF0000000;
            break;

        case SERIAL:
            /* There are two interrupts, but we have only one entry here.
             * So we don't do enable/disable in this API
             */
            retval = SA_FAIL;
            break;

        case SRAM_MEMORY:
            *(unsigned int*)SRAM_CSR =
                (*(unsigned int*)SRAM_CSR | SRAM_IRQ) & 0xFFFFFFFC;
            break;

        case SDRAM_MEMORY:
            *(unsigned int*)SDRAM_CSR |= (SDRAM_EN_INT | SDRAM_EN_IRQ);
            break;

        case PCI_DOORBELL:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DOORBELL_INT;
            break;

        case SOFT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = SOFT_INT;
            break;

        case TIMER1:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = TIMER1_INT;
            break;

        case TIMER2:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = TIMER2_INT;
            break;

        case TIMER3:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = TIMER3_INT;
            break;

        case TIMER4:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = TIMER4_INT;
            break;

        case DMA1:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DMA1_INT;
            break;

        case DMA2:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DMA2_INT;
            break;

        case PCI_IRQ_L:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = PCI_IRQ_L_INT;
            break;

        case DMA1_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DMA1_NOT_BUSY_INT;
            break;

        case DMA2_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DMA2_NOT_BUSY_INT;
            break;

        case START_BIST:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = START_BIST_INT;
            break;

        case SERR:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = SERR_INT;
            break;

        case SDRAM_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = SDRAM_PARITY_INT;
            break;

        case I2O_INPOST:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = I2O_INPOST_INT;
            break;

        case DISCARD_TIMER:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DISCARD_TIMER_INT;
            break;

        case DATA_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = DATA_PARITY_INT;
            break;

        case PCI_MASTER_ABORT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = PCI_MASTER_ABORT_INT;
            break;

        case PCI_TARGET_ABORT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = PCI_TARGET_ABORT_INT;
            break;

        case PCI_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_SET) = PCI_PARITY_INT;
            break;

        default:
            retval = SA_FAIL;
            break;
    }

    SAr_EnableInt(origIntMask);

    return(retval);
}


int SA_IRQ_Disable(IntSource intSource)
{
    int retval = SA_SUCCESS;

    unsigned int origIntMask = SAr_DisableInt();

    switch (intSource)
    {
        case MICRO_THREAD:
            *(unsigned int*)IREG = (*(unsigned int*)IREG & ~UEIRQ_ENABLE) & 0xF0000000;
            break;

        case EXTERNAL_C_PIN:
            *(unsigned int*)IREG = (*(unsigned int*)IREG & ~CINTIRQ_ENABLE) & 0xF0000000;
            break;

        case SERIAL:
            /* There are two interrupts, but we have only one entry here.
             * So we don't do enable/disable in this API
             */
            retval = SA_FAIL;
            break;

        case SRAM_MEMORY:
            *(unsigned int*)SRAM_CSR =
                (*(unsigned int*)SRAM_CSR & ~SRAM_IRQ) & 0xFFFFFFFC;
            break;

        case SDRAM_MEMORY:
            *(unsigned int*)SDRAM_CSR &= ~SDRAM_EN_IRQ;
            break;

        case PCI_DOORBELL:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DOORBELL_INT;
            break;

        case SOFT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = SOFT_INT;
            break;

        case TIMER1:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = TIMER1_INT;
            break;

        case TIMER2:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = TIMER2_INT;
            break;

        case TIMER3:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = TIMER3_INT;
            break;

        case TIMER4:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = TIMER4_INT;
            break;

        case DMA1:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DMA1_INT;
            break;

        case DMA2:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DMA2_INT;
            break;

        case PCI_IRQ_L:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = PCI_IRQ_L_INT;
            break;

        case DMA1_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DMA1_NOT_BUSY_INT;
            break;

        case DMA2_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DMA2_NOT_BUSY_INT;
            break;

        case START_BIST:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = START_BIST_INT;
            break;

        case SERR:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = SERR_INT;
            break;

        case SDRAM_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = SDRAM_PARITY_INT;
            break;

        case I2O_INPOST:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = I2O_INPOST_INT;
            break;

        case DISCARD_TIMER:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DISCARD_TIMER_INT;
            break;

        case DATA_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = DATA_PARITY_INT;
            break;

        case PCI_MASTER_ABORT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = PCI_MASTER_ABORT_INT;
            break;

        case PCI_TARGET_ABORT:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = PCI_TARGET_ABORT_INT;
            break;

        case PCI_PARITY:
            *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = PCI_PARITY_INT;
            break;

        default:
            retval = SA_FAIL;
            break;
    }

    SAr_EnableInt(origIntMask);

    return(retval);
}


/*
**  void SA_IRQ_InitDoorBell()
**
**  DESCRIPTION
**      Initialize doorbell register.
**      Hook up an interrupt service routine to
**      doorbell register.
**      Enable doorbell interrupt.
** 
**  RETURNS
**      None.
*/
void SA_IRQ_InitDoorBell(PFVI intServiceRoutine)
{
    unsigned int origIntMask = SAr_DisableInt();

    SA_IRQ_Disable(PCI_DOORBELL);

    // Setup doorbell register's directions
    //
	*(unsigned int*)(CSR_BASE + DBELL_PCI_MASK) = 0;
	*(unsigned int*)(CSR_BASE + DBELL_SA_MASK)  = 0xFFFFFFFF;

	*(unsigned int*)(CSR_BASE + DOORBELL_SETUP)  = 0xFFFFFFFF;

    SA_IRQ_Connect(PCI_DOORBELL, intServiceRoutine, 0);

    SA_IRQ_Enable(PCI_DOORBELL);

    SAr_EnableInt(origIntMask);
}


/*
**  void SA_IRQ_InitSoft()
**
**  DESCRIPTION
**      Hook up an interrupt service routine to
**      soft interrupt.
**      Enable soft interrupt.
** 
**  RETURNS
**      None.
*/
void SA_IRQ_InitSoft(PFVI intServiceRoutine)
{
    unsigned int origIntMask = SAr_DisableInt();

    SA_IRQ_Disable(SOFT);
    SA_IRQ_Connect(SOFT, intServiceRoutine, 0);
    SA_IRQ_Enable(SOFT);

    SAr_EnableInt(origIntMask);
}

/*
**  void SA_InitSysTimer(void)
**
**  DESCRIPTION
** 
**  RETURNS
**      None.
*/
void SA_SysTimer_ISR(int IGNOR)
{
    *(unsigned int*)(CSR_BASE + TIMER_3_CLEAR) = 0xFFFFFFFF;
    /* OSTimeTick(); */
}


/*
**  void SA_InitSysTimer(void)
**
**  DESCRIPTION
** 
**  RETURNS
**      None.
*/
void SA_InitSysTimer(void)
{
    unsigned int origIntMask = SAr_DisableInt();

    SA_IRQ_Disable(TIMER3);

    /* OSTime = 0; */
    /* Timer 3 used as an OS timer */
    /* *(unsigned int*)(CSR_BASE + TIMER_3_LOAD) = mSEC_10; */
    *(unsigned int*)(CSR_BASE + TIMER_3_LOAD) = 100000000;

    /* Timer is periodic, and enabled */
    *(unsigned int*)(CSR_BASE + TIMER_3_CONTROL) = 0xC0;

    SA_IRQ_Connect(TIMER3, SA_SysTimer_ISR, 0);
    SA_IRQ_Enable(TIMER3);

    SAr_EnableInt(origIntMask);
}




// --------------------
// Fast Interrupt (FIQ)
// --------------------

IntHandler FIQ_Handlers[MAX_NUM_OF_SOURCES];

/*
**  void SA_InitFIQ()
**
**  DESCRIPTION
**      Disable FIQ.
**      Initialize FIQ stack and install a FIQ
**      interrupt vector.
** 
**  RETURNS
**      None.
*/
void SA_InitFIQ(void)
{
    int ii;

    unsigned int origIntMask = SAr_DisableInt();

    /* Initialize internal vector table */
    for (ii = 0; ii < MAX_NUM_OF_SOURCES; ii++)
    {
        FIQ_Handlers[ii].func = 0;
        FIQ_Handlers[ii].data = 0;
    }

    SAr_DisableAllFIQs();

    /* install FIQ Interrupt Service Routine */
    *(unsigned int*)FIQ_VECTOR = (unsigned int)SAir_FIQTrapWrap;

    SAr_EnableInt(origIntMask);
}


/*
**  void SA_FIQ_Connect()
**
**  DESCRIPTION
**      Register ISR for FIQ interrupts.
** 
**  RETURNS
**      success: original interrupt handler;
**      error: IntHandler.func == -1.
*/
IntHandler SA_FIQ_Connect(  IntSource intSource,
                            PFVI intServiceRoutine,
                            int userParam)
{
    IntHandler retval;

    unsigned int origIntMask = SAr_DisableInt();

    if ((intSource >= 0) &&
        (intSource < MAX_NUM_OF_SOURCES) &&
        (intSource != SERIAL))
    {
        /* Return old interrupt handler */
        retval = FIQ_Handlers[intSource];
        /* install the new handler */
        FIQ_Handlers[intSource].func = intServiceRoutine;
        FIQ_Handlers[intSource].data = userParam;
    }
    else
        retval.func = (PFVI)-1;

    SAr_EnableInt(origIntMask);

    return(retval);
}


/*
**  void SA_FIQ_Dispatcher()
**
**  DESCRIPTION
**      Sort out different FIQ interrupts.
**      Read different interrupt sourse status registers
**      to determine what interrupt(s) we have and call
**      their ISR respectively.
** 
**  RETURNS
**      None.
*/
void SA_FIQ_Dispatcher(void)
{
    unsigned int FIQ_Status;
    unsigned int PCI_FIQ_Status;
    FIQ_Status = *(unsigned int*)(FIQ);

    // External C pin interrupt (MAC device interrupt)
    //
    if (FIQ_Status & F_CINT)
    {
        if (FIQ_Handlers[EXTERNAL_C_PIN].func)
            (FIQ_Handlers[EXTERNAL_C_PIN].func)(FIQ_Handlers[EXTERNAL_C_PIN].data);
        else
        {
            // Clear CINT interrupt, if driver is not there.
            *(unsigned int*)IREG |= CINT_BIT;
        }
    }

    // Micro engin/thread interrupt
    //
    if (FIQ_Status & F_UENG)
    {
        if (FIQ_Handlers[MICRO_THREAD].func)
            (FIQ_Handlers[MICRO_THREAD].func)(FIQ_Handlers[MICRO_THREAD].data);
        else
        {
            // Clear UENG interrupt, if driver is not there.
            *(unsigned int*)IREG &= 0xF8000000;
        }
    }

    // Real Time Clock Interrupt
    //
    if (FIQ_Status & F_RTC)
    {
        if (FIQ_Handlers[REAL_TIME_CLOCK].func)
            (FIQ_Handlers[REAL_TIME_CLOCK].func)(FIQ_Handlers[REAL_TIME_CLOCK].data);
    }

    // SRAM Interrupt
    //
    if (FIQ_Status & F_SRAM)
    {
        if (FIQ_Handlers[SRAM_MEMORY].func)
            (FIQ_Handlers[SRAM_MEMORY].func)(FIQ_Handlers[SRAM_MEMORY].data);
        else
            *(unsigned int*)SRAM_CSR &= ~(SRAM_RLRS | SRAM_RLS);
    }

    // SRAM Interrupt
    //
    if (FIQ_Status & F_SDRAM)
    {
        if (FIQ_Handlers[SDRAM_MEMORY].func)
            (FIQ_Handlers[SDRAM_MEMORY].func)(FIQ_Handlers[SDRAM_MEMORY].data);
    }

    // Interrupts generated by PCI unit
    //
    if (FIQ_Status & F_PCI)
    {
        PCI_FIQ_Status = *(unsigned int*)(CSR_BASE + FIQ_STATUS);

        if (PCI_FIQ_Status & DOORBELL_INT)
        {
            if (FIQ_Handlers[PCI_DOORBELL].func)
                (FIQ_Handlers[PCI_DOORBELL].func)(FIQ_Handlers[PCI_DOORBELL].data);
        }

        if (PCI_FIQ_Status & SOFT_INT)
        {
            if (FIQ_Handlers[SOFT].func)
                (FIQ_Handlers[SOFT].func)(FIQ_Handlers[SOFT].data);
			else
				*(unsigned int*)(CSR_BASE + FIQ_SOFT) = 0;
        }

        if (PCI_FIQ_Status & TIMER1_INT)
        {
            if (FIQ_Handlers[TIMER1].func)
                (FIQ_Handlers[TIMER1].func)(FIQ_Handlers[TIMER1].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_1_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_FIQ_Status & TIMER2_INT)
        {
            if (FIQ_Handlers[TIMER2].func)
                (FIQ_Handlers[TIMER2].func)(FIQ_Handlers[TIMER2].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_2_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_FIQ_Status & TIMER3_INT)
        {
            if (FIQ_Handlers[TIMER3].func)
                (FIQ_Handlers[TIMER3].func)(FIQ_Handlers[TIMER3].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_3_CLEAR) = 0xFFFFFFFF;
        }

        if (PCI_FIQ_Status & TIMER4_INT)
        {
            if (FIQ_Handlers[TIMER4].func)
                (FIQ_Handlers[TIMER4].func)(FIQ_Handlers[TIMER4].data);
			else
				*(unsigned int*)(CSR_BASE + TIMER_4_CLEAR) = 0xFFFFFFFF;
        }

#if INT_NOT_USED
        if (PCI_FIQ_Status & DMA1_INT)
        {
            if (FIQ_Handlers[DMA1].func)
                (FIQ_Handlers[DMA1].func)(FIQ_Handlers[DMA1].data);
        }

        if (PCI_FIQ_Status & DMA2_INT)
        {
            if (FIQ_Handlers[DMA2].func)
                (FIQ_Handlers[DMA2].func)(FIQ_Handlers[DMA2].data);
        }

        if (PCI_FIQ_Status & PCI_IRQ_L_INT)
        {
            if (FIQ_Handlers[PCI_IRQ_L].func)
                (FIQ_Handlers[PCI_IRQ_L].func)(FIQ_Handlers[PCI_IRQ_L].data);
        }

        if (PCI_FIQ_Status & DMA1_NOT_BUSY_INT)
        {
            if (FIQ_Handlers[DMA1_NOT_BUSY].func)
                (FIQ_Handlers[DMA1_NOT_BUSY].func)(FIQ_Handlers[DMA1_NOT_BUSY].data);
        }

        if (PCI_FIQ_Status & DMA2_NOT_BUSY_INT)
        {
            if (FIQ_Handlers[DMA2_NOT_BUSY].func)
                (FIQ_Handlers[DMA2_NOT_BUSY].func)(FIQ_Handlers[DMA2_NOT_BUSY].data);
        }

        if (PCI_FIQ_Status & START_BIST_INT)
        {
            if (FIQ_Handlers[START_BIST].func)
                (FIQ_Handlers[START_BIST].func)(FIQ_Handlers[START_BIST].data);
        }

        if (PCI_FIQ_Status & SERR_INT)
        {
            if (FIQ_Handlers[SERR].func)
                (FIQ_Handlers[SERR].func)(FIQ_Handlers[SERR].data);
        }

        if (PCI_FIQ_Status & SDRAM_PARITY_INT)
        {
            if (FIQ_Handlers[SDRAM_PARITY].func)
                (FIQ_Handlers[SDRAM_PARITY].func)(FIQ_Handlers[SDRAM_PARITY].data);
        }

        if (PCI_FIQ_Status & I2O_INPOST_INT)
        {
            if (FIQ_Handlers[I2O_INPOST].func)
                (FIQ_Handlers[I2O_INPOST].func)(FIQ_Handlers[I2O_INPOST].data);
        }

        if (PCI_FIQ_Status & DISCARD_TIMER_INT)
        {
            if (FIQ_Handlers[DISCARD_TIMER].func)
                (FIQ_Handlers[DISCARD_TIMER].func)(FIQ_Handlers[DISCARD_TIMER].data);
        }

        if (PCI_FIQ_Status & DATA_PARITY_INT)
        {
            if (FIQ_Handlers[DATA_PARITY].func)
                (FIQ_Handlers[DATA_PARITY].func)(FIQ_Handlers[DATA_PARITY].data);
        }

        if (PCI_FIQ_Status & PCI_MASTER_ABORT_INT)
        {
            if (FIQ_Handlers[PCI_MASTER_ABORT].func)
                (FIQ_Handlers[PCI_MASTER_ABORT].func)(FIQ_Handlers[PCI_MASTER_ABORT].data);
        }

        if (PCI_FIQ_Status & PCI_TARGET_ABORT_INT)
        {
            if (FIQ_Handlers[PCI_TARGET_ABORT].func)
                (FIQ_Handlers[PCI_TARGET_ABORT].func)(FIQ_Handlers[PCI_TARGET_ABORT].data);
        }

        if (PCI_FIQ_Status & PCI_PARITY_INT)
        {
            if (FIQ_Handlers[PCI_PARITY].func)
                (FIQ_Handlers[PCI_PARITY].func)(FIQ_Handlers[PCI_PARITY].data);
        }
#endif // #if INT_NOT_USED
    }
}


/*
**  int SA_FIQ_Enable(IntSource intSource)
**  int SA_FIQ_Disable(IntSource intSource)
**
**  DESCRIPTION
**      Enable/disable a FIQ interrupt.
** 
**  RETURNS
**      success:    SA_SUCCESS
**      error:      SA_FAIL
*/
int SA_FIQ_Enable(IntSource intSource)
{
    int retval = SA_SUCCESS;

    unsigned int origIntMask = SAr_DisableInt();

    switch (intSource)
    {
        case MICRO_THREAD:
            *(unsigned int*)IREG = (*(unsigned int*)IREG | UEFIQ_ENABLE) & 0xF0000000;
            break;

        case EXTERNAL_C_PIN:
            *(unsigned int*)IREG = (*(unsigned int*)IREG | CINTFIQ_ENABLE) & 0xF0000000;
            break;

        case REAL_TIME_CLOCK:
            *(unsigned int*)RTC_DIV |= (RTC_INT_ENABLE | RTC_WRITE_ENABLE);
            break;

        case SRAM_MEMORY:
            *(unsigned int*)SRAM_CSR =
                (*(unsigned int*)SRAM_CSR | SRAM_FIQ) & 0xFFFFFFFC;
            break;

        case SDRAM_MEMORY:
            *(unsigned int*)SDRAM_CSR |= (SDRAM_EN_INT | SDRAM_EN_FIQ);
            break;

        case PCI_DOORBELL:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DOORBELL_INT;
            break;

        case SOFT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = SOFT_INT;
            break;

        case TIMER1:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = TIMER1_INT;
            break;

        case TIMER2:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = TIMER2_INT;
            break;

        case TIMER3:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = TIMER3_INT;
            break;

        case TIMER4:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = TIMER4_INT;
            break;

        case DMA1:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DMA1_INT;
            break;

        case DMA2:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DMA2_INT;
            break;

        case PCI_IRQ_L:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = PCI_IRQ_L_INT;
            break;

        case DMA1_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DMA1_NOT_BUSY_INT;
            break;

        case DMA2_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DMA2_NOT_BUSY_INT;
            break;

        case START_BIST:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = START_BIST_INT;
            break;

        case SERR:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = SERR_INT;
            break;

        case SDRAM_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = SDRAM_PARITY_INT;
            break;

        case I2O_INPOST:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = I2O_INPOST_INT;
            break;

        case DISCARD_TIMER:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DISCARD_TIMER_INT;
            break;

        case DATA_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = DATA_PARITY_INT;
            break;

        case PCI_MASTER_ABORT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = PCI_MASTER_ABORT_INT;
            break;

        case PCI_TARGET_ABORT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = PCI_TARGET_ABORT_INT;
            break;

        case PCI_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_SET) = PCI_PARITY_INT;
            break;

        default:
            retval = SA_FAIL;
            break;
    }

    SAr_EnableInt(origIntMask);

    return(retval);
}


int SA_FIQ_Disable(IntSource intSource)
{
    int retval = SA_SUCCESS;

    unsigned int origIntMask = SAr_DisableInt();

    switch (intSource)
    {
        case MICRO_THREAD:
            *(unsigned int*)IREG =
                (*(unsigned int*)IREG & ~UEFIQ_ENABLE) & 0xF0000000;
            break;

        case EXTERNAL_C_PIN:
            *(unsigned int*)IREG =
                (*(unsigned int*)IREG & ~CINTFIQ_ENABLE) & 0xF0000000;
            break;

        case REAL_TIME_CLOCK:
            *(unsigned int*)RTC_DIV =
                (*(unsigned int*)RTC_DIV & ~RTC_INT_ENABLE) | RTC_WRITE_ENABLE;
            break;

        case SRAM_MEMORY:
            *(unsigned int*)SRAM_CSR =
                (*(unsigned int*)SRAM_CSR & ~SRAM_FIQ) & 0xFFFFFFFC;
            break;

        case SDRAM_MEMORY:
            *(unsigned int*)SDRAM_CSR &= ~SDRAM_EN_FIQ;
            break;

        case PCI_DOORBELL:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DOORBELL_INT;
            break;

        case SOFT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = SOFT_INT;
            break;

        case TIMER1:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = TIMER1_INT;
            break;

        case TIMER2:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = TIMER2_INT;
            break;

        case TIMER3:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = TIMER3_INT;
            break;

        case TIMER4:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = TIMER4_INT;
            break;

        case DMA1:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DMA1_INT;
            break;

        case DMA2:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DMA2_INT;
            break;

        case PCI_IRQ_L:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = PCI_IRQ_L_INT;
            break;

        case DMA1_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DMA1_NOT_BUSY_INT;
            break;

        case DMA2_NOT_BUSY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DMA2_NOT_BUSY_INT;
            break;

        case START_BIST:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = START_BIST_INT;
            break;

        case SERR:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = SERR_INT;
            break;

        case SDRAM_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = SDRAM_PARITY_INT;
            break;

        case I2O_INPOST:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = I2O_INPOST_INT;
            break;

        case DISCARD_TIMER:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DISCARD_TIMER_INT;
            break;

        case DATA_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = DATA_PARITY_INT;
            break;

        case PCI_MASTER_ABORT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = PCI_MASTER_ABORT_INT;
            break;

        case PCI_TARGET_ABORT:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = PCI_TARGET_ABORT_INT;
            break;

        case PCI_PARITY:
            *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = PCI_PARITY_INT;
            break;

        default:
            retval = SA_FAIL;
            break;
    }

    SAr_EnableInt(origIntMask);

    return(retval);
}


/*
**  void SA_FIQ_InitDoorBell()
**
**  DESCRIPTION
**      Initialize doorbell register.
**      Hook up an interrupt service routine to
**      doorbell register.
**      Enable doorbell interrupt.
** 
**  RETURNS
**      None.
*/
void SA_FIQ_InitDoorBell(PFVI intServiceRoutine)
{
    SA_FIQ_Disable(PCI_DOORBELL);

    // Setup doorbell register's directions
    //
	*(unsigned int*)(CSR_BASE + DBELL_PCI_MASK) = 0;
	*(unsigned int*)(CSR_BASE + DBELL_SA_MASK)  = 0xFFFFFFFF;

	*(unsigned int*)(CSR_BASE + DOORBELL_SETUP)  = 0xFFFFFFFF;

    SA_FIQ_Connect(PCI_DOORBELL, intServiceRoutine, 0);

    SA_FIQ_Enable(PCI_DOORBELL);
}


/*
**  void SA_FIQ_InitSoft()
**
**  DESCRIPTION
**      Hook up an interrupt service routine to
**      soft interrupt.
**      Enable soft interrupt.
** 
**  RETURNS
**      None.
*/
void SA_FIQ_InitSoft(PFVI intServiceRoutine)
{
    SA_FIQ_Disable(SOFT);
    SA_FIQ_Connect(SOFT, intServiceRoutine, 0);
    SA_FIQ_Enable(SOFT);
}




/*
**  void SAr_DisableAllIRQs()
**  void SAr_DisableAllFIQs()
**
**  DESCRIPTION
**      Disable all the interrupts IRQ & FIQ.
** 
**  RETURNS
**      None.
*/
void SAr_DisableAllIRQs(void)
{
    unsigned int origIntMask = SAr_DisableInt();

    *(unsigned int*)IREG &= ~(UEIRQ_ENABLE + CINTIRQ_ENABLE);
    *(unsigned int*)UART_CR = 0;
    *(unsigned int*)(CSR_BASE + IRQ_ENABLE_CLEAR) = ALL_BITS;

    *(unsigned int*)SRAM_CSR &= ~SRAM_IRQ;
    *(unsigned int*)SDRAM_CSR &= ~SDRAM_EN_IRQ;

    SAr_EnableInt(origIntMask);
}

void SAr_DisableAllFIQs(void)
{
    unsigned int origIntMask = SAr_DisableInt();

    *(unsigned int*)IREG &= ~(UEFIQ_ENABLE + CINTFIQ_ENABLE);
    *(unsigned int*)(CSR_BASE + FIQ_ENABLE_CLEAR) = ALL_BITS;

    SAr_EnableInt(origIntMask);
}


/*
**  void SAr_OsNotUseUART(void)
**  void SAr_OsUseUART(void)
**
**  DESCRIPTION
**      In the situation that OS works with Angel,
**      Angel has the whole control of the only UART.
**      When UART IRQ interrupt comes, OS need yield control to
**      Angel's ISR instead of processing it itself.
**
**      SAr_OsUseUART() let user have a choice later not to use Angel,
**      and return UART control back to OS.
** 
**  RETURNS
**      None.
*/
unsigned int SAv_OsNotUseUART = 0;

void SAr_OsNotUseUART(void)
{
    SAv_OsNotUseUART = 1;
}

void SAr_OsUseUART(void)
{
    SAv_OsNotUseUART = 0;
}

// end of file
