/* NetTrafficSimInterface.h
 * 
 * Defines the interface between the Developer Workbench and a dynamic-
 * link library (DLL) that simulates network traffic.
 *
 * ---------------------------------------------------------------------	
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ---------------------------------------------------------------------
 *
    This file must be included in the source file for a network traffic 
    simulation DLL which can be assigned to a IX bus device's port when 
    debugging using the Developer Workbench.

    When you Start Debugging, the Workbench loads the DLL and calls the
    Initialize() function immediately after the transactor initializes 
    the model.  This is the only time that this function is called. In
    the Initialize() function you can register console functions with the
    transactor.  You can then call these functions from a script in order
    configure your traffic simulation.  To call the transactor you must
    include the file xact_dll.h and link against xact_dll.lib.

    When the Workbench is activated to start receiving packets, it calls 
    the DLL to initialize each port.  If you have checked the option for
    packets to be received at the first clock cycle, then the initialization
    will occur at the first clock cycle.  Otherwise, it will occur when
    the start_packet_receive() function is invoked from your startup script.
    In either case, the Workbench iterates through all the ports on every 
    IX bus device.  If the DLL is connected to the receive side of the 
    port, then the function InitializeRxPort() is  called.  If it is 
    connected to the transmit side, then the function InitializeTxPort() 
    is called.  See the file NetTrafficSimInterface.h for the contents 
    of the port configuration data structures that are passed to these 
    initialization functions.

    As the simulation progresses, the Workbench calls function 
    GetNextByte() whenever a byte of data is required on a receive port.
    In addition to the byte of data, the DLL must also return a flag 
    indicating whether the byte is the last byte (EOP) in the frame/cell.  
    When the Workbench receives the last byte, it will call the 
    GetInterpacketTime() function in order to determine how long it waits 
    before asking for the first byte of the next frame/cell.  This allows 
    the DLL to randomize the arrival of frames/cells.
    
    On the transmit side, the Workbench calls the function SendNextByte() 
    when it takes a byte of data is out of the transmit buffer to be sent 
    over the network.  An EOP flag is asserted along with the last byte
    in the frame/cell.

    When the user Stops Debugging, the functions CloseRxPort() and 
    CloseTxPort() are called for each connected receive and transmit
    port, respectively.

    When the user closes the project or exits the Workbench, the function
    Close() is called just before the DLL is freed.
 */

#if !defined(__NetTrafficSimInterface_H)
#define __NetTrafficSimInterface_H

// Defines the data structure that is passed as a parameter when initializing
// a receive port connection.
typedef struct
{
    unsigned int m_DataRate;           // port's data rate in Mhz
    unsigned int m_InterPktTime;       // interpacket time in nanosecs
    unsigned int m_IxBusClockFreq;     // IX bus clock frequency, in Mhz
} RX_PORT_CONFIG_DATA;

// Defines the data structure that is passed as a parameter when initializing
// a transmit port connection.
typedef struct
{
    unsigned int m_DataRate;           // port's data rate in Mhz
    unsigned int m_InterPktTime;       // interpacket time in nanosecs
    unsigned int m_IxBusClockFreq;     // IX bus clock frequency, in Mhz
} TX_PORT_CONFIG_DATA;

/*-----------------------------------------------------------------
    Initialize

    This function is called once right after the DLL is loaded.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
Initialize();

/*-----------------------------------------------------------------
    Close

    This function is called once just before the DLL is freed.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
Close();

/*-----------------------------------------------------------------
    InitializeRxPort

    This function is called when debugging is started but before any
    simulation cycles are executed.  It is called once for each IX bus receive 
    port that is being simulated.  A unique handle is returned by the DLL to the
    Workbench.  The Workbench passes this handle along with other parameters
    when it requires data from the DLL for the specific port.
    
*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
InitializeRxPort(
    unsigned int DeviceNumber,             // IX bus device number
    unsigned int PortNumber,               // # of port to be initialized
    RX_PORT_CONFIG_DATA *pPortConfigData,  // pointer to port configuration data
    void **ppRxPortHandle);                // pointer to where unique handle is to be returned

/*-----------------------------------------------------------------
    GetNextByte

    This function returns the next byte of data coming from the network.
    If there is no data, the function returns false. Else, it returns true.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) bool __cdecl
GetNextByte(
    void *pPortHandle,          // pointer to unique port handle
    unsigned int Cycle,         // IX bus cycle count
    unsigned char *pDataByte,   // pointer to where byte is returned
    bool *pbEop);               // pointer to where boolean flag is returned to indicate if the byte
                                // is the last byte in packet

/*-----------------------------------------------------------------
    GetInterpacketTime

    This function returns the interpacket gap to be used before the next
    packet is requested from the DLL. The value is specified in # of nanoseconds. 

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) unsigned int __cdecl
GetInterpacketTime(
    void *pPortHandle);          // pointer to unique port handle

/*-----------------------------------------------------------------
    CloseRxPort

    This function is called when the user stops debugging in the Workbench.
    It is called once for each IX bus receive port that is being simulated.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
CloseRxPort(
    void *pPortHandle);          // pointer to unique port handle

/*-----------------------------------------------------------------
    InitializeTxPort

    This function is called when debugging is started but before any
    simulation cycles are executed.  It is called once for each IX bus receive 
    port that is being simulated.  A unique handle is returned by the DLL to the
    Workbench.  The Workbench passes this handle along with other parameters
    when it requires data from the DLL for the specific port.
    
*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
InitializeTxPort(
    unsigned int DeviceNumber,             // IX bus device number
    unsigned int PortNumber,               // # of port to be initialized
    TX_PORT_CONFIG_DATA *pPortConfigData,  // pointer to port configuration data
    void **ppTxPortHandle);                // pointer to where unique handle is to be returned

/*-----------------------------------------------------------------
    SendNextByte

    This function is called when a byte of data is transmitted 
    out of the IX bus device onto the network.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
SendNextByte(
    void *pTxPortHandle,        // pointer to unique port handle
    unsigned int Cycle,         // IX bus cycle count
    unsigned char DataByte,     // byte of frame/cell data
    bool bEop);                 // true if byte represents the last byte in a frame/cell

/*-----------------------------------------------------------------
    CloseRxPort

    This function is called when the user stops debugging in the Workbench.
    It is called once for each IX bus transmit port that is being simulated.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
CloseTxPort(
    void *pPortHandle);          // pointer to unique port handle


#endif      // #if !defined(__NetTrafficSimInterface_H)
