//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// rec_sch_fastest.uc
// receive scheduler 12 100Mbit ports that demonstrates receiving and storing data at 2gpbs
//
// Version = 1.0.NoBldNum
//
// system: SA1200
// subsystem: receive microcode Base Level 3
// usage: example
// author: dfh 10/8/98
// revisions:
//		dfh		12/16/98	BL4 remove include config.h
//
//
// ---------------------------SA1200 microcode--------------------------

// note1:	This is only demo code to prove that the receive bandwidth through fbus,
//			through receive fifo and to sdram can achieve 2 gigabits per second.

// note2:	do not assemble with optimize (there is an assembler bug that swaps 2
//			CSR reads, causing a signal error). Assemble with:
//			uca rec_sch_fastest
//	
// note3:	This is compatible with pfdv_nostat.ind. This script run 32 bit mode
//			with no status cycle over f bus. To set this mote, it does:
//			dep fbi.fbus.rdybus_template_ctl = 0x40
//			The script drives long packets 20 elements long.
//			This script sets fbus clock rate to 15ns with
//
// note4:	to run:
//	rtm.exe
//	@pfwdv_nostat
//	run_receives(50);				; to get past headers into pure data xfers
//	run_receives(100);				; run 100 elements through to sdram
//  calc_rate(100);					; get element rate (Oct 8, 98: 3952569 meps)
//									; gig bit rate is 512 bits * meps = 2.01 gbps
//
//			Using pfwd_watch.ind, you can see fbus cycles
//				cycle 1 first 32 bits of 64 byte element
//				cycle 16 last 32 bits of 64 byte element
//			It runs 16 cycles of data transfer, 1 idle cyle, and repeats, for maximum
//			fbus rate.
//
// caveat:  This code does speculative requests based on the ready bit sampling
//			there is no masking to prevent going to a port that looks ready at
//			the receive scheduler, but has just gone not ready. Usually this type of
//			speculative request is used for fast ports only (and fbi generates a
//			cancel message in receive control word 31:30, based on the fast port 
//			ready pins).
//			Every 50 cycles or so you can see ~6 aborts where the scheduler is waiting
//			for fbi to complete the previous receive request. During this time it
//			is possible to perform port masking without impacting the receive rate. 

// FBOX microcode assignment
//
//    1. the receive threads are thread ids 0-11
//    2. receive thread 0 always gets element 0, receive thread 1 gets element 1, etc.
//	  3. receive thread 0 always gets port 0, receive thread 1 gets port 1, etc.
//    3. receive scheduler is thread 16

// general design flow:

// key registers:
//
// @thread_done_capture		captured thd_done_reg0 CSR from CSR, read by context 1
// @rec_rdy					receive ready bits CSR from fbi, read by context 2
// thread_done_copy			context 0: copy of thread done being worked on
// current_thread			context 0: current thread index
// this_rr					context 0: receive request while being assembled

// context 0
// 1. get port and control store configuration from memory
// 2. initialize registers used in the main loop
// 3. main loop
//
//  a) calculate a new thread_done_copy from thread_done_skips OR'd with @thread_done_capture
//
//	b) minor loop
//      1) find first bit set in thread_done_copy
//                if found, look at 2 bit encode for that thread.
//		2) assemble and send receive request 
//		3) clear the thread done encode for thread just assign
//		4) if port was not ready, 0x3 into thread_done_skips for that thread
//		4) if thread_done_copy not yet 0, repeat minor loop              
//			else repeat main loop

// context 1
// 1. read thread_done1 register from fbi, repeat
//	

// context 2
// 1. read receive ready bits lo register and receive ready count from fbi, repeat
//


	
;-----------------------begin---------------------
#include "mem_map.h"
#include "stdmac.uc"
#include "refdes_macros.uc"

.xfer_order $xfer0 $xfer1 $xfer2
.operand_synonym $thread_done_reg0 $xfer2

; startup
	br=ctx[0, rec_scheduler#]		; receive scheduler
	br=ctx[1, tdone_reader#]		; context to read thread done
	br=ctx[2, rdybit_reader#]		; context to read ready bits and ready count
	ctx_arb[kill]					; kill contexts 1-3



rec_scheduler#:



// call	#macro freelist_create[freelist_id, base_addr, stride, count]
//		initialize freelist 0 for BUFFER_COUNT 2KB_packet_descriptors
//		see also: mem_map.h, stdmac.uc, stdfunc.uc
//
#ifndef BUFFER_COUNT
	#define BUFFER_COUNT FREELIST_BTYPE_SRAM8_SDRAM2K_COUNT
#endif
//
freelist_create[0, SRAM_BUFF_DESCRIPTOR_BASE, 4, BUFFER_COUNT]

freelist_created#:



// program fbi bus mode and rdybus sequencer program for transmit/receive autopush and rdy bit polling
// note: Core can do this by writing to the FBI CSRs
// prerequisite: define constants as described in 
//		Fbus_SetupRdyProg,  refdes_macros.uc
//
#ifdef REAL_RDYBUS_PROG
setup_rdyprg#:
	Fbus_SetupRdyProg
#endif



//----------------------------macros ---------------------------


#macro RxSched_Select[thread_fail_label, port_fail_label]		; 15 instr
	find_bset[thread_done_copy, >>16]							; find first bit set in 31:16
	alu[rec_rdy_true, --, B, @rec_rdy]
	nop
	nop
	load_bset_result1[current_threadx2]
	br=0[thread_fail_label], defer[1]							; no thread done set
	alu_shf[current_threadx2, current_threadx2, AND~, 1, <<8]	; clear the valid bit
	alu_shf[current_thread, --, B, current_threadx2, >>1]		; current threadx2 / 2

// clear the wait encode 2 bits of the thread_done_copy
    alu_shf[current_inv_threadx2, current_threadx2, B-A, 1, <<5]; setup left indirect shift
	alu[--, current_inv_threadx2, B, 0]
	alu[thread_done_copy, thread_done_copy, AND~, 0x3, <<indirect]

	alu[--, current_thread, B, 0]            					; setup shift
    alu[--, 1, AND, rec_rdy_true, >>indirect]    				; is port ready (port = thread)
	br>0[done#], guess_branch
set_skips#:														; this is entered if port is busy 
	alu[--, current_inv_threadx2, B, 0]
	alu[thread_done_skips, thread_done_skips, OR, 0x3, <<indirect]
	br[port_fail_label]												; no assignment because port is not ready
done#:
#endm


#macro RxSched_SelectNew[thread_fail_label, port_fail_label]	; 28
	find_bset[thread_done_copy], clr_results					; find first bit set in 15:0
	RxSched_Select[thread_fail_label, port_fail_label]
#endm
											


//----------------------------------------------------------------
// present format of RCV_REQ (3/13/98 transactor) is
// +-+------+-----+----+-----+-----+-----+----+-----+------+-------+------+
// | |fetch9| msg |stat| elem| elem|seq# |1or2|misc | sig  | thread| port |
// | |      | pkt |    | #2  | #1  |     |    |     | sched|       |      |
// | |   29 |28:27| 26 |25:22|21:18|17:16| 15 |14:12|  11  | 10:6  | 5:0  |
// +-+------+-----+----+-----+-----+-----+----+-----+------+-------+------+
//

// RxSched_WriteFirstRR
//		write a receive request where there is no previous receive request in progress
//
#macro RxSched_WriteFirstRR

// test to see if a receive request register has been made available by the FBI
rec_req_avail_check#:
	br_inp_state[rec_req_avail, _got_rr_avail#], defer[1]
#ifdef FETCH9
	alu_shf[this_rr, current_thread, OR, 1, <<29]	; fetch 9, insert port num into new receive request
#else
	alu[this_rr, --, B, current_thread]				; insert port number (= current thread)
#endif
	br[rec_req_avail_check#]

_got_rr_avail#:

// write receive request register
    alu_shf[this_rr, this_rr, OR, current_thread, <<6]				; insert thread num into new receive request
// this code does one element xfers
 
write_rr#:
    alu_shf[$xfer0, this_rr, OR, current_thread, <<18]				; insert element num into new receive request
    csr[write, $xfer0, rcv_req], sig_done							; send req to FBI
rx_assigned#:
#endm



// RxSched_WriteRR
//		write a receive request where there is an outstanding previous receive request
//
#macro RxSched_WriteRR												; 8 instr

// this will cause FBI to set RCV_CNTL

// write receive request register
	ctx_arb[fbi], defer[1]													; wait on
#ifdef FETCH9
	alu_shf[this_rr, current_thread, OR, 1, <<29]	; fetch 9, insert port num into new receive request
#else
	alu[this_rr, --, B, current_thread]				; insert port number (= current thread)
#endif

// test to see if a receive request register has been made available by the FBI
rec_req_avail_check#:
	br_inp_state[rec_req_avail, _got_rr_avail#], defer[1]
    alu_shf[this_rr, this_rr, OR, current_thread, <<6]				; insert thread num into new receive request
	br[rec_req_avail_check#]

_got_rr_avail#:



// this code does one element xfers
 
write_rr#:
    alu_shf[$xfer0, this_rr, OR, current_thread, <<18]				; insert element num into new receive request
    csr[write, $xfer0, rcv_req], sig_done							; send req to FBI
rx_assigned#:
#endm

//------------------------end macros ---------------------------

// context 0: job is to select thread/port  and write receive requests
//
rx_sched_first#:
	ctx_arb[voluntary]											; allow other contexts to run
	alu[--, --, B, @thread_done_capture]
	br=0[rx_sched_first#]
	alu[--, --, B, @rec_rdy]
	br=0[rx_sched_first#]

process_tdone_first#:
	alu[thread_done_copy, thread_done_skips, OR, @thread_done_capture]
	immed[@thread_done_capture, 0] 
	immed[thread_done_skips, 0]

rx_sched_first_loop#:
	ctx_arb[voluntary]											; allow other contexts to run
	RxSched_SelectNew[rx_sched_first#, rx_sched_first_loop#]
	RxSched_WriteFirstRR
	br[rx_sched_select#], defer[1]
	find_bset[thread_done_copy], clr_results					; find first bit set in 15:0

rx_sched_fail_loop#:
	ctx_arb[voluntary]

rx_sched_loop#:
	alu[thread_done_copy, thread_done_skips, OR, @thread_done_capture]
	immed[@thread_done_capture, 0] 
	immed[thread_done_skips, 0]
	RxSched_SelectNew[rx_sched_fail_loop#, rx_sched_select#]	; (26 instr)select with new thread done
	RxSched_WriteRR												; (8 instr)
	find_bset[thread_done_copy], clr_results					; find first bit set in 15:0

rx_sched_select#:
	RxSched_Select[rx_sched_fail_loop#, rx_sched_fail_iterate#]	; (15 instr)
	RxSched_WriteRR												; (8 instr)

rx_sched_iterate#:												; (3 instr major loop) (5 instr minor loop)
	alu[--, --, B, thread_done_copy]
	br=0[rx_sched_loop#]
	br[rx_sched_select#], defer[1]
	find_bset[thread_done_copy], clr_results					; find first bit set in 15:0

rx_sched_fail_iterate#:
	ctx_arb[voluntary]
	alu[--, --, B, thread_done_copy]
	br=0[rx_sched_loop#]
	br[rx_sched_select#], defer[1]
	find_bset[thread_done_copy], clr_results					; find first bit set in 15:0



// context 1: job is to read thread_done1 register to get receive thread done status
//
tdone_reader#:													; 7 instr overhead
	csr[read, $thread_done_reg0, THREAD_DONE_REG0], ctx_swap	; thread done two bits per thread
	alu[$thread_done_reg0, --, B, @thread_done_capture]			; write back to thread done to clear captured bits
	csr[write, $thread_done_reg0, thread_done_reg0], ctx_swap, defer[1]
 	alu[@thread_done_capture, @thread_done_capture, OR, $thread_done_reg0]	; take thread done snapshot
	br[tdone_reader#]



// context 2: job is to read ready bits and ready count
//
.operand_synonym $rec_rdy_count $xfer0
.operand_synonym $rec_rdy $xfer1

rdybit_reader_loop#:
	ctx_arb[voluntary]
rdybit_reader#:
	csr[read, $rec_rdy_count, RCV_RDY_CNT]
	csr[read, $rec_rdy, RCV_RDY_LO], ctx_swap
	ld_field_w_clr[rec_rdy_count, 0001, $rec_rdy_count]			; extract just the count
	alu[@rec_rdy_count, --, B, rec_rdy_count]
	br[rdybit_reader_loop#], defer[1]
	alu[@rec_rdy, --, B, $rec_rdy]
	

