 //////////////////////////////////////////////////////////////////////////////
//           INTEL CORPORATION PROPRIETARY INFORMATION
//  This software is supplied under the terms of a license agreement or
//  nondisclosure agreement with Intel Corporation and may not be copied
//  or disclosed except in accordance with the terms of that agreement.
//      Copyright (c) 1999-2000 Intel Corporation.  All Rights Reserved.
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/FE/FEAPI/IXP1200/MicroCode/src/Attic/IP_Utils.uc,v $
// $Revision: 1 $
// $Date: 4/18/00 5:06p $
//
// Description:
//    This file contains a macro to implement a version of IPv4 longest prefix 
//    match route lookups.
// Entry Points:
//    ip_route_lookup# (macro)

#include "Trie.uc"

//////////////////////////////////////////////////////////////////////////////
// Routine: ip_route_lookup()
//
// Parameters:
//    lastValidRoute - (OUT) resulting route pointer. index to route entry. 
//                  0 if no route entry.
//    ip_da       - (IN) IP destination address
//    xfer_reg    - (IN/OUT) an SRAM read transfer register to be used
//    const_lookup_base - (IN) base address of trie tables.  All trie's 
//                             are allocated from a block of SRAM memory.  
//                             This is the base 
//    head_ptr          - (IN) the pointer to the first trie block.  This
//                             value needs to be added to const_lookup_base
//                             in order to access the actual trie.
//
// Description:
//    Performs a route lookup (longest prefix match) for the given
//    IPv4 destination address.  In this particular implementation, only
//    trie blocks are used to store the route table.  This results in
//    SRAM memory usage proportional to the number of entries in the route
//    table, but worse, worst case performance (i.e., 8 memory references).
//
// Returns:
//    lastValidRoute - a pointer to the last valid route found, zero if none.
//    !!! This value is divided by two from the actual bits within the trie. !!!
//
// Register usage:
//    Uses three thread local variables: offset, bitPos and curPos (see 
//    algorithm).
//    In addition, the given SRAM read transfer register is used for temporary
//    storage
//
// Algorithm: 
//      uint32 curRtPtr;
//      uint32 *curPos = head_ptr;
//      uint32 bitPos = 28;
//
//      lastValidRoute = 0x0;
//      do {
//              // Get the next trie, store the current route ptr and update
//              // curPos
//              trie_next_block(ip_da, bitPos, curRtPtr, curPos, xfer_reg,
//                              const_lookup_base);
//              if(curRtPtr) {           
//                      lastValidRoute = curRtPtr;   
//              }                                      
//             	if (!curPos) {
//             		break;
//      	}
//              bitPos -= 4;
//      } while (bitPos >= 0);
//
//
// Comments:
//    It remains a mystery why I must divide the lastValidRoute pointer by two.
#macro ip_route_lookup[lastValidRoute, ip_da, xfer_reg, const_lookup_base, head_ptr]
.local bitPos curPos curRtPtr

    alu[curPos, head_ptr, -, 0]                 ; curPos = head_ptr;
    immed32[bitPos, 28]                         ; bitPos = 28;
    immed32[lastValidRoute, 0]                  ; lastValidRoute = 0x0;

    ; do {
    ip_route_lookup_do#:
        trie_next_block[ip_da, bitPos, curRtPtr, curPos, xfer_reg, const_lookup_base]

        ; if (curRtPtr) 
        ;     lastValidRoute = curRtPtr;
        alu[--, curRtPtr, -, 0]
        br=0[ip_not_valid_route#]
        alu[lastValidRoute, curRtPtr, -, 0]
        ip_not_valid_route#:

        ; if (!curPos) break;
        alu[--, curPos, -, 0]
        br=0[ip_no_more_tries#]

        ; bitPos -= 4; 
        ; } while (bitPos >= 0);
        alu[bitPos, bitPos, -, 4]
        br>=0[ip_route_lookup_do#]

    ip_no_more_tries#:
	// The following is an unknown hack.  For some reason the remaining pieces
    // of the reference design expect this
	alu_shf[lastValidRoute, --, B, lastValidRoute, >>1]
.endlocal
#endm
