//------------------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  1999-2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//-----------------------------------------------------------------------------------
// hdbm.uc
// IXPblock hash database access macros
//
//
// system: IXP1200
// subsystem: Hash Database Manager microcode
// usage: library macros
// author: drj Aug 6, 2000	
//
// revisions:


#ifndef HDBM_UC
#define HDBM_UC

#include "mem.uc"
#include "field.uc"
#include "stdmac.uc"

#define	HDBM_SUCCESS		0					// operaton successful
#define	HDBM_ERROR			1					// operation failed
#define	HDBM_NO_ENTRY		2					// entry not found

//
// The microcode macros in this module are a part of the Hash Database
// Manager (HDbM) which is  described in Section 2.1.3.4 of the IXP1200
// Software Reference Manual.  These macros allow the searching of a
// database that has been created by HDbM on the ARM-Core.  Please read the
// IXP1200 Network Processor Software Reference Manual for description on HDbM.
//


// API
//
// hdbm_lookup48
// hdbm_dual_lookup48
// hdbm_lookup64
// hdbm_dual_lookup64
// hdbm_lookup
// hdbm_dual_lookup




// hdbm_lookup48
//
// Description:					Search the database for the specified key and if found,
//								return the associated data value. A 48-bit hash is first
//								performed on the search key then the resultant is used
//								to search the database.  The effective bits, of the key,
//								that will be used in the search will be determined by the
//								primary and collision index descriptors:
//									prim_index_start_bit, prim_index_bit_len,
//								    coll_index_start_bit, coll_index_bit_len, and
//									coll_shift_bits.
//
// Outputs:
// 		out_stat				GPR -- Indicates the status of the operation:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data				GPR -- The associated user-data.
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper 16 bits and lower 32 bits of the key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
// Size:
//
// Side effects:
//
// See also:					IXP1200 Programmer's Reference manaual for information on
//								the HASH1_48 instruction and the affected bits.
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_lookup48(status, out_data, key_lo, key_hi, DB_HANDLE);
//
#macro hdbm_lookup48(out_stat, out_data, in_key_hi, in_key_lo,  prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $hash_reg $hash_reg1
		.xfer_order $hash_reg $hash_reg1

		move($hash_reg1, in_key_lo);
		move($hash_reg, in_key_hi);
		hash1_48[$hash_reg], ctx_swap		; hash the key
		move(in_key_lo, $hash_reg1);
		field_clr_insert(in_key_hi, 0011, $hash_reg);
	.endlocal

	hdbm_lookup(out_stat, out_data, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);

#endm


// hdbm_dual_lookup48
//
// Description:					The function of this macro is simillar to hdbm_lookup48,
//								except the two keys are searched for in parallel in the
//								primary table.  The status and their associated data are
//								returned in their respective out_stat and out_data values.
//
// Outputs:
// 		out_stat1, out_stat2	GPR -- the status of the lookup on key1 and key2:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data1, out_data2		GPR -- The data associated with key1 and key2.
// Inputs:
//		in_key1_hi, in_key1_lo	GPR -- Upper 16 bits and lower 32 bits of the first key
//		in_key2_hi, in_key2_lo	GPR -- Upper 16 bits and lower 32 bits of the second key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
// Size:
//
// Side effects:
//
// See also:					IXP1200 Programmer's Reference manaual for information on
//								the HASH1_48 instruction and the affected bits.
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_dual_lookup48(stat1, stat2, data1, data2, key1Hi, key1Lo,
//												   key2Hi, key2Lo, DB_HANDLE);
//
#macro hdbm_dual_lookup48(out_stat1, out_stat2, out_data1, out_data2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $hash_reg $hash_reg1 $hash_reg2 $hash_reg3
		.xfer_order $hash_reg $hash_reg1 $hash_reg2 $hash_reg3

		move($hash_reg1, in_key1_lo);
		move($hash_reg, in_key1_hi);
		move($hash_reg3, in_key2_lo);
		move($hash_reg2, in_key2_hi);
		hash2_48[$hash_reg], ctx_swap		; hash the key
		move(in_key1_lo, $hash_reg1);
		field_clr_insert(in_key1_hi, 0011, $hash_reg);
		move(in_key2_lo, $hash_reg3);
		field_clr_insert(in_key2_hi, 0011, $hash_reg2);
	.endlocal

	hdbm_dual_lookup(out_stat1, out_stat2, out_data1, out_data2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
#endm



// hdbm_lookup64
//
// Description:					Search the database for the specified key and if found,
//								return the associated data value.  A 64-bit hash is first
//								performed on the input key then the resultant is used to
//								search the database.  The effective bits, of the key, that
//								will be used in the search will be determined by the
//								primary and collision index descriptors:
//									prim_index_start_bit, prim_index_bit_len,
//									coll_index_start_bit, coll_index_bit_len, and
//									coll_shift_bits.
//
// Outputs:
// 		out_stat				GPR -- Indicates the status of the operation:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data				GPR -- The associated user-data.
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
// Size:
//
// Side effects:				
//
// See also:					IXP1200 Programmer's Reference manaual for information on the
//								HASH1_64 instruction.
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_lookup64(status, out_data, keyHi, keyLo, DB_HANDLE);
//
#macro hdbm_lookup64(out_status, out_data, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $hash_reg $hash_reg1
		.xfer_order $hash_reg $hash_reg1

		move($hash_reg1, in_key_lo);
		move($hash_reg, in_key_hi);
		hash1_64[$hash_reg], ctx_swap		; hash the key
		move(in_key_lo, $hash_reg1);
		move(in_key_hi, $hash_reg);
	.endlocal

	.local $table_val
		_hdbm_read_prim_table($table_val, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len);
		_hdbm_resolve(status, out_data, $table_val, in_key_hi, in_key_lo, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
	.endlocal
#endm




// hdbm_dual_lookup64
//
// Description:					The function of this macro is simillar to hdbm_lookup64,
//								except the two keys are searched for in parallel in the
//								primary table.  The status and their associated data are
//								returned in their respective out_stat and out_data values.
//
// Outputs:
// 		out_stat1, out_stat2	GPR -- the status of the lookup on key1 and key2:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data1, out_data2		GPR -- The data associated with key1 and key2.
// Inputs:
//		in_key1_hi, in_key1_lo	GPR -- Upper and lower 32 bits of the first search key
//		in_key2_hi, in_key2_lo	GPR -- Upper and lower 32 bits of the second search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
// Size:
//
// Side effects:
//
// See also:					IXP1200 Programmer's Reference manaual for information on
//								the HASH1_48 instruction and the affected bits.
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_dual_lookup64(stat1, stat2, data1, data2, key1Hi, key1Lo,
//												   key2Hi, key2Lo, DB_HANDLE);
//
#macro hdbm_dual_lookup64(out_stat1, out_stat2, out_data1, out_data2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $hash_reg $hash_reg1 $hash_reg2 $hash_reg3
		.xfer_order $hash_reg $hash_reg1 $hash_reg2 $hash_reg3

		move($hash_reg1, in_key1_lo);
		move($hash_reg, in_key1_hi);
		move($hash_reg3, in_key2_lo);
		move($hash_reg2, in_key2_hi);
		hash2_64[$hash_reg], ctx_swap		; hash the key
		move(in_key1_lo, $hash_reg1);
		move(in_key1_hi, $hash_reg);
		move(in_key2_lo, $hash_reg3);
		move(in_key2_hi, $hash_reg2);
	.endlocal
	hdbm_dual_lookup(out_stat1, out_stat2, out_data1, out_data2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
#endm



// hdbm_lookup
//
// Description:					Searches the database for a specified key and returns
//								the associated data value if the key was found.
//								Note: no hash is performed on the key.
//
// Outputs:
// 		out_stat				GPR -- Indicates the status of the operation:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data				GPR -- The associated user-data.
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used
//												for key/data storage
//
//
// Size: 						? instructions
//
// Side effects:				Cannot be the last statement -- another instruction must
//								follow the invocation of this macro.
//
// See also:
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_lookup(stat, data, keyHi, keyLo, DB_HANDLE);
//
#macro hdbm_lookup(out_status, out_data, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $table_val
		_hdbm_read_prim_table($table_val, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len);
		_hdbm_resolve(out_status, out_data, $table_val, in_key_hi, in_key_lo, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
	.endlocal
#endm



// hdbm_dual_lookup
//
// Description:					Searches the database for the specified keys and returns
//								their status and associated data if the key was found.
//								Note: no hash is performed on the key.
//
// Outputs:
// 		out_stat				GPR -- Indicates the status of the operation:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data				GPR -- The associated user-data.
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
//
// Size: 						? instructions
//
// Side effects:
//
// See also:
//
// Example Usage:				#define DB_HANDLE prim_table_base, PRIM_INDX_START_BIT, \
//								PRIM_INDX_BIT_LEN, COL_INDX_START_BIT, COL_INDX_BIT_LEN, \
//								COL_SHF_BITS, COL_TABLE_BASE, KEYDATA_SD_BASE
//								hdbm_lookup(stat, data, keyHi, keyLo, DB_HANDLE);
//
#macro hdbm_dual_lookup(out_status1, out_status2, out_data1, out_data2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local $table_val1 $table_val2
		_hdbm_bi_read_prim_table($table_val1, $table_val2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len);
		_hdbm_resolve(out_status1, out_data1, $table_val1, in_key1_hi, in_key1_lo, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
		_hdbm_resolve(out_status2, out_data2, $table_val2, in_key2_hi, in_key2_lo, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base);
	.endlocal
#endm



// _hdbm_resolve
//
// Description:					Determine the status of the value read from the primary
//								table and return the data if there was not conflict.
//								If there was a conflict, then search the secondary table
//								for the specified key.
//
// Outputs:
// 		out_status				GPR -- Indicates the status of the operation:-
//									HDBM_SUCCESS -- key found and out_data valid.
//									HDBM_NO_ENTRY -- key not found and out_data invalid.
//									HDBM_ERROR -- an error occured and out_data invalid.
//		out_data				GPR -- The associated user-data, if found.
// Inputs:
//		$table_val				SRAM XFER -- The value read from the primary table.
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		coll_index_start_bit	GPR or CONST -- Collision table index starting-bit position
//		coll_index_bit_len		GPR or CONST -- Collision table index bit length
//		coll_shift_bits			GPR or CONST -- Number of bits to step the collision table
//												index after a collision
//		coll_freelist_base		GPR or CONST -- List of available SRAM space to be used for
//												collision tables -- currently being used
//		key_data_freelist_base	GPR or CONST -- List of available SDRAM space to be used for
//												key/data storage
//
//
// Size: 						? instructions
//
// Side effects:
//
// See also:
//
// Example Usage:
//
#macro _hdbm_resolve(out_status, out_data, $table_val, in_key_hi, in_key_lo, coll_index_start_bit, coll_index_bit_len, coll_shift_bits, coll_freelist_base, key_data_freelist_base)
	.local fwd_entry _hdbm_index table_base col_sht_amt $$xfer_a $$xfer_b $$xfer_c $$xfer_d
		.xfer_order $$xfer_a $$xfer_b $$xfer_c $$xfer_d

		move(col_sht_amt, 0);

		top_of_loop#:
			.if($table_val != 0)
				br[entry_or_collision#]
			.else
				// if 0, entry not found
				br[not_found#]
			.endif


		// if collision bit set, add the entry to collision table
		//
		entry_or_collision#:				
			and_shf_left(--, $table_val, 1, 31);
			br=0[entry_and_no_collision#]

			// collision and existing entry 
			_get_entry_index(table_base, $table_val);				;get collision table reference
			.local start_bit
				add(start_bit, coll_index_start_bit, col_sht_amt);
				.if(start_bit > coll_index_bit_len)
					move(out_status, HDBM_ERROR);
					br[exit#]
				.endif

				add(col_sht_amt, coll_shift_bits, col_sht_amt);			;incr shft amt to avoid doing mult
				_get_index_field(_hdbm_index, in_key_lo, in_key_hi, start_bit, coll_shift_bits);
			.endlocal

		next_read#:
			// read the table
		    sram[read, $table_val, table_base, _hdbm_index, 1], ctx_swap, optimize_mem ; lookup hash table
			br[top_of_loop#]

		// no collision and existing entry,
		entry_and_no_collision#:
			_get_entry_index(fwd_entry, $table_val);					;get collision table reference

			// read and compare keys
			add(fwd_entry, fwd_entry, key_data_freelist_base);
			sub(fwd_entry, fwd_entry, 1);
			sdram[read, $$xfer_a, fwd_entry, 0, 2], ctx_swap		;read entry key
			.if(($$xfer_a == in_key_lo) && ($$xfer_b == in_key_hi))
				move(out_data, $$xfer_c);
				move(out_status, HDBM_SUCCESS);
				br[exit#]
			.endif

		not_found#:
			move(out_status, HDBM_NO_ENTRY);

		exit#:
	.endlocal
#endm


// _hdbm_bi_read_prim_table
//
// Description: 				Perform a parallel search of the primary table for the
//								specified keys.
//
// Outputs:
//		$out_value1				GPR or SRAM XFER
//		$out_value2				GPR or SRAM XFER
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//
// Size: 					? instructions
//
// Side effects:			none
//
// See also:
//
// Example Usage:			_get_entry_index(index, reg);
//
#macro _hdbm_bi_read_prim_table($out_value1, $out_value2, in_key1_hi, in_key1_lo, in_key2_hi, in_key2_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len)
	.local _hdbm_index _hdbm_index2
		.local start_bit
			//get the index from the key based on the starting bit position and the length
			move(start_bit, prim_index_start_bit);
			_get_index_field(_hdbm_index, in_key1_lo, in_key1_hi, start_bit, prim_index_bit_len);
			_get_index_field(_hdbm_index2, in_key2_lo, in_key2_hi, start_bit, prim_index_bit_len);
		.endlocal

		// read the primary table
		#if((isnum(prim_table_base)))
			.local table_base
				move(table_base, prim_table_base);
				sram[read, $out_value1, table_base, _hdbm_index, 1]
				sram[read, $out_value2, table_base, _hdbm_index2, 1], ctx_swap
			.endlocal
		#else
			sram[read, $out_value1, prim_table_base, _hdbm_index, 1]
			sram[read, $out_value2, prim_table_base, _hdbm_index2, 1], ctx_swap
		#endif
	.endlocal
#endm


// _hdbm_read_prim_table
//
// Description: 			Perform a parallel search of the primary table for the
//							specified key.
//
// Outputs:
//		$out_value				GPR or SRAM XFER
// Inputs:
//		in_key_hi, in_key_lo	GPR -- Upper lower 32 bits of the search key
//		prim_table_base			GPR or CONST -- SRAM base of the primary table
//		prim_index_start_bit	GPR or CONST -- Primary table index starting-bit position
//		prim_index_bit_len		GPR or CONST -- Primary table index bit length
//
// Size: 					? instructions
//
// Side effects:			none
//
// See also:
//
// Example Usage:
//
#macro _hdbm_read_prim_table($out_value, in_key_hi, in_key_lo, prim_table_base, prim_index_start_bit, prim_index_bit_len)
	.local _hdbm_index
		.local start_bit
			//get the index from the key based on the starting bit position and the length
			move(start_bit, prim_index_start_bit);
			_get_index_field(_hdbm_index, in_key_lo, in_key_hi, start_bit, prim_index_bit_len);
		.endlocal

		// read the primary table
		#if((isnum(prim_table_base)))
			.local table_base
				move(table_base, prim_table_base);
				sram[read, $out_value, table_base, _hdbm_index, 1], ctx_swap
			.endlocal
		#else
			sram[read, $out_value, prim_table_base, _hdbm_index, 1], ctx_swap
		#endif
	.endlocal
#endm


// _get_entry_index
//
// Description: Extract the index-entry bits <29:0> from the a 32-bit word
//
// Outputs:
//		out_index			Resulting index or entry value
// Inputs:
//		in_value			A register containing the input value
//
// Size: 					1-3 instructions
//
// Side effects:			none
//
// See also:
//
// Example Usage:			_get_entry_index(index, reg);
//
#macro _get_entry_index(out_index, in_value)
	alu_op(out_index, in_value, AND, 0x3fffffff);
#endm



// _get_index_field
//
// Description:				Obtain a 1 thru 32 bit-field from a quad-word.
//
// Output:
//		out_index 			The resulting index field
// Input:
///		in_word_lo			The lower 32 bits of the 64-bit value
//		in_word_hi			The upper 32 bits of the 64-bit value
//		in_start_bit		Starting bit position of the field
//		in_field_len		Number of bit in the field
//
// Size: 					5-8 instructions
//
// Side effects:			No checks are made to insure that the arguments are of
//							reasonable values.
//
// See also:
//
// Example Usage: _get_index_field(index, word_lo, word_hi, 0, 16); // extracts the lower 16 bits
//
#macro _get_index_field(out_index, in_word_lo, in_word_hi, in_start_bit, in_field_len)
	#define BITS_IN_WORD		32					// number of bits in the word
	#define LAST_BIT_IN_WORD	31					// the last bit the word

	// if(start > LAST_BIT_IN_WORD)
	sub(--, LAST_BIT_IN_WORD, in_start_bit);
	br!=cout[wrd_hi#]

	// start <= LAST_BIT_IN_WORD
	alu[--, in_start_bit, B, 0]
	dbl_shf[out_index, in_word_hi, in_word_lo, >>indirect]
	br[trim_to_len#]

	wrd_hi#:
		// start > LAST_BIT_IN_WORD
		shf_right(out_index, in_word_hi, in_start_bit);

	trim_to_len#:
		#if((isnum(in_field_len)))
			#define_eval _hdbm_tmp (BITS_IN_WORD - in_field_len)
			shf_left(out_index, out_index, _hdbm_tmp);		;out_index <<= (BITS_IN_WORD - _field_len)	
			shf_right(out_index, out_index, _hdbm_tmp);		;out_index <<= (BITS_IN_WORD - _field_len)
		#else
			.local _hdbm_tmp
				sub(_hdbm_tmp, BITS_IN_WORD, in_field_len);
				shf_left(out_index, out_index, _hdbm_tmp);	;out_index <<= (BITS_IN_WORD - _field_len)	
				shf_right(out_index, out_index, _hdbm_tmp);	;out_index <<= (BITS_IN_WORD - _field_len)
			.endlocal
		#endif

	#undef BITS_IN_WORD
	#undef LAST_BIT_IN_WORD
#endm

#endif // HDBM_UC
