//------------------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  1999-2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//-----------------------------------------------------------------------------------
// mem.uc
// IXPblock memory allocation macros
//
//
// system: IXP1200
// subsystem: IO microcode
// usage: library macros
// author: dfh Jan 12, 1999	
//
// revisions:	dfh Apr 6, 2000		separated from stdmac.uc


#ifndef MEM_UC
#define MEM_UC

// API

// mem_freelist_create
// mem_pop
// mem_wait
// mem_ds_alloc
// mem_alloc
// mem_free



// mem_freelist_create
//		description: Create freelist of available memory buffers using push/pop regs.
//
//		inputs:
//			NUM_BUFFERS			number of buffers put on the freelist
//			STACK_ID			0-7, to identify one of 8 hardware push/pop stacks
//			S_BASE				Base address of SRAM buffer
//			S_SIZE				Size of SRAM buffers
//			EOL_VALUE			End of list value.  When this value is Poped from the freelist
//								there are no buffers left.
//		size:	9 instr
//		example usage:
//			#define FREELIST0_HANDLE 0,100, 4, 0
//			mem_freelist_create(300, FREELIST0_HANDLE);
///
#macro mem_freelist_create[NUM_BUFFERS, STACK_ID, S_BASE, S_SIZE, EOL_VALUE]
.local  buf_addr ii count $_xfer_reg

	immed32[buf_addr, S_BASE]						; base address for memory region
	immed[ii, 0]
	immed32[count, NUM_BUFFERS]						; num entries

freelist_loop#:

	sram[push, --, buf_addr, ii, STACK_ID]			; push to freelist
	alu[ii, ii, +, S_SIZE]							; add sram buffer size
	alu[count, count, -, 1]
	br>0[freelist_loop#]

	immed32[$_xfer_reg, EOL_VALUE]
	sram[write, $_xfer_reg, buf_addr, 0, 1], ctx_swap	; end of list is set to EOL_VALUE

.endlocal
#endm


// mem_pop
// 		description: Pop a pointer from the push/pop stack identified by STACK_ID.
//		outputs:
//			USE_XFER			which sram read xfer to place the popped pointer, 0-7
//		inputs:
//			STACK_ID			0-7, to identify one of 8 hardware push/pop stacks
//			EOL_VALUE			End of list value.  When this value is Poped from the freelist
//								there are no buffers left.
//			sig_spec			pend:  waiting on signal, context swap
//								sig:   requesting signal, but continuing
//								nosig: not allowed. error
//		size: 1 - 6 instr
//		example usage:
//			mem_pop(X6, STACK_ID, EOL_VALUE, sig);
//
#macro mem_pop[USE_XFER, STACK_ID, EOL_VALUE, sig_spec]
.local _temp
	#if (isnum (sig_spec))
		#error 2  "Error: mem_pop. pend or sig is required."
	#endif
wait_for_good_pop#:
	// note: STACK_ID is a dummy unused argument, it prevents an assembler warning
	sram[pop, USE_XFER, USE_XFER, 0, STACK_ID], sig_spec
	#if (strstr(sig_spec, ctx_arb))		; if we waited for the pop
		#if (isnum(EOL_VALUE))			; if immed
			#if((EOL_VALUE > 0x1f) || (EOL_VALUE < 0))
				immed32[_temp, EOL_VALUE]
				alu[--, USE_XFER, -, _temp]
			#else	; immed < 0x1f and > 0
				alu[--, USE_XFER, -, EOL_VALUE]
			#endif
		#else ; they passed in a register
			alu[--, USE_XFER, -, EOL_VALUE]
		#endif
		br=0[wait_for_good_pop#]	; if popped value = end of list marker, loop back
	#endif ; ctx_arb on pop
.endlocal
#endm


// mem_wait
//		description: Wait on a signal.
//		inputs:
//			mem_type	sram, sdram, or other signal
//		size: 1  instr
//		example usage:
//			mem_wait[sram]
//
#macro mem_wait[mem_type]
	ctx_arb[mem_type]
#endm

// sig_wait
// later: to be moved to sig.uc
#macro sig_wait[mem_type]
	ctx_arb[mem_type]
#endm


// mem_ds_alloc
// 		description: Given a freelist handle and popped ptr, calculate SDRAM and SRAM buffer addresses
//		outputs:
//			out_sd_addr			quadword address of buffer in SDRAM
//			out_s_addr				longword address of buffer in SRAM
//		inputs:
//			STACK_ID			which freelist to use for retry
//			D_BASE				base address of sdram buffers
//			D_SIZE				size of sdram buffers
//			S_BASE				base address of sram buffers
//			S_SIZE				size of sram buffers	
//			EOL_VALUE			End of list value.  When this value is Poped from the freelist
//								there are no buffers left.
//			in_pop_xfer			popped pointer (from mem_pop), which transfer reg to use for retry
//		size: 9 instr
//		example usage:
//			#define FREELIST0 0
//			#define EOL_VALUE 0
//			#define FREELIST0_HANDLE 0,100,2048,500,4, 0
//			mem_pop(X6, FREELIST0, EOL_VALUE);
//			mem_ds_alloc(packet_buf_addr, descriptor_addr, FREELIST0_HANDLE, X6);
//
#macro mem_ds_alloc[out_sd_addr, out_s_addr, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE, EOL_VALUE, in_pop_xfer]
.local buffer_base descriptor_base relative_addr
started#:
// test free buffer pool delivered by the pop 
//
    alu[--, --, B, in_pop_xfer]									; if zero, the free list is empty
    br=0[no_buf_available#], defer[1]							; wait for buffer to free up

    immed32[buffer_base, D_BASE]
	immed32[descriptor_base, S_BASE]
	alu[relative_addr, in_pop_xfer, -, descriptor_base]

// note: D_SIZE is quadwords, S_SIZE is longwords
// calculate ratio, to get correct shift amount
// this will support up to 64kB sdram buffer with 4 word sram buffer
#define_eval SD_MULTIPLIER (D_SIZE / S_SIZE)
#if (SD_MULTIPLIER == 1)
#define_eval SD_SHIFT (0)
#elif (SD_MULTIPLIER == 2)
#define_eval SD_SHIFT (1)
#elif (SD_MULTIPLIER == 4)
#define_eval SD_SHIFT (2)
#elif (SD_MULTIPLIER == 8)
#define_eval SD_SHIFT (3)
#elif (SD_MULTIPLIER == 16)
#define_eval SD_SHIFT (4)
#elif (SD_MULTIPLIER == 32)
#define_eval SD_SHIFT (5)
#elif (SD_MULTIPLIER == 64)
#define_eval SD_SHIFT (6)
#elif (SD_MULTIPLIER == 128)
#define_eval SD_SHIFT (7)
#elif (SD_MULTIPLIER == 256)
#define_eval SD_SHIFT (8)
#elif (SD_MULTIPLIER == 512)
#define_eval SD_SHIFT (9)
#elif (SD_MULTIPLIER == 1024)
#define_eval SD_SHIFT (10)
#elif (SD_MULTIPLIER == 2048)
#define_eval SD_SHIFT (11)
#else
#error 2 "Error: mem_ds_alloc. sdram buffer size/ sram buffer size too large"
#endif
    alu[out_sd_addr, buffer_base, +, relative_addr, <</**/SD_SHIFT/**/]	; calculate sdram address

	br[done#], defer[1]
	alu[out_s_addr, descriptor_base, +, relative_addr]		; calculate sram address
no_buf_available#:
// retry until buffer is available
// note: third argument here is ignored for sram pop
	sram[pop, in_pop_xfer, in_pop_xfer, 0, STACK_ID], ctx_swap
	br[started#]
done#:
.endlocal
nop
#endm


// pre v1.0. THIS IS TO BE REMOVED in v1.1
// freelist_create
//		description: Create freelist of available memory buffers using push/pop regs.
//
//		inputs:
//			STACK_ID_id				0-7, identify which of 8 hardware push/pop stacks
//			CONST_BASE_ADDR			starting address in sram for the memory region
//			CONST_STRIDE			number of longwords per entry, including a first longword
//									which hardware uses as the node pointer
//			CONST_NUM_ENTRIES		number of nodes to be created
//		size:	9 - 10 instr
//		example usage:
//			mem_freelist_create(FREELIST0, SRAM_BUFFdescriptor_base, LWCOUNT4, 300);
//
#macro freelist_create[STACK_ID_id, CONST_BASE_ADDR, CONST_STRIDE, CONST_NUM_ENTRIES]
.local  buf_addr ii count $_xfer_reg temp

	immed32[buf_addr, CONST_BASE_ADDR]					; base address for memory region
	immed[ii, 0]
	immed32[count, CONST_NUM_ENTRIES]					; num entries

freelist_loop#:

	sram[push, --, buf_addr, ii, STACK_ID_id]		; push to freelist
	#if ((CONST_STRIDE < 0x1f) && (CONST_STRIDE > 0))
		alu[ii, ii, +, CONST_STRIDE]				; add stride
	#else ; need to use a temp GPR
		immed32[temp, CONST_STRIDE]
		alu[ii, ii, +, temp]
	#endif
	alu[count, count, -, 1]
	br>0[freelist_loop#]

	immed[$_xfer_reg, 0]
	sram[write, $_xfer_reg, buf_addr, 0, 1], ctx_swap		; end of list is set to 0

.endlocal
#endm

// pre v1.0. THIS IS TO BE REMOVED in v1.1
// malloc
//		description: Get next available address (mem_ptr) from freelist (freelist_id)
//					 using the read transfer reg read_xfer_reg
//
#macro malloc[mem_ptr, freelist_id, read_xfer_reg]
.local tempa tempb
	sram[pop, read_xfer_reg, tempa, 0, freelist_id], ctx_swap

#define_eval FREELIST_IDENT freelist_id
#if (FREELIST/**/FREELIST_IDENT/**/_TYPE == STORAGE_CLASS_SDRAM)
	immed32[tempa, FREELIST/**/FREELIST_IDENT/**/_PTR_BASE]						; get sram ptr base
	alu[tempa, read_xfer_reg, -, tempa]											; calculate pointer offset
	immed32[tempb, FREELIST/**/FREELIST_IDENT/**/_BASE]							; get sdram base
	alu[mem_ptr, tempb, +, tempa, <<FREELIST/**/FREELIST_IDENT/**/_PTR_SHIFT]	; add shifted offset to base
#else
	move[mem_ptr, read_xfer_reg]
#endif
.endlocal
#endm

// pre v1.0. THIS IS TO BE REMOVED in v1.1
// free
//		description: Free an address (mem_ptr) by pushing it onto freelist (freelist_id)
//					 using the write transfer reg write_xfer_reg
//
#macro free[mem_ptr, freelist_id]
.local tempa tempb
#define_eval FREELIST_IDENT freelist_id
#if (FREELIST/**/FREELIST_IDENT/**/_TYPE == STORAGE_CLASS_SDRAM)
	immed32[tempa, FREELIST/**/FREELIST_IDENT/**/_BASE]							; get sdram base
	alu[tempa, mem_ptr, -, tempa]												; subtract base from sdram memptr
	alu[tempa, --, B, tempa, >>FREELIST/**/FREELIST_IDENT/**/_PTR_SHIFT]		; shift offset
	immed32[tempb, FREELIST/**/FREELIST_IDENT/**/_PTR_BASE]						; get sram ptr base
	sram[push, --, tempa, tempb, freelist_id], ctx_swap							; push ptr base + shifted offset
#else
	sram[push, --, mem_ptr, 0, freelist_id], ctx_swap
#endif

.endlocal
#endm


#endif // MEM_UC