//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// sdram.uc
// macros to access sdram
//
//
// system: IXP1200
// subsystem: utility microcode
// usage: library macros
// author: dfh	May 1, 2000
//
// revisions:


#ifndef SDRAM_UC
#define SDRAM_UC


// API:
//	sdram_read[out_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT, sig_spec]
//	sdram_write[in_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT, sig_spec]
//  sdram_rfifo_read[in_sd_qw_addr, in_sd_offset, in_rfifo_base_addr, CONST_RFIFO_OFFSET, QW_COUNT, sig_spec]


// sdram_read
//		description: Read 1-4 quadwords from sdram to io_xfer registers.
//		outputs:
//			out_xfer		out_xfer register with first longword of write data
//		inputs:
//			in_base_qw_addr	quadword base address to be added to form address
//			in_qw_offset	quadword offset to be added to form address
//			CONST_QW_COUNT	number of quadwords to write
//			sig_spec		pend:  waiting on signal, context swap
//							sig:   requesting signal, but continuing
//							nosig: not requesting signal, continuing
//		size: 1 instr
//		see also: constants.uc
//		example usage:
//			sdram_read(DX0, packet_buf_addr, QWOFFSET0, QWCOUNT4, nosig);
//
#macro sdram_read[out_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT, sig_spec]
#if (isnum(sig_spec))
	sdram[read, out_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT]
#else
	sdram[read, out_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT], sig_spec
#endif
#endm


// sdram_write
// 		description: Write 1-4 quadwords from in_xfer registers to sdram.
//		outputs:
//		inputs:
//			in_xfer			in_xfer register with first longword of read data
//			in_base_qw_addr	quadword base address to be added to form address
//			in_qw_offset	quadword offset to be added to form address
//			CONST_QW_COUNT	number of quadwords to write
//			sig_spec		pend:  waiting on signal, context swap
//							sig:   requesting signal, but continuing
//							nosig: not requesting signal, continuing
//		size: 1 instr
//		see also: constants.uc
//		example usage:
//			sdram_read(DX0, packet_buf_addr, QWOFFSET0, QWCOUNT4, nosig);
//
#macro sdram_write[in_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT, sig_spec]
#if (isnum(sig_spec))
	sdram[write, in_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT]
#else
	sdram[write, in_xfer, in_base_qw_addr, in_qw_offset, CONST_QW_COUNT], sig_spec
#endif
#endm


// sdram_rfifo_read
// 		description: Copy 1-16 quadwords from rfifo(base+offset) to sdram.
//		outputs:
//		inputs:
//			in_sd_qw_addr		starting quadword address at sdram
//			in_rfifo_base_addr	quadword base address to be added to form rfifo address
//			CONST_RFIFO_OFFSET	quadword offset to be added to form rfifo address
//			in_qw_count			GPR or constant. number of quadwords to write
//			sig_spec			pend:  waiting on signal, context swap
//								sig:   requesting signal, but continuing
//								nosig: not requesting signal, continuing
//		size: 1-5 instr
//		see also: ixplib.uc
//		example usage:immed[temp1, 0x0]			; start sdram addr 0
//                    immed[temp2, 0x0]			; start r_fifo addr 0
//                    sdram_rfifo_read[temp1, 0, temp2, 0, 8, pend]
//
#macro sdram_rfifo_read[in_sd_qw_addr, in_sd_offset, in_rfifo_base_addr, CONST_RFIFO_OFFSET, in_qw_count, sig_spec]
.local _tempa

#if (isnum(in_qw_count))
#if (CONST_RFIFO_OFFSET == 0)
	alu_shf[--, --, B, in_rfifo_base_addr, <<4]
#else
	alu_shf[_tempa, --, B, CONST_RFIFO_OFFSET, <<4]
	alu_shf[--, _tempa, +, in_rfifo_base_addr, <<4]
#endif
#else	// in_qw_count is a register
// use indirect count (bit 20 overide, 16-19 count)
	alu[_tempa, --, B, 1, <<20]
	alu[_tempa, _tempa, +, in_qw_count, <<16]//
#if (CONST_RFIFO_OFFSET == 0)
	alu_shf[--, _tempa, OR, in_rfifo_base_addr, <<4]
#else
	alu_shf[_tempa, _tempa, OR, CONST_RFIFO_OFFSET, <<4]
	alu_shf[--, _tempa, +, in_rfifo_base_addr, <<4]
#endif
#endif	// in_qw_count is a register
#if (isnum(in_qw_count))

#if (isnum(sig_spec))
	sdram[r_fifo_rd, --, in_sd_qw_addr, in_sd_offset, in_qw_count], indirect_ref
#else
	sdram[r_fifo_rd, --, in_sd_qw_addr, in_sd_offset, in_qw_count], indirect_ref, sig_spec
#endif
#else
#if (isnum(sig_spec))
	sdram[r_fifo_rd, --, in_sd_qw_addr, in_sd_offset, 1], indirect_ref
#else
	sdram[r_fifo_rd, --, in_sd_qw_addr, in_sd_offset, 1], indirect_ref, sig_spec
#endif

#endif
.endlocal
#endm



// DO NOT USE. pre v1.0. This will be removed in v1.1
//
//sdram_r_fifo_rd
//		description: Transfer r_fifo fixed count to sdram.
//		inputs:
//			sd_addr			sdram quadword address
//			r_fifo_addr		r_fifo quadword address
//			const_offset	quadowrd offset applied to both addresses
//			CONST_QW_COUNT	number of quadwords to transfer
//			const_sig_spec	SYNC	ctx_swap
//							ASYNC	sig_done
//						    NOSYNC	(no signal)
//		size:	
//		example usage:
//			// transfer 8 quadwords to/from address offset 4
//			sdram_r_fifo_rd[sdaddr, raddr, 4, 8, SYNC]
//
#macro sdram_r_fifo_rd[sd_addr, r_fifo_addr, const_offset, CONST_QW_COUNT, const_sig_spec]

#if (const_offset == 0)
	alu_shf[--, --, B, r_fifo_addr, <<4]
#else
.local _tempa
	alu_shf[_tempa, --, B, const_offset, <<4]
	alu_shf[--, _tempa, +, r_fifo_addr, <<4]
.endlocal
#endif

#if (const_sig_spec == SYNC)
	sdram[r_fifo_rd, --, sd_addr, const_offset, CONST_QW_COUNT], indirect_ref, ctx_swap
#elif (const_sig_spec == ASYNC)	
	sdram[r_fifo_rd, --, sd_addr, const_offset, CONST_QW_COUNT], indirect_ref, sig_done
#else
	sdram[r_fifo_rd, --, sd_addr, const_offset, CONST_QW_COUNT], indirect_ref
#endif

#endm


#endif // SDRAM_UC