//------------------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//-----------------------------------------------------------------------------------
// buf.uc
// IXP memory allocation macros
//
//
// system: IXP1200
// subsystem:  microcode
// usage: library macros
// author: dfh, Aug 14, 2000	
//


#ifndef BUF_UC
#define BUF_UC

//	Buffer Management Macros
//
//	There are two ways to use these macros:
//
//	SRAM buffer descriptors describing DRAM data buffers:
//
//		Buffer descriptors are stored in an array in SRAM.
//		Data buffers live in a parallel array in DRAM.
//		An SRAM descriptor at index i refers to DRAM
//		data buffer	at the same index.
//
//		Free DRAM data buffers are allocated by allocating
//		the associated SRAM buffer descriptor.
//		Access to the SRAM buffer descriptor free pool
//		is achieved via buf_pop() and buf_push(), which
//		utilize the hardware top of stack registers.
//		The benefit of this implementation is not the LIFO
//		stack ordering, but the implicit synchronization
//		built into the top of stack hardware.
//
//	SRAM data buffers.
//
//		The data is held directly in the SRAM buffers, and the
//		relationship with corresponding DRAM buffers is ignored.
//		Here just the SRAM addresses are used -- D_BASE, D_SIZE,
//		and the macros below with _dram in them are not used.
//
//	SRAM and DRAM addresses may be stored as indexes
//	to save space, so macros to convert between
//	indexes and addresses are provided.

// API

// buf_freelist_create(NUM_BUFFERS, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_pop(out_xfer, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE, signal_token)
// buf_wait()
// buf_push(in_relative_ptr, in_ptr_base, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_dram_addr_from_sram_addr(out_dram_addr, in_sram_addr, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_index_from_sram_addr(out_index, in_address, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_index_from_dram_addr(out_index, in_address, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_sram_addr_from_index(out_address, in_index, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
// buf_dram_addr_from_index(out_address, in_index, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)


#include "stdmac.uc"


// buf_freelist_create
//
//		Description:
//			Create SRAM buffer (descriptor) freelist using push/pop regs.
//			The freelist is terminated with the base SRAM address, S_BASE,
//			so the number of buffers available to pop is NUM_BUFFERS - 1.
//
//		Inputs:
//			NUM_BUFFERS			number of buffers put on the freelist
//			FREELIST_HANDLE, composed of:
//				STACK_ID			which freelist to create
//				D_BASE				base 64 bit (quadword) address of dram buffers
//				D_SIZE				size of dram buffers, in 64 bit words (quadwords)
//				S_BASE				base 32 bit (longword) address of sram buffers
//				S_SIZE				size of sram buffers, in 32 bit words (longwords)
//	
//		Implementation:
//			To terminate the list with a self-referencing element at S_BASE,
//			we execute the initial push twice.
//			The inner loop runs without ctx_swap, limited only by SRAM speed,
//			The last push has a ctx_swap to guarantee all elements are in RAM
//			before completing the macro.
//
//		Size:	9 instructions typical (32-bit S_BASE, 16-bit offset, 5-bit S_SIZE)
//				10 instructions for large S_SIZE
//
//		Example usage:
//			#define FREELIST_HANDLE 0,100,2048,500,4
//			buf_freelist_create(300, FREELIST0_HANDLE);
//

#macro buf_freelist_create(NUM_BUFFERS, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)

	#if ((NUM_BUFFERS) < 2)
		#error 4 "buf_freelist_create() implementation requires NUM_BUFFERS >= 2"
	#endif
	.local  buf_sram_addr, buf_count

		immed32(buf_sram_addr, S_BASE)			; base sram address
		immed32(buf_count, (NUM_BUFFERS - 1))	// loop count

		// initialize Top-Of-Stack register
		sram[push, --, buf_sram_addr, 0, STACK_ID] ; *addr = junk; TOS = addr

		// push (NUM_BUFFERS - 1) elements
		freelist_loop#:
			sram[push, --, buf_sram_addr, 0, STACK_ID]	; *addr = TOS; TOS = addr
			add(buf_sram_addr, buf_sram_addr, S_SIZE)	; add sram buffer size
			alu[buf_count, buf_count, -, 1]			; decrement buf_count
			br>0[freelist_loop#]

		// push the last entry with ctx_swap
		sram[push, --, buf_sram_addr, 0, STACK_ID], ctx_swap	; last entry
	.endlocal
#endm

// buf_pop
//
// 		Description: Pop an sram address from the push/pop stack identified by STACK_ID.
//		Outputs:
//			out_xfer		which sram read xfer to place the popped pointer, 0-7
//							On error S_BASE is returned.
//		Inputs:
//			FREELIST_HANDLE, composed of:
//				STACK_ID			which freelist to use
//				D_BASE				base 64 bit (quadword) address of dram buffers
//				D_SIZE				size of dram buffers, in 64 bit words (quadwords)
//				S_BASE				base 32 bit (longword) address of sram buffers
//				S_SIZE				size of sram buffers, in 32 bit words (longwords)	
//			signal_token		ctx_swap:   waiting on signal, context swap
//								sig_done:   requesting signal, but continuing
//		Size: 1 - 6 instr
//		Example usage:
//			#define FREELIST_HANDLE MY_STACK_ID,100,2048,MY_SRAM_BASE,4
//			xbuf_alloc($pop_xfer, 1);
//			buf_pop($pop_xfer[0], FREELIST_HANDLE);
//			.if ($pop_xfer[0] == MY_SRAM_BASE)
//				no buffers available
//			.endif
//
#macro buf_pop(out_xfer, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE, signal_token)
#if ((!streq (signal_token, ctx_swap)) && (!streq (signal_token, sig_done)))
	#error 2  "buf_pop. ctx_swap or sig_done is required."
#endif
// note: the second out_xfer is a dummy unused argument, it prevents an assembler warning
	sram[pop, out_xfer, out_xfer, 0, STACK_ID], signal_token
#endm



// buf_wait
//		description: Wait on a sram signal. This needs to be invoked if buf_pop used 
//					 the sig_done signal_token.
//		size: 1  instr
//		example usage:
//			buf_wait();
//		se also: buf_alloc
//
#macro buf_wait()
	ctx_arb[sram]
#endm



// buf_dram_addr_from_sram_addr
// 		description: Given a freelist handle and SRAM address, calculate DRAM buffer address
//		outputs:
//			out_dram_addr			quadword address of buffer in SDRAM
//		inputs:
//			in_sram_addr			popped pointer (from buf_pop), or sram address
//			FREELIST_HANDLE, composed of:
//				STACK_ID			which freelist
//				D_BASE				base 64 bit (quadword) address of dram buffers
//				D_SIZE				size of dram buffers, in 64 bit words (quadwords)
//				S_BASE				base 32 bit (longword) address of sram buffers
//				S_SIZE				size of sram buffers, in 32 bit words (longwords)	
//		size: 4 instr
//		example usage:
//			#define FREELIST_HANDLE 0,100,2048,500,4
//			xbuf_alloc($pop_xfer, 1);						// allocate transfer reg buffer
//
//			buf_pop($pop_xfer[0], FREELIST_HANDLE, sig_done);
//			buf_wait();
//			.while ($pop_xfer[0] < 0)
//				buf_pop($pop_xfer[0], FREELIST_HANDLE);    	// retry
//			.endw
//			buf_dram_addr_calc(packet_buf_addr, $pop_xfer[0], FREELIST_HANDLE);
//			// use the buffer
//			buf_push($pop_xfer[0], 0, FREELIST_HANDLE);
//
#macro buf_dram_addr_from_sram_addr(out_dram_addr, in_sram_addr, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
.local buffer_base descriptor_base relative_addr

    immed32[buffer_base, D_BASE]
	immed32[descriptor_base, S_BASE]
	alu[relative_addr, in_sram_addr, -, descriptor_base]

	// note: D_SIZE is quadwords, S_SIZE is longwords
	// calculate ratio, to get correct shift amount
	// this will support up to 64kB sdram buffer with 4 word sram buffer
	#define_eval SD_MULTIPLIER (D_SIZE / S_SIZE)

	#define_eval SD_SHIFT 0
	#define_eval BUF_TWO2SD_SHIFT 1
	#while(SD_MULTIPLIER > BUF_TWO2SD_SHIFT)
		#define_eval BUF_TWO2SD_SHIFT (BUF_TWO2SD_SHIFT * 2)
		#define_eval SD_SHIFT (SD_SHIFT + 1)
	#endloop

	// special case to cover signed compare above
	#if (SD_MULTIPLIER == 0x80000000)
		#define_eval BUF_TWO2SD_SHIFT SD_MULTIPLIER
		#define_eval SD_SHIFT 31
	#endif

	// if power of 2
	#if (BUF_TWO2SD_SHIFT == SD_MULTIPLIER)
		alu[out_dram_addr, buffer_base, +, relative_addr, <</**/SD_SHIFT/**/]	; calculate sdram address
	#else
		#error 2 "Error: buf_dram_addr_calc. dram buffer size/ sram buffer size not a power of 2"
	#endif

	#undef SD_SHIFT				// clean up name space
	#undef BUF_TWO2SD_SHIFT		// clean up name space

.endlocal
#endm	// buf_dram_addr_from_sram_addr()


// buf_push
// 		description: Given push/pop stack pointer (in_relative_ptr, in_ptr_base) and freelist handle, 
//					 push pointer to stack.
//		inputs:
//			in_relative_ptr			GPR relative offset from base
//			in_ptr_base				GPR or constant, push/pop stack pointer
//			FREELIST_HANDLE, composed of:
//				STACK_ID			which freelist
//				D_BASE				base 64 bit (quadword) address of dram buffers
//				D_SIZE				size of dram buffers, in 64 bit words (quadwords)
//				S_BASE				base 32 bit (longword) address of sram buffers
//				S_SIZE				size of sram buffers, in 32 bit words (longwords)	
//		size: 1-2 instr
//		example usage:
//			#define FREELIST_HANDLE 0,100,2048,500,4
//			xbuf_alloc($pop_xfer, 1);
//
//			buf_pop($pop_xfer[0], FREELIST_HANDLE);
//			buf_dram_addr_calc(packet_buf_addr, $pop_xfer[0], FREELIST_HANDLE);
//			// use the buffer
//			buf_push($pop_xfer[0], 0, FREELIST_HANDLE);
//
//			xbuf_free($pop_xfer);
//
#macro buf_push(in_relative_ptr, in_ptr_base, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)

	#if (isnum (in_ptr_base))

		#if (in_ptr_base > 0x1f)
			.local _temp_base
				immed32[_temp_base, in_ptr_base]
				sram[push, --, _temp_base, in_relative_ptr, STACK_ID]
			.endlocal
		#else // (in_ptr_base <= 0x1f)
			sram[push, --, in_ptr_base, in_relative_ptr, STACK_ID]
		#endif

	#else // (!isnum (in_ptr_base))
		sram[push, --, in_ptr_base, in_relative_ptr, STACK_ID]
	#endif // // (isnum (in_ptr_base))

#endm


// buf_index_from_sram_addr
//
//	Description:
//		Calculate Array Index from Buffer Descriptor Address
//		Ie.	out_index = (in_address - S_BASE) / S_SIZE
//
//	Outputs:	out_index
//	Inputs:		in_address		SRAM Buffer Descriptor address
//				FREELIST_HANDLE, composed of:
//					STACK_ID		which freelist
//					D_BASE			base 64 bit (quadword) address of dram buffers
//					D_SIZE			size of dram buffers, in 64 bit words (quadwords)
//					S_BASE			base 32 bit (longword) address of sram buffers
//					S_SIZE			size of sram buffers, in 32 bit words (longwords)	

#macro buf_index_from_sram_addr(out_index, in_address, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
	index_from_addr(out_index, in_address, S_BASE, S_SIZE)
#endm

// buf_index_from_dram_addr
//
//	Description:
//		Calculate Array Index from DRAM Buffer Address
//		Ie.	out_index = (in_address - D_BASE) / D_SIZE
//
//	Outputs:	out_index
//	Inputs:		in_address		DRAM buffer address
//				FREELIST_HANDLE, composed of:
//					STACK_ID		which freelist
//					D_BASE			base 64 bit (quadword) address of dram buffers
//					D_SIZE			size of dram buffers, in 64 bit words (quadwords)
//					S_BASE			base 32 bit (longword) address of sram buffers
//					S_SIZE			size of sram buffers, in 32 bit words (longwords)	

#macro buf_index_from_dram_addr(out_index, in_address, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
	index_from_addr(out_index, in_address, D_BASE, D_SIZE)
#endm

//
// buf_sram_addr_from_index
//
//	Description:
//		Calculate Buffer Descriptor Address from Index.
//		Ie. out_address = (in_index * S_SIZE) + S_BASE
//
//	Outputs:	out_index
//	Inputs:		in_address
//				FREELIST_HANDLE, composed of:
//					STACK_ID		which freelist
//					D_BASE			base 64 bit (quadword) address of dram buffers
//					D_SIZE			size of dram buffers, in 64 bit words (quadwords)
//					S_BASE			base 32 bit (longword) address of sram buffers
//					S_SIZE			size of sram buffers, in 32 bit words (longwords)

#macro buf_sram_addr_from_index(out_address, in_index, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
	addr_from_index(out_address, in_index, S_BASE, S_SIZE)
#endm


//
// buf_dram_addr_from_index
//
//	Description:
//		Calculate DRAM Buffer Address from Index.
//		Ie. out_address = (in_index * S_SIZE) + S_BASE
//
//	Outputs:	out_index
//	Inputs:		in_address
//				FREELIST_HANDLE, composed of:
//					STACK_ID		which freelist
//					D_BASE			base 64 bit (quadword) address of dram buffers
//					D_SIZE			size of dram buffers, in 64 bit words (quadwords)
//					S_BASE			base 32 bit (longword) address of sram buffers
//					S_SIZE			size of sram buffers, in 32 bit words (longwords)

#macro buf_dram_addr_from_index(out_address, in_index, STACK_ID, D_BASE, D_SIZE, S_BASE, S_SIZE)
	addr_from_index(out_address, in_index, D_BASE, D_SIZE)
#endm

#endif // BUF_UC