//------------------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c)  1999-2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//-----------------------------------------------------------------------------------
// cycle.uc
// IXPblock timer macros
//
//
// system: IXP1200
// subsystem: IO microcode
// usage: library macros
// author: dfh Jan 12, 1999	
//
// revisions:	dfh Apr 6, 2000		separated from stdmac.uc
//				sep 11, 2000		renamed from tim.uc to cycle.uc

#ifndef CYCLE_UC
#define CYCLE_UC


// API:

//		cycle_delay(IN_CYCLES)
//		cycle_sleep(IN_CYCLES)
//		cycle_fetch(out_prev_timer, io_xfer_new_timer, optional_token)
//		cycle_elapsed(out_difference, in_prev_timer, in_new_timer)

#include "stdmac.uc"
#include "xbuf.uc"

// cycle_delay
//		description: delay approx 4n cycles, don't swap out
//		inputs:
//			IN_CYCLES				constant. cycles to delay
//		size: 6 instr
//		see also: 
//		example usage:
//			cycle_delay(20);						; delay 80 cycles		
////
#macro cycle_delay(IN_CYCLES)
.local _delay_val
	immed[_delay_val, IN_CYCLES]
	alu_shf[_delay_val, --, B, _delay_val, >>2]		; to account for 4 cycles per iteration
	.while (_delay_val > 0)
		alu[_delay_val, _delay_val, -, 1]
	.endw
.endlocal
#endm
// LEGACY pre v1.0
#define delay cycle_delay


// cycle_sleep
// 		description: sleep for n cycles, let other contests run. will use 2 transfer reg.
//		inputs:
//			IN_CYCLES				constant or GPR. cycles to sleep
//		size: 12 instr
//		see also: 
//		example usage:
//			cycle_sleep(100);						; sleep 100 cycles		
//
#macro cycle_sleep(IN_CYCLES)
xbuf_alloc($cycle_xfer, 2);

csr[read, $cycle_xfer[0], CYCLE_CNT], ctx_swap
#if (((isnum (IN_CYCLES)) && (IN_CYCLES > 20)) || (!isnum(IN_CYCLES)))
	.local _new_timer _saved_timer _diff
	alu[_saved_timer, --, B, $cycle_xfer[0], >>2]	; save timer ignore low 2 bits
	iterate#:
	csr[read, $cycle_xfer[0], CYCLE_CNT], ctx_swap
	alu[_new_timer, --, B, $cycle_xfer[0], >>2]		; new timer ignore low 2 bits	
	.if (_new_timer < _saved_timer)
		alu[_new_timer, _new_timer, +, 1, <<30]
	.endif
	alu[_diff, _new_timer, -, _saved_timer]
	.if (_diff < IN_CYCLES)
		br[iterate#], defer[1]
		nop
	.endif
	.endlocal
#endif
xbuf_free($cycle_xfer);
#endm


// cycle_fetch
// 		description: 		save previous cycle time/4. fetch new cycle time/4
//		inputs:
//			out_prev_timer			previously fetched cycle time
//			io_xfer_new_timer		read xfer reg, newly fetched cycle time
//			optional_token			ctx_swap, sig_done
//		size: 3 instr
//		example usage:
//			cycle_fetch(prev_timer, $timer, sig_done);		
//
#macro cycle_fetch(out_prev_timer, io_xfer_new_timer, optional_token)
	alu[out_prev_timer, --, B, io_xfer_new_timer]
	csr[read, io_xfer_new_timer, CYCLE_CNT], optional_token
#endm


// cycle_elapsed
// 		description: 			compare time interval with (new_timer - current_timer), accounting for wrap
//		outputs:
//			out_difference		positive elapsed cycle count
//		inputs:
//			in_prev_timer		previously fetched cycle time
//			in_new_timer		newly fetched cycle time
//		size: 5 instr 
//		example usage:
//			cycle_fetch(prev_timer, $timer, sig_done);
//			// some other code		
//			sig_wait(fbi,0);
//			cycle_elapsed(diff, prev_timer, $timer);		
//
#macro cycle_elapsed(out_difference, in_prev_timer, in_new_timer)
.local _temp_new_timer
	alu[_temp_new_timer, --, B, in_new_timer, >>2]		; ignore low 2 bits
	alu[--, _temp_new_timer, -, in_prev_timer, >>2]
	br>=0[pos_diff#]
		alu[_temp_new_timer, _temp_new_timer, +, 1, <<30]
	pos_diff#:
	alu[out_difference, _temp_new_timer, -, in_prev_timer, >>2]
	shf_left[out_difference, out_difference, 2]
.endlocal
#endm


#endif // CYCLE_UC