//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// ether.uc
// macros for ethernet protocol header 
//
// system: IXP1200
// subsystem: IP microcode
// usage: library macros
// author: dfh	Apr 20, 2000
//
// revisions:


#ifndef ETHER_UC
#define ETHER_UC 

#define ETHER_MTU 1500					// maximum ethernet packet size
#define ETHER_PROTOCOL_LENGTH 				12, 2		// protocol length field byte offset 12, length 2 bytes
#define ETHER_PROTOCOL_SSAP					14, 1		// SSAP field byte offset 14, length 1 byte
#define ETHER_PROTOCOL_DSAP					15, 1		// DSAP field byte offset 15, length 1 byte
#define ETHER_PROTOCOL_CTRL					16, 1		// CTRL field byte offset 16, length 1 byte
#define ETHER_PROTOCOL_LLC_SNAP_OUI			17, 3		// LLC SNAP OUI field byte offset 16, length 1 byte
#define ETHER_PROTOCOL_LLC_SNAP_ETHERTYPE	20, 2		// LLC SNAP Ethertype field byte offset 16, length 1 byte


// Layer 2 packet types
#define ETHERNET	0xa
#define LLC_802_3	0xb
#define LLC_SNAP	0xc

// ether_parse_802_3
//		description: Parse 802.3 fields to determine if this is a LLC or LLC/SNAP packet. Return ethertype if the
//					 packet is LLC/SNAP.
//					 
//
//		input:
//			Transfer Registers		Contains ethernet header
//		output:
//			Ethertype				Is set to 0 if packet is LLC. If packet is LLC/SNAP, valid ethertype present
//		size:	XXX instr
//		example usage:
//			ether_parse_802_3[packet_ethertype, $pkt_buf]
///
#macro ether_parse_802_3 [out_packet_ethertype, in_xfer]
.local _eth_llc1 _eth_llc2

//	Extract dsap, dsap and ctrl from packet_header
	xbuf_extract(_eth_llc1, in_xfer, BYTEOFFSET0, ETHER_PROTOCOL_SSAP);	
	xbuf_extract(_eth_llc2, in_xfer, BYTEOFFSET0, ETHER_PROTOCOL_DSAP);	
	alu[_eth_llc1, _eth_llc1, AND, _eth_llc2]
	xbuf_extract(_eth_llc2, in_xfer, BYTEOFFSET0, ETHER_PROTOCOL_CTRL);	
	.if (_eth_llc1 == 0xAA)
		.if (_eth_llc2 == 0x03)	; SNAP frame, a valid ethertype can be extracted
//			xbuf_extract(out_packet_oui, io_xfer, BYTEOFFSET0, ETHER_PROTOCOL_LLC_SNAP_OUI);	
			xbuf_extract(out_packet_ethertype, in_xfer, BYTEOFFSET0, ETHER_PROTOCOL_LLC_SNAP_ETHERTYPE);	
		.endif 
	.endif			; Check if this is a SNAP frame
.endlocal
#endm

// Format of port table entry
//			  +-------+-------+------+------+-------+
// 			  |unused | ether |unused|filter| port  |
//long word0  |       |  type |      | type | state |
//			  | 31:24 | 23:8  | 7:6  |  5:4 |  3:0  |
//			  +-------+-------+------+------+-------+
//long word1  |			  MAC	Address				|
//			  |										|
//			  +-----------------+-------------------+
//			  |					| 	MAC Address  	|
//long word2  |					|	   	16:0		|
//			  +-----------------+-------------------+
//			  |										|
//long word3  |										|
//			  +-------------------------------------+	
//
// ether_port_info
//		description: Extract port configuration and device MAC address. The device MAC address is saved here for lack
//					  lack of better place (all port entries have the same device MAC address)
//					 
//
//		input:
//			Port #				Port Number
//		output:
//			MAC address			Device MAC address (same for all ports)
//			Port configuration	Port configuration containing, port state, filter type, and ethertype
//		size:	XXX instr
//		example usage:
//			ether_port_info(mac_addr32, mac_addr16, port_config, port_num)
///
#macro ether_port_info[out_port_mac_addr32, out_port_mac_addr16, out_port_entry, in_port_no]
.local _ethport_tempa _ethport_tempb
		// save port MAC address 
	xbuf_alloc($port_inf, 3);
    alu_shf[_ethport_tempa, --, B, in_port_no, <<BR_PORT_ENTRY_MULTIPLER]	// 4 32bit words 
	immed32[_ethport_tempb, SRAM_PORT_STATE_BASE]
	sram_read($port_inf[0], _ethport_tempb, _ethport_tempa, 3, ctx_swap)

	move[out_port_entry, $port_inf[0]]
	xbuf_extract(out_port_mac_addr32, $port_inf, BYTEOFFSET0, 4, 4);	// 5 MAC addres starts from 2nd 32bit word
	xbuf_extract(out_port_mac_addr16, $port_inf, BYTEOFFSET0, 10, 2);
	xbuf_free($port_inf);

.endlocal
#endm


// ether_init_hash
//		description: Initialize FBI Hash multiplier
//					 
//		input:
//			Hash Multipliers	Hi and Lo 32 bit values.
//		output:
//		size:	XXX instr
//		example usage:
//			ether_init_hash(BR_HASH_LOAD_HI, BR_HASH_LOAD_LO)
///
#macro ether_init_hash(in_HASH_LOAD_HI, in_HASH_LOAD_LO)
.local $hash_mult0, $hash_mult1
move[$hash_mult0, in_HASH_LOAD_HI]
move[$hash_mult1, in_HASH_LOAD_LO]
csr[write, $hash_mult0, HASH_MULTIPLIER_48_HI]
csr[write, $hash_mult1, HASH_MULTIPLIER_48_LO]
.endlocal
#endm

#endif // ETHER_UC
