//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// ip.uc
// macros to verify, modify and extract fields from ip header 
//
//
// system: IXP1200
// subsystem: IP microcode
// usage: library macros
// author: dfh	Apr 3, 2000
//
// revisions:


#ifndef IP_UC
#define IP_UC


// API:
//		ip_proc(out_exception, str_dwxbuf, IPHDR_WR_START_BYTE, str_rxbuf, IPHDR_RD_START_BYTE)
//		ip_verify(out_exception, str_rxbuf, IPHDR_RD_START_BYTE)
//		ip_cksum_verify(str_rxbuf, IPHDR_RD_START_BYTE)
//		ip_modify(str_dwxbuf, IPHDR_WR_START_BYTE, str_rxbuf, IPHDR_RD_START_BYTE)
//		ip_cksum_modify(out_dest, in_source, BYTE_OFFSET)
//		ip_route_lookup(out_rt_ptr, in_ip_da, CONST_LOOKUP_BASE, TRIE_TYPE)


// IP exceptions, range is 0x8-0xf
#define    IP_BAD_TOTAL_LENGTH			0x08
#define    IP_BAD_TTL					0x09
#define    IP_BAD_CHECKSUM				0x0a
#define    IP_NO_ROUTE					0x0b

#define	   IPV4_NO_OPTIONS				0x45			// version 4, header length 5 longwords 

#define IP_VERSION_LENGTH 0, 1          // version/length fields.  offset 0, 1 byte
#define IP_TYPE_OF_SERVICE 1, 1	        // type of service field.  offset 1, 1 byte
#define IP_TOTAL_LENGTH 2, 2            // type of service field.  offset 2, 2 bytes
#define IP_IDENTIFICATION 4, 2          // identification field.   offset 4, 2 bytes
#define IP_FLAGS_FRAGMENT 6, 2          // flags/fragment fields.  offset 6, 2 bytes
#define IP_TIME_TO_LIVE 8, 1            // time to live field.     offset 8, 1 byte
#define IP_PROTOCOL 9, 1                // protocol field.         offset 9, 1 byte
#define IP_CHECKSUM 10, 2               // checksum field.         offset 10, 2 bytes
#define IP_SOURCE_ADDRESS 12, 4         // source ip address.      offset 12, 4 bytes
#define IP_DESTINATION_ADDRESS 16, 4    // destination ip address. offset 16, 4 bytes   
#define IP_SOURCE_PORT 20, 2            // source port.            offset 20, 2 bytes   
#define IP_DESTINATION_PORT 22, 2       // destination port.       offset 22, 2 bytes   


// ip_proc
//		description: Process ip header that starts at read xfer byte pointer. 
//					 1. Verify ip total length, time-to-live and checksum.
//					 2. Modify time-to-live and checksum. Write modified header to sdram write xfers.
//					 Note: ip frag and ip options are not supported
//		outputs:
//			out_exception			0 if no exception, otherwise ip exception code	
//			str_dwxbuf				name of dram write xfer transfer buffer to contain packet data
//		inputs:
//			IPHDR_WR_START_BYTE		constant byte address of ip header start relative to str_dwxbuf
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 28 big-endian, 34 little-endian instr
//		see also: stdmac.uc, endian.uc
//		example usage:
//			// ip header starts at byte 14 of ethernet packet
//			ip_proc(exception, modified_pkt, BYTEOFFSET14, pkt, BYTEOFFSET14);
//			
//
#macro ip_proc(out_exception, str_dwxbuf, IPHDR_WR_START_BYTE, str_rxbuf, IPHDR_RD_START_BYTE)
	ip_verify[out_exception, str_rxbuf, IPHDR_RD_START_BYTE]			; 18-23 instr
	br>0[end#]
	ctx_arb[voluntary]
	ip_modify[str_dwxbuf, IPHDR_WR_START_BYTE, str_rxbuf, IPHDR_RD_START_BYTE]	; 9-11 instr
	alu[out_exception, --, B, out_exception]						; set exit condition code
end#:
#endm


// ip_verify
//		description: Verify ip total length, time-to live and checksum.
//		outputs:
//			out_exception			0 if no exception, otherwise ip exception code	
//		inputs:
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 22-26 instr
//		latency: 19-23 instr
//		see also: stdmac.uc, endian.uc
//		example usage:
//			ip_verify(exception, pkt, BYTEOFFSET14);	ip header starts at byte 14 of ethernet packet
//
#macro ip_verify(out_exception, str_rxbuf, IPHDR_RD_START_BYTE)
.local total_len
total_len_verify#:
// check that total ip packet length in $xfer2 bytes 0-1 is at least 20 bytes
	xbuf_extract(total_len, str_rxbuf, IPHDR_RD_START_BYTE, IP_TOTAL_LENGTH); // get total length field
	alu[--, total_len, -, 0x14]
	br>0[ttl_verify#], guess_branch
	alu[out_exception, --, B, IP_BAD_TOTAL_LENGTH]
	br[end#]
.endlocal
.local ttl
ttl_verify#:
// Check that TTL in byte 8 of ip header is greater than 1.  
	xbuf_extract(ttl, str_rxbuf, IPHDR_RD_START_BYTE, IP_TIME_TO_LIVE); // get ttl field
	br>0[cksum_verify#], defer[1]
	alu[out_exception, --, B,  0]
	alu[out_exception, --, B, IP_BAD_TTL]
	br[end#]
.endlocal
cksum_verify#:
	ip_cksum_verify[str_rxbuf, IPHDR_RD_START_BYTE]
	br=0[end#]
	alu[out_exception, --, B, IP_BAD_CHECKSUM]
end#:

#endm


// ip_cksum_verify
//		description: Verify ip checksum 
//		outputs:
//			out_exception			GPR, 0 if good checksum, non-0 if bad checksum.	
//		inputs:
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 10-14 instr (depending on alignment and endian mode)
//		latency: 10-14 cycles
//		see also: stdmac.uc, endian.uc
//		example usage:
//			ip_cksum_verify(pkt, BYTEOFFSET14);		ip header starts at byte 14 of ethernet packet
//
#macro ip_cksum_verify(str_rxbuf, IPHDR_RD_START_BYTE)
#ifdef LITTLE_ENDIAN
_ip_checksum_verify_le[str_rxbuf, IPHDR_RD_START_BYTE]
#else
_ip_checksum_verify_be[str_rxbuf,  IPHDR_RD_START_BYTE]
#endif
#endm


// _ip_checksum_verify_le
//		description: Verify ip header  checksum, little-endian.
//					 Internal macro used by ip_cksum_verify.
//					 Uses halfword adds to reduce total instructions.
//		outputs:
//			<cond codes>		    Good checksum if condition code == 0, otherwise have bad checksum.	
//		inputs:
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 14 instr
//		latency: 14 cycles
//		example usage:
//			_ip_cksum_verify_le(pkt, BYTEOFFSET14);		ip header starts at byte 14 of ethernet packet	
//
#macro _ip_checksum_verify_le[str_rxbuf, IPHDR_RD_START_BYTE]
.local temp accum

xbuf_xfer_set[str_rxbuf, IPHDR_RD_START_BYTE]	; set xbuf tokens _XFER0 to _XFER7

#define_eval READ_ALIGN (IPHDR_RD_START_BYTE & 0x3)
#if (READ_ALIGN == 0)
	alu[accum, --, B, _XFER0, >>16]
	alu[accum, accum, +16, _XFER0]
#elif (READ_ALIGN == 2)
	alu[accum, 0, +, _XFER0, >>16]
	alu[accum, accum, +16, _XFER5]
#else
	_ip_checksum_verify_oddalign[str_rxbuf, IPHDR_RD_START_BYTE]
#endif
#if ((READ_ALIGN == 0) || (READ_ALIGN == 2))
	alu[accum, accum, +, _XFER1, >>16]
	alu[accum, accum, +16, _XFER1]
	alu[accum, accum, +, _XFER2, >>16]
	alu[accum, accum, +16, _XFER2]
	alu[accum, accum, +, _XFER3, >>16]
	alu[accum, accum, +16, _XFER3]
	alu[accum, accum, +, _XFER4, >>16]
	alu[accum, accum, +16, _XFER4]

    alu[temp, 0, +, accum, >>rot16]				;rotate by 16
	alu[accum, accum, +, temp]
	alu[accum, accum, +carry, 1]				;add 1 to 0xffff ffff to get zero result
#endif
.endlocal
#endm


// _ip_checksum_verify_be
//		description: Verify ip header  checksum, big-endian.
//					 Internal macro used by ip_cksum_verify.
//					 Uses 32 bit add with carry to reduce the total number of instructions.
//		outputs:
//			<cond codes>		    Good checksum if condition code == 0, otherwise have bad checksum.	
//		inputs:
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 10-11 instr
//		latency: 10-11 cycles
//		example usage:
//			_ip_cksum_verify_be(pkt, BYTEOFFSET14);		ip header starts at byte 14 of ethernet packet	
//
#macro _ip_checksum_verify_be[str_rxbuf, IPHDR_RD_START_BYTE]
.local temp accum

xbuf_xfer_set[str_rxbuf, IPHDR_RD_START_BYTE]	; set xbuf tokens _XFER0 to _XFER7

#define_eval READ_ALIGN (IPHDR_RD_START_BYTE & 0x3)
#if (READ_ALIGN == 0)
	alu[accum, --, B, _XFER0]
	alu[accum, accum, +, _XFER1]
#elif (READ_ALIGN == 2)
	alu[accum, 0, +16, _XFER0]
	alu[accum, accum, +, _XFER1]
	alu[accum, accum, +carry, _XFER5, >>16]
#else
	_ip_checksum_verify_oddalign[str_rxbuf, IPHDR_RD_START_BYTE]
#endif
#if ((READ_ALIGN == 0) || (READ_ALIGN == 2))
	alu[accum, accum, +carry, _XFER2]
	alu[accum, accum, +carry, _XFER3]
	alu[accum, accum, +carry, _XFER4]

	alu[accum, accum, +carry, 0]				; add in previous carry
    ld_field_w_clr[temp, 1100, accum]			; get high 16 of the total
    alu_shf[accum, temp, +, accum, <<16]		; add low 16 bits to upper 16
	alu[temp, 1, +carry, temp, <<16]			; add last carry +1, temp B op=0
	alu[accum, accum, +, temp, <<16]			; add 1<<16 to 0xffff to get zero result
#endif
.endlocal
#endm


// _ip_checksum_verify_oddalign
//		description: support for _ip_checksum_verify_be/le, if byte alignment is 1 or 3
//
#macro _ip_checksum_verify_oddalign[str_rxbuf, IPHDR_RD_START_BYTE]
xbuf_xfer_set[str_rxbuf, IPHDR_RD_START_BYTE]	; set xbuf tokens _XFER0 to _XFER7

#define_eval START_BYTE (IPHDR_RD_START_BYTE & 0x3)
#define_eval END_BYTE ((START_BYTE + 3) & 0x3)

.local ck_temp ck_accum
	field_dbl_extract[ck_accum, _XFER0, START_BYTE, _XFER1, END_BYTE]
	field_dbl_extract[ck_temp, _XFER1, START_BYTE, _XFER2, END_BYTE]
	alu[ck_accum, ck_accum, +, ck_temp]
	field_dbl_extract[ck_temp, _XFER2, START_BYTE, _XFER3, END_BYTE]
	alu[ck_accum, ck_accum, +carry, ck_temp]
	field_dbl_extract[ck_temp, _XFER3, START_BYTE, _XFER4, END_BYTE]
	alu[ck_accum, ck_accum, +carry, ck_temp]
	field_dbl_extract[ck_temp, _XFER4, START_BYTE, _XFER5, END_BYTE]
	alu[ck_accum, ck_accum, +carry, ck_temp]

	alu[ck_accum, ck_accum, +carry, 0]				; add in previous carry
    ld_field_w_clr[ck_temp, 1100, ck_accum]			; get high 16 of the total
    alu_shf[ck_accum, ck_temp, +, ck_accum, <<16]	; add low 16 bits to upper 16
	alu[ck_temp, 1, +carry, ck_temp, <<16]			; add last carry +1, ck_temp B op=0
	alu[ck_accum, ck_accum, +, ck_temp, <<16]		; add 1<<16 to 0xffff to get zero result

.endlocal
#endm


// ip_modify
//		description: Modify first 5 longwords of ip header. Modify checksum and ttl
//					 Write modified ip header to dram write transfer registers.
//					 Read and write start bytes must be the same.
//		outputs:
//		inputs:
//			str_dwxbuf				name of dram write xfer transfer buffer to contain packet data
//			IPHDR_WR_START_BYTE		constant byte address of ip header start relative to str_dwxbuf
//			str_rxbuf				name of sram read xfer transfer buffer containing packet data
//			IPHDR_RD_START_BYTE		constant byte address of ip header start relative to str_rxbuf
//
//		size: 9-11 instr
//		latency: 9-11 ccyles
//		see also: stdmac.uc, endian.uc
//		example usage:
//			// ip header starts at byte 14 of ethernet packet
//			ip_modify(modified_pkt, BYTEOFFSET14, pkt, BYTEOFFSET14);
//
#macro 	ip_modify(str_dwxbuf, IPHDR_WR_START_BYTE, str_rxbuf, IPHDR_RD_START_BYTE)
.local temp

xbuf_xfer_set[str_rxbuf, IPHDR_RD_START_BYTE]		; set xbuf tokens _XFER0 to _XFER7
xbuf_xfer_set[str_dwxbuf, IPHDR_WR_START_BYTE]		; set xbuf tokens _D_XFER0 to _D_XFER7

#define_eval READ_ALIGN (IPHDR_RD_START_BYTE & 0x3)
#define_eval WRITE_ALIGN (IPHDR_WR_START_BYTE & 0x3)

#if (READ_ALIGN != WRITE_ALIGN)
	#error 2 "ip_modify. read align != write align."
#else
	#if (READ_ALIGN == 0)								; iphdr starts at byte 0 of longword
		move[_D_XFER0, _XFER0]
		move[_D_XFER1, _XFER1]
		field_decr[temp, _XFER2, 0, 0]					; iphdr byte offset 8, ttl(see endian.uc)
		ip_cksum_modify[_D_XFER2, temp, 2]				; iphdr byte offset 10-11, checksum
		move[_D_XFER3, _XFER3]
		move[_D_XFER4, _XFER4]
	#elif(READ_ALIGN == 1)
		move[_D_XFER0, _XFER0]
		move[_D_XFER1, _XFER1]
		field_decr[temp, _XFER2, 1, 1]				; (see endian.uc) byte 8+1
		_ip_cksum_B3align_modify[_D_XFER2, _D_XFER3, temp, _XFER3]		; byte 10+1, straddles 2 words
		move[_D_XFER4, _XFER4]
	#elif(READ_ALIGN == 2)								; iphdr starts at byte 2 of longword 						3
		move[_D_XFER0, _XFER0]
		move[_D_XFER1, _XFER1]
		field_decr[_D_XFER2, _XFER2, 2, 2]				; (see endian.uc) byte 8+2
		ip_cksum_modify[_D_XFER3, _XFER3, 0]			; byte 10+2
		move[_D_XFER4, _XFER4]
	#else
		move[_D_XFER0, _XFER0]
		move[_D_XFER1, _XFER1]
		field_decr[_D_XFER2, _XFER2, 3, 3]				; (see endian.uc) byte 8+3
		ip_cksum_modify[_D_XFER3, _XFER3, 1]			; byte 10+2
		move[_D_XFER4, _XFER4]
	#endif
#endif
.endlocal
#endm


// ip_cksum_modify
// 		description: Increment checksum to compensate for subtracting 1 from ttl.
// 				 checksum = bytes 10-11 of ip header
//		outputs:
//			out_dest					GPR destination	
//		inputs:
//			in_source		GPR with c
//			BYTE_OFFSET		constant byte offset
//
//		size: 6 instr
//		latency: 6 cycles
//		see also: stdmac.uc, endian.uc. RFC 1141
//		example usage:
//			ip_cksum_modify(new_ip_wd2, prev_ip_wd2, 2);	cksum at bytes 2-3
//
#macro ip_cksum_modify(out_dest, in_source, BYTE_OFFSET)
.local temp0 temp1
#define_eval READ_BYTE0 BYTE_OFFSET
#define_eval READ_BYTE1 (BYTE_OFFSET + 1)
	field_incr[temp0, in_source, READ_BYTE0, READ_BYTE0]
	field_extract[temp1, temp0, READ_BYTE0, READ_BYTE0]
	br=0[add_carry#]
	br[end#], defer[1]
	move[out_dest, temp0]							; can be a write transfer register
add_carry#:
	field_incr[out_dest, temp0, READ_BYTE0, READ_BYTE1]
end#:
.endlocal
#endm

// _ip_cksum_B3align_modify
//		description: support ip_cksum_modify if byte alignment is 3 on source and dest
//
#macro _ip_cksum_B3align_modify(out_dest0, out_dest1, in_source0, in_source1)
.local new_cksum merged_cksum
#define_eval READ_BYTE0 3
#define_eval READ_BYTE1 0
	field_dbl_extract[merged_cksum, in_source0, READ_BYTE0, in_source1, READ_BYTE1];

	alu[new_cksum, merged_cksum, +, 1]
	alu[--, --, B, new_cksum, >>8]
	br=0[add_carry#]
	br[end#]
add_carry#:
	alu[new_cksum, new_cksum, +, 1]			// now new_cksum in bytes 2-3
end#:

// now insert the incremented cksum

#ifdef LITTLE_ENDIAN
	ld_field_w_clr[merged_cksum, 0111, in_source0]
	alu[out_dest0, merged_cksum, +, new_cksum, <<16]		; le byte3
	ld_field_w_clr[merged_cksum, 1110, in_source1]
	alu[out_dest1, merged_cksum, +8, new_cksum]			; le byte0

#else
	ld_field_w_clr[merged_cksum, 1110, in_source0]
	alu[out_dest0, merged_cksum, +8, new_cksum, >>8]		; be byte3
	ld_field_w_clr[merged_cksum, 0111, in_source1]
	alu[out_dest1, merged_cksum, +, new_cksum, <<24]		; be byte0
#endif

.endlocal
#endm


// ip_route_lookup
//		description: Perform ip longest prefix match lookup. Uses tables written by core Route Table Manager.
//					 This uses a dual lookup of hi64k entry table and trie block lookups.
//					 It uses a maximum of 5 sram reference signals.
//		outputs:
//			out_rt_ptr				GPR resulting route pointer. index to route entry. 0 if no route entry.
//		inputs:
//			in_ip_da				GPR with IP destination address
//			CONST_LOOKUP_BASE		constant base address of route lookup tables
//			TRIE_TYPE				trie depth 5: 16bit index table followed by 4 4bit index tables 
//									trie depth 7: 8bit index table followed by 6 4bit index tables 
//
//		see also: IXP Software Reference Manual(SRM) for algorithm description.
//		example usage:
//			ip_route_lookup(route_entry_ptr, ip_da, 0x20000, 5);
//
//
#macro ip_route_lookup(out_rt_ptr, in_ip_da, CONST_LOOKUP_BASE, TRIE_TYPE)
#if (TRIE_TYPE == 5)
	ip_trie5_lookup(out_rt_ptr, in_ip_da, CONST_LOOKUP_BASE);
#else
	#error "ip_route_lookup. unsupported trie type."
#endif
#endm


// ip_trie5_lookup
//		description: support macro ip_route_lookup
//					 This uses a dual lookup of hi64k entry table and trie block lookups.
//					 It uses a maximum of 5 sram reference signals.
//		outputs:
//			out_rt_ptr				GPR resulting route pointer. index to route entry. 0 if no route entry.
//		inputs:
//			in_ip_da				GPR with IP destination address
//			CONST_LOOKUP_BASE		constant base address of route lookup tables
//		size: 77 instr
//		example usage:
//			_ip_trie5_lookup[route_entry_ptr, ip_da, 0x20000]
//
#macro ip_trie5_lookup(out_rt_ptr, in_ip_da, CONST_LOOKUP_BASE)
.local temp_base2 temp_base3, offset, lookup_short, lookup_long, prev_rt_short, prev_rt_long
.local $rd_xfer0 $rd_xfer1


// macro used by route_lookup
#macro next_trie[ipaddr, const_lsb, prevout_rt_ptr, lookup, xfer_reg, trie_base]
#define_eval SHIFT_AMT (const_lsb)
	alu_shf[lookup, lookup, +4, ipaddr, >>SHIFT_AMT]	; form offset from ipaddr and trie block addr
	sram[read, xfer_reg, trie_base, lookup, 1], ctx_swap, optimize_mem, defer[1]
	ld_field_w_clr[prevout_rt_ptr, 1100, xfer_reg]	
	ld_field_w_clr[lookup, 0011, xfer_reg], load_cc		; get trie pointer, set cond code
#endm

// tables_base can be either a static constant or from an .import_var
	immed32(tables_base, CONST_LOOKUP_BASE)					; if static constant
	alu_shf[temp_base2, tables_base, +, 1, <<16]		; add 0x10000, 256 entry table
	alu_shf[temp_base3, temp_base2, +, 1, <<8]			; add 0x100, multiple 16 entry tables

	alu_shf[offset, 0, B, in_ip_da, >>16]					; form offset from 31:16
first_lookup#:
 	sram[read, $rd_xfer0, tables_base, offset, 1], optimize_mem	; direct lookup off addr 31:16
	alu_shf[offset, 0, B, in_ip_da, >>24]					; form offset from 31:24
	sram[read, $rd_xfer1, temp_base2, offset, 1],ctx_swap, optimize_mem, defer[1]	; direct lookup off addr 31:24
	immed[prev_rt_long, 0]
	ld_field_w_clr[lookup_short, 0011, $rd_xfer1], load_cc	; get trie pointer, set cond code
	br=0[long_path_only#]								; lookup_short = 0, no short path
	ld_field_w_clr[lookup_long, 0011, $rd_xfer0], load_cc	; get trie pointer, set cond code
	br=0[short_path_only#]								; lookup_long = 0, short path only
	br[both_paths#]										; follow both paths

short_path_only#:

second_lookup_short#:
	next_trie[in_ip_da, 20, prev_rt_short, lookup_short, $rd_xfer1, temp_base3]
	br=0[set_route_ptr#]
third_lookup_short#:
	next_trie[in_ip_da, 16, prev_rt_short, lookup_short, $rd_xfer1, temp_base3]
	br=0[set_route_ptr#]
	br[set_route_ptr#]

long_path_only#:
	ld_field_w_clr[lookup_long, 0011, $rd_xfer0], load_cc		; get trie pointer, set cond code
	br=0[set_route_ptr#]
second_lookup_long#:
	next_trie[in_ip_da, 12, prev_rt_long, lookup_long, $rd_xfer0, temp_base3]
	br=0[set_route_ptr#]
third_lookup_long#:
	next_trie[in_ip_da, 8, prev_rt_long, lookup_long, $rd_xfer0, temp_base3]
	br=0[set_route_ptr#]
fourth_lookup_long#:
	next_trie[in_ip_da, 4, prev_rt_long, lookup_long, $rd_xfer0, temp_base3]
	br=0[set_route_ptr#]
fifth_lookup_long#:
	next_trie[in_ip_da, 0, prev_rt_long, lookup_long, $rd_xfer0, temp_base3]
	br[set_route_ptr#]

both_paths#:
	alu_shf[lookup_short, lookup_short, +4, in_ip_da, >>20]	; form offset from ipaddr and trie block addr
	sram[read, $rd_xfer1, temp_base3, lookup_short, 1], optimize_mem
	ld_field_w_clr[prev_rt_short, 1100, $rd_xfer1]	
	alu_shf[lookup_long, lookup_long, +4, in_ip_da, >>12]	; form offset from ipaddr and trie block addr
	sram[read, $rd_xfer0, temp_base3, lookup_long, 1], ctx_swap, optimize_mem, defer[1]
	ld_field_w_clr[prev_rt_long, 1100, $rd_xfer0]	
	ld_field_w_clr[lookup_long, 0011, $rd_xfer0], load_cc		; get trie pointer, set cond code
	br=0[second_both_no_long#]
second_both_long#:
	ld_field_w_clr[lookup_short, 0011, $rd_xfer1], load_cc		; get trie pointer, set cond code
	br=0[third_lookup_long#]
	br[third_lookup_both#]
second_both_no_long#:
	ld_field_w_clr[lookup_short, 0011, $rd_xfer1], load_cc		; get trie pointer, set cond code
	br=0[set_route_ptr#]
	br[third_lookup_short#]
	
third_lookup_both#:
	alu_shf[lookup_short, lookup_short, +4, in_ip_da, >>16]	; form offset from ipaddr and trie block addr
	sram[read, $rd_xfer1, temp_base3, lookup_short, 1], optimize_mem
	ld_field_w_clr[prev_rt_short, 1100, $rd_xfer1]	
	alu_shf[lookup_long, lookup_long, +4, in_ip_da, >>8]	; form offset from ipaddr and trie block addr
	sram[read, $rd_xfer0, temp_base3, lookup_long, 1], ctx_swap, optimize_mem, defer[1]
	ld_field_w_clr[prev_rt_long, 1100, $rd_xfer0]	
	ld_field_w_clr[lookup_long, 0011, $rd_xfer0], load_cc		; get trie pointer, set cond code
	br=0[set_route_ptr#]
	br[fourth_lookup_long#]

set_route_ptr#:
	alu[out_rt_ptr, --, B, $rd_xfer0, >>17]				; long match
	br!=0[end#]
	alu[out_rt_ptr, --, B, prev_rt_long, >>17]		; long match at previous trie
	br!=0[end#]
    alu[out_rt_ptr, --, B, $rd_xfer1, >>17]				; short match
	br!=0[end#]
	alu[out_rt_ptr, --, B, prev_rt_short, >>17]		; short match at previous trie
end#:
.endlocal
.endlocal
#endm


#endif // IP_UC