//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// mailbox.uc
// sram or scratch mailbox macros
//
//
// system: IXP1200
// subsystem: microcode
// usage: library macros
// author: dfh	Sep 6, 2000
//
// revisions:


#ifndef MAILBOX_UC
#define MAILBOX_UC

#include "stdmac.uc"

// API:

//	mailbox_init(in_type, in_addr1, in_addr2, optional_token)
//	mailbox_send(in_data, in_type, in_addr1, in_addr2, optional_token)
//	mailbox_receiveout_xfer, in_type, in_addr1, in_addr2, optional_token)


// mailbox_init
//		description: initialize a 32 bit longword at a scratch or sram mailbox
//					 set bit 31 as a valid bit.
//		inputs:
//			in_type			mem type (sram or scratch)
//			in_addr1		first 32 bit longword address to be added to form mailbox address
//			in_addr2		second address to be added to form mailbox address
//			optional_token	sig_done, ctx_swap, or no_signal
//
//		example:
//			mailbox_init(scratch, mailbox_addr, LWOFFSET2, ctx_swap); 
//
#macro mailbox_init(in_type, in_addr1, in_addr2, optional_token)
xbuf_alloc($mailbox_xfer, 1);
	move($mailbox_xfer[0], 0);
	#define_eval _MEM_TYPE in_type
	_MEM_TYPE/**/_write($mailbox_xfer[0], in_addr1, in_addr2, 1, optional_token);
xbuf_free($mailbox_xfer);
#endm


// mailbox_send
//		description: write a 32 bit longword to a scratch or sram mailbox
//					 set bit 31 as a valid bit.
//		inputs:
//			in_data			data bits 30:0 to write
//			in_type			mem type (sram or scratch)
//			in_addr1		first 32 bit longword address to be added to form mailbox address
//			in_addr2		second address to be added to form mailbox address
//			optional_token	sig_done, ctx_swap, or no_signal
//
//		example:
//			mailbox_send(mpacket_state, scratch, mailbox_addr, LWOFFSET2, ctx_swap); 
//
#macro mailbox_send(in_data, in_type, in_addr1, in_addr2, optional_token)
xbuf_alloc($mailbox_xfer, 1);
	#if(isnum(in_data))
		#define_eval _immed_data (in_data | 0x80000000)
		move($mailbox_xfer[0], _immed_data);
	#else
		alu_shf_left($mailbox_xfer[0], in_data, OR, 1, 31);
	#endif
	#define_eval _MEM_TYPE in_type
	_MEM_TYPE/**/_write($mailbox_xfer[0], in_addr1, in_addr2, 1, optional_token);
xbuf_free($mailbox_xfer);
#endm


// mailbox_receive
//		description: read a 32 bit longword from a scratch or sram mailbox
//					 use test and clear on bit 31 valid bit. retry if bit 31 is off
//		outputs:
//			out_xfer		data bits 30:0 to write
//		inputs:
//			in_type			mem type (sram or scratch)
//			in_addr1		first 32 bit longword address to be added to form mailbox address
//			in_addr2		second address to be added to form mailbox address
//			optional_token	
//							retry:		if mailbox not valid, retry reading the mailbox
//							no_option:	read the mailbox once, after which application can check
//										bit 31 to see if message is valid
//
//		example:
//			mailbox_receive(mpacket_state, mailbox_addr, LWOFFSET2); 
//
#macro mailbox_receive(out_xfer, in_type, in_addr1, in_addr2, optional_token)
#if (streq(optional_token, retry))
	repeat#:
	#define_eval _MEM_TYPE in_type
	_MEM_TYPE/**/_bits_test_and_clr(out_xfer, in_addr1, in_addr2, 0, 0x80000000, ctx_swap)
	shf_right(--, out_xfer, 31);
	br=0[repeat#]
#else
	_MEM_TYPE/**/_bits_test_and_clr(out_xfer, in_addr1, in_addr2, 0, 0x80000000, ctx_swap)
#endif
#endm


#endif // MAILBOX_UC