//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// packetq.uc
// macros for queueing packets 
//
//
// system: IXP1200
// subsystem: forwarding microcode
// usage: library macros
// author: dfh	May 1, 2000
//
// revisions:


#ifndef PACKETQ_UC
#define PACKETQ_UC


// API:
//	packetq_send(in_desc_addr, in_q_index, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE)
//	packetq_seq_wait(in_rec_state, WHICH_FAST_PORT)
//	packetq_index(out_xfer_index, in_q_index, Q_BASE, SIZE_OF_ENTRY, NUM_ENTRIES)


// get REC_STATE* definitions
#include "port.uc"

// define queue types
#define Q_LINKED_LIST     0					// generic linked list
#define Q_LINKED_LIST_2 0x1					// linked list queue with 2 sub-queues
#define Q_LINKED_LIST_4 0x2					// linked list queue with 4 sub-queues
#define Q_LINKED_LIST_8 0x3					// linked list queue with 8 sub-queues

#define Q_CIRCULAR   0x10000				// generic circular queue
#define Q_CIRCULAR_2 0x10001				// circular queue with 2 sub-queues
#define Q_CIRCULAR_4 0x10002				// circular queue with 4 sub-queues
#define Q_CIRCULAR_8 0x10003				// circular queue with 8 sub-queues


// define byte locations in common descriptor
// descriptor word 1
#define DESC1_ELE_COUNT 0					// 7:0 element or cell count
#define DESC1_PKT_END_BYTE 8				// 15:8 ending quadword/byte in last mpacket or cell
#define DESC1_PKT_START_BYTE 16				// 23:16 starting quadword/byte in packet buffer
#define DESC1_FREELIST 24					// 26:24 freelist id


// packetq_send
// 		description:	enqueue a packet. 
//						depending on the q interface, it could be for transmit, core stack, or other thread.
//						allocates and uses up to 4 transfer registers 
//		inputs:
//			in_desc_addr		packet descriptor sram address
//			in_q_index			interface selected by route lookup or packet classification
//			Q_TYPE				type of queue. linked list, circular, etc
//			Q_RDY				location in scratch for packet present indication
//			Q_BASE				base address of queue head array
//			DESC_BASE			base address of descriptor buffers
//		size: 
//		xfer regs used:
//			4 sram (3 by this macro, 1 by scratch_bset)
//		example usage:
//			#define QUEUE_HANDLE QX8_LINKED_LIST,XMIT_PWP_VECTOR,SRAM_QUEUE_DESCRIPTOR_BASE,SRAM_BUFF_DESCRIPTOR_BASE
//			packetq_send(descriptor_addr, output_intf, Q_HANDLE);	// at eop, q to xmit
//
#macro packetq_send(in_desc_addr, in_q_index, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE)
	#if ((Q_TYPE & 0xFF0000)== Q_CIRCULAR)
		_packetq_fast_send[in_q_index, in_desc_addr, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE]
	#elif ((Q_TYPE & 0xFF0000) == Q_LINKED_LIST)
		_packetq_slow_send[in_q_index, in_desc_addr, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE]
	#else
		#error "packetq_send. Q type not supported."
	#endif
#endm


// _packetq_slow_send
//		description: enqueue a packet on a linked list queue. support macro packetq_send
//
#macro _packetq_slow_send(in_q_index, in_desc_addr, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE)
.local queue_packet_count tempa tail_ptr head_ptr descriptor_base queue_descriptor_addr relative_buf_addr
xbuf_alloc[$que_buf, 3]				// declare local transfer reg array x4-x7

	immed32[queue_descriptor_addr, Q_BASE]
	immed32[descriptor_base, DESC_BASE]
	alu[queue_descriptor_addr, queue_descriptor_addr, +, in_q_index, <<1]	; add queue index * 2

	sram [read_lock, $que_buf0, queue_descriptor_addr, 0, 2], 
				optimize_mem, ctx_swap, defer[1]							; read the queue descriptor 2 words
	alu[relative_buf_addr, in_desc_addr, -, descriptor_base]
	alu_shf[queue_packet_count, 1, +16, $que_buf1]					; update queue element count
	.if (queue_packet_count == 1)									; if queue was empty before
		#if ((Q_TYPE & 0xFFFF) != 0)
			#define_eval _Q_VECTOR_SHIFT (Q_TYPE & 0xFFFF)
			shf_right(tempa, in_q_index, _Q_VECTOR_SHIFT);			; get primary queue vector bit position
		#endif
		scratch_bits_set[Q_RDY, 0, tempa, 1, sig_done]				; set packet queued bit for output port
		alu_shf [tempa, 0, b, relative_buf_addr, <<16]	
		alu [$que_buf0, tempa, or, relative_buf_addr]				; set head pointer and tail pointer to new buffer descr
		ld_field_w_clr[$que_buf1, 0111, queue_packet_count]
		sram [write_unlock, $que_buf0, queue_descriptor_addr, 0, 2], priority, ctx_swap ; write the updated queue descriptor and release the lock - ctx_swap required
		ctx_arb[fbi]
	.else															; if queue had one or more packet
		ld_field_w_clr[tail_ptr, 0011, $que_buf0]
		ld_field_w_clr[head_ptr, 1100, $que_buf0]
		alu [$que_buf0, head_ptr, or, relative_buf_addr]			; OR in the new tail pointer for queue descr
		alu [$que_buf2, --, B, relative_buf_addr]					; relative_buf_addr is the new tail
		sram [write, $que_buf2, descriptor_base, tail_ptr, 1], priority	; rewrite the current (now old) tail
		ld_field_w_clr[$que_buf1, 0111, queue_packet_count]
		sram [write_unlock, $que_buf0, queue_descriptor_addr, 0, 2], priority, ctx_swap ; write the updated queue descriptor and release the lock - ctx_swap required
	.endif
#ifdef PROFILE
	immed[tempa, TOTAL_RECEIVES]				; instrumentation. from mem_map.h counter at scratch total enqueues
	scratch[incr, --, tempa, 0, 1]

#endif
xbuf_free[$que_buf]
.endlocal
#endm


// packetq_index
//		description: get the next index for a circular buffer, modify index by entry size, 
//
//		inputs:
//			NUM_ENTRIES			number of entries on the ring. must be a power of 2
//			SIZE_OF_ENTRY		size of each entry. in 32 bit longwords
//			Q_BASE				base of an array of queue descriptors. each descriptor is 2 longwords
//			in_q_index			GPR or read transfer reg. which queue
//			out_xfer_index		read transfer reg. next index 	
//
#macro packetq_index(out_xfer_index, in_q_index, Q_BASE, SIZE_OF_ENTRY, NUM_ENTRIES)
.local queue_descriptor_addr packetq_next_tempa
	immed32[queue_descriptor_addr, Q_BASE]
	alu[queue_descriptor_addr, queue_descriptor_addr, +, in_q_index, <<1]	; add queue index * 2

	sram [read_lock, out_xfer_index, queue_descriptor_addr, 0, 1],	ctx_swap
	alu[packetq_next_tempa, SIZE_OF_ENTRY, +, out_xfer_index]			; increment to next descriptor index
	#define_eval _WRAP_MASK ((SIZE_OF_ENTRY * NUM_ENTRIES) -1)
	#if (_WRAP_MASK == 15)
		alu[out_xfer_index, 0, +4, packetq_next_tempa]				; truncate so wrap is = transmit wrap
	#elif (_WRAP_MASK == 255)
		alu[out_xfer_index, 0, +8, packetq_next_tempa]				; truncate so wrap is = transmit wrap
    #else
		alu[out_xfer_index, _WRAP_MASK, AND, packetq_next_tempa]
	#endif
	sram[write_unlock, out_xfer_index, queue_descriptor_addr, 0, 1], priority		; increment index
.endlocal
#endm



// _packetq_fast_send
//		description: increment element count for a circular queue.
//		inputs:
//			Q_RDY			scratch address for the element count
//
#macro _packetq_fast_send(in_q_index, in_desc_addr, Q_TYPE, Q_RDY, Q_BASE, DESC_BASE)
	scratch_incr(Q_RDY, 0);
#ifdef PROFILE
	scratch_incr(TOTAL_RECEIVES, 0);				// counter for packets enqueued
#endif
#endm											



// packetq_seq_wait
// 		description: wait for enqueue sequence to match received sequence number
//					 insure correct enqueue sequence for packets revceived on fast ports
//		inputs:
//			in_rec_state		information from received packet, sopseq#
//			WHICH_FAST_PORT		FAST_PORT1 or FAST_PORT2
//
#macro packetq_seq_wait(in_rec_state, WHICH_FAST_PORT)
.local tempa
	alu_shf[tempa, 0xf, AND, in_rec_state, >>/**/REC_STATE_SOPSEQ]	; extract sop seq number
	#if (WHICH_FAST_PORT == FAST_PORT_LO)							; which fast port we received on
		_packetq_fast_seq_check[tempa, ENQUEUE_SEQ1]			; wait for my seq no. to come up
		fast_wr[3, INCR_ENQ_NUM1]								; increment fast port 1 seq number
	#else
		_packetq_fast_seq_check[tempa, ENQUEUE_SEQ2]			; wait for my seq no. to come up
		fast_wr[3, INCR_ENQ_NUM2]								; increment fast port 1 seq number
	#endif
.endlocal
#endm


// enqueue packets from FAST PORT in order
// read the fbi fast port sequence number until it matches our sequence number
//
#macro _packetq_fast_seq_check[in_seq_num, CSR_NAME]
sleep_wait_for_seq_change#:
.local $seq_read_xfer
	csr[read, $seq_read_xfer, CSR_NAME], ctx_swap		; read the current seq number
	alu[--, in_seq_num, -, $seq_read_xfer]				; is my number up?
	br=0[seq_check_done#]
	br[sleep_wait_for_seq_change#]
.endlocal
seq_check_done#:
#endm


#endif  // PACKETQ_UC