//------------------------------------------------------------------------------------
//                                                                     
//                  I N T E L   P R O P R I E T A R Y                   
//                                                                      
//     COPYRIGHT (c) 2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//     THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                      
//                        INTEL  CORPORATION                            
//                                                                     
//                     2200 MISSION COLLEGE BLVD                        
//                                                                      
//               SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                      
//-----------------------------------------------------------------------------------
// pci.uc
// IXPblock PCI macros
//
//
// system: IXP1200
// subsystem: IO microcode
// usage: library macros
// author: MFF Nov, 2000
//
// revisions:	None
//

#ifndef PCI_MACROS
#define PCI_MACROS

// API
//
//	pci_descriptor_create(descriptor_dram_address, target_system_pci_base_address, target_system_dram_offset, 
//							our_dram_addr, next_descriptor_dram_addrs, END_OF_CHAIN, PCI_TO_DRAM, 
//								MIN_DELAY, MSG_SIZE)
//  pci_transfer(dram_descriptor_address, DRAM_DESCRIPTOR_OFFSET, QUEUE, OPTIONAL_TOKEN)
//
#include "stdmac.uc"



// pci_descriptor_create
//
// Description: This macro creates a PCI DMA descriptor in DRAM.
//
// arguments:
//		in_descriptor_addr		: DRAM address where descriptor is to be written. 
//								  Descriptor length is 16 bytes.
//		in_pci_base_byte_addr	: Base PCI address of targets memory.
//		in_target_dram_addr		: Offset from target's base address.  64 bit address. ie. Byte address / 8.
//		in_our_dram_addr		: 64 bit address of our memory.
//		in_next_descriptor		: 64 bit address of the next descriptor in chain.  If end of chain
//								  is set then this field is a don't care
// 		END_OF_CHAIN_FLAG  		: 0 == There is another link in this DMA chain
//								  1 == This is the end of the chain
//		TRANSFER_DIRECTION 		: 0 == PCI to DRAM
//								  1 == DRAM to PCI
//		CHANNEL_INTERBURST_DELAY :	max value is 0x3f (=63)
//		BYTES_TO_XFER            :	max value is 0xffffff
//
// size:  max = 12 instructions
// side effects: N/A
// see also: N/A
// example usage: 	
//		pci_descriptor_create(descriptor1, target_system_pci_address, target_dram_addr, 
//								msg_queue_addr, null_ptr, END_OF_CHAIN, PCI_TO_DRAM, 
//								MIN_DELAY, MSG_SIZE)
//
#macro pci_descriptor_create(in_descriptor_addr, in_pci_base_byte_addr, in_target_dram_addr, in_our_dram_addr, in_next_descriptor, END_OF_CHAIN_FLAG, TRANSFER_DIRECTION, CHANNEL_INTERBURST_DELAY, BYTES_TO_XFER)
.local $$pci_xfer0 $$pci_xfer1 $$pci_xfer2 $$pci_xfer3 _big_num_holder pci_temp
.xfer_order $$pci_xfer0 $$pci_xfer1 $$pci_xfer2 $$pci_xfer3

	// Error check the inputs for out of bounds values first

	// Descriptor Word 0
	//						31		End of Chain
	//						30		Transfer Direction
	//						29:24	Channel Interburst Delay
	//						23:0	Number of Bytes to Transfer
	alu[pci_temp, --, b, END_OF_CHAIN_FLAG, <<31]
	alu[pci_temp, pci_temp, or, TRANSFER_DIRECTION, <<30]
	#if (isnum(CHANNEL_INTERBURST_DELAY))
		immed32(_big_num_holder, CHANNEL_INTERBURST_DELAY)
		alu[pci_temp, pci_temp, or, _big_num_holder, <<24]
	#else	; they gave us a register
		alu[pci_temp, pci_temp, or, CHANNEL_INTERBURST_DELAY]
	#endif
	#if (isnum(BYTES_TO_XFER))
		immed32(_big_num_holder, BYTES_TO_XFER)
		alu[$$pci_xfer0, pci_temp, or, _big_num_holder]  		; 
	#else	; they gave us a register
		alu[$$pci_xfer0, pci_temp, or, BYTES_TO_XFER]			;
	#endif

	// Descriptor Word 1
	//						31:0	PCI Address (Byte Address)
	alu[$$pci_xfer1, in_pci_base_byte_addr, +, in_target_dram_addr, <<3]	; shift double longword by 3 = mult by 8 to get byte addr

	// Descriptor Word 2
	//						31:28	Reserved
	//						27:0	SDRAM Address (Byte Address)
	alu[$$pci_xfer2, --, b, in_our_dram_addr, <<3]		; create byte address

	// Descriptor Word 3
	//						31:28	Reserved
	//						27:4	Address of the next descriptor
	//						3:0		Reserved (force address to a double quadword)
	alu[$$pci_xfer3, --, b, in_next_descriptor, <<3]	; byte address

	// write the 4 xfer regs to sdram
	sdram[write, $$pci_xfer0, in_descriptor_addr, 0, 2], ctx_swap

.endlocal
#endm 




//  pci_transfer
//
//  Description:  
//		This macro kicks off a PCI DMA transfer using the pci_dma instruction.  The
//		input DMA descriptor address should be a quadword address.  The descriptor offset should
//		be a quadword offset.  This marcro divides both the in_desc_addr and the offset by 2, as
//		the pci_dma instruction requires a double quadword address.
//
//  Arguements:
//		in_desc_addr			 DRAM quadword descriptor address.  This value is added with 
//								descriptor_dram_offste to determine the address of the descriptor.
//								NOTE: descriptor_dram_address + descriptor_dram_offset must be an EVEN
//								dram memory location. ie. bit[0] of the address MUST be 0.
//		CONST_DESCR_OFFSET		Offset from descriptor_dram_address.  5 bit zero filled immediate, 
//		queue					order_queue: DMA channel 2
//								any_queue	 DMA channel 1 or DMA channel 2
//		optional_token			ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								voluntary_swap: allow other threads to execute and then this thread continues
//								no_signal or no_option: not requesting signal, continuing
//
// Limitation: dram_descriptor_address and dram_descriptor_offset must be EVEN values
//								ie. bit[0] MUST be 0.
//
// size:  2 instruction
// side effects: N/A
// see also: N/A
// example usage: 	
//		pci_transfer(descriptor1, 0, order_queue, ctx_swap) 
//
#macro pci_transfer(in_desc_addr, CONST_DESCR_OFFSET, CONST_QUEUE, CONST_TOKENS)
#define_eval _REAL_OFFSET (CONST_DESCR_OFFSET/2)		; descriptor address in pci_dma inst = address >> 1
.local pt_temp
	alu[pt_temp, --, b, in_desc_addr, >>1]
	pci_dma[pt_temp, _REAL_OFFSET, CONST_QUEUE], CONST_TOKENS
.endlocal
#undef _REAL_OFFSET
#endm

#endif ; PCI_MACROS