//------------------------------------------------------------------------------------
//                                                                      
//                   I N T E L   P R O P R I E T A R Y                   
//                                                                       
//      COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
//      RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
//      BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
//      RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
//      LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
//      MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
//      THE PRIOR WRITTEN PERMISSION OF :                                
//                                                                       
//                         INTEL  CORPORATION                            
//                                                                      
//                      2200 MISSION COLLEGE BLVD                        
//                                                                       
//                SANTA  CLARA,  CALIFORNIA  95052-8119                  
//                                                                       
//------------------------------------------------------------------------------------
// sram.uc
// macros to access sram 
//
//
// system: IXP1200
// subsystem: utility microcode
// usage: library macros
// author: dfh	May 1, 2000
//
// revisions:
//		8/8/00	dfh		accept register or constant args, traverse xbufs


#ifndef SRAM_UC
#define SRAM_UC


// API:
//	sram_read(out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
//	sram_read_lock(out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
//	sram_write(in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
//	sram_write_unlock(in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
//	sram_unlock(in_base_lw_addr, in_lw_offset, optional_token)
//	sram_bits_set(in_base_lw_addr, in_lw_offset, in_start_bit_pos, optional_token)
//	sram_bits_test_and_set(out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
//	sram_bits_clr(in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
//	sram_bits_test_and_clr(out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)

#include "stdmac.uc"
#include "xbuf.uc"


// sram_read
//		description: Read from sram and lock address of first longword.
//		outputs:
//			out_xfer			first sram xfer reg of sequence to read to, array notation
//								must be in xbuf array notation, index range  0-7
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_lw_count			register or constant longword count
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//								optimize_mem	optimize queue selection and sig_done
//								priority		priority queue and sig_done
//		temp register usage:	0-2 	uses regs if constant addr args > MAX_IMMEDIATE, or register length
//		example:
//			sram_read($packet[2], addr, 0, LWCOUNT3, ctx_swap);
//
//
#macro	sram_read(out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
_sram_op(read, out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
#endm


// sram_read_lock
//		description: Read from sram.
//		outputs:
//			out_xfer			first sram xfer reg of sequence to read to, array notation
//								must be in xbuf array notation, index range  0-7
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_lw_count			register or constant longword count
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//								optimize_mem	optimize queue selection and sig_done
//								priority		priority queue and sig_done
//		temp register usage:	0-2 	uses regs if constant addr args > MAX_IMMEDIATE, or register length
//		example:
//			sram_read_lock($packet[2], addr, 0, LWCOUNT3, ctx_swap);
//
//
#macro	sram_read_lock(out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
_sram_op(read_lock, out_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
#endm


// sram_write
//		description: Write to sram.
//		inputs:
//			in_xfer				first sram xfer reg of sequence to write from, array notation
//								must be in xbuf array notation, index range  0-7
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_lw_count			register or constant longword count
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//								optimize_mem	optimize queue selection and sig_done
//								priority		priority queue and sig_done
//		temp register usage:	0-2 	uses regs if constant addr args > MAX_IMMEDIATE, or register or length
//		example:
//			sram_write($packet[2], addr, 0, LWCOUNT3, priority);
//
#macro	sram_write(in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
_sram_op(write, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
#endm


// sram_write_unlock
//		description: Write to sram and unlock address of first longword.
//		inputs:
//			in_xfer				first sram xfer reg of sequence to write from, array notation
//								must be in xbuf array notation, index range  0-7
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_lw_count			register or constant longword count
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//								optimize_mem	optimize queue selection and sig_done
//								priority		priority queue and sig_done
//		temp register usage:	0-2 	uses regs if constant addr args > MAX_IMMEDIATE, or register length
//		example:
//			sram_write_unlock($packet[2], addr, 0, LWCOUNT3, priority);
//
#macro	sram_write_unlock(in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
_sram_op(write_unlock, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
#endm


// sram_unlock
//		description: Write to sram and unlock address of first longword.
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//								optimize_mem	optimize queue selection and sig_done
//								priority		priority queue and sig_done
//		temp register usage:	0-2 	uses regs if constant addr args > MAX_IMMEDIATE, or register length
//		example:
//			sram_unlock(addr, 0, priority);
//
#macro	sram_unlock(in_base_lw_addr, in_lw_offset, optional_token)
_sram_op(unlock, no_xfer, in_base_lw_addr, in_lw_offset, 1, optional_token)
#endm


#macro	_sram_op(operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
.local _temp_addr

#if ((isnum(in_base_lw_addr))&&(!isnum(in_lw_offset)))
	#if ((in_base_lw_addr > MAX_IMMEDIATE))
		immed32[_temp_addr, in_base_lw_addr]
		_sram_op_addr_asis[operation, in_xfer, _temp_addr, in_lw_offset, in_lw_count, optional_token)
 	#else
		_sram_op_addr_asis[operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
	#endif
	// ((isnum(in_base_lw_addr))&&(!isnum(in_lw_offset)))

#elif ((isnum(in_lw_offset))&&(!isnum(in_base_lw_addr)))
	#if ((in_lw_offset > MAX_IMMEDIATE))
		immed32[_temp_addr, in_lw_offset]
		_sram_op_addr_asis[operation, in_xfer, _temp_addr, in_base_lw_addr, in_lw_count, optional_token)
	#else
		_sram_op_addr_asis[operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
	#endif  // ((in_lw_offset > MAX_IMMEDIATE))
 	// ((isnum(in_lw_offset))&&(!isnum(in_base_lw_addr)))

#elif ((isnum(in_lw_offset))&&(isnum(in_base_lw_addr)))
	#define_eval _addr_sum (in_lw_offset + in_base_lw_addr)
	immed32[_temp_addr,_addr_sum]
	_sram_op_addr_asis[operation, in_xfer, _temp_addr, 0, in_lw_count, optional_token)

// else both addresses are GPRs
#else ((!isnum(in_lw_offset))&&(!isnum(in_base_lw_addr)))
	_sram_op_addr_asis[operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)

#endif // ((isnum(in_lw_offset))&&(isnum(in_base_lw_addr)))
.endlocal  // _temp_addr
#endm


// _sram_op_addr_asis
// 		description: sram operation where addresses are acceptable as is to Assembler
//					 support macros sram_read, sram_read_lock, sram_write, sram_write_unlock
//
#macro _sram_op_addr_asis[operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
#if (!streq(in_xfer, no_xfer))
	xbuf_find[in_xfer]		// if found, #define _xbuf_found, _cur_d_xbuf_name, _cur_d_xbuf_size
	#if (isnum (in_lw_count))
		#ifdef _xbuf_found
			#define _perform_xbuf_op
		#endif
	#endif
#endif
#ifdef _perform_xbuf_op
 	_sram_op_xbufs[operation, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token]
	#undef _perform_xbuf_op
#elif (isnum (in_lw_count))
	_sram_wtok[operation, in_xfer, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token]
#else
	.local _temp_ind 

	immed32 [_temp_ind, 0x100000]							; set bit 20 indirect refcount override bit
	alu_shf_left(--, _temp_ind, OR, in_lw_count, 16)		; insert indirect refcount
	_sram_ind_wtok[operation, in_xfer, in_base_lw_addr, in_lw_offset, 1, optional_token]
	.endlocal //  _temp_ind
#endif	// (isnum (in_lw_count))

#endm


// _sram_op_xbufs
//		description: perform sram operation with one or more(linked) xbufs.  
//					 support macros sram_read, sram_read_lock, sram_write, sram_write_unlock
//		
#macro _sram_op_xbufs(operation, in_base_lw_addr, in_lw_offset, in_lw_count, optional_token)
.local new_base_addr

#define_eval _rem_buf (_cur_xbuf_size - _cur_xbuf_index)
#define_eval _cur_lw_offset (in_lw_offset)
#define_eval _cur_lw_count (in_lw_count)

#if (_cur_lw_count > _rem_buf)
	#define _use_new_base
#endif

#if (!isnum(in_lw_offset))
	#ifndef _use_new_base
		#define _use_new_base
	#endif
#endif

#ifdef _use_new_base
	alu_op[new_base_addr, in_base_lw_addr, +, in_lw_offset]
	#define_eval _cur_lw_offset (0)
#endif

#while (_cur_lw_count > _rem_buf)
	sram[operation, _cur_xbuf_name/**/_cur_xbuf_index, new_base_addr, _cur_lw_offset, _rem_buf]
    #define_eval _cur_lw_offset (_cur_lw_offset + _rem_buf)
    #define_eval _cur_lw_count (_cur_lw_count - _rem_buf)
	
	#ifdef _cur_xbuf_next
		 xbuf_param_set[_cur_xbuf_next]
		#define_eval _cur_xbuf_index 0
		#define_eval _rem_buf _cur_xbuf_size
	#else
		#error	"_sram_op_xbufs. out of xbuf range"
	#endif	//  _cur_xbuf_next

#endloop

// perform last operation
#ifdef _use_new_base
	_sram_wtok[operation, _cur_xbuf_name/**/_cur_xbuf_index, new_base_addr, _cur_lw_offset, _cur_lw_count, optional_token]
	#undef _use_new_base
#else
	_sram_wtok[operation, _cur_xbuf_name/**/_cur_xbuf_index, in_base_lw_addr, _cur_lw_offset, _cur_lw_count, optional_token]
#endif
.endlocal
#endm


// sram_bits_set
//		description: Set in_vector bits shifted left by in_start_bit_pos at sram.
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_start_bit_pos	register or constant<32, starting bit position
//								in_vector is shifted left by this amount.
//			in_vector			register or constant, mask of bits to set							
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//		size: 2-6 instr
//		xfer regs used:
//			1
//		example usage:
//			sram_bits_set(0x111, 0, bit_position, ctx_swap);
//
//
#macro sram_bits_set(in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
xbuf_alloc($bset_vector, 1);
_sram_bits_op(set_bits, $bset_vector[0], in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
xbuf_free($bset_vector);
#endm


// sram_bits_test_and_set
//		description: Set in_vector bits shifted left by in_start_bit_pos at sram.
//					 Read contents of sram address prior to the write.
// 		outs:
//			out_xfer			read xfer register result
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_start_bit_pos	register or constant<32, starting bit position
//								in_vector is shifted left by this amount.
//			in_vector			register or constant, mask of bits to set							
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//		size: 2-6 instr
//		xfer regs used:
//			1
//		example usage:
//			sram_bits_test_and_set(0x111, 0, bit_position, ctx_swap);
//
//
#macro sram_bits_test_and_set(out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
_sram_bits_op(test_and_set_bits, out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
#endm


// sram_bits_clr
//		description: Clear in_vector bits shifted left by in_start_bit_pos at sram.
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_start_bit_pos	register or constant<32, starting bit position
//								in_vector is shifted left by this amount.
//			in_vector			register or constant, mask of bits to set							
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//		size: 2-6 instr
//		xfer regs used:
//			1
//		example usage:
//			sram_bits_clr(0x111, 0, bit_position, ctx_swap);
//
//
#macro sram_bits_clr(in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
xbuf_alloc($bset_vector, 1);
_sram_bits_op(clear_bits, $bset_vector[0], in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
xbuf_free($bset_vector);
#endm


// sram_bits_test_and_clr
//		description: Clear in_vector bits shifted left by in_start_bit_pos at sram.
//					 Read contents of sram address prior to the write.
// 		outs:
//			out_xfer			read xfer register result
//		inputs:
//			in_base_lw_addr		register or constant base longword address
//			in_lw_offset		register or constant longword address offset
//			in_start_bit_pos	register or constant<32, starting bit position
//								in_vector is shifted left by this amount.
//			in_vector			register or constant, mask of bits to set							
//			optional_token		ctx_swap:  waiting on completion signal, context swap
//								sig_done:  requesting signal when done, continuing execution
//								no_signal or no_option: not requesting signal, continuing
//		size: 2-6 instr
//		xfer regs used:
//			1
//		example usage:
//			sram_bits_test_and_clr(0x111, 0, bit_position, ctx_swap);
//
//
#macro sram_bits_test_and_clr(out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
_sram_bits_op(test_and_clear_bits, out_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)
#endm


// _sram_bits_op
//		description: support macros sram_bits_set, sram_bits_clr
//
#macro _sram_bits_op(operation, io_xfer, in_base_lw_addr, in_lw_offset, in_start_bit_pos, in_vector, optional_token)

.local _temp_shift _temp_addr
#if (!isnum(in_start_bit_pos))
	alu[_temp_shift, in_start_bit_pos, B-A, 1, <<5]	; invert shift amount (32 - output_port) for left indirect
	alu[--, _temp_shift, B, 0]					; feed to A operand for indirect shift
	alu_shf[io_xfer, --, B, in_vector, <<indirect]
#else
	#define_eval _SHIFT_AMT (in_start_bit_pos)
	shf_left[io_xfer, in_vector, _SHIFT_AMT]
#endif

#if ((isnum(in_base_lw_addr))&&(!isnum(in_lw_offset)))
	#if ((in_base_lw_addr > MAX_IMMEDIATE))
		immed32[_temp_addr, in_base_lw_addr]
		_sram_wtok[bit_wr, io_xfer, _temp_addr, in_lw_offset, operation, optional_token]
	#else
		_sram_wtok[bit_wr, io_xfer, in_base_lw_addr, in_lw_offset, operation, optional_token]
	#endif  // ((in_base_lw_addr > MAX_IMMEDIATE))

#elif ((isnum(in_lw_offset))&&(!isnum(in_base_lw_addr)))
	#if ((in_lw_offset > MAX_IMMEDIATE))
		immed32[_temp_addr, in_lw_offset]
		_sram_wtok[bit_wr, io_xfer, _temp_addr, in_base_lw_addr, operation, optional_token]
	#else
		_sram_wtok[bit_wr, io_xfer, in_base_lw_addr, in_lw_offset, operation, optional_token]
	#endif  // ((in_lw_offset > MAX_IMMEDIATE))

#elif ((isnum(in_lw_offset))&&(isnum(in_base_lw_addr)))
	#define_eval _addr_sum (in_lw_offset + in_base_lw_addr)
	immed32[_temp_addr, _addr_sum]
	_sram_wtok[bit_wr, io_xfer, _temp_addr, 0, operation, optional_token]

// else both addresses are GPRs
#else
	_sram_wtok[bit_wr, io_xfer, in_base_lw_addr, in_lw_offset, operation, optional_token]

#endif // ((isnum(in_lw_offset))&&(isnum(in_base_lw_addr)))

.endlocal
#endm

// support for optional_token
//	for optimize_mem and priority, force sig_done also
//
#macro _sram_wtok(op_code, io_xfer, in_base_lw_addr, in_lw_offset, op_count, optional_token)
	#if (streq(optional_token, optimize_mem))
		#define _combined_token	optimize_mem, sig_done
	#elif (streq(optional_token, priority))
		#define _combined_token priority, sig_done
	#else
		#define_eval _combined_token (optional_token)
	#endif

#if (streq(io_xfer, no_xfer))
	sram[op_code, --, in_base_lw_addr, in_lw_offset, op_count], _combined_token 
#else
	sram[op_code, io_xfer, in_base_lw_addr, in_lw_offset, op_count], _combined_token 
#endif

	#undef _combined_token
#endm

// support for optional_token
//	for optimize_mem and priority, force sig_done also
//
#macro _sram_ind_wtok(op_code, io_xfer, in_base_lw_addr, in_lw_offset, op_count, optional_token)
	#if (streq(optional_token, optimize_mem))
		#define _combined_token	optimize_mem, sig_done
	#elif (streq(optional_token, priority))
		#define _combined_token priority, sig_done
	#else
		#define_eval _combined_token (optional_token)
	#endif
#if (streq(io_xfer, no_xfer))
	sram[op_code, --, in_base_lw_addr, in_lw_offset, op_count], indirect_ref, _combined_token 
#else
	sram[op_code, io_xfer, in_base_lw_addr, in_lw_offset, op_count], indirect_ref, _combined_token 
#endif

	#undef _combined_token

#endm

#endif // SRAM_UC