// Transmit FIFO macros
#ifndef TFIFO_MACROS
#define TFIFO_MACROS


// API
//
// tfifo_status_write[in_tfifo_entry, in_xfr_tfifo_status_wd0]
// tfifo_skipstatus_write[in_tfifo_entry, in_pkt_buffer_base, in_bit15on, CONST_FID, CONST_SDRAM_VALIDATE_FLAG]
// tfifo_validate_write[in_tfifo_entry, in_bit15on]
// tfifo_element_skip[out_abs_mpkt_tx_ok_cnt, in_tfifo_entry, in_pkt_buffer_base, in_port_rdy_bit, in_next_thread_ir, in_bit15on, CONST_SDRAM_VALIDATE_FLAG, FID]
// tfifo_element_skip_nordy[in_tfifo_entry,  in_pkt_buffer_base, in_bit15on, CONST_SDRAM_VALIDATE_FLAG, FID, THREAD1, THREAD2, THREAD3]
// tfifo_ptr_wait[in_tfifo_entry, XMIT_PTR_WINDOW_SIZE]
// tfifo_100m_validate[in_tfifo_entry, in_xfr_tfifo_ctl_wd0, in_bit15on, CONST_PORT_TO_TFIFO_MULT, CONST_SDRAM_VALIDATE_FLAG, CONST_THREAD_1, CONST_THREAD_2, CONST_THREAD_3]
// tfifo_validate[io_mpkt_tx_ok_cnt, in_xfr_tfifo_ctl_wd0, in_tfifo_entry, in_port_rdy_bit, in_bit15on, in_const1, in_next_thread_ir, CONST_SDRAM_VALIDATE_FLAG]
// tfifo_validate_sop[out_mpkt_tx_ok_cnt, in_xfer_tfifo_ctl_wd0, in_tfifo_entry, in_port_rdy_bit, in_bit15on, in_next_thread_ir, CONST_SDRAM_VALIDATE_FLAG]
//
//


// tfifo_status_write
//
// Description: This macro writes status to the transmit FIFO.
//
// arguments:
//		in_tfifo_entry			- Transmit FIFO element number (0 through 15)
//		in_xfr_tfifo_status_wd0	- SRAM transfer register that contains the
//								  status that is to be writting to the TFIFO
//
// size: 3 instructions
// side effects: N/A
// see also: N/A
// example usage: 	tfifo_status_write[tfifo_entry, $tfifo_ctl_wd0]
//
#macro tfifo_status_write[in_tfifo_entry, in_xfr_tfifo_status_wd0]
.local tsw_temp 
	alu_shf[tsw_temp, 0, +8, in_tfifo_entry, <<1]					; status address = element no. x 2
	alu_shf[tsw_temp, tsw_temp, +, 1, <<7]							;	+ 128
    t_fifo_wr [in_xfr_tfifo_status_wd0, tsw_temp, 0, 1], ctx_swap	; write the status word
.endlocal
#endm


// tfifo_skipstatus_write
//
// Description: Write a status control word to our tfifo element that indicates to the HW
//				that our tfifo element should be skipped.  This routine supports SDRAM dual-
//				validation through the use of the CONST_SDRAM_VALIDATE_FLAG input paramenter.
//
// arguments:
//		in_tfifo_entry				- our assigned tfifo element number. must be >=0, and <=15
//		in_pkt_buffer_base			- SDRAM packet buffer base address
//		in_bit15on					- constant 0x8000
//		CONST_FID					- this Fbox's ID
//		CONST_SDRAM_VALIDATE_FLAG	- SDRAM dual validation flag
//
// size:  13 instructions
// side effects: None
// see also: tfifo_status_write, which is actually used to write the status to the tfifo.
// example usage: 	tfifo_skipstatus_write[tfifo_entry, pkt_buffer_base, 
//			bit15on, FID, SDRAM_VALIDATE_FLAG]
//
#macro tfifo_skipstatus_write[in_tfifo_entry, in_pkt_buffer_base, in_bit15on, CONST_FID, CONST_SDRAM_VALIDATE_FLAG]
.local _bank $_temp_xfr_reg0 $_temp_xfr_reg1 _const_1
.xfer_order  $_temp_xfr_reg0 $_temp_xfr_reg1

	alu[$_temp_xfr_reg1, --, b, 0]		; second status word is always zero
	// when sdram is setting valid bits, we have to issue a dummy request
	#if (CONST_SDRAM_VALIDATE_FLAG)
		#if (VERSION_MIN == 0)			; will not work on A3 chip
			#error 3 "Dual validate not supported on chip version 0"
		#endif
		//cause skips to go to different banks for different fboxes
		alu[_const_1, --, b, 1]
		alu[_bank, _const_1, and, FID]
		alu[_bank, --, b, _bank, <<20]						; assumes 8 MB bank size
		alu_shf[--, in_bit15on, or, in_tfifo_entry, <<7]	; put element no. in 10:7
		sdram[t_fifo_wr, --, in_pkt_buffer_base, _bank, 1], indirect_ref, sig_done	; must send 1 auto validate
	#endif													; SDRAM_VALIDATE_FLAG
	alu[$_temp_xfr_reg0, in_tfifo_entry, OR, 1, <<7]		; set skip bit
	tfifo_status_write[in_tfifo_entry, $_temp_xfr_reg0]
.endlocal
#endm


// tfifo_validate_write
//
// Description: Write validate for the input tfifo element.
//
// arguments:
// 		in_tfifo_entry - Target tfifo element number. Must be >=0 and <=15.
//		in_bit15on     - Constant used in indirect reference.
//
// size: 2 instructions 
// side effects: None.
// see also: Programmers reference manual
// example usage: 	tfifo_validate_write[tfifo_entry, bit15on]
//
#macro tfifo_validate_write[in_tfifo_entry, in_bit15on]
	alu_shf[--, in_bit15on, OR, in_tfifo_entry, <<5]		; setup indirect data from element no.
	fast_wr[0, XMIT_VALIDATE], indirect_ref					; tell fbi to run with that element no.
#endm






// tfifo_element_skip
//
// Description: This macro performs the sequence required when a skip command
//				is received from the tx scheduler.
//
// arguments:
//		out_abs_mpkt_tx_ok_cnt		- Count of the number of mpkts that can 
//									  be output before checking thx xmit_rdy bit again
//		in_tfifo_entry				- Transmit FIFO element assigned to this fill thread
//		in_pkt_buffer_base			- Base address of packets in SDRAM
//		in_port_rdy_bit				- Register with bit set for our port
//		in_next_thread_ir			- thread number of the thread that is to be
//									  inter_thread signaled when this thread has 
//									  validated its tfifo element.  The thread number is 
//									  saved as (32-thread number) so that this register
//									  can be used as in indirect reference.
//		in_bit15on					- Constant required for indirect reference
//		CONST_SDRAM_VALIDATE_FLAG   - Used to support dual-validate mode
//		FID							- This FBox ID
//
// size: 13+1+7+5+2+2=30 instructions
// side effects: <global assumptions or side effects>
// see also: <reference to other libs or doc)
// example usage: <use of macro in a line of code>
//
#macro tfifo_element_skip[out_abs_mpkt_tx_ok_cnt, in_tfifo_entry, in_pkt_buffer_base, in_port_rdy_bit, in_next_thread_ir, in_bit15on, CONST_SDRAM_VALIDATE_FLAG, FID]

	tfifo_skipstatus_write[in_tfifo_entry, in_pkt_buffer_base, 
			in_bit15on, FID, CONST_SDRAM_VALIDATE_FLAG]
	ctx_arb[inter_thread]				; wait for our turn to validate
	tfifo_ptr_wait[in_tfifo_entry, 5]	; wait for xmit ptr to be within 5 of our fifo element
	port_txrdy_chk[out_abs_mpkt_tx_ok_cnt, in_port_rdy_bit]		; check output port and reset cnt
	tfifo_validate_write[in_tfifo_entry, in_bit15on]

	alu[--, --, b, in_next_thread_ir]			; next_thread_ir set during thread initialization
	fast_wr[0, inter_thd_sig], indirect_ref		; signal next thread

#endm


// tfifo_element_skip_nordy
//
// Description:  This macro performs the sequence required when a skip command
//				 is received from the tx scheduler.  This routine differs from
//				 the tfifo_element_skip routine in that this one does not check 
//				 the xmit_rdy register, and so it does not set the mpkt_tx_ok_cnt
//				 register.  This routine also differs in that it does not require 
//				 a register to be allocated for storing the next thread to be signaled.
//
// arguments:
//		in_tfifo_entry				- Transmit FIFO element assigned to this fill thread
//		in_pkt_buffer_base			- Base address of packets in SDRAM
//		in_bit15on					- Constant required for indirect reference
//		CONST_SDRAM_VALIDATE_FLAG   - Used to support dual-validate mode
//		FID							- This FBox ID
//		THREAD1						- Constant containing thread 1's thread ID
//		THREAD2						- Constant containing thread 2's thread ID
//		THREAD3						- Constant containing thread 3's thread ID
//
// size: 13+1+7+2+7=30 instructions
// side effects: None
// see also: N/A
// example usage: tfifo_element_skip_nordy[tfifo_entry, pkt_buffer_base, bit15on, 
//						SDRAM_VALIDATE, FID, THREAD1, THREAD2, THREAD3]
//
#macro tfifo_element_skip_nordy[in_tfifo_entry,  in_pkt_buffer_base, in_bit15on, CONST_SDRAM_VALIDATE_FLAG, FID, THREAD1, THREAD2, THREAD3]

#define TFIFO_XMIT_PTR_WINDOW 5

	tfifo_skipstatus_write[in_tfifo_entry, in_pkt_buffer_base, 
							in_bit15on, FID, CONST_SDRAM_VALIDATE_FLAG]
	ctx_arb[inter_thread]					; wait for our turn to validate
	tfifo_ptr_wait[in_tfifo_entry, TFIFO_XMIT_PTR_WINDOW]		; wait for xmit ptr to be within TFIFO_XMIT_PTR_WINDOW of our fifo element
	tfifo_validate_write[in_tfifo_entry, in_bit15on]

	//signal next thread telling it that it is its turn to validate
	br=ctx[1, ctx1#]
	br=ctx[2, ctx2#]
ctx3#:
	fast_wr[THREAD1, inter_thd_sig]
	br[end_tx_skip#]
ctx2#:
	fast_wr[THREAD3, inter_thd_sig]
	br[end_tx_skip#]
ctx1#:
	fast_wr[THREAD2, inter_thd_sig]
end_tx_skip#:
#undef TFIFO_XMIT_PTR_WINDOW
#endm

// tfifo_ptr_wait
//
// Description: Wait for the xmit_ptr to get within IN_XMIT_PTR_WINDOW_SIZE of our tfifo element.
//
// arguments:
//		in_tfifo_entry 			- TFIFO element that this fill thread has been assigned.
//		IN_XMIT_PTR_WINDOW_SIZE - Constant.  Maximum difference between xmit_ptr and in_tifo_entry.
//									This macro waits for the xmit_ptr to be  within 
//									IN_XMIT_PTR_WINDOW_SIZE of in_tifo_entry
//									

// size: 7 instructions
// side effects: None
// see also: N/A
// example usage: tfifo_ptr_wait[in_tfifo_entry, IN_XMIT_PTR_WINDOW_SIZE]
//
#macro tfifo_ptr_wait[in_tfifo_entry, IN_XMIT_PTR_WINDOW_SIZE]
.local $_xmit_ptr _temp
#define_eval WRAP_INCREMENT 16 - IN_XMIT_PTR_WINDOW_SIZE
wait_for_xmit_ptr#:

	csr[read, $_xmit_ptr, XMIT_PTR], ctx_swap

	alu[_temp, $_xmit_ptr, -, in_tfifo_entry]		; figure out where the xmit_ptr is
	br=0[endm#] 									; if xmit_ptr = tfifo_entry then make sure port is rdy before validate
	br>0[ptr_wrapped#]								; xmit ptr > t_fifo_entry -> wrap condition

	// No wrap condition
	alu[_temp, _temp, +, IN_XMIT_PTR_WINDOW_SIZE]
	br>=0[endm#]
	br[wait_for_xmit_ptr#]							; the xmit_ptr is not close enough yet

ptr_wrapped#:

	// Wrap condition
	alu[_temp, _temp, -, WRAP_INCREMENT]
	br<0[wait_for_xmit_ptr#]								; the xmit_ptr is not close enough yet
endm#:


.endlocal
#endm



// tfifo_100m_validate
//
// Description: This macro:
//					1. Waits for xmit_ptr to be within one element of our
//						target element.
//					2. Waits for the target port to be ready for data.
//					3. Validates our target tfifo element.
//					4. Signals the next thread indicating that it is its
//					 	turn to validate a tfifo element.
//
// arguments:
//		in_tfifo_entry				- Transmit FIFO to be validated
//		in_xfr_tfifo_ctl_wd0		- Status word to be written to the tfifo
//		in_bit15on					- Constant
//		CONST_PORT_TO_TFIFO_MULT	- Target port times this constant = tfifo element
//									  This design supports two possible values for this constant.
//									  For 16 port design, port to tfifo mult = 1. 
//									  ie port 3 is mapped to tfifo element 3
//       							  For 8_1f design, port to tfifo mult = 2. 
//									  ie port 3 is mapped to tfifo element 6
//		CONST_SDRAM_VALIDATE_FLAG	- Flag that indicates if the dual validate
//									  mode is being used
//		CONST_THREAD_1				- This fbox's thread id for thread 1
//		CONST_THREAD_2				- This fbox's thread id for thread 2
//		CONST_THREAD_3				- This fbox's thread id for thread 3
//
// size: 30 instructions
// side effects: N/A
// see also: N/A
// example usage: 	tfifo_100m_validate[tfifo_entry, $tfifo_ctl_wd0, bit15on, 
//						PORT_TO_TFIFO_MULT, SDRAM_VALIDATE, THREAD1, THREAD2, THREAD3]
//
#macro tfifo_100m_validate[out_return_status, in_tfifo_entry, in_xfr_tfifo_ctl_wd0, in_bit15on, CONST_PORT_TO_TFIFO_MULT, CONST_SDRAM_VALIDATE_FLAG, CONST_THREAD_1, CONST_THREAD_2, CONST_THREAD_3, IN_PASS, IN_FAIL]
.local _port _tfifo_entry_m1 _temp $_xmit_ptr $_tx_rdy_copy

	tfifo_status_write[in_tfifo_entry, in_xfr_tfifo_ctl_wd0]
	ctx_arb[inter_thread]		;wait for our turn

	xmit_ptr_wait#:
		csr[read, $_xmit_ptr, XMIT_PTR]
		csr[read, $_tx_rdy_copy, XMIT_RDY_LO], ctx_swap
		alu[_temp, $_xmit_ptr, -, in_tfifo_entry]
		br=0[port_wait_loop#] 
		br>0[ptr_wrapped#]								; xmit ptr > t_fifo_entry -> wrap condition
		// No wrap condition
		alu[_temp, _temp, +, 5] 
		br>=0[port_wait_loop#]
		br[xmit_ptr_wait#]							; the xmit_ptr is not close enough yet
		ptr_wrapped#:
		// Wrap condition
		alu[_temp, _temp, -, 11]
		br<0[xmit_ptr_wait#]								; the xmit_ptr is not close enough yet

port_wait_loop#:



	#if (CONST_PORT_TO_TFIFO_MULT == 1)
		alu[--, in_tfifo_entry, OR, 0]
	#endif
	#if (CONST_PORT_TO_TFIFO_MULT == 2)
		alu[_port, --, b, in_tfifo_entry, >>1] 
		alu[--, _port, OR, 0]
	#endif

	alu[--, 1, AND, $_tx_rdy_copy, >>indirect]
	br>0[write_validate#], defer[1]
	alu[out_return_status, --, b, PASS] ; set PASS flag

	alu[in_xfr_tfifo_ctl_wd0, in_tfifo_entry, OR, 1, <<7]		; set skip bit
	tfifo_status_write[in_tfifo_entry, in_xfr_tfifo_ctl_wd0]
	alu[out_return_status, --, b, FAIL] ; set fail flag

write_validate#:
	#if (!CONST_SDRAM_VALIDATE_FLAG)
		ctx_arb[sdram]											; wait for packet data to be transferred to tfifo
	#endif

	tfifo_validate_write[in_tfifo_entry, in_bit15on]

	//signal next thread telling it that it is its turn to write a valid bit
	br=ctx[1, ctx1#]
	br=ctx[2, ctx2#]
ctx3#:
	fast_wr[CONST_THREAD_1, inter_thd_sig]
	br[end_tfifo_validate#]
ctx2#:
	fast_wr[CONST_THREAD_3, inter_thd_sig]
	br[end_tfifo_validate#]
ctx1#:
	fast_wr[CONST_THREAD_2, inter_thd_sig]
end_tfifo_validate#:

.endlocal
#endm



// tfifo_validate
//
// Description: This routine first writes the fifo status word. Then it waits 
//				for an inter thread signal from the fill thread that was 
//				assigned to fill fifo element tfifo_entry - 2.  (The other tx 
//				scheduler fills  tfifo_entry - 1.)  We must limit the number 
//				of mpackets we have in the tfifo to 7.  If both schedulers 
//				allowed 8 mpackets in the tfifo then we could overwrite data 
//				that is being output by the hardware.  Variable @mpkt_tx_ok_cnt 
//				is used to keep track of how many mpackets can be written to the 
//				MAC without reading the xmit_rdy bits.  Before reading the xmit_rdy 
//				bits, this routine first must make sure that the xmit_ptr is not 
//				more than 4 fifo elements away from our tfifo_entry.  This must 
//				be done for a few reasons: 1. Make sure the MAC is still accepting 
//				data from our fifo.  2. Make sure we do not overwrite data in our tfifo.  
// 				3.  In order to set @mpkt_tx_ok_cnt.  For performance reasons we do 
//				not want to wait for the xmit_ptr to equal our tfifo_entry.  In comparing 
//				xmit_ptr to tfifo_entry there are 3 possibilities:  they are equal, the 
// 				xmit_ptr > tfifo_entry (wrap condition), and xmit_ptr < tfifo_entry 
//				(no wrap condition).
//
// 				xmit_ptr = tfifo_entry condition:
// 				Reset @mpkt_tx_ok_cnt to 6 because max mpackets in our tfifo = 7 - 1 
//				for the element we are validating here.
//
// 				No wrap condition: eg. $xmit_ptr = 4, and tfifo_entry = 8
// 				check $xmit_ptr - (tfifo_entry - 4) >= 0.  which is to say: 
//				$xmit_ptr - tfifo_entry + 4 >= 0  At most we can have 7 mpackets for this 
//				port in our TFIFO at a time.   Therefore, if our tfifo currently has less 
//				than or equal to 2 mpackets for our port right now (we don't want to wait 
//				until xmit_ptr = tfifo_entry for performance reasons), then we can send at 
//				least 5 more.  And since we are sending an mpacket right now, 
//				5 - 1 = 4 = @mpkt_tx_ok_cnt reset value. 
//
//				Q. Why check xmit_ptr - 4?   
//				A. Since this port fills every other tfifo element, max of 2 mpackets in 
//				our tfifo for our port right now * 2 (every other fifo element) = 4.
//
// 				Wrap condition: eg. $xmit_ptr could be 12, and tfifo_entry could be 0
// 				check $xmit_ptr - [(tfifo_entry + 16) - 4] >= 0.  or 
//				$xmit_ptr - tfifo_entry - 12 >= 0.  The 16 is to account for the wrap.  
//				The 4 is for the same reasons listed in the no wrap condition above.
// 				@mpkt_tx_ok_cnt is reset to 4 for same reasons as no wrap condition.
//
// arguments:
//		io_mpkt_tx_ok_cnt			- Number of mpkts that can be transmitted before
//									  checking the tx_rdy bit again.
//		in_xfr_tfifo_ctl_wd0		- Transmit FIFO control word that is to be written
//									  to the transmit FIFO
//		in_tfifo_entry				- Transmit FIFO element to be used
//		in_port_rdy_bit				- Register that has a bit set corresponding to the
//									  target port.  Bit 0 -> port 0.  Bit 1 -> port 1...
//		in_bit15on					- Constant
//		in_const1					- Constant
//		in_next_thread_ir			- thread number of the thread that is to be
//									  inter_thread signaled when this thread has 
//									  validated its tfifo element.  the thread number is 
//									  saved as (32-thread number) so that this register
//									  can be used only as in indirect reference.
//		CONST_SDRAM_VALIDATE_FLAG	- Immediate used in support of dual validate mode.
//
// size: 3+6+13+3+2+2=29 instructions
// side effects: The MAC's xmit_rdy threshold must be set to at least 7 mpackets * 64 bytes.  
//				 Also, This macro only works in dual transmit schedulers mode. 
// see also: N/A
// example usage: tfifo_validate[@mpkt_tx_ok_cnt, $tfifo_ctl_wd0, tfifo_entry,
//						port_rdy_bit, bit15on, const_1, next_thread_ir, 
//						SDRAM_VALIDATE_FLAG]
//
#macro tfifo_validate[io_mpkt_tx_ok_cnt, in_xfr_tfifo_ctl_wd0, in_tfifo_entry, in_port_rdy_bit, in_bit15on, in_const1, in_next_thread_ir, CONST_SDRAM_VALIDATE_FLAG]

	tfifo_status_write[in_tfifo_entry, in_xfr_tfifo_ctl_wd0]
	#if (CONST_SDRAM_VALIDATE_FLAG == 0)
		sig_wait[sdram, OR, 1]					; wait for packet data to be transferred to tfifo
	#endif
	sig_wait[inter_thread, OR, 1]				; wait for our turn to validate

	alu[io_mpkt_tx_ok_cnt, io_mpkt_tx_ok_cnt, -, in_const1]		; check for room in MAC's tfifo
	br>0[write_validate#]										; if there is enough room, then validate

	tfifo_ptr_wait[in_tfifo_entry, 4]	; wait for xmit ptr to be within 4 of our fifo element
	port_txrdy_wait[in_port_rdy_bit]
	immed32(io_mpkt_tx_ok_cnt, 4)		; reset number of mpackets that can be xmit without looking at xmit_rdy_lo

write_validate#:
	tfifo_validate_write[in_tfifo_entry, in_bit15on]
	alu[--, --, b, in_next_thread_ir]								; next_thread_ir set during thread initialization
	fast_wr[0, inter_thd_sig], indirect_ref							; signal next thread

#endm




// tfifo_validate_sop
//	
// Description: This macro is a MAC bug work-around.  It works very much like 
//				the tfifo_validate macro.  The only difference is that before
//				sending this mpacket, we must wait for all of the other 
//				mpackets that were sent to our port to actually get to the MAC,
//				and then, and only then, do we check the xmit_rdy bit.
//
// arguments:
//		out_mpkt_tx_ok_cnt			- Number of mpkts that can be transmitted before
//									  checking the tx_rdy bit again.
//		in_xfer_tfifo_ctl_wd0		- Transmit FIFO control word that is to be written
//									  to the transmit FIFO
//		in_tfifo_entry				- Transmit FIFO element to be used
//		in_port_rdy_bit				- Register that has a bit set corresponding to the
//									  target port.  Bit 0 -> port 0.  Bit 1 -> port 1...
//		in_bit15on					- Constant
//		in_next_thread_ir			- thread number of the thread that is to be
//									  inter_thread signaled when this thread has 
//									  validated its tfifo element.  the thread number is 
//									  saved as (32-thread number) so that this register
//									  can be used only as in indirect reference.
//		CONST_SDRAM_VALIDATE_FLAG	- Immediate used in support of dual validate mode.
//
// size: 3+4+7+3+2+4=23 instructions
// side effects: N/A
// see also: N/A
// example usage: tfifo_validate_sop[@mpkt_tx_ok_cnt, $tfifo_ctl_wd0, tfifo_entry, 
//						port_rdy_bit, bit15on, next_thread_ir, SDRAM_VALIDATE_FLAG]
//
#macro tfifo_validate_sop[out_mpkt_tx_ok_cnt, in_xfer_tfifo_ctl_wd0, in_tfifo_entry, in_port_rdy_bit, in_bit15on, in_next_thread_ir, CONST_SDRAM_VALIDATE_FLAG]
.local tvs_temp 
#define SOP_XMIT_PTR_WINDOW 5

	tfifo_status_write[in_tfifo_entry, in_xfer_tfifo_ctl_wd0]

	#if (CONST_SDRAM_VALIDATE_FLAG == 0)
		sig_wait[sdram, OR, 1]					; wait for packet data to be transferred to tfifo
	#endif

	sig_wait[inter_thread, OR, 1]				; wait for our turn to validate

	tfifo_ptr_wait[in_tfifo_entry, SOP_XMIT_PTR_WINDOW]
	port_txrdy_wait[in_port_rdy_bit]			; wait for the xmit_ptr to be within 1 of our tfifo element

	tfifo_validate_write[in_tfifo_entry, in_bit15on]
	alu[--, --, b, in_next_thread_ir]								; next_thread_ir set during thread initialization
	fast_wr[0, inter_thd_sig], indirect_ref							; signal next thread

	// Set number of non-SOP mpackets that can be sent to the tfifo before checking the 
	// xmit_rdy bit again.  
	// The MAC xmit_rdy threshold must be at least 7*64 bytes.  The ref_des sets it to 8*64
	alu[tvs_temp, --, b, 5]
	alu[out_mpkt_tx_ok_cnt, --, b, tvs_temp]
#undef SOP_XMIT_PTR_WINDOW
.endlocal
#endm



#endif ; TFIFO_MACROS