/* devnode.c
*
*
* --------------------------------------------------------------------
*                                                                      
*                  I N T E L   P R O P R I E T A R Y                   
*                                                                      
*     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
*     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
*     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
*     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
*     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
*     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
*     THE PRIOR WRITTEN PERMISSION OF :                                
*                                                                      
*                        INTEL  CORPORATION                            
*                                                                     
*                     2200 MISSION COLLEGE BLVD                        
*                                                                      
*               SANTA  CLARA,  CALIFORNIA  95052-8119                  
*                                                                      
*---------------------------------------------------------------------
* system: SA1200
* subsystem: uHAL 
* author: Henry Qian 4/9/99
* revisions:
*
*/

#include "pci.h"

devNodeType StaticDevNodes[MAX_DEVNODE];
devNodeType *freeNodes = 0;
devNodeType *devNodes = 0;

/*
**  __addToQueue(q, n)
**
**  DESCRIPTION
**      Insert a filled out devNode into a devNode list.
**  
**  RETURNS
**      None.
*/
#define __addToQueue(q, n) {(n)->next = (q); (q) = (n);}

/*
**  void initDevNodes(void)
**
**  DESCRIPTION
**      Initializes the devNode data structures.
**  
**  RETURNS
**      None.
*/
void initDevNodes(void)
{
    int ii;
    devNodeType *newNode;
    
    for (ii = 0; ii < MAX_DEVNODE; ii++) {
        newNode = &StaticDevNodes[ii];
        __addToQueue(freeNodes, newNode);
    }
}

/*
**  devNodeType *SAir_AllocateDevNode(void)
**
**  DESCRIPTION
**      Allocate new devnode structure.
**  
**  RETURNS
**      Pointer to a devNodeType or 0 if it fails.
*/
devNodeType* SAir_AllocateDevNode(void)
{
    devNodeType *newNode;
    
    newNode = freeNodes;
    if (freeNodes != 0) {
        freeNodes = freeNodes->next;
        newNode->next = 0;
    }
    
    return newNode;
}


/*
**  void SAr_BuildDevNodeTree(void)
**
**  DESCRIPTION
**      It will then scan the PCI buses looking for valid devices
**      and insert all these devices into the device node tree.
**  
**  RETURNS
**      None.
*/
void SAr_BuildDevNodeTree(void)
{
    int bus, slot, func = 0;
    unsigned int vendor, handle;
    unsigned int status;
    
    /* Initialize Device Node Tree */
    initDevNodes();
    
    /* add all the fixed devices on the platform into the devNode tree */
    /* SAir_AddFixedDevNodes(); */
    
    if (SAr_PciHost())
    {
        /* Our SA1200 will be the first devnode - bus 0, slot 0 */
        /* now scan all possible buses and slots looking for devices */
		for (bus = MAX_PCI_BUS - 1; bus >= 0; bus--)
        {
		    for (slot = MAX_PCI_SLOT - 1; slot >= 0; slot--)
            {
                /* make handle for device at this slot */
                handle = MAKE_HANDLE(PCI_TYPE, bus, slot, func);
                
                /* read the device configuration space */
                vendor = SAr_GetPciVendorID(handle, &status);
		if(status) continue;
                
                /* check if it responds */
                if (vendor != PCI_NOTFITTED)
                {
                    /*
                     * pci drivers are required to read the
                     * BARs for base addr
                     */ 
                    vendor <<= 16;
                    vendor |= SAr_GetPciDeviceID(handle, &status);
                    SAr_AddDevNode(PCI_TYPE, vendor, handle, 0);
                }
            }
        }
    }
}


/*
**  void SAr_PrintDevNodeTree(void)
**
**  DESCRIPTION
**      Displays the current devNode tree.
**  
**  RETURNS
**      None.
*/
void SAr_PrintDevNodeTree(void)
{
    devNodeType *current = devNodes;
    
    if (current == 0)
        printf("No devNodes defined\n");
    
    while (current != 0)
    {
        printf("devNode:\n");
        printf("  id:     0x%08X\n", current->id);
        printf("  handle: 0x%08X\n", current->handle);
        printf("  intDma: 0x%08X\n", current->intDma);
        current = current->next;
    }
}


/*
**  void SAr_AddDevNode(    unsigned int type,
**                          unsigned int id,
**                          unsigned int handle,
**                          unsigned int intDma)
**
**  DESCRIPTION
**      this routine will insert this device into the devNode tree. 
**      It will keep track of available devNodes
**
**      type   = bus type (0=PCI, 1=xbus, 2=CMA bus)
**      id     = id used to uniquely identify device in devNode tree
**      handle = handle of device to add
**      intDma = interrupt information and DMA channel assigned to 
**               device
**  
**  RETURNS
**      None.
*/
void SAr_AddDevNode(    unsigned int type,
                        unsigned int id,
                        unsigned int handle,
                        unsigned int intDma)
{
    devNodeType *newNode;
    
    /* first allocate a new one */
    newNode = SAir_AllocateDevNode();
    if (newNode != 0)
    {
        newNode->id = id;
        newNode->handle = handle | type;
        newNode->intDma = intDma;
        __addToQueue(devNodes, newNode);
    }
}       



/*
**  unsigned int SAr_GetDevNodeHandle(  unsigned int id,
**                                      unsigned int index)
**
**  DESCRIPTION
**      Search for a valid device with a matching
**      id in the devNode tree.
**
**      id     = id used to uniquely identify device in devNode tree
**      index  = iteration of device in the devNode tree. This is used
**               if there is more than one instance of a device. 
**  
**  RETURNS
**      handle for the device if found, returns 0 if failed.
*/
unsigned int SAr_GetDevNodeHandle(  unsigned int id,
                                    unsigned int index)
{
    devNodeType *node;
    unsigned int      numDev = 0;
    
    for (node = devNodes; node != 0; node = node->next)
    {
        if (node->id == id)
        {
            if (numDev++ == index)
                return node->handle;
        }
    }
    
    /*  failure to find device in devNode tree */
    return 0;
}       


/*
**  devNodeType *SAr_FindDevNode(unsigned int handle)
**
**  DESCRIPTION
**      this routine will search for a valid device with a matching
**      handle in the devNode tree.
**
**      handle = handle used to uniquely identify device in tree
**  
**  RETURNS
**      handle for the device if found, returns 0 if failed.
*/
devNodeType *SAr_FindDevNode(unsigned int handle)
{
    devNodeType *node;
    
    /* Don't waste time on bad handles */
    if (handle == 0)
        return 0;
    
    for (node = devNodes; node != 0; node = node->next)
    {
        if (node->handle == handle)
            return node;
    }
    
    /*  failure to find device in devNode tree */
    return 0;
}


/*
**  char *getDeviceClass(   unsigned int id,
**                          unsigned int device,
**                          unsigned int devClass)
**
**  DESCRIPTION
**      This routine will take the parameters it needs to determine
**      the actual class so that it can pass back a string decribing
**      the device. Some devices do not yet support the class usage,
**      so we also pass in the id and device so that they aren't
**      described as just 'PCI Device'. Soon this will not be needed.
**
**      id       = vendor ID code
**      device   = device ID code
**      devClass = device Class code
**  
**  RETURNS
**      char *descriptionString
*/
char *getDeviceClass(   unsigned int id,
                        unsigned int device,
                        unsigned int devClass)
{
    unsigned int baseClass = devClass >> 16 ;
    unsigned int subClass = ((devClass & 0xff00) >> 8) ;
    
    switch(baseClass)
    {
        case PCI_HISTORY_CLASS:
            /* Early revisions of these devices have no class value.
             * Future versions hopefully will.
             */
            if ((id == 0x8086) && (device == 0x484))
                return (char *) "ISA Bridge";
            if ((id == 0x8086) && (device == 0x483))
                return (char *) "Host Bridge";
            if ((id == 0x8086) && (device == 0x482))
                return (char *) "EISA Bridge";
            if ((id == 0x8086) && (device == 0x453))
                return (char *) "PCI Test Card";
            if ((id == 0x1000) && (device == 0x1))
                return (char *) "SCSI Bus";

            switch (subClass)
            {
                case 0:     return (char *) "PCI Device";
                case 1:     return (char *) "VGA Device";
            }

        case PCI_STORAGE_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "SCSI Bus";
                case 1:     return (char *) "IDE";
                case 2:     return (char *) "Floppy Disk";
                case 3:     return (char *) "IPI Disk";
                case 4:     return (char *) "RAID Controller";
                default:    return (char *) "Mass Storage";
            }

        case PCI_NETWORK_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Ethernet";
                case 1:     return (char *) "Token Ring";
                case 2:     return (char *) "FDDI";
                case 3:     return (char *) "ATM";
                default:    return (char *) "Network";
            }

        case PCI_DISPLAY_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "VGA";
                case 1:     return (char *) "XGA";
                default:    return (char *) "Display";
            }

        case PCI_MMEDIA_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Multimedia Video";
                case 1:     return (char *) "Multimedia Audio";
                default:    return (char *) "Multimedia";
            }

        case PCI_MEMORY_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "RAM";
                case 1:     return (char *) "FLASH";
                default:    return (char *) "Memory";
            }

        case PCI_BRIDGE_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Host Bridge";
                case 1:     return (char *) "ISA Bridge";
                case 2:     return (char *) "EISA Bridge";
                case 3:     return (char *) "MC Bridge";
                case 4:     return (char *) "PCI->PCI Bridge";
                case 5:     return (char *) "PCMCIA Bridge";
                case 6:     return (char *) "NuBus Bridge";
                case 7:     return (char *) "CardBus Bridge";
                default:    return (char *) "Bridge Device";
            }

        case PCI_COMMS_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Serial port";
                case 1:     return (char *) "Parallel port";
                default:    return (char *) "Comm Device";
            }

        case PCI_PERIPH_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "PIC";
                case 1:     return (char *) "DMA Controller";
                case 2:     return (char *) "System Timer";
                case 3:     return (char *) "RTC";
                default:    return (char *) "Peripheral";
            }

        case PCI_INPUT_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Keyboard";
                case 1:     return (char *) "Digitizer (pen)";
                case 2:     return (char *) "Mouse";
                default:    return (char *) "Input Controller";
            }

        case PCI_DOCKING_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "Docking Station";
                default:    return (char *) "Docking Station";
            }

        case PCI_PROCESS_CLASS:
            if ((id == 0x1011) && (device == 0x1065))
                return (char *) "DIGITAL DC21285";

            switch (subClass)
            {
                case 0:     return (char *) "386 Processor";
                case 1:     return (char *) "486 Processor";
                case 2:     return (char *) "Pentium";
                case 0x10:  return (char *) "Alpha Processor";
                case 0x20:  return (char *) "Power PC";
                case 0x40:  return (char *) "Co-processor";
                default:    return (char *) "Unknown Device";
            }

        case PCI_SERIAL_CLASS:
            switch (subClass)
            {
                case 0:     return (char *) "FireWire";
                case 1:     return (char *) "ACCESS.bus";
                case 2:     return (char *) "SSA";
                case 3:     return (char *) "USB";
                case 4:     return (char *) "Fibre Channel";
                default:    return (char *) "Unknown Device";
            }

        case PCI_MISC_CLASS:
            if ((id == 0x103C) && (device == 0x2925))
                return (char *) "HP PCI Exerciser";
            return (char *) "PCI Device";

        default:
            if ((id == 0x8086) && (device == 0x4248))
                return (char *) "PCI Dlock Card";
            return (char *) "Unknown Device";
    }
}


/*
**  void mapResources(unsigned int handle, unsigned int *sizePtr)
**
**  DESCRIPTION
**      This routine will attempt to gleen resource assignments
**      from the device pointed to by handle.
**
**      handle = handle of device that we are assigning resources to
**      *sizePtr = pointer to an array which can hold the BAR sizes.
**  
**  RETURNS
**      None.
*/
void mapResources(unsigned int handle, unsigned int *sizePtr)
{
    unsigned int data, offset = PCI_MEM_BAR ;
    int i;
	unsigned int status ;
    unsigned int addr ;
    unsigned int bus, slot, func ;
    unsigned int vendor, device, rev, devClass, cmd ;
    
    printf("\n");
    printf("Bus Slot Func Vendor Device Rev      Class         Cmd\n") ;
    printf("=== ==== ==== ====== ====== === ================   ===\n") ;
    bus = (unsigned int)GET_BUS(handle) ;
    slot = (unsigned int)GET_DEVICE(handle) ;
    func = (unsigned int)GET_FUNCTION(handle) ;
    printf(" %02d   %02d   %02d  ", bus, slot, func) ;
    
    vendor = SAr_GetPciVendorID(handle, &status);
    device = SAr_GetPciDeviceID(handle, &status);
    rev = SAr_GetPciRevID(handle, &status);
    devClass = SAr_GetPciClassID(handle, &status);
    cmd = SAr_GetPciCommand(handle, &status);

    printf("%04xh  %04xh %02xh ", vendor, device, rev) ;
    printf("%-16s %4xh\n\n", getDeviceClass(vendor, device, devClass), cmd) ;
    printf("     Reg     Size      Address     Type\n") ;
    printf("     ===  ==========  =========   ======\n") ;
    
    /* Scan each of the BARS for this device */
    for (i = 0; i < PCI_MAX_BAR; i++)
    {
        if (*(sizePtr + i) != 0)
        {
            printf("     %02xh  %10u  ", offset, *(sizePtr + i));

            /* Read back address assigned for this BAR */
            data = SAr_GetPciBAR(handle, i, &status) ;

            /* Mask off BAR bits */
            addr = data & (~0x0F) ;

            if (data & PCI_IO_ENABLE)
                printf("%8xh   IO\n", addr);        /* IO space BAR */
            else
                printf("%8xh   Memory\n", addr);    /* memory space BAR */
        }

        /* increment the BAR offset value */
        offset += 4 ;
    }
}


/* end of file */
