/* pciMsgQ.cpp
 *
 *------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------
 * system: SA1200
 * subsystem: PCI Device Communication
 * author: Henry Qian 07/30/98
 * revisions:
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <semLib.h>
#include <cacheLib.h>
#include <taskLib.h>
#include "pciComm.h"
#include "assert.h"

/*
**  int Pci_MsgQInit();
**
**  DESCRIPTION
**      Allocates non-cacheable memory for messages queues.
**      These queues are in-bound messages queues, which
**      are accessed by other devices on the PCI bus.
**      After the queues are created, a queue control block
**      structure is generated, which will be an entry in
**      the System Config Table.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_MsgQInit(   unsigned int Pci_NumOfDevices,
                    QueueControlBlock* queueCB)
{
    unsigned int ii;

    // Clean QueueCB, get ready for checksum
    //
    for (ii = 0; ii < (sizeof(QueueControlBlock) * PCI_MAX_DEVICES); ii++)
        *((unsigned char*)queueCB + ii) = 0;

    // Create Incoming Message Queues
    // Message queues will be accessed by PCI devices,
    // so it should not be cached.
    //
    Pci_InMsgQ = (InMsgQ*)cacheDmaMalloc(Pci_NumOfDevices * sizeof(InMsgQ));
    Assert(Pci_InMsgQ != 0);

    // initialize incoming message queues
    //
    unsigned int devIndex;
    for (devIndex = 0; devIndex < Pci_NumOfDevices; devIndex++)
    {
        Pci_InMsgQ[devIndex].head = 0;
        Pci_InMsgQ[devIndex].tail = 0;
        Pci_InMsgQ[devIndex].statusZero = 0;
        Pci_InMsgQ[devIndex].statusOne = 0;

        queueCB[devIndex].base = Pci_InMsgQ[devIndex].queue;
        queueCB[devIndex].headPtr = &(Pci_InMsgQ[devIndex].head);
        queueCB[devIndex].tailPtr = &(Pci_InMsgQ[devIndex].tail);
        queueCB[devIndex].statusZeroPtr = &(Pci_InMsgQ[devIndex].statusZero);
        queueCB[devIndex].statusOnePtr = &(Pci_InMsgQ[devIndex].statusOne);
    }

    return(PCI_SUCCESS);
}

/*
**  int Pci_CheckSum();
**
**  DESCRIPTION
**      Generates a simple checksum.  It starts at address
**      'data' and 'dataLength' bytes.
**
**  RETURNS
**      checksum.
*/
unsigned int Pci_CheckSum(char* data, unsigned int dataLength)
{
    unsigned int checksum = 0;
    unsigned int ii;
    for (ii = 0; ii < dataLength; ii++)
        checksum += *(unsigned char*)(data + ii);

    return(checksum);

}

/*
**  int Pci_GetQueueCB();
**
**  DESCRIPTION
**      Queue control block has all the info about the queue.
**      All accesses to the queue are thru its queueCB, which
**      includes accessing queues over the PCI bus.
**      Queue Control Block is identified by its device number
**      and queue number.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_GetQueueCB( unsigned int deviceNo,
                    unsigned int queueNo,
                    QueueControlBlock* qcb)
{
    Assert((deviceNo < Pci_NumOfDevices) && (queueNo < Pci_NumOfDevices));
    Assert(Pci_SystemConfigTableDone);

    unsigned int memoryBase = Pci_SystemConfigTable[deviceNo].memoryBase;
    unsigned int memorySize = Pci_SystemConfigTable[deviceNo].memorySize;

    *qcb = Pci_SystemConfigTable[deviceNo].msgQCB[queueNo];

    Assert((unsigned int)(qcb->base + PCI_MAX_Q_MSG) < memorySize);
    Assert((unsigned int)(qcb->headPtr) < memorySize);
    Assert((unsigned int)(qcb->tailPtr) < memorySize);
    Assert((unsigned int)(qcb->statusZeroPtr) < memorySize);
    Assert((unsigned int)(qcb->statusOnePtr) < memorySize);

    // If it is not a local device.
    //
    if (deviceNo != Pci_DeviceID)
    {
        qcb->base = (ShrimpMessage*)
            ((unsigned int)(qcb->base) + memoryBase + PCIMemBase);
        qcb->headPtr = (unsigned int*)
            ((unsigned int)(qcb->headPtr) + memoryBase + PCIMemBase);
        qcb->tailPtr = (unsigned int*)
            ((unsigned int)(qcb->tailPtr) + memoryBase + PCIMemBase);
        qcb->statusZeroPtr = (unsigned int*)
            ((unsigned int)(qcb->statusZeroPtr) + memoryBase + PCIMemBase);
        qcb->statusOnePtr = (unsigned int*)
            ((unsigned int)(qcb->statusOnePtr) + memoryBase + PCIMemBase);
    }

    return(PCI_SUCCESS);
}

/*
**  int Pci_QueueShow();
**
**  DESCRIPTION
**      Display a particalar queue info specified by
**      device number and queue number.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_QueueShow(  unsigned int deviceNo,
                    unsigned int queueNo)
{
    QueueControlBlock qcb;
    int retval = Pci_GetQueueCB(deviceNo, queueNo, &qcb);
    Assert(retval == PCI_SUCCESS);

    printf(" ==========================================\n");
    printf(" -- Device %d Queue %d\n", deviceNo, queueNo);
    printf(" ------------------------------------------\n");
    printf(" -- Base: %08X pStat0: %08X pStat1: %08X\n",
            qcb.base,
            qcb.statusZeroPtr,
            qcb.statusOnePtr);
    printf(" -- pHead: %08X pTail: %08X\n",
            qcb.headPtr,
            qcb.tailPtr);
    printf(" -- Head: %04X  Tail: %04X  Stat0: %08X Stat1: %08X\n",
            *(qcb.headPtr),
            *(qcb.tailPtr),
            *(qcb.statusZeroPtr),
            *(qcb.statusOnePtr));
    printf(" ------------------------------------------\n");

    unsigned int head = *(qcb.headPtr);
    unsigned int tail = *(qcb.tailPtr);

    Assert(head < PCI_MAX_Q_MSG);
    Assert(tail < PCI_MAX_Q_MSG);

    ShrimpMessage* msg;

    if (head == tail)
        printf(" -- empty\n");
    else
    {
        while(head != tail)
        {
            msg = &(qcb.base[head]);
            printf(" -- Pos %04X 0x%08X 0x%08X 0x%08X 0x%08X\n",
                head,
                msg->word_0, msg->word_1, msg->word_2, msg->word_3);

            head = (head + 1) % PCI_MAX_Q_MSG;
        }
    }

    printf(" ------------------------------------------\n");
    return(PCI_SUCCESS);
}

/*
**  int Pci_AllQueueShow();
**
**  DESCRIPTION
**      Display all the queue in the PCI system.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_AllQueueShow(void)
{
    int retval;
    unsigned int devIndex;
    unsigned int queueIndex;

    for (devIndex = 0; devIndex < Pci_NumOfDevices; devIndex++)
    {
        for (queueIndex = 0; queueIndex < Pci_NumOfDevices; queueIndex++)
        {
            retval = Pci_QueueShow(devIndex, queueIndex);
            Assert(retval == PCI_SUCCESS);
        }
    }

    return(PCI_SUCCESS);
}


/*
**  int Pci_QPutMsg();
**
**  DESCRIPTION
**      Send a message to a queue specified by device
**      number and queue number.  Signal/interrupt is not
**      used.  This function is used more often for test purpose.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_QPutMsg(unsigned int deviceNo,
                unsigned int queueNo,
                ShrimpMessage* message)
{
    int retval;

    QueueControlBlock qcb;
    retval = Pci_GetQueueCB(deviceNo, queueNo, &qcb);
    Assert(retval == PCI_SUCCESS);

    Assert(*(qcb.headPtr) < PCI_MAX_Q_MSG);
    Assert(*(qcb.tailPtr) < PCI_MAX_Q_MSG);

    // if the queue is full
    //
    if ((retval = Pci_QueueFull(&qcb)) != PCI_MSGQ_FULL)
    {
        Pci_ToQueue(&qcb, message);
        retval = PCI_SUCCESS;
    }

    return(retval);
}

/*
**  int Pci_QGetMsg();
**
**  DESCRIPTION
**      Receive a message from a queue specified by device
**      number and queue number.  Signal/interrupt is not
**      used.  This function is used more often for test purpose.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_QGetMsg(unsigned int deviceNo,
                unsigned int queueNo,
                ShrimpMessage* message)
{
    int retval;

    QueueControlBlock qcb;
    retval = Pci_GetQueueCB(deviceNo, queueNo, &qcb);
    Assert(retval == PCI_SUCCESS);

    Assert(*(qcb.headPtr) < PCI_MAX_Q_MSG);
    Assert(*(qcb.tailPtr) < PCI_MAX_Q_MSG);

    // if the queue is empty
    //
    if ((retval = Pci_QueueEmpty(&qcb)) != PCI_MSGQ_EMPTY)
    {
        Pci_FromQueue(&qcb, message);
        retval = PCI_SUCCESS;
    }

    return(retval);
}

/*
**  int Pci_QueueEmpty();
**
**  DESCRIPTION
**      Test if the queue has no message.
**
**  RETURNS
**      PCI_MSGQ_EMPTY
**      PCI_MSGQ_NOT_EMPTY
*/
int Pci_QueueEmpty(QueueControlBlock* qcb)
{
    int retval;

    if (*(qcb->headPtr) == *(qcb->tailPtr))
        retval = PCI_MSGQ_EMPTY;
    else
        retval = PCI_MSGQ_NOT_EMPTY;

    return(retval);
}


/*
**  int Pci_AllQueueEmpty();
**
**  DESCRIPTION
**      Assert all the queues in PCI system are at
**      reset stage.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_AllQueueEmpty(void)
{
    QueueControlBlock qcb;
    int retval;
    unsigned int devIndex;
    unsigned int queueIndex;

    for (devIndex = 0; devIndex < Pci_NumOfDevices; devIndex++)
    {
        for (queueIndex = 0; queueIndex < Pci_NumOfDevices; queueIndex++)
        {
            retval = Pci_GetQueueCB(devIndex, queueIndex, &qcb);
            Assert(retval == PCI_SUCCESS);
            Assert(Pci_QueueEmpty(&qcb) == PCI_MSGQ_EMPTY);
            Assert(Pci_GetQueueStatusZero(&qcb) == 0);
            Assert(Pci_GetQueueStatusOne(&qcb) == 0);
        }
    }

    return(PCI_SUCCESS);
}


/*
**  int Pci_QueueFull();
**
**  DESCRIPTION
**      Test if the queue has no room for new message.
**
**  RETURNS
**      PCI_MSGQ_FULL
**      PCI_MSGQ_NOT_FULL
*/
int Pci_QueueFull(QueueControlBlock* qcb)
{
    int retval;

    if ((*(qcb->tailPtr) + 1) % PCI_MAX_Q_MSG == *(qcb->headPtr))
        retval = PCI_MSGQ_FULL;
    else
        retval = PCI_MSGQ_NOT_FULL;

    return(retval);
}

/*
**  int Pci_QueueReset();
**
**  DESCRIPTION
**      Reset a queue specified by device
**      number and queue number.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_QueueReset( unsigned int deviceNo,
                    unsigned int queueNo)
{
    QueueControlBlock qcb;
    int retval = Pci_GetQueueCB(deviceNo, queueNo, &qcb);
    Assert(retval == PCI_SUCCESS);

    *(qcb.headPtr) = 0;
    *(qcb.tailPtr) = 0;
    *(qcb.statusZeroPtr) = 0;
    *(qcb.statusOnePtr) = 0;

    return(PCI_SUCCESS);
}

/*
**  int Pci_AllQueueReset();
**
**  DESCRIPTION
**      Reset all queues in system.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_AllQueueReset(void)
{
    unsigned int devIndex;
    unsigned int queueIndex;

    for (devIndex = 0; devIndex < Pci_NumOfDevices; devIndex++)
    {
        for (queueIndex = 0; queueIndex < Pci_NumOfDevices; queueIndex++)
        {
            Pci_QueueReset(devIndex, queueIndex);
        }
    }

    return(PCI_SUCCESS);
}

/*
**  unsigned int Pci_GetQueueStatusZero()
**  void Pci_SetQueueStatusZero()
**  unsigned int Pci_GetQueueStatusOne()
**  void Pci_SetQueueStatusOne()
**
**  DESCRIPTION
**      Read/write queue status variables.  These status
**      variables are used for signaling on demand.
**
**  RETURNS
**      Status bytes/none.
*/
unsigned int Pci_GetQueueStatusZero(QueueControlBlock* qcb)
{
    return(*(unsigned char*)(qcb->statusZeroPtr));
}

void Pci_SetQueueStatusZero(    QueueControlBlock* qcb,
                                unsigned int flag)
{
    *(unsigned char*)(qcb->statusZeroPtr) = flag;
}

unsigned int Pci_GetQueueStatusOne(QueueControlBlock* qcb)
{
    return(*(unsigned char*)(qcb->statusOnePtr));
}

void Pci_SetQueueStatusOne( QueueControlBlock* qcb,
                            unsigned int flag)
{
    *(unsigned char*)(qcb->statusOnePtr) = flag;
}

/*
**  int Pci_ToQueue()
**  void Pci_FromQueue()
**
**  DESCRIPTION
**      Put/Get message to the queue specified by
**      queue control block (qcb).
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
void Pci_ToQueue(   QueueControlBlock* qcb,
                    ShrimpMessage* message)
{
    qcb->base[*(qcb->tailPtr)] = *message;
    *(qcb->tailPtr) = (*(qcb->tailPtr) + 1) % PCI_MAX_Q_MSG;
}

void Pci_FromQueue( QueueControlBlock* qcb,
                    ShrimpMessage* message)
{
    *message = qcb->base[*(qcb->headPtr)];
    *(qcb->headPtr) = (*(qcb->headPtr) + 1) % PCI_MAX_Q_MSG;
}

/*
**  int Pci_PutMsg();
**
**  DESCRIPTION
**      Send message to a device specified by device number
**      'deviceNo'.  The message is put into the in-bound queue
**      on that device.  If the queue is full and the call mode
**      is block, then register with the queue status byte and
**      wait on semaphore.  When the queue becomes availabe,
**      we will be informed and the semaphore will be released.
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
int Pci_PutMsg( unsigned int deviceNo,
                ShrimpMessage* message,
                unsigned int callMode)
{
    int retval;

    Assert(Pci_InitDone == PCI_TRUE);

    // do not sent message to myself
    Assert(deviceNo != Pci_DeviceID);

    Assert((callMode == PCI_CALL_BLOCK) || (callMode == PCI_CALL_UNBLOCK));

    // Get queue control block of device 'Pci_DeviceID' 's
    // queue number 'deviceNo'.
    QueueControlBlock qcb;
    retval = Pci_GetQueueCB(deviceNo, Pci_DeviceID, &qcb);
    Assert(retval == PCI_SUCCESS);

    // Only one thread at a time is allowed to access a particular
    // device.
    if (callMode == PCI_CALL_BLOCK){
        retval = semTake(putAccessSem[deviceNo], WAIT_FOREVER);
        Assert(retval == OK);
    }
    else {
        retval = semTake(putAccessSem[deviceNo], NO_WAIT);
        if (retval != OK) return(PCI_MSGQ_BUSY);
    }

    Assert(*(qcb.headPtr) < PCI_MAX_Q_MSG);
    Assert(*(qcb.tailPtr) < PCI_MAX_Q_MSG);

    // if the queue is full
    //
    if (Pci_QueueFull(&qcb) == PCI_MSGQ_FULL){
        if (callMode == PCI_CALL_UNBLOCK){
            retval = semGive(putAccessSem[deviceNo]);
            Assert(retval == OK);
            return(PCI_MSGQ_FULL);
        }

        // Let other device know that queue is full and we are
        // waiting for sending. When the other device takes
        // message out of queue and queue has room for new
        // message, it will check this status variable and
        // signal us on demand.
        Pci_SetQueueStatusZero(&qcb, PCI_WAIT_TO_SEND);

        // Reset semaphore and get ready to wait on it.
        semTake(outQueueFullSem[deviceNo], NO_WAIT);

        // recheck to eliminate the vulnerable window
        while (Pci_QueueFull(&qcb) == PCI_MSGQ_FULL){
            // Once we get signal that queue is available, ISR
            // will release this semaphore.
            semTake(outQueueFullSem[deviceNo], WAIT_FOREVER);
            Assert(Pci_QueueFull(&qcb) != PCI_MSGQ_FULL);
        }
    }

    // Add message to queue
    Pci_ToQueue(&qcb, message);

    // Always signal for incoming message.
    retval = Pci_Signal(deviceNo, PCI_INCOMING_MSG_SIGNAL);
    Assert(retval == PCI_SUCCESS);

    semGive(putAccessSem[deviceNo]);
    return(PCI_SUCCESS);
}


/*
**  int Pci_GetMsg();
**
**  DESCRIPTION
**      Get message from the in-bound queues at local device.
**      Each function call will get exactly one incoming message.
**      Successive calls will get message from multiple queues in
**      round robin turn.  If all the queues are empty and the call mode
**      is PCI_CALL_BLOCK, then register with the queue status byte and
**      wait on a semaphore.  When the message comes
**      we will be informed and the semaphore will be released.
**
**      Psudo code:
**          while (1)
**          {
**              RESET semaphore;
**
**              for (deviceQueue = 0; deviceQueue < numOfDevice; deviceQueue++)
**              {
**                  if (deviceQueue is not empty)
**                      get message and leave;
**                  else
**                      continue;
**              }
**
**              WAIT on semaphore;
**          }
**
**  RETURNS
**      PCI_SUCCESS		the function succeeded.
**      PCI_ERROR		the function failed.
*/
unsigned int devRdRobinNum = 0;

int Pci_GetMsg( ShrimpMessage* message,
                unsigned int callMode)
{
    int retval;

    Assert(Pci_InitDone == PCI_TRUE);

    // Sanity check
    Assert((callMode == PCI_CALL_BLOCK) || (callMode == PCI_CALL_UNBLOCK));

    // Only one thread at a time is allowed to use this function.
    if (callMode == PCI_CALL_BLOCK){
        retval = semTake(getAccessSem, WAIT_FOREVER);
        Assert(retval == OK);
    }
    else{
        retval = semTake(getAccessSem, NO_WAIT);
        if (retval != OK)
            return(PCI_MSGQ_BUSY);
    }

    // Get queue control block of device 'Pci_DeviceID' 's
    // queue number 'deviceNo'.
    QueueControlBlock qcb;

    while (1){
        if (callMode == PCI_CALL_BLOCK){
            // Reset semaphore and get ready to wait on it.
            // During following check of each queue for message
            // new message might come and this semaphore will
            // reflect it.
            semTake(inQueueEmptySem, NO_WAIT);
        }

        for (int devIndex = 0; devIndex < Pci_NumOfDevices; devIndex++){
            // Generate round robin queueNo to get message from.
            devRdRobinNum = (devRdRobinNum + 1) % Pci_NumOfDevices;

            // Listen to all other devices except our own.
            if (Pci_DeviceID == devRdRobinNum)
                continue;

            retval = Pci_GetQueueCB(Pci_DeviceID, devRdRobinNum, &qcb);
            Assert(retval == PCI_SUCCESS);

            // Sanity check if queue is good.
            Assert(*(qcb.headPtr) < PCI_MAX_Q_MSG);
            Assert(*(qcb.tailPtr) < PCI_MAX_Q_MSG);

            if (Pci_QueueEmpty(&qcb) != PCI_MSGQ_EMPTY){
                // Get message
                Pci_FromQueue(&qcb, message);

                // if other people are waiting for queue is not full
                if (Pci_GetQueueStatusZero(&qcb) == PCI_WAIT_TO_SEND){
                    // reset status
                    Pci_SetQueueStatusZero(&qcb, 0);
                    retval = Pci_Signal(devRdRobinNum, PCI_MSGQ_AVAILABLE_SIGNAL);
                    Assert(retval == PCI_SUCCESS);
                }

                semGive(getAccessSem);
                return(PCI_SUCCESS);
            }
        }

        // At this point, all the queues are empty
        //
        if (callMode == PCI_CALL_UNBLOCK){
            retval = semGive(getAccessSem);
            Assert(retval == OK);
            return(PCI_MSGQ_EMPTY);
        }
        else{
            // Once we get signal for incoming message, ISR
            // will release this semaphore.
            retval = semTake(inQueueEmptySem, WAIT_FOREVER);
            Assert(retval == OK);
        }
    }
}

// end of file
