#ifndef	__IX_ASL_INTERNAL_HASH_H__
#define	__IX_ASL_INTERNAL_HASH_H__

/*
**			INTEL CORPORATION PROPRIETARY INFORMATION
**	This software is supplied under the terms of a license agreement or
**	nondisclosure agreement with Intel Corporation and may not be copied
**	or disclosed except in accordance with the terms of that agreement.
**		Copyright (C) 2000 Intel Corporation. All rights Reserved.
*/

#include <sys/types.h>

#include "ix_types.h"
/*#include "ix_error.h"*/

#define	IX_ERROR_CHECK_ARG_NULL(__v)	\
	if ((__v) == 0) \
		return 0

#ifdef	__cplusplus
extern "C" {
#endif

/** @file ix/asl/internal/internal-hash.h
    @brief INTERNAL Calculation of HASH values.
*/

/** Bit Mask Construction.
 */
#define	IX_HASH48_TAP(n)	(((ix_48bit)1)<<(n))

/** Feedback Tap Vector for 48-bit hash.
    This bit vector represents the tap locations where the Q
    bits are pulled in to XOR with the shifted R bit vector.
    These tap points are based on the transactor simulation
    of the hash unit in the IXP-1200, with the intent that
    this code can be used to maintain tables that are searched
    with the assistance of the IXP-1200 hash unit.
*/
#define	IX_HASH48_TAPS		(IX_HASH48_TAP(0)|IX_HASH48_TAP(10)|IX_HASH48_TAP(25)|IX_HASH48_TAP(36))

/** 48-bit Hash State Storage.
 */
struct ix_s_hash48 {

/** Q portion of multiplier output.
    This array contains the various values to XOR with
    the Q bit vector. Each entry corresponds to a different
    value of the B input; the entry contents are calculated
    from the single common hash multiplier.
*/
    unsigned char           qm[256];

/** R portion of multiplier output.
    This array contains the various values to XOR with
    the shifted R bit vector. Each entry corresponds to a different
    value of the B input; the entry contents are calculated
    from the single common hash multiplier.
*/
    ix_48bit                rm[256];

/** Feedback Tap Masks.
    This array contains the XOR contribution to XOR with
    the shifted R bit vector. Each entry corresponds to a different
    value of the Q vector; the entry contents are calculated
    from the single common tap list.
*/
    ix_48bit                rt[256];
};

/** Predigest 48-bit Hash Polynomial.
    This function predigests the coefficients of the 48-bit
    hash polynomial into a table that will be indexed by
    the eight Q bits shifted eoff the top of the shift register
    during each byte input cycle.

    Should the list of tap locations be an input parameter?
*/
int                ix_hash48_init(ix_hash48 *h);

/** Predigest 48-bit Hash Multiplier.
    This function predigests the hash multiplier for the 48-bit hash
    function into two tables that will be indexed by the eight input
    bits shifted into the unit each input cycle. The smaller table
    gives the effect of the input bits and multipier on the Q bits and
    the larger table gives the effect on the shifted register.
 */
int                ix_hash48_set_mult(ix_hash48 *h, ix_48bit m);

/** Step the 48-bit Hash unit across an input byte.
    This function executes eight bits worth of input state updates for
    the hash unit, using previously calculated tables to find the
    cumulative state change based on all eight bits at once, returning
    the final state of the hash unit at the end.
*/
ix_48bit                ix_hash48_inline_step(ix_hash48 *h, ix_48bit r,
					      int b);

/** Step the 48-bit Hash unit across an input byte.
    This function executes eight bits worth of input state updates for
    the hash unit, using previously calculated tables to find the
    cumulative state change based on all eight bits at once, returning
    the final state of the hash unit at the end.
*/
int                ix_hash48_run_byte(ix_hash48 *h, ix_48bit *p,
					   int b);

/** Step the 48-bit Hash unit across a ix_48bit input.
    This function executes eight bytes worth of input state updates
    for the hash unit, returning the final state of the hash unit at
    the end.
*/
int                ix_hash48_run_wide(ix_hash48 *h, ix_48bit *p,
					   ix_48bit b);

/** Destroy 48-bit Hash Context.
 */
int                ix_hash48_fini(ix_hash48 *h);

/** Allocate and initialize a 48-bit Hash Engine.
    This function allocates the local storage used to contain the
    predigested images of the hash tap configuration and a specific
    hash multiplier, and initializes it.
*/
int                ix_hash48_new(ix_hash48 **hp);

/** Free a 48-bit Hash Engine.
    This function frees all storage associated with the specified
    48-bit Hash Engine, including the outermost ix_hash48 structure.
 */
int                ix_hash48_del(ix_hash48 *h);

/** Bit Mask Construction.
 */
#define	IX_HASH64_TAP(n)	(((ix_64bit)1)<<(n))

/** Feedback Tap Vector for 64-bit hash.
    This bit vector represents the tap locations where the Q
    bits are pulled in to XOR with the shifted R bit vector.
    These tap points are based on the transactor simulation
    of the hash unit in the IXP-1200, with the intent that
    this code can be used to maintain tables that are searched
    with the assistance of the IXP-1200 hash unit.
*/
#define	IX_HASH64_TAPS		(IX_HASH64_TAP(0)|IX_HASH64_TAP(17)|IX_HASH64_TAP(35)|IX_HASH64_TAP(54))

/** 64-bit Hash State Storage.
 */
struct ix_s_hash64 {

/** Q portion of multiplier output.
    This array contains the various values to XOR with
    the Q bit vector. Each entry corresponds to a different
    value of the B input; the entry contents are calculated
    from the single common hash multiplier.
*/
    unsigned char           qm[256];

/** R portion of multiplier output.
    This array contains the various values to XOR with
    the shifted R bit vector. Each entry corresponds to a different
    value of the B input; the entry contents are calculated
    from the single common hash multiplier.
*/
    ix_64bit                rm[256];

/** Feedback Tap Masks.
    This array contains the XOR contribution to XOR with
    the shifted R bit vector. Each entry corresponds to a different
    value of the Q vector; the entry contents are calculated
    from the single common tap list.
*/
    ix_64bit                rt[256];
};

/** Predigest 64-bit Hash Polynomial.
    This function predigests the coefficients of the 64-bit
    hash polynomial into a table that will be indexed by
    the eight Q bits shifted eoff the top of the shift register
    during each byte input cycle.

    Should the list of tap locations be an input parameter?
*/
int                ix_hash64_init(ix_hash64 *h);

/** Predigest 64-bit Hash Multiplier.
    This function predigests the hash multiplier for the 64-bit hash
    function into two tables that will be indexed by the eight input
    bits shifted into the unit each input cycle. The smaller table
    gives the effect of the input bits and multipier on the Q bits and
    the larger table gives the effect on the shifted register.
 */
int                ix_hash64_set_mult(ix_hash64 *h, ix_64bit m);

/** Step the 64-bit Hash unit across an input byte.
    This function executes eight bits worth of input state updates for
    the hash unit, using previously calculated tables to find the
    cumulative state change based on all eight bits at once, returning
    the final state of the hash unit at the end.
*/
ix_64bit                ix_hash64_inline_step(ix_hash64 *h, ix_64bit r,
					      int b);

/** Step the 64-bit Hash unit across an input byte.
    This function executes eight bits worth of input state updates for
    the hash unit, using previously calculated tables to find the
    cumulative state change based on all eight bits at once, returning
    the final state of the hash unit at the end.
*/
int                ix_hash64_run_byte(ix_hash64 *h, ix_64bit * p,
					   int b);

/** Step the 64-bit Hash unit across a ix_64bit input.
    This function executes eight bytes worth of input state updates
    for the hash unit, returning the final state of the hash unit at
    the end.
*/
int                ix_hash64_run_wide(ix_hash64 *h, ix_64bit * p,
					   ix_64bit b);

/** Destroy 64-bit Hash Context.
 */
int                ix_hash64_fini(ix_hash64 *h);

/** Allocate and initialize a 64-bit Hash Engine.
    This function allocates the local storage used to contain the
    predigested images of the hash tap configuration and a specific
    hash multiplier, and initializes it.
*/
int                ix_hash64_new(ix_hash64 **hp);

/** Free a 64-bit Hash Engine.
    This function frees all storage associated with the specified
    64-bit Hash Engine, including the outermost ix_hash64 structure.
 */
int                ix_hash64_del(ix_hash64 *h);

/** Reset ix_hash module.
    This function clears all system state retained by the hash
    module when all processes using IX-ASL have been terminated.
    This function should be used to reset the state of this module
    after running test code, or code under development, that may
    have left this module with inconsistant retained state.
 */
void                    ix_hash_reset(void);

#ifdef	__cplusplus
};
#endif

#endif/*__INTERNAL_HASH_H__*/
