/* hashDbmExt.cpp
 * Hash Database Manager
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: IXP1200
 * subsystem: Hash Database Manager
 * author: drj Sept 13, 2000 Initial version
 * revisions:
 * $History: hashDbmExt.cpp $
 * 
 * *****************  Version 1  *****************
 * User: Drjohns1     Date: 10/20/00   Time: 5:03p
 * Created in $/dev_1200/SA1_CoreLibs/hdbm
 * Dev update - storing user data into database table
 * 
 */
#include <string.h>
#ifdef HDBM_DEBUG
#include <assert.h>
#endif
#include <stdio.h>
#include "hdbm.h"

static const unsigned int UserDataTrue = 1, UserDataFalse = 0;


/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::getKeyIndex

   Description: Extract the table index from the database value

  Arguments:
				dbv		-- a value read from the database table

  Returns:		the extracted index

  Uses:
  Modifies:
-----------------------------------------------------------------------------*/
unsigned int HashDbmExt::getKeyIndex(unsigned int dbv)
{
	return((dbv >> (HDBM_INDEX_POS)) & HDBM_INDEX_MASK);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::setKeyIndex

   Description: Insert the table index into the value

  Arguments:
				dbv		-- a value read from the database table
				ei		-- the data index

  Returns:		the value with the inserted index

  Uses:
  Modifies:
-----------------------------------------------------------------------------*/
unsigned int HashDbmExt::setKeyIndex(unsigned int dbv, unsigned int ei)
{
	return((dbv & ~(HDBM_INDEX_MASK << HDBM_INDEX_POS)) | ((ei & HDBM_INDEX_MASK) << HDBM_INDEX_POS));
}



/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::getEntryData

   Description: Extract data from the database value

  Arguments:
				dbv	-- a value read from the database table

  Returns:		the extracted data value

  Uses:
  Modifies:
-----------------------------------------------------------------------------*/
unsigned int HashDbmExt::getEntryData(unsigned int dbv)
{
	return((dbv >> HDBM_DATA_POS) & HDBM_DATA_MASK);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::setEntryData

   Description: Insert the data into the value

  Arguments:
				dbv		-- a value read from the database table
				ed		-- the entry data

  Returns:		the value with the inserted index

  Uses:
  Modifies:
-----------------------------------------------------------------------------*/
unsigned int HashDbmExt::setEntryData(unsigned int dbv, unsigned int ed)
{
	unsigned int tmp;
	unsigned int mask = HDBM_DATA_MASK, pos = HDBM_DATA_POS;

	tmp = 	(dbv & ~(mask << pos)) | ((ed & mask) << pos);
	return(tmp);
//	return((dbv & ~HDBM_DATA_MASK) | ((ed & HDBM_DATA_MASK) << HDBM_DATA_POS));
}




/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::Add

   Description: Add an entry to the database using the specified key and
				associate userData with the entry if the operation was
				successfully.

  Arguments:
				key			-- key to the entry
				userData	-- data to be associated with the entry

  Returns:		
				HDBM_SUCCESS			-- operation successfully completed
				HDBM_LOCK_ERROR			-- entry is locked
				HDBM_EMPTY_COL_LST		-- collision table freelist empty
				HDBM_EMPTY_KEYDATA_LST	-- dataKey freelist empty
				HDBM_DUPL_ENTRY			-- duplicate entry
				HDBM_ERROR				-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDbmExt::Add(HdbmKey_T *key, void* userData)
{
	unsigned int tableValue, colIndex, i, dbv;
    unsigned int collisionLevel=0, tableBase, index;

    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
			case HDBM_LOCK_ERROR :			//  Lock bit is set -- let upper layer recover
				return(HDBM_LOCK_ERROR);
				break;

			case HDBM_NO_ENTRY :						// no entry - add to main index table
				unsigned int keyDataIndex;

				if(!(keyDataIndex = pKeyDataFreeLst->allocEle()))
					return(HDBM_EMPTY_KEYDATA_LST);


				// add entry -- write the key entry
				writeKeyData(UserDataFalse, keyDataIndex, key, NULL);

				// write the forward entry index
				//note:- 1 is added to keyDataIndex, normalized to 1, so that the entry
				//at the base will not have an index of zero. An index of zero indicates
				//no entry found.
				keyDataIndex -= pKeyDataFreeLst->getBase();
				keyDataIndex >>= 1;
				if(keyDataIndex & (~HDBM_INDEX_MASK)) return(HDBM_ERROR);


				dbv = setKeyIndex(0, keyDataIndex+1);
				dbv = setEntryData(dbv, (unsigned int)userData);
				writeFwdIndex(tableBase + index, dbv);	//update the table index
				return(HDBM_SUCCESS);
				break;

			case HDBM_ENTRY_NO_COL:				// entry with no collision
				unsigned int fwdEntryAddr, startBit;
				HdbmKey_T fwdEntryKey;

				// check for duplicate keys
				fwdEntryAddr = getKeyIndex(tableValue);
				readKeyData(UserDataFalse, ((fwdEntryAddr - 1) << 1) + pKeyDataFreeLst->getBase(), &fwdEntryKey, NULL);
				if((fwdEntryKey.B3210 == key->B3210) && (fwdEntryKey.B7654 == key->B7654)){
					return(HDBM_DUPL_ENTRY);
				}

				// alloc and init a new collision table
				unsigned int newTable;

				if(!(newTable = pColTabFreeLst->allocEle())) return(HDBM_EMPTY_COL_LST);

				// init new collision table
				for(i = 0; i < pColTableSize; i++) writeFwdIndex(newTable + i, 0);

				// add old entry to the new table
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)){
					pColTabFreeLst->freeEle(newTable);
					return(HDBM_ERROR);
				}
				colIndex = getIndexField(&fwdEntryKey, startBit, pColShiftBits);
				writeFwdIndex(newTable + colIndex, fwdEntryAddr);	//update the new table index

				// add the reference to collision table
				colIndex = setTableIndex(setCollision(0), newTable);				
				writeFwdIndex(tableBase + index, colIndex);	//update the table index


				// add the new entry to collision table
				tableBase = newTable;
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				collisionLevel++;
				break;

			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				collisionLevel++;

			break;
			default:
				break;
		}
	}
    return(HDBM_ERROR);
}



/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::traverseDb

   Description: Traverse database and call the user specified function for each
				entry. The traversal will be terminated if the user function
				by returns a non-zero value.

  Arguments:
				tableBase	-- the SRAM base of the primary table.
				tableSize	-- the size of the primary table
				funcPtr		-- pointer to a user function that will be called for
							   each entry. The key, dataEntry, and a user specified
							   value will be passed to the function.
				userData	-- user specific value

  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_ERROR			-- invalid database value encountered
				The non-zero value returned from the user function.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDbmExt::traverseDb(unsigned int tableBase, unsigned int tableSize,
					   int (*funcPtr)(HdbmKey_T *key, void *entry, void *userData), void *userData)
{
	unsigned int priIndex, tableValue, fwdEntry;
	int status;

	if(!funcPtr) return(HDBM_ERROR);

	for(priIndex = 0; priIndex < tableSize; priIndex++){
		switch(getState(tableBase, priIndex, &tableValue)){
			case HDBM_ENTRY_NO_COL:				// entry with no collision
				unsigned int dataEntry;
				HdbmKey_T key;

				fwdEntry = pKeyDataFreeLst->getBase() + ((getKeyIndex(tableValue)-1) << 1);
				readKeyData(UserDataFalse, fwdEntry, &key, NULL);
				dataEntry = getEntryData(tableValue);
				if((status = (*funcPtr)(&key, (void *)dataEntry, userData))) return(status);
				break;
	        case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				unsigned int colTableBase;

				colTableBase = getTableIndex(tableValue);
				traverseDb(colTableBase, pColTableSize, funcPtr, userData);
				break;
			default:
				break;
		}
	}
	return(HDBM_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::Lookup

   Description: Search the database using the specified key return the
				associated userData if the operation was successfully.

  Arguments:
				key			-- pointer to the key.
				userData	-- pointer to where the associated data will be stored.
  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_NO_ENTRY		-- entry not found
				HDBM_ERROR			-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDbmExt::Lookup(HdbmKey_T *key, void *userData)
{
	unsigned int tableValue, tableBase;
    unsigned int index=0, colLevel = 0, startBit = 0;
    
    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
	        case HDBM_NO_ENTRY :					// no entry - add to main index table
				return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_NO_COL:					// entry with no collision
				*(unsigned int *)userData = getEntryData(tableValue);
			    return(HDBM_SUCCESS);
				break;
			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * colLevel))) >=
					(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				colLevel++;

			break;
			default:
				break;
		}
	}
    return(HDBM_ERROR);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDbmExt::Remove

   Description: Remove an entry from the database.

  Arguments:
				key			-- pointer to the key.
  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_NO_ENTRY		-- entry not found
				HDBM_ERROR			-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDbmExt::Remove(HdbmKey_T *key)
{
	unsigned int tableValue, tableBase, prevTable, prevIndex;
    unsigned int index=0, colLevel = 0, startBit;
    
    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
	        case HDBM_NO_ENTRY :				// no entry - add to main index table
				return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_NO_COL:				// entry with no collision
				HdbmKey_T keyInDb;
				unsigned int fwdEntry;

				fwdEntry = pKeyDataFreeLst->getBase() + ((getKeyIndex(tableValue)-1) << 1);
				readKeyData(UserDataFalse, fwdEntry, &keyInDb, NULL);
				if((key->B3210 == keyInDb.B3210) && (key->B7654 == keyInDb.B7654)){
					unsigned int i, indx, numOtherEntries = 0, otherEntryValue;

					pKeyDataFreeLst->freeEle(fwdEntry);
					// determine if any other entry is in this table
					if(colLevel){
						for(i = 0; i < pColTableSize && (numOtherEntries < 2); i++){
							if(i == index) continue;
							readFwdIndex(tableBase + i, &indx);
							if(indx){
								numOtherEntries++;
								otherEntryValue = indx;
							}
						}
						if((numOtherEntries == 1) && !(otherEntryValue & COLLISION_MASK)){
							// only one other entry with no collision. So, remove target entry;
							// move remaining entry to a lower collison level; update prev
							// table; and free current table
							writeFwdIndex(prevTable + prevIndex, otherEntryValue);
							pColTabFreeLst->freeEle(tableBase);
						}
						else{
							// there are at least two other entries or a single entry with collision.
							// So, just remove target entry from the table
							writeFwdIndex(tableBase + index, 0);
						}
					}
					else{
						// no collision -- so, just initialize the table
						writeFwdIndex(tableBase + index, 0);
					}
				    return(HDBM_SUCCESS);
				}
			    else return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table
				prevTable = tableBase;
				prevIndex = index;
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * colLevel))) >=
					(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				colLevel++;
				break;
			default:
				break;
		}
	}

    return(HDBM_NO_ENTRY);
}


