/* hdbm.cpp
 * Hash Database Manager
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: IXP1200
 * subsystem: Hash Database Manager
 * author: drj Sept 13, 2000 Initial version
 * revisions:
 * $History: hdbm.cpp $
 * 
 * *****************  Version 8  *****************
 * User: Drjohns1     Date: 12/13/00   Time: 4:13p
 * Updated in $/dev_1200/SA1_CoreLibs/hdbm
 * Fixed sdram overwrite bug
 * 
 * *****************  Version 7  *****************
 * User: Drjohns1     Date: 12/05/00   Time: 10:17a
 * Updated in $/dev_1200/SA1_CoreLibs/hdbm
 * Updated to fix build problem
 * 
 * *****************  Version 6  *****************
 * User: Drjohns1     Date: 10/20/00   Time: 5:01p
 * Updated in $/dev_1200/SA1_CoreLibs/hdbm
 * Dev update - storing user data into database table
 * 
 * *****************  Version 5  *****************
 * User: Drjohns1     Date: 10/12/00   Time: 3:32p
 * Updated in $/dev_1200/SA1_CoreLibs/hdbm
 * Fixed bug in deleEntries
 * 
 *---------------------------------------------------------------------------
 */
#include <string.h>
#ifdef HDBM_DEBUG
#include <assert.h>
#endif
#include <stdio.h>
#include "hal_sdram.h"
#include "hdbm.h"


static const unsigned BitsInWord = 32;
#define HDBM_VALID_SDBASE(a) (!((a) & 0xC0000000))

//dj: static unsigned int const TABLE_INDEX_POS = 0, TABLE_INDEX_MASK = 0x3fffffff;
static const unsigned int UserDataTrue = 1, UserDataFalse = 0;




/*-----------------------------------------------------------------------------
   Function: 	SwQ::push

   Description:	Push an address to on the queue

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SwQ::push(unsigned int eleAddr){
	if(pMemType == FREELST_SRAM){
		SRAM_WRITE(eleAddr, pQhead);
	}
	else{
		SDRAM_WRITE(eleAddr<<1, pQhead);
	}
	pQhead = eleAddr;
	return(0);
}

/*-----------------------------------------------------------------------------
   Function: 	SwQ::pop

   Description:	Push an address from the queue

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int SwQ::pop(void){
	unsigned int addr;

	addr = pQhead;
	if(pQhead != pNull){
		if(pMemType == FREELST_SRAM){
			SRAM_READ(pQhead, &pQhead);
		}
		else{
			SDRAM_READ(pQhead<<1, &pQhead);
		}
	}
	return(addr);
}

#ifdef HDBM_DEBUG
/*-----------------------------------------------------------------------------
   Function: 	SwQ::dumpQue

   Description:	Display the queue entries without removing them.

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void SwQ::dumpQue(void){
	unsigned int addr, head = pQhead, i=0;

	while(head != pNull){
		if(pMemType == FREELST_SRAM){
			SRAM_READ(head, &addr);
		}
		else{
			SDRAM_READ(head<<1, &addr);
		}
		if(addr == head){
			printf("Seems we are in an infinite loop...\n");
			break;
		}
		printf("***dumpQue: q[%d], Addr=0x%x = 0x%x\n", i++, head, addr);
		head = addr;
	}
	return;
}


/*-----------------------------------------------------------------------------
   Function: 	SwQ::chkQue

   Description:	Check the validity of the queue entries without removing them.

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SwQ::chkQue(unsigned int hiAddr, unsigned int loAddr){
	unsigned int addr, head = pQhead, i=0;

	while(head != pNull){
		if((head < loAddr) || (head > hiAddr)){
			printf("***chkQue: q[%d], Invalid address 0x%x\n", i, head);
			return -1;
		}
		if(pMemType == FREELST_SRAM){
			SRAM_READ(head, &addr);
		}
		else{
			SDRAM_READ(head<<1, &addr);
		}
		if(addr == head){
			printf("***chkQue: q[%d], Seems to be in an infinite loop\n", i);
			return -2;
		}
		head = addr;
		i++;
	}
	return 0;
}
#endif


/*-----------------------------------------------------------------------------
   Function: 	FreeLst::FreeLst

   Description:	Construct a free-list of elements in SRAM or SDRAM utilizing
				either a hardware or software queue.

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
FreeLst::FreeLst(unsigned short qId, FreeLstMemTyp_T memType, unsigned int lstBase,
				 unsigned int eleWordSize, unsigned int maxEle)
{
	if(qId <= HDBM_HWQ_7) pQ = new HwQ(qId);
	else pQ = new SwQ(EOQ_MARKER, memType);

	pEleWordSize = 	eleWordSize;
	pLstBase = lstBase;
	pMaxEle = maxEle;

	for(pNumEle =0; pNumEle < pMaxEle; pNumEle++){
		pQ->push(lstBase);
		lstBase += pEleWordSize;
	}
	pLastAddr = lstBase;
	sem = Sem_Init(1);
	return;
}


/*-----------------------------------------------------------------------------
   Function: 	FreeLst::allocEle

   Description:	Allocate/remove an element from the freelist.

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int FreeLst::allocEle(unsigned int *eleAddr)
{
	int stat = HDBM_SUCCESS;
	Sem_Wait(sem, SEM_WAITFOREVER);
	if(pNumEle){
		if((*eleAddr = pQ->pop()) != EOQ_MARKER) pNumEle--;
		else stat = HDBM_ERROR;
	}else stat= HDBM_ERROR;
	Sem_Post(sem);
	return(stat);
}



/*-----------------------------------------------------------------------------
   Function: 	FreeLst::freeEle

   Description:	Reclaim an element that was allocated from the freelist.

   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int FreeLst::freeEle(unsigned int eleAddr)
{
	int stat = HDBM_SUCCESS;

#ifdef HDBM_DEBUG
	// check if addr is within range
	if(((unsigned int)eleAddr < pLstBase) || ((unsigned int)eleAddr > pLastAddr)){
		printf("Error: FreeLst::freeEle; invalid address 0x%x -- out of range\n", eleAddr);
		return(HDBM_BAD_LSTADDR);
	}
	// check if addr is aligned on eleWordSize boundary
	if((eleAddr - pLstBase) % pEleWordSize){
		printf("Error: FreeLst::freeEle; invalid address 0x%x -- unaligned for eleSize %d\n",
			eleAddr, pEleWordSize);
		return(HDBM_BAD_LSTADDR);
	}
#endif
	Sem_Wait(sem, SEM_WAITFOREVER);
	if(pNumEle < pMaxEle){
		pQ->push(eleAddr);
		pNumEle++;
	}else stat = HDBM_BAD_LSTFULL;
	Sem_Post(sem);
	return(stat);
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::~HashDBM

   Description:	Reclaimed resources used by the HashDBM object

   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
HashDBM::~HashDBM()
{
	Ixp_Hash48Del(pHash48);
	Ixp_Hash64Del(pHash64);
	delete pColTabFreeLst;
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::HashDBM

   Description: Creates an instance of the HashDBM object.

   Arguments:
				hashType			-- an enum indicating the type of hash to be performed
				hashMultHi			-- hash function multiplier bits<63:32>
				hashMultLo			-- hash function multiplier bits<31:0>
				priTableDesc		-- primary table descriptor
				HdbmColTableDesc	-- collision table descriptor
				keyDataFreeLst		-- key and data free list
   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
HashDBM::HashDBM(HdbmHashType_T hashType, unsigned int hashMultHi, unsigned int hashMultLo,
				 HdbmPriTableDesc *priTableDesc, HdbmColTableDesc *colTableDesc,
				 FreeLst *keyDataFreeLst)
{
	unsigned int i;

#ifdef HDBM_DEBUG
	assert((priTableDesc->numIndxBits + priTableDesc->indxStartBit) <= (BitsInWord*2));
	assert((colTableDesc->numIndxBits + colTableDesc->indxStartBit) <= (BitsInWord*2));
	assert(colTableDesc->colShiftBits <= 32);
#endif

	pColTableSize = 1 << colTableDesc->colShiftBits;
	pPriTableSize = 1 << priTableDesc->numIndxBits;


	pHashType = hashType;

	// contiguous set of 32-bit words of sram used as hash table
	// priSize = 1 << priIndxBitLen
	pPriTableBase = priTableDesc->baseAddr;
	pPriStartBit = priTableDesc->indxStartBit;
	pPriIndxBitLen = priTableDesc->numIndxBits;


	// Init collision table parameters
	pColStartBit = colTableDesc->indxStartBit;
	pColIndxBitLen = colTableDesc->numIndxBits;
	pColShiftBits = colTableDesc->colShiftBits;
	pColShiftMask =  0xffffffff >> (BitsInWord - colTableDesc->colShiftBits);
	pColTabFreeLst = colTableDesc->freeLst;
	
	sram = SRAM_Attach();

	// clear the tables
	for(i = 0; i < pPriTableSize; i++){
		SRAM_WRITE(priTableDesc->baseAddr + i, 0);

#ifdef CREATE_IND
		logSram(priTableDesc->baseAddr + i, 0);
#endif
	}


	pKeyDataFreeLst = keyDataFreeLst;

	// init hash function
	if(pHashType == HASH48){
		ix_48bit h48Mult;
		h48Mult = hashMultHi;
		h48Mult = h48Mult << 32 | hashMultLo;
		Ixp_Hash48New(&pHash48);
		Ixp_Hash48SetMult(pHash48, h48Mult);
	}
	else{
		ix_64bit h64Mult;
		h64Mult = hashMultHi;
		h64Mult = h64Mult << 32 | hashMultLo;
		Ixp_Hash64New(&pHash64);
		Ixp_Hash64SetMult(pHash64, h64Mult);
	}
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::HashDBM

   Description: Creates an instance of the HashDBM object.

   Arguments:
				priTableDesc		-- primary table descriptor
				HdbmColTableDesc	-- collision table descriptor
				keyDataFreeLst		-- key and data free list
   Returns:		n/a
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
HashDBM::HashDBM(HdbmPriTableDesc *priTableDesc, HdbmColTableDesc *colTableDesc,
				 FreeLst *keyDataFreeLst)
{
	HashDBM(HASH_NONE, 0, 1, priTableDesc, colTableDesc, keyDataFreeLst);
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::getIndexField

   Description: Retreive the number of bits from the key as specified by the
				startBit and fieldLen arguments. The number of bits to be
				extracted must be less than or equal to 32 bits.
   Arguments:
				key			-- a 64-bit key to extract the field
				startBit	-- starting bit position to extract the field
				fieldLen	-- number of bits to extract from the key

   Returns:		an unsigned integer containing the extracted bits
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int HashDBM::getIndexField(HdbmKey_T *key, unsigned int startBit, unsigned int fieldLen)
{
	unsigned int val;
	ix_64bit	bit64;

#ifdef HDBM_DEBUG
	assert(((BitsInWord >= fieldLen) && ((BitsInWord*2) >= (startBit + fieldLen))));
#endif
	bit64 = key->B7654;
	bit64 = (bit64 << 32) | key->B3210; 
	val = (unsigned int)(bit64 >> startBit);
	val <<= (BitsInWord - fieldLen);
	val >>= (BitsInWord - fieldLen);
	return(val);
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::hashEncode

   Description: Perform either a 48-bit or 64-bit hash on the key.

  Arguments:
				key		-- a 64-bit key hash. The hash result is returned in
						   in the key.
   Returns:		HDBM_SUCCESS
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::hashEncode(HdbmKey_T *key)
{	
	if(pHashType == HASH48){
		ix_48bit toHash48 = 0L, fromHash48 = 0L;
		toHash48 = key->B7654 & 0xffff;
		toHash48 = (toHash48 << 32) | (key->B3210);
		Ixp_Hash48(pHash48, &fromHash48, toHash48);
		key->B3210 = (unsigned int)fromHash48;
		key->B7654 = (unsigned int)(fromHash48 >> 32) & 0xffff;
		}
	else{
		ix_64bit toHash64 = 0L, fromHash64 = 0L;
		toHash64 = key->B7654;
		toHash64 = (toHash64 << 32) | key->B3210;
		Ixp_Hash64(pHash64, &fromHash64, toHash64);
		key->B3210 = (unsigned int)fromHash64;
		key->B7654 = (unsigned int)(fromHash64 >> 32);
	}
    return(HDBM_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::getState

   Description: Determine the state of the database entry.

  Arguments:
				base		-- the SRAM base of the table
				index		-- offset into the table
				ret_ptr		-- the value of the entry

  Returns:		
				HDBM_NO_ENTRY		-- no entry
				HDBM_LOCK_ERROR		-- entry is locked
				HDBM_ENTRY_COL		-- entry with collision
				HDBM_ENTRY_NO_COL	-- entry without collision
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::getState(unsigned int base, unsigned int index, unsigned int *ret_ptr)
{
    SRAM_READ(base + index, ret_ptr);

	// no entry or 30 bit is set (entry is getting added)
    if(*ret_ptr == 0)
		return(HDBM_NO_ENTRY);		// no entry
	else if((*ret_ptr) & LOCK_MASK)
		return(HDBM_LOCK_ERROR);		// entry locked or error
    else if ((*ret_ptr) & COLLISION_MASK)
        return(HDBM_ENTRY_COL);		// entry with collision bit is set
    else 
        return(HDBM_ENTRY_NO_COL);	// entry with no collision
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::readKeyData

   Description: Read the key and user data from sdram.

  Arguments:
				numArg		-- the number of longwords to be read from sdram
				sdAddr		-- the sdram quad-word address to be read
				key			-- pointer to storage for key
				userData	-- pointer to the user data

  Returns:		0
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void HashDBM::readKeyData(unsigned int userDataArg, unsigned int sdAddr,
					   HdbmKey_T *key, unsigned int *userData)
{
	unsigned int longWordaddr = sdAddr << 1;

	SDRAM_READ(longWordaddr, &key->B3210);
	SDRAM_READ(longWordaddr+1, &key->B7654);
	if(userDataArg) SDRAM_READ(longWordaddr+2, userData);
	return;
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::writeKeyData

   Description: Write the key and user data to sdram.

  Arguments:
				numArg		-- the number of longwords to be written to sdram
				sdAddr		-- the sdram quad-word address to be written
				keyLo		-- lower 32 bit of the key
				keyHi		-- higher 32 bits of the key
				userData	-- the user data

  Returns:		0
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void HashDBM::writeKeyData(unsigned int userDataArg, unsigned int sdAddr,
					   HdbmKey_T *key, unsigned int userData)
{
	unsigned int longWordaddr = sdAddr << 1;

	SDRAM_WRITE(longWordaddr, key->B3210);
	SDRAM_WRITE(longWordaddr+1, key->B7654);
	if(userDataArg) SDRAM_WRITE(longWordaddr+2, userData);

#ifdef CREATE_IND
	//dj: debug
	logSdram(1, sdAddr, key->B3210, key->B7654);
	if(userDataArg) logSdram(0, sdAddr+1, userData, 0);
#endif
	return;
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::Add

   Description: Add an entry to the database using the specified key and
				associate userData with the entry if the operation was
				successfully.

  Arguments:
				key			-- key to the entry
				userData	-- data to be associated with the entry

  Returns:		
				HDBM_SUCCESS			-- operation successfully completed
				HDBM_LOCK_ERROR			-- entry is locked
				HDBM_EMPTY_COL_LST		-- collision table freelist empty
				HDBM_EMPTY_KEYDATA_LST	-- dataKey freelist empty
				HDBM_DUPL_ENTRY			-- duplicate entry
				HDBM_ERROR				-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::Add(HdbmKey_T *key, void* userData)
{
	unsigned int tableValue, colIndex, i;
    unsigned int collisionLevel=0, tableBase, index;

    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
			case HDBM_LOCK_ERROR :			//  Lock bit is set -- let upper layer recover
				return(HDBM_LOCK_ERROR);
				break;

			case HDBM_NO_ENTRY :						// no entry - add to main index table
				unsigned int keyDataIndex;

				if(pKeyDataFreeLst->allocEle(&keyDataIndex))
					return(HDBM_EMPTY_KEYDATA_LST);
				// add entry -- write the key entry
				writeKeyData(UserDataTrue, keyDataIndex, key, (unsigned int)userData);

				// write the forward entry index
				//note:- 1 is added to keyDataIndex, normalized to 1, so that the entry
				//at the base will not have an index of zero. An index of zero indicates
				//no entry found.
				keyDataIndex -= pKeyDataFreeLst->getBase();

				writeFwdIndex(tableBase + index, (keyDataIndex+1));	//update the table index
				return(HDBM_SUCCESS);
				break;

			case HDBM_ENTRY_NO_COL:				// entry with no collision
				unsigned int fwdEntryAddr, startBit;
				HdbmKey_T fwdEntryKey;

				// check for duplicate keys
				fwdEntryAddr = getKeyIndex(tableValue);			// fwd table entry index <29:0>
				readKeyData(UserDataFalse, (fwdEntryAddr - 1) + pKeyDataFreeLst->getBase(), &fwdEntryKey, NULL);
				if((fwdEntryKey.B3210 == key->B3210) && (fwdEntryKey.B7654 == key->B7654)){
					return(HDBM_DUPL_ENTRY);
				}

				// alloc and init a new collision table
				unsigned int newTable;
				if(pColTabFreeLst->allocEle(&newTable))
					return(HDBM_EMPTY_COL_LST);

				// init new collision table
				for(i = 0; i < pColTableSize; i++) writeFwdIndex(newTable + i, 0);

				// add old entry to the new table
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)){
					pColTabFreeLst->freeEle(newTable);
					return(HDBM_ERROR);
				}
				colIndex = getIndexField(&fwdEntryKey, startBit, pColShiftBits);
				writeFwdIndex(newTable + colIndex, fwdEntryAddr);	//update the new table index

				// add the reference to collision table
   				colIndex = setCollision(0);
				colIndex = setTableIndex(colIndex, newTable);
				writeFwdIndex(tableBase + index, colIndex);	//update the table index


				// add the new entry to collision table
				tableBase = newTable;
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				collisionLevel++;
				break;

			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * collisionLevel))) >= 
						(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				collisionLevel++;

			break;
			default:
				break;
		}
	}
    return(HDBM_ERROR);
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::Traverse

   Description: Traverse database and call the user specified function for each
				entry. The traversal will be terminated if the user function
				by returns a non-zero value.

  Arguments:
				funcPtr		-- pointer to a user function that will be called for
							   each entry. The key, dataEntry, and a user specified
							   value will be passed to the function.
				userData	-- user specific value

  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_ERROR			-- invalid database value encountered
				The non-zero value returned from the user function.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::Traverse(int (*funcPtr)(HdbmKey_T *key, void *dataEntry, void *userData), void *userData)
{
	return(traverseDb(pPriTableBase, pPriTableSize, funcPtr, userData));
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::traverseDb

   Description: Traverse database and call the user specified function for each
				entry. The traversal will be terminated if the user function
				by returns a non-zero value.

  Arguments:
				tableBase	-- the SRAM base of the primary table.
				tableSize	-- the size of the primary table
				funcPtr		-- pointer to a user function that will be called for
							   each entry. The key, dataEntry, and a user specified
							   value will be passed to the function.
				userData	-- user specific value

  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_ERROR			-- invalid database value encountered
				The non-zero value returned from the user function.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::traverseDb(unsigned int tableBase, unsigned int tableSize,
					   int (*funcPtr)(HdbmKey_T *key, void *entry, void *userData), void *userData)
{
	unsigned int priIndex, tableValue, fwdEntry;
	int status;

	if(!funcPtr) return(HDBM_ERROR);

	for(priIndex = 0; priIndex < tableSize; priIndex++){
		switch(getState(tableBase, priIndex, &tableValue)){
			case HDBM_ENTRY_NO_COL:				// entry with no collision
				unsigned int dataEntry;
				HdbmKey_T key;

				fwdEntry = pKeyDataFreeLst->getBase() + getKeyIndex(tableValue);
				readKeyData(UserDataTrue, fwdEntry-1, &key, &dataEntry);
				if((status = (*funcPtr)(&key, (void *)dataEntry, userData))) return(status);
				break;
	        case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				unsigned int colTableBase;

				colTableBase = getTableIndex(tableValue);
				traverseDb(colTableBase, pColTableSize, funcPtr, userData);
				break;
			default:
				break;
		}
	}
	return(HDBM_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::Lookup

   Description: Search the database using the specified key return the
				associated userData if the operation was successfully.

  Arguments:
				key			-- pointer to the key.
				userData	-- pointer to where the associated data will be stored.
  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_NO_ENTRY		-- entry not found
				HDBM_ERROR			-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::Lookup(HdbmKey_T *key, void *userData)
{
	unsigned int tableValue, fwdEntry, tableBase;
    unsigned int index=0, colLevel = 0, startBit = 0;
    
    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
	        case HDBM_NO_ENTRY :					// no entry - add to main index table
				return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_NO_COL:				// entry with no collision
				HdbmKey_T keyInDb;

				fwdEntry = pKeyDataFreeLst->getBase() + getKeyIndex(tableValue);
				readKeyData(UserDataTrue, fwdEntry-1, &keyInDb, (unsigned int*)userData);

				if((keyInDb.B3210 == key->B3210) && (keyInDb.B7654 == key->B7654))
				    return(HDBM_SUCCESS);
				return(HDBM_NO_ENTRY);	// entry not found
				break;
			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table 
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * colLevel))) >=
					(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				colLevel++;

			break;
			default:
				break;
		}
	}
    return(HDBM_ERROR);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::Remove

   Description: Remove an entry from the database.

  Arguments:
				key			-- pointer to the key.
  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
				HDBM_NO_ENTRY		-- entry not found
				HDBM_ERROR			-- invalid database value encountered
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::Remove(HdbmKey_T *key)
{
	unsigned int tableValue, tableBase, prevTable, prevIndex;
    unsigned int index=0, colLevel = 0, startBit;
    
    // Hash encode the key
	if(pHashType != HASH_NONE) hashEncode(key);

	index = getIndexField(key, pPriStartBit, pPriIndxBitLen);
	tableBase = pPriTableBase;


	while(1){
		switch(getState(tableBase, index, &tableValue)){
	        case HDBM_NO_ENTRY :				// no entry - add to main index table
				return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_NO_COL:				// entry with no collision
				HdbmKey_T keyInDb;
				unsigned int fwdEntry;

				fwdEntry = pKeyDataFreeLst->getBase() + getKeyIndex(tableValue) - 1;
				readKeyData(UserDataFalse, fwdEntry, &keyInDb, NULL);
				if((key->B3210 == keyInDb.B3210) && (key->B7654 == keyInDb.B7654)){
					unsigned int i, indx, numOtherEntries = 0, otherEntryValue;

					pKeyDataFreeLst->freeEle(fwdEntry);
					// determine if any other entry is in this table
					if(colLevel){
						for(i = 0; i < pColTableSize && (numOtherEntries < 2); i++){
							if(i == index) continue;
							readFwdIndex(tableBase + i, &indx);
							if(indx){
								numOtherEntries++;
								otherEntryValue = indx;
							}
						}
						if((numOtherEntries == 1) && !(otherEntryValue & COLLISION_MASK)){
							// only one other entry with no collision. So, remove target entry;
							// move remaining entry to a lower collison level; update prev
							// table; and free current table
							writeFwdIndex(prevTable + prevIndex, otherEntryValue);
							pColTabFreeLst->freeEle(tableBase);
						}
						else{
							// there are at least two other entries or a single entry with collision.
							// So, just remove target entry from the table
							writeFwdIndex(tableBase + index, 0);
						}
					}
					else{
						// no collision -- so, just initialize the table
						writeFwdIndex(tableBase + index, 0);
					}
				    return(HDBM_SUCCESS);
				}
			    else return(HDBM_NO_ENTRY);
				break;
			case HDBM_ENTRY_COL:					// entry but collision -- enter next collision table
				prevTable = tableBase;
				prevIndex = index;
				tableBase = getTableIndex(tableValue);
				if((startBit = (pColStartBit + (pColShiftBits * colLevel))) >=
					(pColStartBit + pColIndxBitLen)) return(HDBM_ERROR);
				index = getIndexField(key, startBit, pColShiftBits);
				colLevel++;
				break;
			default:
				break;
		}
	}

    return(HDBM_NO_ENTRY);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::RemoveAllEntries

   Description: Remove all entries from the database.

  Arguments:	n/a

  Returns:		
				HDBM_SUCCESS	-- operation successfully completed
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::RemoveAllEntries(void)
{
	return(deleEntries(pPriTableBase, pPriTableSize, 0));
}



/*-----------------------------------------------------------------------------
   Function: 	HashDBM::deleEntries

   Description: Remove all entries from the database.

  Arguments:	
				tableBase	-- the SRAM base of the primary table.
				tableSize	-- the size of the primary table

  Returns:		
				HDBM_SUCCESS		-- operation successfully completed
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::deleEntries(unsigned int tableBase, unsigned int tableSize, unsigned int level)
{
	unsigned int index, tableValue;

	if(level > HDBM_MAXLEVEL) return(HDBM_ERROR);

	for(index = 0; index < tableSize; index++){
		switch(getState(tableBase, index, &tableValue)){
			case HDBM_ENTRY_NO_COL:						// entry with no collision
				unsigned int fwdEntry;
				fwdEntry = pKeyDataFreeLst->getBase() + getKeyIndex(tableValue) -1;
				pKeyDataFreeLst->freeEle(fwdEntry);
				writeFwdIndex(tableBase + index, 0);	// init the table
				break;
	        case HDBM_ENTRY_COL:						// entry but collision -- enter next collision table 
				unsigned int colTableBase;
				colTableBase = getTableIndex(tableValue);
				deleEntries(colTableBase, pColTableSize, level+1);
				pColTabFreeLst->freeEle(colTableBase);
				writeFwdIndex(tableBase + index, 0);	// init the table
				break;
			default:
				break;
		}
	}
	return(HDBM_SUCCESS);
}


#ifdef HDBM_DEBUG
/*-----------------------------------------------------------------------------
   Function: 	HashDBM::getStats

   Description: 

   Returns:		
				HDBM_SUCCESS		-- operation successfully completed
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int	HashDBM::getStats(unsigned int tableBase, unsigned int tableSize, unsigned int level)
{
	unsigned int index, tableValue;

	pStats[level].totSpace += tableSize;
	for(index = 0; index < tableSize; index++){
		switch(getState(tableBase, index, &tableValue)){
			case HDBM_ENTRY_NO_COL:			// entry with no collision
				unsigned int fwdEntry;
				pStats[level].occupied++;
				pTotEntries++;
				fwdEntry = pKeyDataFreeLst->getBase() + getKeyIndex(tableValue) -1;
				break;
	        case HDBM_ENTRY_COL:				// entry but collision -- enter next collision table 
				unsigned int colTableBase;
				pStats[level].collisions++;
				colTableBase = getTableIndex(tableValue);
				getStats(colTableBase, pColTableSize, level+1);
				break;
	        case HDBM_NO_ENTRY :				// no entry - add to main index table
				pStats[level].empty++;
				break;
			default:
				break;
		}
	}
	return(HDBM_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::printStats

   Description: 

   Returns:		
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void HashDBM::printStats()
{
	unsigned int totalSpace = 0, totalCollision=0, totalEmpty=0;
	float storeEff, utilEff;

	printf("\n\n-------|------------|-------------------------------\n");
	printf("Search |   Search   |      Table Utilization\n");
	printf("Level  | Proability | Data %%   Collision %%  Empty %%\n");
	printf("-------|------------|-------------------------------\n");
	for(int i = 0; i < HDBM_MAXLEVEL; i++){
		if(pStats[i].occupied || pStats[i].collisions){
			printf("%-6d | %-10.4f | %7.4f  %11.4f  %7.4f\n",
				i, (float)pStats[i].occupied/pTotEntries,
				(((float)pStats[i].occupied/pStats[i].totSpace)) * 100,
				(((float)pStats[i].collisions/pStats[i].totSpace)) * 100,
				(((float)pStats[i].empty/pStats[i].totSpace)) * 100);
		}
		totalSpace += pStats[i].totSpace;
		totalCollision += pStats[i].collisions;
		totalEmpty += pStats[i].empty;
	}

	printf("\nPrimary Table: Size=%d, StartBit=%d, IndxLen=%d\n",
		pPriTableSize, pPriStartBit, pPriIndxBitLen);

	printf("Collision Table: Size=%d, StartBit=%d, IndxLen=%d, IndexStep=%d\n",
		pColTableSize, pColStartBit, pColIndxBitLen, pColShiftBits);

	printf("Total: Size=%d, Collisions=%d (%0.4f%%), Empty=%d (%0.4f%%), Entries=%d (%0.4f%%)\n",
		totalSpace, totalCollision, ((float)totalCollision/totalSpace) * 100,
		totalEmpty, ((float)totalEmpty/totalSpace) * 100,
		pTotEntries, ((float)pTotEntries/totalSpace) * 100);

	if(pTotEntries){
		storeEff = ((pTotEntries /((float)totalCollision + pTotEntries))) * 100;
		utilEff = (((pTotEntries << 1) /((float)pTotEntries + ((float)totalSpace - pPriTableSize)))) * 100;
		printf("Efficiency: Storage=%0.4f%%, Utilization=%0.4f%%\n",storeEff, utilEff);
	}
	
	return;
}


/*-----------------------------------------------------------------------------
   Function: 	HashDBM::analyze

   Description: 

   Returns:		
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int HashDBM::analyze(void)
{
	for(int i = 0; i < HDBM_MAXLEVEL; i++){
		pStats[i].collisions = 0;
		pStats[i].empty = 0;
		pStats[i].occupied = 0;
		pStats[i].totSpace = 0;
	}
	pTotEntries = 0;
	getStats(pPriTableBase, pPriTableSize, 0);
	printStats();

	printf("\nFreeList Availability: Collision=%d, Key/Data=%d\n",
		pColTabFreeLst->getAvailEle(), pKeyDataFreeLst->getAvailEle());
	return(0);
}

#endif
