/* ------------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------------------------*/

/* fdrive.cpp
 * fbus mac receive side
 *
 *
 * system: SA1200
 * subsystem: packet_gen
 * author: Don Hooper 3/10/99
 * revisions:
 * $History fdrive.cpp $
 * 
 * 
 */ 
#include "packet.h"
#include "fmon.h"
#include "fdrive.h"
#include "endian.h"

FDrive fd;




/*-----------------------------------------------------------------
	PacketPool::Init
	description: Flush packet pool.
	returns:
	uses:
	modifies: pkendbuf, pkend, pkcount, pkstarts[0] to 0
  -----------------------------------------------------------------*/
int PacketPool::PkPoolInit()
{
	pkendbuf = 0;
	pkend = 0;
	pkcount = 0;
	pkstarts[0] = 0;
	return(FDRIVE_SUCCESS);			
}

extern "C" int pkpool_init()
{
	return(fd.PkPoolInit());
}

/*-----------------------------------------------------------------
	PacketPool::PacketPool
	description: Call Init to initialize parameters for empty pool.
	returns: FDRIVE_SUCCESS or FDRIVE_FAIL
	uses:
	modifies:
  -----------------------------------------------------------------*/
PacketPool::PacketPool()
{
	PkPoolInit();
}


/*-----------------------------------------------------------------
	PacketPool::Add
	description: Add a packet to the packet pool by copying len bytes from buf
		buf data must be in network order.
	returns: FDRIVE_SUCCESS or FDRIVE_POOL_FULL
	uses:
	modifies:
	see also: packet.h
  -----------------------------------------------------------------*/
int PacketPool::Add(void **buf, int len)
{
	int i, j, wdsize;
	unsigned int *pkbuf = (unsigned int *)buf;

	wdsize = len/4;					// word size
	if (len%4) wdsize++;

	if ((wdsize + pkendbuf) >= POOL_SIZE) return(FDRIVE_POOL_FULL);

	for(i=0, j=0; i < len; i+=4, j++){
		pksource[pkendbuf+j] = ntohl(pkbuf[j]);
	}
	// get last bytes if len not on word boundary
	if ((i-len) > 0){
		switch(i-len){
		case 1: pksource[pkendbuf+j] = ntohl(pkbuf[j]) & 0xff000000; break;
		case 2: pksource[pkendbuf+j] = ntohl(pkbuf[j]) & 0xffff0000; break;
		case 3: pksource[pkendbuf+j] = ntohl(pkbuf[j]) & 0xffffff00; break;
		default:		// cannot get here
			break;
		}
	}

	pksizes[pkend] = len;
	pkcount++;

	pkendbuf = pkendbuf + wdsize;
	pkend++;
	pkstarts[pkend] = pkendbuf;
	return(FDRIVE_SUCCESS);			
}


/*-----------------------------------------------------------------
	pkpool_add
	description: Add a packet created by Packet::Create to the packet pool.
		this assumes the packet created is in Network Order.
	returns: FDRIVE_SUCCESS or FDRIVE_POOL_FULL
	uses: packet P created by Packet::Create
	modifies:
	see also: packet.h
  -----------------------------------------------------------------*/
extern "C" int pkpool_add_from_buf(void **buf, int len)
{
	return(fd.Add(buf, len));
}

extern "C" int pkpool_add()
{
	int retval;

#ifdef LITTLE_ENDIAN
	P.OtherEndian();					// swap to network order if host is little-endian
#endif
	retval = fd.Add((void **)&P.pk_buf, P.size);

#ifdef LITTLE_ENDIAN
	P.OtherEndian();					// swap back to host order
#endif
	return(retval);
}


/*-----------------------------------------------------------------
	PacketPool::Show
	description: Print to screen. Show a specific packet n in the packet pool.
	uses:
	returns: FDRIVE_SUCCESS or FDRIVE_NO_SUCH_PACKET
	modifies:
  -----------------------------------------------------------------*/
int PacketPool::Show(int n)
{

	unsigned int i, size, qw, rem_qw_bytes;

	if (n >= pkend){
		XACT_printf("Packet not in pool.\n");
		return (FDRIVE_NO_SUCH_PACKET);
	}

	size = pksizes[n];			// byte size

	qw = size >> 3;				// byte size * 8 = full quadwords
	rem_qw_bytes = size & 0x7;	// remaining bytes in last quadword

	XACT_printf("bytes 0-3 \tbytes 4-7 \tbytes 8-11 \tbytes 12-15\n");
	for (i = pkstarts[n]; i < pkstarts[n]+qw*2; i+=2) {
			XACT_printf("0x%.8x\t0x%.8x\t", pksource[i] , pksource[i+1]);
			i+=2;
			if(i < pkstarts[n]+qw*2){
				XACT_printf("0x%.8x\t0x%.8x\n", pksource[i] , pksource[i+1]);
			} else break;
	}
	if (rem_qw_bytes > 4){
			XACT_printf("0x%.8x\t0x%.8x\n", pksource[i] , pksource[i+1]);
	} else if (rem_qw_bytes > 0){
			XACT_printf("0x%.8x\t0x%.8x\n", pksource[i] , 0);
	}

	XACT_printf("size: %d bytes\n\n", size);

	return(FDRIVE_SUCCESS);
}

extern "C" int pkpool_show(unsigned int n)
{
	fd.Show(n);
	return(n);
}


extern "C" int pkpool_show_last()
{
	int temp = (fd.pkend)-1;

	XACT_printf("Id of last packet in PkPool is: %d\n", temp);
	return(fd.Show(temp));
}


/*-----------------------------------------------------------------
	RxPort::Init
	description: Initialize the port to default settings.
	returns: FDRIVE_SUCCESS or FDRIVE_FAIL
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxPort::Init()
{
	state = IDLE;
	enabled = 0;
	current_packet = 0;
	last_packet = fd.pkcount;
	current_pksource = 0;
	first_pksource = 0;
	last_pksource = fd.pkstarts[fd.pkend];
	rem_size = fd.pksizes[0];

	fast_port = 0;
	wire_bytes_per_clock = 0;
	overflow = 0;
	underflow = 0;
	buf_bytes = 0;
	current_wire_packet = 0;
	rdy_threshold = DEFAULT_RDY_THRESHOLD;
	rdy = 0;

	return;			
}


/*-----------------------------------------------------------------
	RxPort::Enable
	description: Enable to drive packets.
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxPort::Enable()
{
	enabled = 1;
	// if rate hasn't been set, make it always ready (max rate)
	if (!wire_bytes_per_clock) rdy = 1;

	return;
}

extern "C" int rport_on(unsigned int mac, unsigned int port)
{
	fd.macs[mac].enabled_vector = fd.macs[mac].enabled_vector | (1<<port);
	fd.macs[mac].ports[port].Enable();
	return(FDRIVE_SUCCESS);
}


/*-----------------------------------------------------------------
	RxPort::Disable
	description: Drive pkcount packets from this port then disable this port after packet/status send.
	returns:
	uses:
	modifies: pk_rem_count, prep_to_disable
  -----------------------------------------------------------------*/
void RxPort::Disable(unsigned int pkcount)
{
	pk_rem_count = pkcount;
	prep_to_disable = 1;
	return;
}

extern "C" int rport_off(unsigned int mac, unsigned int port, unsigned int afterpkcount)
{
	fd.macs[mac].enabled_vector = fd.macs[mac].enabled_vector & ~(1<<port);
	fd.macs[mac].ports[port].Disable(afterpkcount);
	return(FDRIVE_SUCCESS);
}


/*-----------------------------------------------------------------
	RxPort::BufBytesPut
	description: every F Bus Clock add wire_bytes_per_clock to current_wire_pksource
		add wire_bytes_per_clock to buf_bytes. If buf_bytes > hardware_buf_size, overflow.
		If current_wire_pksource > pksizes[current_wire_packet] set a delay to ignore
		IPG, preamble, SPD, SFD bytes (next 20 wire bytes).
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxPort::BufBytesPut()
{
	double temp;
	unsigned int val;

	if(inter_packet_wait > 0){
		inter_packet_wait--;
	}else{
		buf_pk_bytes += wire_bytes_per_clock;
		temp = buf_pk_bytes - fd.pksizes[(int)current_wire_packet];
		if (temp>0){
			buf_pk_bytes = temp;
			current_wire_packet++;
			current_wire_packet = (current_wire_packet > fd.pkend) ? 0 : current_wire_packet;
			inter_packet_wait = INTER_PACKET_BYTES / wire_bytes_per_clock;
		}

		buf_bytes += wire_bytes_per_clock;

	

		if (buf_bytes > buf_size){
			XACT_printf("Receive OverFlow, mac: %d \tport: %d\n", my_id >> 3, my_id &0x7);
			overflow = 1;
			val = 1;
			XACT_set_state_value(fm.sim_halt_handle, &val);
		}
	}
	if (buf_bytes >= rdy_threshold){
		rdy = 1;
	} else rdy = 0;

	if(fast_port == 1){
		val = rdy;
		XACT_set_state_value(fd.fast_rx1_handle, &val);
	} else if (fast_port == 2){
		val = rdy;
		XACT_set_state_value(fd.fast_rx2_handle, &val);
	}

	return;
}


/*-----------------------------------------------------------------
	RxPort::BufBytesTake
	description: Every Data Drive Cycle for this port decrement n buf_bytes. 
		If buf_bytes < 0	set underflow.
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxPort::BufBytesTake(int n)
{
	unsigned int halt_val;
	buf_bytes -=n;

	//if (buf_bytes < not_rdy_threshold) rdy = 0;

	if (buf_bytes >= rdy_threshold){
		rdy = 1;
	} else rdy = 0;

	if (buf_bytes < 0){
		XACT_printf("Receive UnderFlow, mac: %d \tport: %d\n", my_id >> 3, my_id &0x7);
		underflow = 1;
		XACT_set_state_value(fm.sim_halt_handle, &halt_val);
	}

	return;
}


/*-----------------------------------------------------------------
	RxPort::UpdateOutptr
	description: Update pointer into packet pool source and decrement size of remaining packet.
	returns:
	uses:
	modifies:	rem_size and outptr
  -----------------------------------------------------------------*/
void RxPort::UpdateOutptr()
{	
	if (fm.pr_enabled) XACT_printf("Rx packet remaining bytes=%d\n", rem_size);
	if (wire_bytes_per_clock) BufBytesTake((rem_size > 4) ? 4 : rem_size);

	current_pksource++;
	rem_size -=4;
	// get next packet if rem_size <= 0
	if (rem_size <=0){
		current_packet++;
		if (current_packet >= last_packet){
			current_pksource = first_pksource;			//wrap to beginning
			current_packet = first_packet;
		}

		rem_size = fd.pksizes[current_packet];
		if (prep_to_disable == 1) pk_rem_count--;
	}

	return;			
}


/*-----------------------------------------------------------------
	RxPort::DriveFbeLo
	description: Drive eop cycle byte enables for fdat_lo.
	returns: XACT_set_state_value return value(1 if success, 0 if otherwise)
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int RxPort::DriveFbeLo()
{
	unsigned int val;

	if (fm.fbus_big_endian){

		if (rem_size > 3){
			val = 0;
		}else if (rem_size == 3){
			val = 0x1;
		}else if (rem_size == 2){
			val = 0x3;
		}else val = 0x7; 						// 0 can't happen

	} else {

		if (rem_size > 3){
			val = 0;
		}else if (rem_size == 3){
			val = 0x8;
		}else if (rem_size == 2){
			val = 0xc;
		}else val = 0xe; 						// 0 can't happen

	}
	
	return(XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val));
}


/*-----------------------------------------------------------------
	RxPort::DriveFbeHi
	description: Drive eop cycle byte enables for fdat_hi.
	returns: XACT_set_state_value return value(1 if success, 0 if otherwise)
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int RxPort::DriveFbeHi()
{
	unsigned int val;

	if (fm.fbus_big_endian){

		if (rem_size < 5){
			val = 0xf;
		}else if (rem_size == 5){
			val = 0x7;
		}else if (rem_size == 6){
			val = 0x3;
		}else if (rem_size == 7){
			val = 0x1;
		} else val = 0;

	} else {

		if (rem_size < 5){
			val = 0xf;
		}else if (rem_size == 5){
			val = 0xe;
		}else if (rem_size == 6){
			val = 0xc;
		}else if (rem_size == 7){
			val = 0x8;
		} else val = 0;

	}


	return(XACT_set_state_value(fd.fbe_hi_l_bus_handle, &val));
}


/*-----------------------------------------------------------------
	RxPort::DriveFBus
	description: Drive fbus byte enables, data, eop, sop, rxfail.
	returns: the bit AND of all XACT_set_state_value return values(1 if success, 0 if otherwise)
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int RxPort::DriveFBus()
{
	unsigned int val =1;
	int retval = 1;

	if (state == IDLE)
		state = IN_PACKET;			// starting new packet

	val = 0;
	retval &= XACT_set_state_value(fd.rxfail_bus_handle, &val);		// must drive this or else will be like EOP

	// if at start of packet rem_size will not have been decremented yet
	// deposit SOP
	if ((state == IN_PACKET) && (rem_size == fd.pksizes[current_packet])){
		val = 1;
		retval &= XACT_set_state_value(fd.sop_bus_handle, &val);

	} else {val = 0; retval &= XACT_set_state_value(fd.sop_bus_handle, &val);}

	// deposit data, EOP, byte enables
	if (fm.fbus_bidirectional == 1){
		switch(state){
		case IN_PACKET:
			if (rem_size <= 8){
				if (fm.fbus_no_status) state = STATUS2_SENT; else state = EOP_SENT;
				val = 1;
				retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
				DriveFbeLo();
				DriveFbeHi();
			}else{
				val = 0;
				retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
				val = 0xf;
				retval &= XACT_set_state_value(fd.fbe_hi_l_bus_handle, &val);
				retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			}
			val = fd.pksource[current_pksource];
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);

			if (rem_size <= 4){ 
				val = 0;
			}else{
				UpdateOutptr();
				val = fd.pksource[current_pksource];
			}
			retval &= XACT_set_state_value(fd.fdat_hi_bus_handle, &val);
			
			UpdateOutptr();
			break;

		case EOP_SENT:
			val = RCV_OK_STATUS;
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			val = 0;
			retval &= XACT_set_state_value(fd.fdat_hi_bus_handle, &val);
			retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
			val = 0xf;
			retval &= XACT_set_state_value(fd.fbe_hi_l_bus_handle, &val);
			retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			state = STATUS2_SENT;
			break;

		case STATUS2_SENT:
			val = 0;
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			retval &= XACT_set_state_value(fd.fdat_hi_bus_handle, &val);
			retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
			val = 0xf;
			retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			retval &= XACT_set_state_value(fd.fbe_hi_l_bus_handle, &val);
		}
	} else{
		switch(state){
		case IN_PACKET:
			if (rem_size <= 4){
				if (fm.fbus_no_status) state = STATUS2_SENT; else state = EOP_SENT;
				val = 1;
				retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
				DriveFbeLo();
			}else {
				val = 0;
				retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
				val = 0xf; retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			}
			val = fd.pksource[current_pksource];
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			UpdateOutptr();
			break;

		case EOP_SENT:
			val = RCV_OK_STATUS;
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			val = 0xf;
			retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			val = 0;
			retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
			state = STATUS1_SENT;
			break;

		case STATUS1_SENT:
			val = RCV_OK_STATUS;
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			val = 0xf;
			retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
			val = 0;
			retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
			state = STATUS2_SENT;
			break;

		case STATUS2_SENT:
			val = 0;
			retval &= XACT_set_state_value(fd.eop_bus_handle, &val);
			retval &= XACT_set_state_value(fd.fdat_lo_bus_handle, &val);
			val = 0xf;
			retval &= XACT_set_state_value(fd.fbe_lo_l_bus_handle, &val);
		}

	}

	return(retval);
}


/*-----------------------------------------------------------------
	RxPort::NoDriveFBus
	description: Update port state and possibly disable port.
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxPort::NoDriveFBus()
{
	if ((state == STATUS1_SENT)||(state == STATUS2_SENT)){
		state = IDLE;
		// disable the port when pk_rem_count = 0
		if ((prep_to_disable == 1) && (pk_rem_count <= 0)){
			enabled = 0;
			prep_to_disable = 0;
		}
		// count packets if simulator set to stop after n packets
		if (fd.rx_stop_count > 0) fd.rx_count++;
	}
	return;
}


/*-----------------------------------------------------------------
	RxPort::SetFirst
	description: Set first packet this port should work on after reaching last_pksource.
	returns: FDRIVE_SUCCESS or FDRIVE_NO_SUCH_PACKET
	uses:
	modifies:
  -----------------------------------------------------------------*/
int RxPort::SetFirst(unsigned int n)
{
	if (n > fd.pkend) return (FDRIVE_NO_SUCH_PACKET);

	first_pksource = fd.pkstarts[n];
	first_packet = n;
	return(FDRIVE_SUCCESS);
}

extern "C" int rport_first(unsigned int mac, unsigned int port, unsigned int n)
{
	return(fd.macs[mac].ports[port].SetFirst(n));
}


/*-----------------------------------------------------------------
	RxPort::SetLast
	description: Set last packet this port should work on before wrapping to first_pksource.
		This should be done when port is disabled or IDLE.
	returns: FDRIVE_SUCCESS or FDRIVE_NO_SUCH_PACKET
	uses:
	modifies:
  -----------------------------------------------------------------*/
int RxPort::SetLast(unsigned int n)
{
	if (n+1 > fd.pkend) return (FDRIVE_NO_SUCH_PACKET);

	last_pksource = fd.pkstarts[n+1];
	last_packet = n+1;

	return(FDRIVE_SUCCESS);
}

extern "C" int rport_last(unsigned int mac, unsigned int port, unsigned int n)
{
	return(fd.macs[mac].ports[port].SetLast(n));
}


/*-----------------------------------------------------------------
	RxPort::SetCurrent
	description: Set current packet this port should work on next.
	returns: FDRIVE_SUCCESS or FDRIVE_NO_SUCH_PACKET
	uses:
	modifies:
  -----------------------------------------------------------------*/
int RxPort::SetCurrent(unsigned int n)
{
	if (n > fd.pkend) return (FDRIVE_NO_SUCH_PACKET);

	current_pksource = fd.pkstarts[n];
	current_packet = n;
	rem_size = fd.pksizes[n];
	return(FDRIVE_SUCCESS);
}

extern "C" int rport_next(unsigned int mac, unsigned int port, unsigned int n)
{
	return(fd.macs[mac].ports[port].SetCurrent(n));
}


/*-----------------------------------------------------------------
	RxPort::ShowCurrent
	description: Show current packet this port is working on.
	returns: FDRIVE_SUCCESS or FDRIVE_NO_SUCH_PACKET
	uses:
	modifies:
  -----------------------------------------------------------------*/
int RxPort::ShowCurrent()
{
	XACT_printf("Receive status for port \t%d\n\n", my_id);
	XACT_printf("\tBuffer bytes:   \t%.0f\n", buf_bytes);
	XACT_printf("\tBuffer size:    \t%d\n", buf_size);
	XACT_printf("\tReady:          \t%d\n", rdy);
	XACT_printf("\tPresent receive packet id: %d\n", current_packet);
	XACT_printf("\n");
	return(fd.Show(current_packet));
}

extern "C" int rport_show(unsigned int mac, unsigned int port)
{
	fd.macs[mac].ports[port].ShowCurrent();

	return(fd.macs[mac].ports[port].current_packet);
}


/*-----------------------------------------------------------------
	RxMAC::Init
	description: Make RxMAC not ready and not enabled, initialize all ports.
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxMAC::Init()
{
	int j;

	enabled_vector = 0;
	rdy_vector = 0;

	for (j=0;j<PORTS_PER_MAC; j++){
		ports[j].Init();
		ports[j].my_id = (my_id<<3) + j;
	}

	return;
}


/*-----------------------------------------------------------------
	RxMAC::SetRate
	description: Set receive rate for MAC
	returns: 
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int RxMAC::SetRate(int Mbits_per_sec, int MFClocks_per_sec)
{
	int j;
	double temp = (double)Mbits_per_sec / (double)(MFClocks_per_sec*8);

	if(Mbits_per_sec >= 600){
		for (j=0;j<MAX_FAST_PORTS_PER_MAC; j++){
			ports[j].fast_port = j+1;
			ports[j].wire_bytes_per_clock = temp;
			ports[j].buf_size = BUF_SIZE_GIG_PORT;
		}

	} else{
		for (j=0;j<PORTS_PER_MAC; j++){
			ports[j].fast_port = 0;
			ports[j].wire_bytes_per_clock = temp;
			ports[j].buf_size = BUF_SIZE_SLOW_PORT;
		}
	}

	return(FDRIVE_SUCCESS);
}

extern "C" int rmac_set_rate(int mac, int Mbits_per_sec, int MFClocks_per_sec)
{
	return(fd.macs[mac].SetRate(Mbits_per_sec, MFClocks_per_sec));
}

extern "C" int rmac_set_rdy_threshold(int mac, int rdy)
{
	int j;

	for (j=0;j<PORTS_PER_MAC; j++){
		fd.macs[mac].ports[j].rdy_threshold = rdy;
	}
	return(FDRIVE_SUCCESS);
}



/*-----------------------------------------------------------------
	RxMAC::UpdatePortVectors
	description: Poll ports to determined if they are enabled and ready.
	returns:
	uses:
	modifies: enabled_vector and rdy_vector
  -----------------------------------------------------------------*/
void RxMAC::UpdatePortVectors()
{
	int i;

	enabled_vector = rdy_vector = 0;
	for(i=0;i<PORTS_PER_MAC;i++){
		enabled_vector = enabled_vector | (ports[i].enabled<<i); 
		rdy_vector = rdy_vector | (ports[i].rdy<<i); 
	}
	return;
}


/*-----------------------------------------------------------------
	RxMAC::DecodeMACSel
	description: Determine which Rx MAC has been selected.
	returns:
	uses:
	modifies: rx_mac_sel
  -----------------------------------------------------------------*/
void RxMAC::DecodeMACSel()
{
	switch(fm.mode){
	case BI2MAC:
		switch(port_ctl_l_dly3){
		case 0xE: rx_mac_sel = 0; break;
		case 0xD: rx_mac_sel = 1; break;
		default: rx_mac_sel = 8;		// no select
		}
		break;
	case UNI2MAC:
		switch(port_ctl_l_dly3 & 0x3){
		case 0x2: rx_mac_sel = 0; break;
		case 0x1: rx_mac_sel = 1; break;
		default: rx_mac_sel = 8;		// no select
		}
		break;
	case BI4MAC:
		switch(port_ctl_l_dly3){
		case 0x1: rx_mac_sel = 0; break;
		case 0x3: rx_mac_sel = 1; break;
		case 0x5: rx_mac_sel = 2; break;
		case 0x7: rx_mac_sel = 3; break;
		case 0x9: rx_mac_sel = 4; break;
		case 0xB: rx_mac_sel = 5; break;
		case 0xD: rx_mac_sel = 6; break;
		default: rx_mac_sel = 8;		// no select
		}
		break;
	case UNI4MAC:
		if (rdyctl_4_l_dly3){			// if rdyctl<4> = 1, no select
			rx_mac_sel = 8;				// no select
		} else {
			switch(port_ctl_l_dly3 & 0x3){
			case 0x0: rx_mac_sel = 0; break;
			case 0x1: rx_mac_sel = 1; break;
			case 0x2: rx_mac_sel = 2; break;
			case 0x3: rx_mac_sel = 3; break;
			default: rx_mac_sel = 8;	// no select
			}
		}
	}
	return;

}


/*-----------------------------------------------------------------
	RxMAC::DecodeMACRdySel
	description: Decode which mac has been selected to drive rdybus.
		If no mac selected, rx_mac_rdy_sel = 8.
	returns:
	uses:
	modifies: rx_mac_rdy_sel
  -----------------------------------------------------------------*/
void RxMAC::DecodeMACRdySel()
{
	switch(fm.mode){
	case BI2MAC:
	case UNI2MAC:
		switch(rdyctl_l_dly3){
		case 0xE: rx_mac_rdy_sel = 0; break;
		case 0xB: rx_mac_rdy_sel = 1; break;
		default: rx_mac_rdy_sel = 8;	// no select
		}
		break;
	case BI4MAC:
		switch(rdyctl_l_dly3 | (rdyctl_4_l_dly3<<4)){
		case 0x1B: rx_mac_rdy_sel = 0; break;
		case 0x1A: rx_mac_rdy_sel = 1; break;
		case 0x19: rx_mac_rdy_sel = 2; break;
		case 0x18: rx_mac_rdy_sel = 3; break;
		case 0x0B: rx_mac_rdy_sel = 4; break;
		case 0x0A: rx_mac_rdy_sel = 5; break;
		case 0x09: rx_mac_rdy_sel = 6; break;
		default: rx_mac_rdy_sel = 8;	// no select
		}
  		break;
	case UNI4MAC:
		switch(rdyctl_l_dly3){
		case 0xB: rx_mac_rdy_sel = 0; break;
		case 0xA: rx_mac_rdy_sel = 1; break;
		case 0x9: rx_mac_rdy_sel = 2; break;
		case 0x8: rx_mac_rdy_sel = 3; break;
		default: rx_mac_rdy_sel = 8;	// no select
		}
	}
	return;

}

/*-----------------------------------------------------------------
	RxMAC::ShowBufs
	description: Print port buffer byte counts.
	returns: FDRIVE_SUCCESS
	uses:
	modifies:
  -----------------------------------------------------------------*/
void RxMAC::ShowBufs()
{	
	int i;

	XACT_printf("RxMAC %d bytes: ", my_id);
	for (i=0;i<PORTS_PER_MAC;i++){
		if (ports[i].enabled){
			XACT_printf("p%d: %.4d  ", i, (int) ports[i].buf_bytes);
		}
	}
	XACT_printf("\n");
	
	return;
}


/*-----------------------------------------------------------------
	RxMAC::Cycle
	description: Execute RxMAC logic. Called by Fmon fm every fbus cycle
		if Fdrive has been initialized via RxMAC_Init.
	returns: FDRIVE_SUCCESS or return valuse of XACT_set_state_value
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int RxMAC::Cycle()
{
	unsigned int val;
	int retval = FDRIVE_SUCCESS;
	int i;

	PinDelayChains();

	port_sel_dly3 = fps_dly3 & 0x7;

	UpdatePortVectors();					// update whether ports are enabled and ready
	
	DecodeMACSel();							// set rx_mac_sel to selected MAC
	if (rx_mac_sel == my_id){
		if (fm.pr_enabled) XACT_printf("Rx++++++++DRIVING IXBUS mac: %d, port: %d\n", rx_mac_sel, port_sel_dly3);
		retval &= ports[port_sel_dly3].DriveFBus();
			
		// not rx selected, tell other port they are not selected
		// so they can update their state
		for(i=0;i<PORTS_PER_MAC;i++){
			if (i != port_sel_dly3) ports[i].NoDriveFBus(); 
		}
	} else {
		for(i=0;i<PORTS_PER_MAC;i++){
			ports[i].NoDriveFBus(); 
		}
	}

	// simulate bringing bytes in from wire
	for (i=0;i<PORTS_PER_MAC;i++){
		if((ports[i].enabled)&&(ports[i].wire_bytes_per_clock)) ports[i].BufBytesPut();
	}
	if ((fm.pr_enabled)&&(ports[0].wire_bytes_per_clock)){
		ShowBufs();						// print port buffer counts
	}

	DecodeMACRdySel();					// set rx_mac_rdy_sel to selected MAC
	// tell the selected port to drive the rdybus
	if (rx_mac_rdy_sel == my_id){
		val = rdy_vector & enabled_vector;
		if (fm.pr_enabled) XACT_printf("Rx DRIVING RDYBUS mac: %d\n", my_id);
		retval &= XACT_set_state_value(fd.rdybus_bus_handle, &val);
	}

	return(retval);
}


/*-----------------------------------------------------------------
	FDrive::Init
	description: Initialize all RxMACs and set their id.
	returns: FDRIVE_SUCCESS or FDRIVE_FAIL
	uses:
	modifies:
  -----------------------------------------------------------------*/
void FDrive::Init()
{
	int i;

	initialized = 1;

	for(i=0;i<fm.fbus_mac_count;i++){
		macs[i].my_id = i;
		macs[i].Init();
	}

	return;
}

extern "C" void FDrive_Init()
{
	fd.Init();
	return;
}


/*-----------------------------------------------------------------
	FDrive::AddBusClients
	description: Insert user defined states into transactor for f bus clients.
		This must be called from foreign_model_initialize()
	returns: FDRIVE_SUCCESS or return value of XACT_add_bus_client
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int FDrive::AddBusClients()
{
	int retval = FDRIVE_SUCCESS;

	if ((retval &= XACT_add_bus_client("fdat_lo", "fdrive_fdat_lo")) == FALSE){
		XACT_printf("could not add bus client fdat_lo\n");
	}
	if ((retval &= XACT_add_bus_client("fdat_hi", "fdrive_fdat_hi")) == FALSE){
		XACT_printf("could not add bus client fdat_hi\n");
	}
	if ((retval &= XACT_add_bus_client("fbe_lo_l", "fdrive_fbe_lo_l")) == FALSE){
		XACT_printf("could not add bus client fbe_lo_l\n");
	}
	if ((retval &= XACT_add_bus_client("fbe_hi_l", "fdrive_fbe_hi_l")) == FALSE){
		XACT_printf("could not add bus client fbe_hi_l\n");
	}
	if ((retval &= XACT_add_bus_client("rdybus", "fdrive_rdybus")) == FALSE){
		XACT_printf("could not add bus client rdybus\n");
	}
	if ((retval &= XACT_add_bus_client("sop", "fdrive_sop")) == FALSE){
		XACT_printf("could not add bus client sop\n");
	}
	if ((retval &= XACT_add_bus_client("eop", "fdrive_eop")) == FALSE){
		XACT_printf("could not add bus client eop\n");
	}
	if ((retval &= XACT_add_bus_client("rxfail", "fdrive_rxfail")) == FALSE){
		XACT_printf("could not add bus client rxfail\n");
	}
	fdat_lo_bus_handle = XACT_get_handle("fdrive_fdat_lo", -1);
	fdat_hi_bus_handle = XACT_get_handle("fdrive_fdat_hi", -1);
	fbe_lo_l_bus_handle = XACT_get_handle("fdrive_fbe_lo_l", -1);
	fbe_hi_l_bus_handle = XACT_get_handle("fdrive_fbe_hi_l", -1);
	rdybus_bus_handle = XACT_get_handle("fdrive_rdybus", -1);
	sop_bus_handle = XACT_get_handle("fdrive_sop", -1);
	eop_bus_handle = XACT_get_handle("fdrive_eop", -1);
	rxfail_bus_handle = XACT_get_handle("fdrive_rxfail", -1);
	// these 2 are not buses
	fast_rx1_handle = XACT_get_handle("fast_rx1", -1);
	fast_rx2_handle = XACT_get_handle("fast_rx2", -1);

	return(retval);
}


/*-----------------------------------------------------------------
	FDrive::Cycle
	description: Execute all RxMACs logic. Called by Fmon fm every fbus cycle
		if Fdrive has been initialized via FDrive_Init.
	returns: FDRIVE_SUCCESS or XACT fail return value
	uses:
	modifies: rx_stop_count
	see also: xact.h
  -----------------------------------------------------------------*/
int FDrive::Cycle()
{
	int i;
	unsigned int val;
	int retval = FDRIVE_SUCCESS;

	for(i=0;i<fm.fbus_mac_count;i++){
			retval &= macs[i].Cycle();
		}
	if ((rx_stop_count)&&(rx_count == rx_stop_count)){
		val = 1;
		retval &= XACT_set_state_value(fm.sim_halt_handle, &val);
		rx_stop_count = 0;
	}

	return(retval);
}


/*-----------------------------------------------------------------
	FDrive::Go
	description: Go until last EOP of pkcount receive packets
	returns: FDRIVE_SUCCESS or XACT fail return value
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
int FDrive::Go(int pkcount)
{
	unsigned int val;

	rx_count = 0;
	rx_stop_count = pkcount;
	val = 0;
	return(XACT_set_state_value(fm.sim_halt_handle, &val));
}

extern "C" int fd_go(int pkcount)
{
	return(fd.Go(pkcount));
}



/*-----------------------------------------------------------------
	FDrive_fb_help;
	description: list FDrive functions
	returns:
	uses:
	modifies:
  -----------------------------------------------------------------*/
int FDrive_Help()
{
	XACT_printf("\n");
	XACT_printf("***FDrive command-line functions:\n\n");
	XACT_printf("pkpool_init();                   // re-initialize the packet pool (to no packets, start at 0)\n");
	XACT_printf("pkpool_add();                    // add most recent packet created by Packet_Create to PkPool\n");
	XACT_printf("pkpool_show(n);                  // show the nth packet in PkPool\n");
	XACT_printf("pkpool_show_last();              // show last packet in PkPool\n");	
	XACT_printf("rport_on(mac, port);             // enable port to drive packets\n");	
	XACT_printf("rport_off(mac, port, n);         // disable port after n packets\n");	
	XACT_printf("rport_first(mac, port, n);       // set first(next) packet to drive\n");
	XACT_printf("rport_last(mac, port, n);        // if at 1+ this packet, wrap to drive first\n");
	XACT_printf("rport_next(mac, port, n);        // set next packet to drive\n");
	XACT_printf("rport_show(mac, port);           // show current/next packet to be driven\n");
	XACT_printf("rmac_set_rate(mac, Mbits_per_sec, MFClocks_per_sec);      // set RxMAC rate\n");
	XACT_printf("\t if Mbits_per_sec <= 0, set port to always be ready(run at request rate)\n");
	XACT_printf("\t if Mbits_per_sec >= 600, set port 0 of this MAC as a fast port\n");
	XACT_printf("\t e.g. 100Mbit Ether port Mbits_per_sec = 100, typical MFClocks_per_sec = 66\n");
	XACT_printf("rmac_set_rdy_threshold(mac, rdy);    // set buf_bytes at which RxMAC becomes ready\n");
	XACT_printf("fd_go(n);                            // setup to stop transactor after driving n packets\n");
	XACT_printf("\t follow this with transactor go <n cycles> or goto <label>\n");
	XACT_printf("\t after n packets have been driven, sim.halt = 1\n");
	XACT_printf("\t you can then issue fd_go(0), which will set sim.halt = 0\n");
	XACT_printf("\t or fd_go(n), which will prime it to drive n more packets\n");

	return(1);
}



/*-----------------------------------------------------------------
	fdrive_model_initialize
	description: Called by foreign_model_initialize to register console functions.
	returns: 1
	uses:
	modifies:
	see also: xact.h
  -----------------------------------------------------------------*/
extern "C" int fdrive_model_initialize()
{
	if (!XACT_register_console_function("pkpool_init", pkpool_init, 0, 0) ){
            fprintf(stderr,"XACT_register_console_function() pkpool_init failed\n");
			return(0);
	}
	if (!XACT_register_console_function("pkpool_add", pkpool_add, 0, 0) ){
            fprintf(stderr,"XACT_register_console_function() pkpool_add failed\n");
			return(0);
	}
	if (!XACT_register_console_function("pkpool_show", pkpool_show, 0, 1) ){
            fprintf(stderr,"XACT_register_console_function() pkpool_show failed\n");
			return(0);
	}
	if (!XACT_register_console_function("pkpool_show_last", pkpool_show_last, 0, 0) ){
            fprintf(stderr,"XACT_register_console_function() pkpool_show_last failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_on", rport_on, 0, 2) ){
            fprintf(stderr,"XACT_register_console_function() rport_on failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_off", rport_off, 0, 3) ){
            fprintf(stderr,"XACT_register_console_function() rport_off failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_first", rport_first, 0, 3) ){
            fprintf(stderr,"XACT_register_console_function() rport_first failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_last", rport_last, 0, 3) ){
            fprintf(stderr,"XACT_register_console_function() rport_last failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_next", rport_next, 0, 3) ){
            fprintf(stderr,"XACT_register_console_function() rport_next failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rport_show", rport_show, 0, 2) ){
            fprintf(stderr,"XACT_register_console_function() rport_show failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rmac_set_rate", rmac_set_rate, 0, 3) ){
            fprintf(stderr,"XACT_register_console_function() rmac_set_rate failed\n");
			return(0);
	}
	if (!XACT_register_console_function("rmac_set_rdy_threshold", rmac_set_rdy_threshold, 0, 2) ){
            fprintf(stderr,"XACT_register_console_function() rmac_set_rdy_threshold failed\n");
			return(0);
	}
	if (!XACT_register_console_function("fd_go", fd_go, 0, 1) ){
            fprintf(stderr,"XACT_register_console_function() fd_go failed\n");
			return(0);
	}

	fd.AddBusClients();

	return(1);
}