/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/PortManager.cpp,v $
// $Revision: 2 $
// $Date: 5/02/00 3:35p $
//
// Description:
//     This file implements the PortManager class.  It is used by the
//     RouteTable class to associate routing tables with particular ports.
//     This association means that packets coming in on ports will be routed
//     using their associated routing table.
//
//////////////////////////////////////////////////////////////////////////////

#include "rtm_cfg.h"
#include "PortManager.h"
#include <hal_sram.h>
#include "Trie.h"

//////////////////////////////////////////////////////////////////////////////
//  Method: PortManager::PortManager
//
//  Parameters:
//      location        The location of the table of first tries in SRAM
//      numPorts        The number of ports (table entries) for which to manage
//                      their first trie.
//
//  Description:
//      This is the constructor for the PortManager class.  It initializes
//      data members and zeroes the memory for the table of first tries in
//      SRAM.
//
//  Context:
//      Called by RTM_Init() during initialization and passed to a RouteTable
//      instance.
//////////////////////////////////////////////////////////////////////////////
PortManager::PortManager(SRAM_ADDRESS location, 
                         unsigned int numPorts) :
                         m_location(location),
                         m_numPorts(numPorts) {
    for (unsigned int i = 0; i < numPorts; i++) {
        g_pSram->write(location + i, 0);
    }
}

PortManager::~PortManager() {
    // Meant to be blank.  Placeholder for future use.
}

//////////////////////////////////////////////////////////////////////////////
//  Method: PortManager::SetPortFirstTrie
//
//  Parameters:
//      port            The port whose first trie is being set.  Must be
//                      less than m_numPorts.
//      firstTrie       The first trie for this port.
//
//  Description:
//      Changes the first trie to be used in route lookups for packets coming
//      in on a particular ingress port.
//
//  Context:
//      Called by RouteTable::AddPort.
//////////////////////////////////////////////////////////////////////////////
void PortManager::SetPortFirstTrie(PORT_NUMBER port, const Trie& firstTrie) {
    if (port >= m_numPorts) {
        throw InvalidPortException();
    }

    g_pSram->write(m_location + port, firstTrie.GetLocation());
}

//////////////////////////////////////////////////////////////////////////////
//  Method: PortManager::ClearPortFirstTrie
//
//  Parameters:
//      port            The port whose first trie is being set.  Must be
//                      less than m_numPorts.
//
//  Description:
//      Changes the first trie to be used in route lookups for packets
//      coming in on a particular ingress port to NULL.  This will cause 
//      all incoming packets to this port to be dropped.
//
//  Context:
//      Called by RouteTable::RemovePort.
//////////////////////////////////////////////////////////////////////////////
void PortManager::ClearPortFirstTrie(PORT_NUMBER port) {
    if (port >= m_numPorts) {
        throw InvalidPortException();
    }

    g_pSram->write(m_location + port, NULL);
}

//////////////////////////////////////////////////////////////////////////////
//  Method: PortManager::RemoveFirstTrieFromAll
//
//  Parameters:
//      firstTrie       A trie that could be a first trie for one or more ports
//                      being managed by this PortManager.
//
//  Description:
//      Searches the list of ports and sets the first trie to NULL if its
//      current first trie matches the specified one.
//
//  Context:
//      Called by the RouteTable destructor.
//////////////////////////////////////////////////////////////////////////////
void PortManager::RemoveFirstTrieFromAll(const Trie& firstTrie) {
    // Iterate through the ports we control
    for (PORT_NUMBER i = 0; i < m_numPorts; i++) {
        unsigned int firstTrieLoc = 0;
        g_pSram->read(m_location + i, &firstTrieLoc);

        // If the first trie for this port matches the one passed in, set it to
        // NULL
        if (firstTrieLoc == firstTrie.GetLocation()) {
            g_pSram->write(m_location + i, NULL);
        }
    }
}
