/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/RouteTable.cpp,v $
// $Revision: 2 $
// $Date: 5/02/00 3:35p $
//
// Description:
//     This file implements the RouteTable class.  It is used by command line
//     users and Forwarding Element objects to maintain one instance of a 
//     routing table within a Forwarding Element.  It contains methods for 
//     adding and removing ports, adding and removing routes, and displaying
//     debug information.
//
//////////////////////////////////////////////////////////////////////////////


#include "hal_sdram.h"
#include "rtm_cfg.h"
#include "RouteTable.h"
#include <endian.h>
#if (IOSTYLE != HARDWARE)
#include "Xact_API.h"
#include "xact.h"
#endif
#include <strstream>

#ifndef NONAMESPACES
using namespace std;
#endif


//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::RouteTable
//
//  Parameters:
//      trieManager     The trie manager that will control the allocation and
//                      deallocation of trie blocks for this route table.
//      portManager     The port manager that will control the setting and 
//                      clearing of first trie entries for the ports this 
//                      route table is allowed to use.
//
//  Description:
//      This is the constructor for the RouteTable class.
//
//  Context:
//      Called by rt_create_table() or Forwarding Element classes that would
//      create these objects directly.
//////////////////////////////////////////////////////////////////////////////
RouteTable::RouteTable(TrieManager& trieManager, 
                       PortManager& portManager) :
                       m_trieManager(trieManager),
                       m_portManager(portManager),
                       m_firstTrie(*trieManager.AllocateTrie()) {
// EMPTY
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::~RouteTable
//
//  Description:
//      This is the destructor for the RouteTable class.  It removes itself
//      from any ports that might be attached to it.
//
//  Context:
//      Called by rt_delete_table() or Forwarding Element classes that would
//      delete these objects directly.
//////////////////////////////////////////////////////////////////////////////
RouteTable::~RouteTable() {
    // Remove the first trie from any ports it might be associated with.
    m_portManager.RemoveFirstTrieFromAll(m_firstTrie);
    // Free the first trie.  This will recurse the trie structure and free any
    // sub-tries
    m_trieManager.FreeTrie(&m_firstTrie);
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::AddPort
//
//  Parameters:
//      portNum     The port number to associate with this route table.
//
//  Description:
//      This method associates a port with a route table instance.  It does
//      this using its PortManager object.  If this port is already 
//      associated with a route table, it will be disassociated from that
//      route table and associated with this one.
//
//  Context:
//      Called by rt_add_port() or Forwarding Element classes that would
//      add ports directly.
//////////////////////////////////////////////////////////////////////////////
void RouteTable::AddPort(PORT_NUMBER portNum) {
    m_portManager.SetPortFirstTrie(portNum, m_firstTrie);
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::RemovePort
//
//  Parameters:
//      portNum     The port number to disassociate with this route table.
//
//  Description:
//      This method disassociates a port from a route table instance.  It does
//      this using its PortManager object.  Packets received on the specified
//      port will be dropped until this port as added to another route table.
//
//  Context:
//      Called by rt_remove_port() or Forwarding Element classes that would
//      remove ports directly.
//////////////////////////////////////////////////////////////////////////////
void RouteTable::RemovePort(PORT_NUMBER portNum) {
    m_portManager.ClearPortFirstTrie(portNum);
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::CountZeroFields
//
//  Parameters:
//      mask          The netmask being considered.
//      firstNonZero  An out-param for to hold the first non-zero mask field
//
//  Description:
//      This function counts the number of fields in a netmask that are 
//      completely zero.  It also returns the lowest order field that 
//      contains a one.  This information is useful in determining how many
//      trie blocks are needed for a route with a given mask.
//
//  Context:
//      Called by AddRoute() and DeleteRoute().
//////////////////////////////////////////////////////////////////////////////
unsigned int RouteTable::CountZeroFields(NETMASK mask, 
                                         NETMASK_FIELD& firstNonZero) {
    unsigned int numZeroFields = 0;
    NETMASK currentMask = mask;
    for (numZeroFields = 0; numZeroFields < 8; numZeroFields++) { 
        if ((currentMask & 0x0000000f) !=0) {
            currentMask = currentMask & 0x0000000f;
            break;
        }
        currentMask >>= 4;
    }

    firstNonZero = currentMask;
    return numZeroFields;
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::AddRoute
//
//  Parameters:
//      routeInfo   A structure containing the route information.
//
//  Description:
//      Adds the specified route to the route table.  Packets coming in on the
//      ports added to this route table will be routed using this entry if
//      this entry is the most-specific matching entry.
//
//  Context:
//      Called by rt_add(), rt_add_full() or Forwarding Element classes that 
//      would add routes directly.
//////////////////////////////////////////////////////////////////////////////
void RouteTable::AddRoute(const RouteTable::RouteInfo& routeInfo) {
    IP_ADDRESS dest = ntohl(routeInfo.destination);   
    NETMASK mask = ntohl(routeInfo.netmask);

    // first find out how many nibbles of netmask are zero.  This will 
    // determine how many tries are needed for this route.
    NETMASK_FIELD finalMask = 0;
    unsigned int num0nibs = CountZeroFields(mask, finalMask);

    unsigned int currentBit = 28;         // currentBit is the start bit of the current
                                 // four bit field we are working on
    unsigned int endBit = num0nibs * 4;
    // resulting trieCount can be 0-7.  The trie count does not include the
    // first trie
    unsigned int trieCount = ((currentBit - endBit) >>2);

    Trie* pCurrentTrie = &m_firstTrie;
    Trie* tries[8];           // array to hold writes so they can be done in
                              // reverse order
    tries[trieCount] = &m_firstTrie;

    // First, we go through and find/create the tries necessary for this 
    // route.  We don't connect them until we have them all so we can connect
    // them in reverse order.  This prevents a temporary condition in which
    // a packet comes in and not all of the tries are connected properly.
    unsigned int writeCount = 0;
    if (trieCount >= 0){
        // for each trie, get lookup offset
        unsigned int currentIndex = (dest >> currentBit) & 0x0000000f;
        int i;
        for (i = trieCount-1; i >= 0; --i){
            // See if there is already a sub-trie for this route index.
            pCurrentTrie = pCurrentTrie->GetSubTrie(currentIndex);
            if (!pCurrentTrie){   // if not previously existing
                // get a new one if it is 0
                pCurrentTrie = m_trieManager.AllocateTrie();
                ++writeCount;
            }
            tries[i] = pCurrentTrie;         // keep existing route ptr
                        
            currentBit -= 4;
            currentIndex = (dest >> currentBit) & 0x0000000f;
        }


        // write multiple route entry pointers based on mask and key nibble
        // bits into the final trie in the list
        WriteMultipleRtPtr(routeInfo, *pCurrentTrie, currentIndex, finalMask);

        // now connect the tries in reverse order
        currentBit = endBit + 4;
        currentIndex = (dest >> currentBit) & 0x0000000f;

        for (i = 1; i <= writeCount; i++) {
            tries[i]->AddSubTrie(currentIndex, *tries[i - 1]);
            currentBit += 4;
            currentIndex = (dest >> currentBit) & 0x0000000f;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::WriteMultipleRtPtr
//
//  Parameters:
//      routeInfo   A structure containing the route information.
//      trie        The trie to add the specified route to.
//      destField   The four bits of the destination address associated with
//                  this route and trie.
//      maskField   The four bits of the destination mask associated with this
//                  route and trie.
//
//  Description:
//      Adds the specified route to the specified trie using the four bits of
//      the destination address and mask that apply to this trie.
//
//  Context:
//      Called by AddRoute().
//////////////////////////////////////////////////////////////////////////////
void RouteTable::WriteMultipleRtPtr(const RouteTable::RouteInfo &routeInfo,
                                 Trie& trie,               
                                 IP_ADDRESS_FIELD destField,
                                 NETMASK_FIELD maskField) 
                                     throw (OutOfRoutesException)  
{
    // See if there is a route for this route destination and mask already.
    // If there isn't one, we create one.  Either way, we write into one.
    unsigned int index = trie.GetSubRoute(destField, maskField);
    if (!index) {
        index = RTM_AllocRoute();
        if (!index) {
            throw OutOfRoutesException();
        }
        trie.AddSubRoute(index, destField, maskField);
    }

    WriteRouteEntry(index, routeInfo);
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::WriteRouteEntry
//
//  Parameters:
//      routeIndex  The index into the SDRAM table of routes where the route
//                  should be written.
//      routeInfo   A structure containing the route information.
//
//  Description:
//      Writes the specified data into the specified route index in SDRAM.
//
//  Context:
//      Called by WriteMultipleRtPtr().
//////////////////////////////////////////////////////////////////////////////
void RouteTable::WriteRouteEntry(unsigned int routeIndex, 
                                 const RouteTable::RouteInfo& routeInfo)
{
    int sdramAddr = RTM_config.route_table_base+routeIndex;

    g_pSdram->write(sdramAddr, routeInfo.itf);
    g_pSdram->write(sdramAddr+1, routeInfo.nextHopDA.hi32);
    g_pSdram->write(sdramAddr+2, routeInfo.nextHopDA.lo16);
    g_pSdram->write(sdramAddr+3, routeInfo.destination);
    g_pSdram->write(sdramAddr+4, routeInfo.netmask);
    g_pSdram->write(sdramAddr+5, routeInfo.gateway);
    for (unsigned int i = 0; i < SIZE_OF_REST_ROUTE_ENTRY; i++){
        g_pSdram->write(sdramAddr+6+i, routeInfo.rest[i]);
    }
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::DeleteRoute
//
//  Parameters:
//      destination The network-byte-order address of the route to delete.
//      netmask     The network-byte-order address of the netmask for this
//                  route.
//
//  Description:
//      Removes the specified route entry from the route table if it exists.
//
//  Context:
//      Called by rt_delete() or Forwarding Element classes that want to do 
//      this directly.
//////////////////////////////////////////////////////////////////////////////
void RouteTable::DeleteRoute(IP_ADDRESS destination, NETMASK netmask)
{
    IP_ADDRESS_FIELD currentIndex;               // temporary index ip_dest nibble
    IP_ADDRESS dest = ntohl(destination);   
    NETMASK mask = ntohl(netmask);

    // first find out  how many nibbles of netmask are zero
    // resulting mask is the nibble mask to be used in the redundant trie
    NETMASK_FIELD finalMask = 0;
    unsigned int num0nibs = CountZeroFields(mask, finalMask);

    unsigned int currentBit = 28; // currentBit is the start bit of the current
                                  // four bit field we are working on
    unsigned int endBit = num0nibs * 4;
    // resulting trie_count can be 0-7.  The trie count does not include the
    // first trie
    unsigned int trieCount = ((currentBit - endBit) >>2);

    Trie* pCurrentTrie = &m_firstTrie;
    Trie* tries[8];
    tries[0] = &m_firstTrie;

    if (trieCount >= 0){
        // for each trie, get lookup offset
        for (unsigned int i = 1; i <= trieCount; i++){
            currentIndex = (dest >> currentBit) & 0x0000000f;
            // reuse and follow existing one
            pCurrentTrie = pCurrentTrie->GetSubTrie(currentIndex);
            if (!pCurrentTrie){   // if not existing
                throw InvalidRouteException();
            }
            tries[i] = pCurrentTrie;         // keep existing route ptr
                        
            currentBit -= 4;
        }
        currentIndex = (dest >> currentBit) & 0x0000000f;

        // Make sure there is a route for this mask and dest first
        unsigned int routeIndex = pCurrentTrie->GetSubRoute(currentIndex, finalMask);
        if (routeIndex){
            pCurrentTrie->DeleteSubRoute(currentIndex, finalMask);
            RTM_FreeRoute(routeIndex);

            // Now after removing the route we go through from the bottom up
            // and remove the tries from the chain if their population is zero
            for (i = trieCount; i > 0; i--) {
                if (0 == tries[i]->GetPopulation()) {
                    tries[i - 1]->RemoveSubTrie(*tries[i]);
                    m_trieManager.FreeTrie(tries[i]);
                }
                else {
                    break;
                }
            }
        }
    }  //if (trieCount >= 0)
}

//////////////////////////////////////////////////////////////////////////////
//  Method: RouteTable::TrieShow
//
//  Description:
//      Prints the contents of the first trie.  The Trie class uses a bredth-
//      first search to print the contents of all of the connected Tries.
//
//  Context:
//      Called by rt_trie_show()
//////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
void RouteTable::TrieShow() const {
    strstream stream;
    m_firstTrie.PrintContents(stream);
    stream << '\0';

#if (IOSTYLE != HARDWARE)
    XACT_printf(stream.str());
#else
    printf(stream.str());
#endif
}

#endif

