/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/RouteTable.h,v $
// $Revision: 2 $
// $Date: 5/02/00 3:35p $
//
// Description:
//     This file declares the RouteTable class.  It is used by command line
//     users and Forwarding Element objects to maintain one instance of a 
//     routing table within a Forwarding Element.  It contains methods for 
//     adding and removing ports, adding and removing routes, and displaying
//     debug information.
//
//////////////////////////////////////////////////////////////////////////////


#ifndef _ROUTETABLE_H_
#define _ROUTETABLE_H_

#include "rtm_mvr.h"
#include "PortManager.h"
#include "TrieManager.h"
#include "Trie.h"
#include "RTMException.h"

/////////////////////////////////////////////////////////////////////////////
//  Name: RouteTable - Manages an instance of a route table of a Forwarding
//                     Element.
//
//  Description:
//      This class manages the routes and ports associated with a route
//      table.  Adding and removing ports to this route table affects how
//      packets incoming to these ports are routed.  This class is the 
//      primary external interface to the routing table manager.
/////////////////////////////////////////////////////////////////////////////
class RouteTable {
public:
    typedef unsigned int IP_ADDRESS;
    typedef unsigned int NETMASK;
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::RouteTable
    //
    //  Description:
    //      This is the constructor for the RouteTable class.
    //////////////////////////////////////////////////////////////////////////
    RouteTable(TrieManager& trieManager, PortManager& portManager);
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::~RouteTable
    //
    //  Description:
    //      This is the destructor for the RouteTable class.  It removes
    //      itself from any ports that might be attached to it.
    //////////////////////////////////////////////////////////////////////////
    virtual ~RouteTable();

// We have to undef AddPort because a Windows header #defines it to something 
// else.
#undef AddPort
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::AddPort
    //
    //  Description:
    //      This method associates a port with a route table instance.  It 
    //      does this using its PortManager object.  If this port is already 
    //      associated with a route table, it will be disassociated from that
    //      route table and associated with this one.
    //////////////////////////////////////////////////////////////////////////
    void AddPort(PORT_NUMBER portNum)
        throw (InvalidPortException);
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::RemovePort
    //
    //  Description:
    //      This method disassociates a port from a route table instance.  It 
    //      does this using its PortManager object.  Packets received on the 
    //      specified port will be dropped until this port as added to another
    //      route table.
    //////////////////////////////////////////////////////////////////////////
    void RemovePort(PORT_NUMBER portNum)
        throw (InvalidPortException);
    
    /////////////////////////////////////////////////////////////////////////
    //  Name: RouteTable::RouteInfo
    //
    //  Description:
    //      The RouteInfo structure is used to pass around information about
    //      a particular route entry between the methods of the RouteTable
    //      class.
    /////////////////////////////////////////////////////////////////////////
    struct RouteInfo {
        PORT_NUMBER itf;                         // transmit port
        MAC_ADDRESS nextHopDA;                   // next hop MAC address
        IP_ADDRESS destination;                  // ip destination
        NETMASK netmask;                         // ip netmask (net order)
        IP_ADDRESS gateway;                      // ip next hop gateway 
                                                 // (net order)
        unsigned int rest[SIZE_OF_REST_ROUTE_ENTRY];
                                                 // unspecified rest of route 
                                                 // entry
    };

    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::AddRoute
    //
    //  Description:
    //      Adds the specified route to the route table.  Packets coming in on 
    //      the ports added to this route table will be routed using this 
    //      entry if this entry is the most-specific matching entry.
    //////////////////////////////////////////////////////////////////////////
    void AddRoute(const RouteTable::RouteInfo& routeInfo)
        throw (OutOfRoutesException, OutOfTriesException);
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::DeleteRoute
    //
    //  Description:
    //      Removes the specified route entry from the route table if it 
    //      exists.
    //////////////////////////////////////////////////////////////////////////
    void DeleteRoute(IP_ADDRESS dest, NETMASK mask)
        throw (InvalidRouteException);

    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::TrieShow
    //
    //  Description:
    //      Prints the contents of the first trie.  The Trie class uses a 
    //      bredth-first search to print the contents of all of the connected 
    //      Tries to stdout.
    //////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
    void TrieShow() const;
#endif

protected:
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::WriteMultipleRtPtr
    //
    //  Description:
    //      Adds the specified route to the specified trie using the four bits
    //      of the destination address and mask that apply to this trie.
    //////////////////////////////////////////////////////////////////////////
    void WriteMultipleRtPtr(const RouteTable::RouteInfo &routeInfo,
                            Trie& trie,               // which trie block
                            IP_ADDRESS_FIELD dest_field,  // 4 bits of ip dest
                            NETMASK_FIELD mask_field) // 4 bits of netmask
                                throw (OutOfRoutesException);
    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::WriteRouteEntry
    //
    //  Description:
    //      Writes the specified data into the specified route index in SDRAM.
    //////////////////////////////////////////////////////////////////////////
    void WriteRouteEntry(unsigned int route_index, 
                         const RouteTable::RouteInfo& routeInfo);

    //////////////////////////////////////////////////////////////////////////
    //  Method: RouteTable::CountZeroFields
    //
    //  Description:
    //      This function counts the number of fields in a netmask that are 
    //      completely zero.  It also returns the lowest order field that 
    //      contains a one.  This information is useful in determining how
    //      many trie blocks are needed for a route with a given mask.
    //////////////////////////////////////////////////////////////////////////
    static unsigned int CountZeroFields(NETMASK mask, 
                                        NETMASK_FIELD& firstNonZero);

    TrieManager& m_trieManager;
    PortManager& m_portManager;
    Trie&        m_firstTrie;
};

#endif //_ROUTETABLE_H_
