/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/Trie.cpp,v $
// $Revision: 2 $
// $Date: 5/02/00 3:35p $
//
// Description:
//      This file implements the Trie class.  A trie is a data structure that
//      facilitates a fast longest-prefix match lookup.  It is basically an 
//      array of entries that are indexed by four bits of an IP address.  The
//      entries contain a pointer to an SDRAM route entry, a next trie 
//      structure to continue the lookup in, or both.  There are also some 
//      side structures associated with the trie that are used for correctly
//      adding and removing route entries.
//
//      The trie entries in the table occupy 32 bits.  The upper 16 bits is an
//      index into the SDRAM route table, or zero if there is no sub-route for
//      the entry.  The lower 16 bits is an index into the SRAM area used for
//      tries, indicating the next trie to be used in the lookup, or zero to 
//      indicate there is no next trie for this entry.
//////////////////////////////////////////////////////////////////////////////

#include "Trie.h"
#include "rtm_cfg.h"

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::Trie
//
//  Parameters:
//      offset          The offset into SRAM that all tries are located at
//      sramLocation    The SRAM location for this particular trie, 
//                      relative to the offset
//
//  Description:
//      This is the constructor for the Trie class.  It initializes data
//      members and side structures.
//
//  Context:
//      Called by TrieManager::AllocateTrie().
//////////////////////////////////////////////////////////////////////////
Trie::Trie(SRAM_ADDRESS offset, 
           unsigned int sramLocation) :
           m_offset(offset),
           m_sramLocation(sramLocation),
           m_population(0) {
    for (int i = 0; i < 16; i++) {
        m_mask[i] = 0;
        m_nextTrie[i] = NULL;
        g_pSram->write(offset + sramLocation + i, 0);
    }
    for (i = 0; i < 8; i++) {
        m_prefix3[i] = 0;
    }
    for (i = 0; i < 4; i++) {
        m_prefix2[i] = 0;
    }
    for (i = 0; i < 2; i++) {
        m_prefix1[i] = 0;
    }
}

Trie::~Trie() {
    // Meant to be blank.  Placeholder for future use.
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::DeleteSubRoute
//
//  Parameters:
//      dest        The four bits of the destination address for this
//                  particular trie.
//      mask        The four bits of the mask for this trie.
//
//  Description:
//      This method removes a sub-route stored with this trie.  The sub-
//      route is specified using the correct four bits of the destination
//      address and mask.
//
//  Context:
//      Called by RouteTable::DeleteRoute().
//////////////////////////////////////////////////////////////////////////
void Trie::DeleteSubRoute(IP_ADDRESS_FIELD dest, NETMASK_FIELD mask) 
                throw (InvalidRouteException) {
    IP_ADDRESS_FIELD maskedIndex = dest & mask;

    unsigned int memContents = 0;
    unsigned int prevRoute = 0;
    NETMASK_FIELD prevMask = 0;

    // First, find an existing route in this trie that matches this one but is
    // less specific.  This will replace whatever entries we remove later.
    if (m_prefix3[dest>>1] && (mask > 0xe)){
        prevRoute = m_prefix3[dest>>1];
        prevMask = 0xe;
    } else if (m_prefix2[dest>>2] && (mask > 0xc)){
        prevRoute = m_prefix2[dest>>2];
        prevMask = 0xc;
    } else if (m_prefix1[dest>>3] && (mask > 0x8)){
        prevRoute = m_prefix1[dest>>3];
        prevMask = 0x8;
    }

    // Now, clear the prefix entries for this mask length.  Before we do that,
    // we determine if there is a route installed with this dest and mask.
    // If there isn't, throw an exception.  numEntries holds the number of 
    // entries that must be overwritten in this operation.
    unsigned int numEntries = 1;
    switch (mask) {
        case 0xe:
            if (!m_prefix3[dest >> 1]) {
                throw InvalidRouteException();
            }
            m_prefix3[dest >> 1] = 0;
            numEntries = 2;
            break;
        case 0xc:
            if (!m_prefix2[dest >> 2]) {
                throw InvalidRouteException();
            }
            m_prefix2[dest >> 2] = 0;
            numEntries = 4;
            break;
        case 0x8:
            if (!m_prefix1[dest >> 3]) {
                throw InvalidRouteException();
            }
            m_prefix1[dest >> 3] = 0;
            numEntries = 8;
            break;
    }

    // Now we overwrite the route pointers with the previously determined less
    // specific route.
    for (int i = 0; i < numEntries; i++) {
        IP_ADDRESS_FIELD index = maskedIndex + i;

        // Only write this route entry if this route is more specific than
        // one that is already in this trie slot
        if (mask >= m_mask[index]){
            // Read the trie entry that is there now and just change the
            // route_index in the upper 16 bits
            unsigned int memContents = 0;
            g_pSram->read(m_offset + m_sramLocation + index, &memContents);
            memContents = (prevRoute<<16) | (memContents & 0xffff);
            g_pSram->write(m_offset + m_sramLocation + index, memContents);
            m_mask[index] = prevMask;
        }
    }

    m_population--;

    return;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::AddSubRoute
//
//  Parameters:
//      routeIndex  The index into SDRAM where the route information is
//                  stored.
//      dest        The four bits of the destination address for this
//                  particular trie.
//      mask        The four bits of the mask for this trie.
//
//  Description:
//      This method adds a sub-route stored with this trie.  The sub-
//      route is specified using the correct four bits of the destination
//      address and mask, as well as the index into SDRAM where the route.
//      information is stored.
//
//  Context:
//      Called by RouteTable::AddRoute().
//////////////////////////////////////////////////////////////////////////
void Trie::AddSubRoute(unsigned int routeIndex,
                       IP_ADDRESS_FIELD dest,
                       NETMASK_FIELD mask) {
    IP_ADDRESS_FIELD maskedIndex = dest & mask;

    // First we set the correct prefix values for the specified mask length.
    // We also check to see if there is already an equivalent route installed.
    // If there is, we decrement the population before adding this route. 
    unsigned int count = 1;
    switch (mask) {
        case 0xe:
            count = 2;
            if (m_prefix3[dest >> 1]) {
                m_population--;
            }
            m_prefix3[dest >> 1] = routeIndex;
            break;
        case 0xc:
            count = 4;
            if (m_prefix2[dest >> 2]) {
                m_population--;
            }
            m_prefix2[dest >> 2] = routeIndex;
            break;
        case 0x8:
            count = 8;
            if (m_prefix1[dest >> 3]) {
                m_population--;
            }
            m_prefix1[dest >> 3] = routeIndex;
            break;
        default:
            break;
    }

    // Now we write the correct route indexes into the trie entries
    for (int i = 0; i<count; i++) {
        IP_ADDRESS_FIELD index = maskedIndex+i;

        if (mask >= m_mask[index]){
            // Only write this route entry if this route is more specific than
            // one that is already in this trie slot.  Also write it if it is
            // the same length as the original.  This is needed if one route
            // replaces another.

            // Read the trie slot that is there now and just change the
            // route_index in the upper 16 bits

            unsigned int memContents = 0;
            g_pSram->read(m_offset + m_sramLocation + index, &memContents);
            memContents = (routeIndex<<16) | (memContents & 0xffff);
            g_pSram->write(m_offset + m_sramLocation + index, memContents);
            m_mask[index] = mask;
        }

    }

    m_population++;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::GetSubRoute
//
//  Parameters:
//      dest        The four bits of the destination address for this
//                  particular trie.
//      mask        The four bits of the mask for this trie.
//
//  Description:
//      Gets the SDRAM index of the route that matches this destination 
//      and mask.  Only finds exact matches, not most-specific matches.
//
//  Returns:
//      The SDRAM index of the sub-route or 0 if the route does not exist.
//
//  Context:
//      Called by RouteTable::DeleteRoute() and 
//      RouteTable::WriteMultipleRtPtr.
//////////////////////////////////////////////////////////////////////////
unsigned int Trie::GetSubRoute(IP_ADDRESS_FIELD dest, NETMASK_FIELD mask) const {
    unsigned int routeIndex = 0;

    switch (mask) {
    case 0xf:
        // If the mask is 0xf, the route is stored in SRAM.
        if (m_mask[dest] == 0xf) {
            routeIndex = GetSubRoute(dest);
        }
        break;
    case 0xe:
        routeIndex = m_prefix3[dest >> 1];
        break;
    case 0xc:
        routeIndex = m_prefix2[dest >> 2];
        break;
    case 0x8:
        routeIndex = m_prefix1[dest >> 3];
        break;
    }

    return routeIndex;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::GetSubRoute
//
//  Parameters:
//      dest        The four bits of the destination address for this
//                  particular trie.
//
//  Description:
//      Gets the SDRAM index of the most specific route that matches this 
//      destination and mask.  
//
//  Returns:
//      The SDRAM index of the sub-route or 0 if the route does not exist.
//
//  Context:
//      Called by TrieManager::FreeTrie.
//////////////////////////////////////////////////////////////////////////
unsigned int Trie::GetSubRoute(IP_ADDRESS_FIELD dest) const {
    unsigned int routeIndex = 0;

    g_pSram->read(m_offset + m_sramLocation + dest, &routeIndex);

    return routeIndex >> 16;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::GetSubTrie
//
//  Parameters:
//      dest        The four bits of the destination address for this
//                  particular trie.
//
//  Description:
//      Gets a pointer to the trie associated with the destination 
//      address.  Can be used to traverse connected tries.
//
//  Returns:
//      A pointer to the sub-trie or NULL if there is not one.
//
//  Context:
//      Called by TrieManager::FreeTrie, RouteTable::AddRoute, and
//      RouteTable::DeleteRoute.
//////////////////////////////////////////////////////////////////////////
Trie* Trie::GetSubTrie(IP_ADDRESS_FIELD dest) const {
    if (dest < 16) {
        return m_nextTrie[dest];
    }
    else {
        return 0;
    }
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::AddSubTrie
//
//  Parameters:
//      dest        The four bits of the destination address for this
//                  particular trie.
//      subTrie     The sub-trie to associate with these four bits of
//                  address.
//
//  Description:
//      Adds a sub-trie to this trie with the specified destination 
//      address.
//
//  Context:
//      Called by RouteTable::AddRoute.
//////////////////////////////////////////////////////////////////////////
void Trie::AddSubTrie(IP_ADDRESS_FIELD dest, Trie& subTrie) {
    if (m_nextTrie[dest]) {
        m_population--;
    }

    m_nextTrie[dest] = &subTrie;
    
    // We have to preserve the route index if there is one
    unsigned int routeIndex = 0;
    g_pSram->read(m_offset + m_sramLocation + dest, &routeIndex);
    routeIndex |= subTrie.m_sramLocation;
    g_pSram->write(m_offset + m_sramLocation + dest, routeIndex);
    
    m_population++;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::RemoveSubTrie
//
//  Parameters:
//      subTrie     The sub-trie to associate with these four bits of
//                  address.
//
//  Description:
//      Removes the specified sub-trie from this trie.  This method will
//      Search for the sub-trie in all of of this trie's entries and
//      remove the specified one if it is found.
//
//  Context:
//      Called by RouteTable::DeleteRoute and TrieManager::FreeTrie
//////////////////////////////////////////////////////////////////////////
void Trie::RemoveSubTrie(const Trie& subTrie) {
    IP_ADDRESS_FIELD dest = 0;

    // First, figure out the destination bits for this sub-trie
    for (dest = 0; dest < 16; dest++) {
        if (m_nextTrie[dest] == &subTrie) {
            m_nextTrie[dest] = NULL;
            
            unsigned int routeIndex = 0;
            g_pSram->read(m_offset + m_sramLocation + dest, &routeIndex);
            routeIndex &= 0xffff0000;
            g_pSram->write(m_offset + m_sramLocation + dest, routeIndex);

            m_population--;
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::GetLocation
//
//  Description:
//      Retrieves the location of this trie in an SRAM table.  This 
//      location is relative to the offset specified in the constructor.
//////////////////////////////////////////////////////////////////////////
unsigned int Trie::GetLocation() const {
    return m_sramLocation;
}

//////////////////////////////////////////////////////////////////////////
//  Method: Trie::GetPopulation
//
//  Description:
//      Returns the total number of sub-tries and sub-routes in this trie.
//////////////////////////////////////////////////////////////////////////
short Trie::GetPopulation() const {
    return m_population;
}

//////////////////////////////////////////////////////////////////////////////
//  Method: Trie::PrintContents
//
//  Description:
//      Prints the contents of this trie.  This uses a bredth-first search to 
//      print the contents of all of the connected Tries.
//
//  Context:
//      Called by RouteTable::TrieShow()
//////////////////////////////////////////////////////////////////////////////
void Trie::PrintContents(ostream& str) const {
    str << hex << "Trie " << this << " Contents:" << endl;
    str << "-------------------------" << endl;
    
    for(int i = 0; i <= 15; i++) {
        unsigned int routeIndex;
        g_pSram->read(m_offset + m_sramLocation + i, &routeIndex);
        str << "#" << i << ": Sub-Trie: " << m_nextTrie[i] << 
        "     Sub-Route: " << (void*)(routeIndex>>16) << endl;
    }

    for(i = 0; i <= 15; i++) {
        if(m_nextTrie[i]) {
            m_nextTrie[i]->PrintContents(str);
        }
    }
}
