/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/Trie.h,v $
// $Revision: 2 $
// $Date: 5/03/00 9:07a $
//
// Description:
//      This file declares the Trie class.  A trie is a data structure that
//      facilitates a fast longest-prefix match lookup.  It is basically an 
//      array of entries that are indexed by four bits of an IP address.  The
//      entries contain a pointer to an SDRAM route entry, a next trie 
//      structure to continue the lookup in, or both.  There are also some 
//      side structures associated with the trie that are used for correctly
//      adding and removing route entries.
//
//      The trie entries in the table occupy 32 bits.  The upper 16 bits is an
//      index into the SDRAM route table, or zero if there is no sub-route for
//      the entry.  The lower 16 bits is an index into the SRAM area used for
//      tries, indicating the next trie to be used in the lookup, or zero to 
//      indicate there is no next trie for this entry.
//////////////////////////////////////////////////////////////////////////////


#ifndef _TRIE_H_
#define _TRIE_H_

#include <hal_sram.h>
#include "TrieManager.h"
#include <iostream>
#include "RTMException.h"

#ifndef NONAMESPACES
using namespace std;
#endif

/////////////////////////////////////////////////////////////////////////////
//  Name: Trie - This class implements a piece of a route lookup algorithm.
//
//  Description:
//      This class implements a trie structure.  A trie is a data structure 
//      that facilitates a fast longest-prefix match lookup.  It is basically
//      an array of entries that are indexed by four bits of an IP address.  
//      The entries contain a pointer to an SDRAM route entry, a next trie 
//      structure to continue the lookup in, or both.  There are also some 
//      side structures associated with the trie that are used for correctly
//      adding and removing route entries.
/////////////////////////////////////////////////////////////////////////////
class Trie {
public:
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::DeleteSubRoute
    //
    //  Description:
    //      This method removes a sub-route stored with this trie.  The sub-
    //      route is specified using the correct four bits of the destination
    //      address and mask.
    //////////////////////////////////////////////////////////////////////////
    void DeleteSubRoute(IP_ADDRESS_FIELD dest, NETMASK_FIELD mask)
        throw (InvalidRouteException);
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::AddSubRoute
    //
    //  Description:
    //      This method adds a sub-route stored with this trie.  The sub-
    //      route is specified using the correct four bits of the destination
    //      address and mask, as well as the index into SDRAM where the route.
    //      information is stored.
    //////////////////////////////////////////////////////////////////////////
    void AddSubRoute(unsigned int routeIndex,
                     IP_ADDRESS_FIELD dest,
                     NETMASK_FIELD mask);

    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::GetSubRoute
    //
    //  Description:
    //      Gets the SDRAM index of the route that matches this destination 
    //      and mask.  Only finds exact matches, not most-specific matches.
    //////////////////////////////////////////////////////////////////////////
    unsigned int GetSubRoute(IP_ADDRESS_FIELD dest, NETMASK_FIELD mask) const;
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::GetSubRoute
    //
    //  Description:
    //      Gets the SDRAM index of the most specific route that matches this 
    //      destination and mask.  
    //////////////////////////////////////////////////////////////////////////
    unsigned int GetSubRoute(IP_ADDRESS_FIELD dest) const;

    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::GetSubTrie
    //
    //  Description:
    //      Gets a pointer to the trie associated with the destination 
    //      address.  Can be used to traverse connected tries.
    //////////////////////////////////////////////////////////////////////////
    Trie* GetSubTrie(IP_ADDRESS_FIELD dest) const;
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::AddSubTrie
    //
    //  Description:
    //      Adds a sub-trie to this trie with the specified destination 
    //      address.
    //////////////////////////////////////////////////////////////////////////
    void AddSubTrie(IP_ADDRESS_FIELD dest, Trie& subTrie);
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::RemoveSubTrie
    //
    //  Description:
    //      Removes the specified sub-trie from this trie.  This method will
    //      Search for the sub-trie in all of of this trie's entries and
    //      remove the specified one if it is found.
    //////////////////////////////////////////////////////////////////////////
    void RemoveSubTrie(const Trie& subTrie);

    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::GetLocation
    //
    //  Description:
    //      Retrieves the location of this trie in an SRAM table.  This 
    //      location is relative to the offset specified in the constructor.
    //////////////////////////////////////////////////////////////////////////
    unsigned int GetLocation() const;
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::GetPopulation
    //
    //  Description:
    //      Returns the total number of sub-tries and sub-routes in this trie.
    //////////////////////////////////////////////////////////////////////////
    short GetPopulation() const;

    //////////////////////////////////////////////////////////////////////////////
    //  Method: Trie::PrintContents
    //
    //  Description:
    //      Prints the contents of this trie.  This uses a bredth-first search to 
    //      print the contents of all of the connected Tries.
    //////////////////////////////////////////////////////////////////////////////
    void PrintContents(ostream& str) const;

protected:
    friend class TrieManager;

    // The constructor and destructor are protected because the TrieManager
    // will handle allocation and deallocation
    //////////////////////////////////////////////////////////////////////////
    //  Method: Trie::Trie
    //
    //  Description:
    //      This is the constructor for the Trie class.  It initializes data
    //      members and side structures.
    //////////////////////////////////////////////////////////////////////////
    Trie(SRAM_ADDRESS offset, unsigned int sramLocation);
    virtual ~Trie();

    SRAM_ADDRESS m_offset;
    unsigned int m_sramLocation;
    short m_population;

    // The m_mask array holds the four-bit masks for the route entries stored 
    // in the trie entries.  This is mainly used to determine if an entry
    // should be overwritten when another matching entry is added.
    unsigned char  m_mask[16];

    // The m_nextTrie array holds pointers to the sub-tries for particular
    // addresses.  This is used when adding and deleting routes, as well as
    // printing sub-tries.
    Trie* m_nextTrie[16];

    // The m_prefixn arrays are weird.  They hold route indexes, and are 
    // indexed by the masked bits of the destination address of a route.  The
    // 'n' in m_prefixn indicates how many ones in the mask.  For
    // example, a route index for a route with a destination address of 0xc
    // and a mask of 0xe would be stored in m_prefix3[6] (6 == 0xc >> 1).
    // A route index for a route with a destination address of 0x4 and a mask
    // of 0xc would be stored in m_prefix2[1].
    unsigned int m_prefix3[8];               
    unsigned int m_prefix2[4];
    unsigned int m_prefix1[2];
};

#endif //_TRIE_H_
