/* ---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *----------------------------------------------------------------------------*/

//////////////////////////////////////////////////////////////////////////////
// $Source: /mnt/bigpine/src/oem/IXP1200/CoreLibs/rtm-mvr/Attic/TrieManager.cpp,v $
// $Revision: 2 $
// $Date: 5/02/00 3:35p $
//
// Description:
//      This file implements the TrieManager class.  This class acts as a
//      Factory for Trie objects.  When a TrieManager is created, it is
//      assigned a block of memory from which it can allocate trie blocks.
//////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
// This warning is given because the MS compiler cannot handle sybols greater
// than 255 characters.  STL symbols, when expanded, are greater than 255
// characters.
#pragma warning (disable: 4786)
#endif

#include "TrieManager.h"
#include "rtm_cfg.h"
#include "rtm_mvr.h"


#include <set>
#ifndef NONAMESPACES
using namespace std;
#endif

//////////////////////////////////////////////////////////////////////////
//  Method: TrieManager::TrieManager
//
//  Parameters:
//      trieBase        The offset into SRAM that all tries are located at
//      maxNumTries     The maximum number of tries this manager can have
//                      allocated at any time + 1.
//
//  Description:
//      This is the constructor for the TrieManager class.  It 
//      initializes data members and zeroes the SRAM to be used by tries.
//      It also creates a freelist of tries to use for allocating and 
//      freeing them.
//
//  Context:
//      Called by RTM_Init().
//////////////////////////////////////////////////////////////////////////
TrieManager::TrieManager(SRAM_ADDRESS trieBase, unsigned int maxNumTries) :
                m_trieBase(trieBase),
                m_numTries(maxNumTries) {

    for (unsigned int i=0; i < SIZE_OF_TRIE_SET * maxNumTries ; i++) {
        g_pSram->write(trieBase+i, 0);
    }

    // Create the freelist entries.  This is just a list of free Trie blocks.
    // They are put in the list as indexes into a large array of SRAM
    // Index 0 is not used because it is written in Trie entries to indicate 
    // no next trie.
    for (i = 1; i < maxNumTries; i++) {
        m_trieFreelist.push_back(i * 16);
    }
}

TrieManager::~TrieManager() {
    //EMPTY
}

//////////////////////////////////////////////////////////////////////////
//  Method: TrieManager::GetTrieBase
//
//  Description:
//      Retrieves the base address for all tries managed by this 
//      TrieManager
//////////////////////////////////////////////////////////////////////////
SRAM_ADDRESS TrieManager::GetTrieBase() {
    return m_trieBase;
}

//////////////////////////////////////////////////////////////////////////
//  Method: TrieManager::AllocateTrie
//
//  Description:
//      This method creates a Trie and assigns a section of SRAM to it.
//
//  Returns:
//      A pointer to the new Trie or NULL if no more memory.
//
//  Context:
//      Called by RouteTable::RouteTable and RouteTable::AddRoute().
//////////////////////////////////////////////////////////////////////////
Trie* TrieManager::AllocateTrie() throw (OutOfTriesException) {
    // This method finds a free sram pointer on the freelist and initializes
    // a new Trie object with this pointer

    if (m_trieFreelist.empty()) {
        throw OutOfTriesException();
    }
    unsigned int newTrie = m_trieFreelist.back();

    Trie* pTrie = new Trie(m_trieBase, newTrie);
    if (!pTrie) {
        throw OutOfTriesException();
    }

    m_trieFreelist.pop_back();
    m_numTries--;
    return pTrie;
}

//////////////////////////////////////////////////////////////////////////
//  Method: TrieManager::FreeTrie
//
//  Parameters:
//      pTrie       The trie block to be freed.
//
//  Description:
//      This method destroys a trie block and makes it's SRAM available 
//      for future use.
//
//  Context:
//      Called by RouteTable::~RouteTable and RouteTable::DeleteRoute().
//////////////////////////////////////////////////////////////////////////
void TrieManager::FreeTrie(Trie* pTrie)
{
    // Now we take the sram pointer from the Trie object and put that pointer
    // on the freelist.  Then, we delete the Trie object

    if (!pTrie) {
        return;
    }

    // Before deleting the Trie object, we enumerate and free all of its sub-
    // tries and sub-routes.  We don't do this in the Trie destructor because 
    // the tries don't know which TrieManager owns them.
    for (int i = 0; i < 16; i++) {
        set<unsigned int> foundRoutes;
        unsigned int route = pTrie->GetSubRoute(i);
        if (route) {
            // Since we only want to call RTM_FreeRoute once, we put the
            // pointer in a map andmake sure we haven't already freed it.
            if(foundRoutes.find(i) == foundRoutes.end()) {
                foundRoutes.insert(i);
                RTM_FreeRoute(route);   // found in rtm.cpp
            }
        }

        Trie* pSubTrie = pTrie->GetSubTrie(i);
        if (pSubTrie) {
            pTrie->RemoveSubTrie(*pSubTrie);
            FreeTrie(pSubTrie);
        }
    }

    m_numTries++;
    m_trieFreelist.push_back(pTrie->m_sramLocation);
    delete pTrie;
}

//////////////////////////////////////////////////////////////////////////
//  Method: TrieManager::GetNumFreeTries
//
//  Description:
//      Retrieves the number of free tries.
//////////////////////////////////////////////////////////////////////////
unsigned int TrieManager::GetNumFreeTries() {
    return m_numTries;
}

