/* uclo_pci.cpp
 *
 *------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *------------------------------------------------------------
 * system: SA1200
 * subsystem: Ucode Dispatcher Library
 * author: Henry Qian & Desmond Johnson 11/02/98
 * revisions:
 *
 */
#ifdef UCLO_PCI
#include "thread.h"
#include "uclo.h"
#include "libd.h"
#include "pciCommAPI.h"

typedef struct LibD_LibHandle_S {
   void *objHandle;
   int   sourceId;
} LibD_LibHandle_T;

extern "C" int LibD_PutMsgQue(void *msg);

static int UcLoPci_InitDone = 0;
static LibD_LibHandle_T *LibdHandle = NULL;


/*-----------------------------------------------------------------------------
   Function: 	UcLoPci_LoadUof
   Description: Load UOF to remote system via PCI.
   Returns:		UCLO_SUCCESS, UCLO_FAILURE
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int UcLoPci_LoadUof(unsigned int destPciDevNum, int loadUofFuncId,
			unsigned int shrmpSrcId, unsigned int shrmpDestId,
			unsigned int *uofBufPtr, unsigned int uofSize)
{
	ShrimpMessage msg;

	if(!uofBufPtr) return(UCLO_FAILURE);

   // construct a one-way-async-request shrimp message
   SET_SYNC(msg.word_0, ASYNC);
   SET_ACK_REQ(msg.word_0, REQ);
   SET_FORMAT(msg.word_0, 3);
   SET_ERROR(msg.word_0, MSG_OK);
   SET_TRANS_ID(msg.word_0, 0);
   SET_SOURCE_ID(msg.word_0, shrmpSrcId);
   SET_DEST_ID(msg.word_0, shrmpDestId);
   SET_FUNC_ID(msg.word_0, loadUofFuncId);

   msg.word_1 = Pci_DeviceID;		// arg1; the PCI device number
   msg.word_2 = uofBufPtr;			// arg2: pointer to the UOF object
   msg.word_3 = uofSize;			// arg3: the size of the UOF object

   // send shrimp message to the server
   if(Pci_PutMsg(destPciDevNum, &msg, PCI_CALL_UNBLOCK) != PCI_SUCCESS){
	   return(UCLO_FAILURE);
   }

   return(UCLO_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	UcLoPci_GetMsg
   Description: Receive SHRIMP message from PCI and placed it on the dispatcher's
				queue.
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void UcLoPci_GetMsg(void *noArg)
{
   ShrimpMessage message;

   while(1){
	   if(Pci_GetMsg(&message, PCI_CALL_BLOCK) == PCI_SUCCESS){
			LibD_RouteMsg((void *)&message);
	   }
   }
   return;
}



/*-----------------------------------------------------------------------------
   Function: 	UcLoPci_LoadObj
   Description: Load uof object via PCI, and write the images to their assigned
				micro-engines.  The input arguments are long-word in the order
				of:- PCI device number, PCI buffer address,	and the size of the
				PCI buffer.  There is no output argument.
   Returns:		UCLO_SUCCESS, UCLO_ERROR
   Uses:		Global variable LibdHandle -- not thread-safe
   Modifies:
-----------------------------------------------------------------------------*/
int UcLoPci_LoadObj(void *inArgs, void *outArgs)
{
   int deviceNo, uofSize, readOnly = 0;
   char *uofPciAddr, *bufPtr = NULL;
   static int bufSize = 0;
   unsigned int *args = (unsigned int *)inArgs;

   if(!LibdHandle) return(UCLO_FAILURE);	/* lib not ready */

   /* get the pci args -- deviceNo, bufferAddr, bufferSize */
   deviceNo = args[0];
   uofPciAddr = (char *)args[1];
   uofSize = args[2];

   if(LibdHandle->objHandle){
	  /* remove previous handle */
	  UcLo_DeleObj(LibdHandle->objHandle);
	  LibdHandle->objHandle = NULL;
   }

   /* alloc only if previous buffer size is less than current */
   if(uofSize > bufSize){
	  if(bufSize && bufPtr) free(bufPtr);
      if(!(bufPtr = (char *)malloc(uofSize))){
		 bufSize = 0;
	     return(UCLO_FAILURE);
	  }
	  bufSize = uofSize;
   }

   Pci_ReadMemBlock(deviceNo, uofPciAddr, bufPtr, uofSize);

   if(UcLo_MapObjAddr(&LibdHandle->objHandle, bufPtr, uofSize, readOnly) != UCLO_SUCCESS)
      return(UCLO_FAILURE);
   if(UcLo_WriteUimageAll(LibdHandle->objHandle) != UCLO_SUCCESS) return(UCLO_FAILURE);

   return(UCLO_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	UcLoPci_Init
   Description: Initialize the library to enable the loading of *.uof object
				via PCI.  This function should only be called once.
   Returns:		UCLO_SUCCESS, UCLO_BADARG, or UCLO_ERROR
   Uses:		Global variable LibdHandle -- not thread-safe
   Modifies:
-----------------------------------------------------------------------------*/
int UcLoPci_Init(void *libdHandle, int loadFuncId)
{
   if(UcLoPci_InitDone == 1) return(UCLO_SUCCESS);
   if(!libdHandle) return(UCLO_BADARG);

   Pci_Init();									// init the PCI library
   UcLo_InitLib();								// initialize the loader library

   LibdHandle = (LibD_LibHandle_T *)libdHandle;		// save to global variable
   if(LibD_RegisterCallback(LibdHandle, loadFuncId, UcLoPci_LoadObj, 12, 0) != LIBD_SUCCESS)
      return(UCLO_FAILURE);

   /* Thread to handle PCI incoming messages. */
   Thread_Create(HIGH_PRIORITY, (threadFunc_T)UcLoPci_GetMsg, (void *)NULL);
   UcLoPci_InitDone = 1;
   return(UCLO_SUCCESS);
}

#endif				/* UCLO_PCI */
