/* SimExtDLL.cpp

    This file contains sample code demonstrating how to create a dynamic-link
    library (DLL) for a simulation extension to be used in conjunction with the 
    transactor DLL when running the SA1200 Developer Workbench.  The Workbench
    provides a menu item in which you can select a DLL to be used
    as an extension to the transactor model.  In your DLL you must provide
    the exported function GetForeignModelFunctions(), which the Workbench calls
    to get the addresses of the four functions that the transactor calls to
    interact with your foreign model.  The Workbench registers these functions 
    with the DLL version of the transactor.

    Since your DLL will be running in conjunction with the DLL version of the
    transactor, you must include xact_dll.h instead of xact.h in your source
    files, if needed, and you must link against xact_dll.lib.
 * 
 * ---------------------------------------------------------------------	
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ---------------------------------------------------------------------
 *
 * system: SA1200
 * subsystem: Simulation Extension DLL example
 * author: rdm
 * revisions:
 * $History: SimExtDLL.cpp $
 * 
 * *****************  Version 9  *****************
 * User: Lhui         Date: 10/21/99   Time: 2:50p
 * Updated in $/dev_1200/Share/Samples/RemoteSimInterface
 * 
 * *****************  Version 8  *****************
 * User: Lhui         Date: 10/21/99   Time: 2:18p
 * Updated in $/dev_1200/Share/Samples/RemoteSimInterface
 * Add new reset callback argument to GetForeignModelFunctions()
 * 
 * *****************  Version 7  *****************
 * User: Qabuild      Date: 7/26/99    Time: 6:48p
 * Updated in $/dev_1200/Share/Samples/RemoteSimInterface
 * Cleanup
 * 
 * *****************  Version 5  *****************
 * User: Qabuild      Date: 5/21/99    Time: 7:58p
 * Updated in $/dev_1200/Share/Samples/RemoteSimInterface
 * 
 * *****************  Version 3  *****************
 * User: Lhui         Date: 4/30/99    Time: 6:03p
 * Updated in $/dev_1200/Share/Samples/RemoteSimInterface
 * 
 * *****************  Version 1  *****************
 * User: Lhui         Date: 4/09/99    Time: 5:38p
 * Created in $/development/Samples/RemoteSimExt
 * 
 * *****************  Version 6  *****************
 * User: C-tclancy    Date: 3/23/99    Time: 4:16p
 * Updated in $/development/Samples/SimExtensionDLL
 * update COPYRIGHT
 * 
 * *****************  Version 5  *****************
 * User: Lhui         Date: 10/12/98   Time: 4:16p
 * Updated in $/base_level3/Samples/SimExtensionDLL
 * Add comments
 * 
 * *****************  Version 4  *****************
 * User: Lhui         Date: 10/12/98   Time: 9:51a
 * Updated in $/base_level3/Samples/SimExtensionDLL
 * Add ..\common\sim.c into the project
 * 
 * *****************  Version 3  *****************
 * User: Lhui         Date: 9/18/98    Time: 9:55p
 * Updated in $/base_level3/Samples/SimExtensionDLL
 * Add support for remote transactor calls
 * 
 * *****************  Version 2  *****************
 * User: Lhui         Date: 8/07/98    Time: 8:00p
 * Updated in $/base_level3/Samples/SimExtensionDLL
 * Add remote supports
 * 
 * *****************  Version 1  *****************
 * User: Rdmurato     Date: 8/03/98    Time: 10:27a
 * Created in $/base_level3/Samples/SimExtensionDLL
 * Initial version
 *
 */

static const char __copyright[]=
        "Copyright (c) Intel Corporation, 1999. All rights reserved";

static const char __SimExtDLL_version__ []=
        "\nFILEINFO: $Workfile: SimExtDLL.cpp $ $Date: 10/21/99 2:50p $ $Revision: 9 $\n";

// You must include xact_dll.h, NOT xact.h in order to link correctly against
// the DLL version of the transactor.
//
#include "xact_dll.h"
#include <OncRpc32.h>
#include <pmap_clnt.h>
#include "remote_services.h"
#include "RemoteSample.h" 
#include "rssimext.h"

/*-----------------------------------------------------------------
    foreign_model_initialize

    This function will be called to initialize the foreign model after
    the transactor "init" command has successfully executed.  Returning 0
    will result in a transactor error.
    
     returns: 
        uses: 
    modifies: 
*/
int foreign_model_initialize()
{
    if (!XACT_register_console_function("RemoteInit", RemoteInit, 1, 1) ){
            fprintf(stderr,"XACT_register_console_function() failed\n");
			return(0);
	}
 

	return(1);
}

/*-----------------------------------------------------------------
    foreign_model_pre_sim

    This function will be called prior to each transactor simulation event.
    It can be used to deposit state values into the transactor model prior
    to simulating the next event.  Returning 0 results in an error.
    
     returns: 
        uses: 
    modifies: 
*/
int foreign_model_pre_sim()
{

    rs_foreign_model_pre_sim();
    XACT_output_to_console("foreign_model_pre_sim called\n");
    return(1);
}

/*-----------------------------------------------------------------
    foreign_model_post_sim

    This function will be called subsequent to each transactor simulation event.
    It can be used to query transactor simulation state, in order to copy it
    into the foreign model simulator.
    
     returns: 
        uses: 
    modifies: 
*/
int foreign_model_post_sim()
{
    rs_foreign_model_post_sim();
    XACT_output_to_console("foreign_model_post_sim called\n");
	return(1);

}

/*-----------------------------------------------------------------
    foreign_model_exit

    This function will be called just prior to exiting the simulator.
    The routine allows the foreign model to clean up, close files, etc
    before shutting down the program.

     returns: 
        uses: 
    modifies: 
*/
int foreign_model_exit()
{
    rs_foreign_model_exit();
    XACT_output_to_console("foreign_model_exit called\n");
    return(1);

}

/*-----------------------------------------------------------------
    foreign_model_reset

    This function will be called just prior to the simulator executing 
    a sim_reset command. The routine allows the foreign model to perform
    required reset actions.

     returns: 
        uses: 
    modifies: 
*/
int
foreign_model_reset()
{
    XACT_output_to_console("foreign_model_reset called\n");
    return(1);
}


/*-----------------------------------------------------------------
    GetForeignModelFunctions

    This function is exported as the sole entry point into the DLL version
    of this package.  The Developer Workbench calls it in order to get the
    foreign model entry points for the transactor.  The Workbench thens
    registers these entry points with the transactor.
    
     returns: 
        uses: 
    modifies: 
*/
#if defined(__cplusplus)
extern "C"      // the exported function must be of generic C type (not CPP),
                // so that its name doesn't get adorned.
#endif
__declspec(dllexport) void __cdecl
GetForeignModelFunctions(
    int (**ForeignModelInitialize)(),
    int (**ForeignModelPreSim)(),
    int (**ForeignModelPostSim)(),
    int (**ForeignModelExit)(),
    int (**ForeignModelReset)())

{

    *ForeignModelInitialize = foreign_model_initialize;
    *ForeignModelPreSim = foreign_model_pre_sim;
    *ForeignModelPostSim = foreign_model_post_sim;
	*ForeignModelReset = foreign_model_reset;
    *ForeignModelExit = foreign_model_exit;

}
#if defined(__cplusplus)
extern "C"      // the exported function must be of generic C type (not CPP),
                // so that its name doesn't get adorned.
#endif
__declspec(dllexport) void __cdecl
SetRemoteNodeName(char *Server){
//"1099" is the remote program ID which is defined in RsSimExt.h 
RemoteInit(Server, 1099);

}

