/* scomm.cpp
 * Serial Communication Library
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: SA1200
 * subsystem: Serial Communication Library
 * author: drj Mar 16, 99 Initial version
 * revisions:
 * $History: scomm.cpp $
 * 
 * *****************  Version 25  *****************
 * User: Ewalker      Date: 9/20/00    Time: 9:58a
 * Updated in $/dev_1200/Share/scomm
 * Performance cleanup
 * 
 * *****************  Version 24  *****************
 * User: Drjohns1     Date: 8/30/00    Time: 2:06p
 * Updated in $/dev_1200/Share/scomm
 * 
 * *****************  Version 23  *****************
 * User: Qabuild      Date: 8/24/00    Time: 5:35p
 * Updated in $/dev_1200/Share/scomm
 * Changed Scomm_InstallDev return type
 * 
 * *****************  Version 22  *****************
 * User: Drjohns1     Date: 6/29/00    Time: 3:18p
 * Updated in $/dev_1200/Share/scomm
 * Updated for HPC and to handle multiple devices
 * 
 * *****************  Version 21  *****************
 * User: Drjohns1     Date: 4/04/00    Time: 6:43p
 * Updated in $/dev_1200/Share/scomm
 * OS type check and cleanup
 * 
 * 
 * ---------------------------------------------------------------------
 */
#include "core_io.h"
#include "utils.h"
#include "scomm.h"
#include "scomm_buff.h"
#include "scomm_llc.h"
#include "ostype.h"


static const char __copyright[] =
"Copyright Intel Corp, 1999. All rights reserved";


#define SCOMM_MAX_CONN	32

enum {
	SCOMM_SNDBUF = 0,
    SCOMM_RCVBUF,
	SCOMM_TRLBUF,
	SCOMM_PUTREQ,
	SCOMM_GETREQ,
	SCOMM_DATAVAIL,
	SCOMM_SETERR,
	SCOMM_GETERR,
	SCOMM_SETXOFF,
	SCOMM_GETXOFF,
	SCOMM_SPACEAVAIL,
	SCOMM_RESET
};

#define VERIFY_SCOMMID(sid) if(sid < 0 || sid >= SCOMM_MAX_CONN) return(SCOMM_INVLDID)

const int DEF_BUFSIZE =  2048;			// buffer size per connection
const int MAX_COMDEVS = 4;				// number of comm devices
const int MAX_PKT_SIZE = DEF_BUFSIZE;	// max packet size
const int PKT_HDR_SIZE = 2;
const int CHKSUM_SIZE = 2;
const int MIN_PKT_SIZE = PKT_HDR_SIZE + CHKSUM_SIZE;
const int SCOMM_XMTRETRYS = 5;

ComDev *ComDevices[MAX_COMDEVS];
enum {PKT_DATA = 0, PKT_CTRL, PKT_BCAST, PKT_INFO};


typedef struct{
   int			sConnId;
   short		portNum;
   short		err;
   int			devId;
   char			xmit;
   Buffer		*recvBuf;
}commResource_t;

// packet format:
// checksum(2 bytes), payload(>= 0 bytes), ctl(srcPort<15:9>, destPort<8:2>, type<1:0>)

enum {CTL_ERROR = 0, CTL_FLOW, CTL_REQTOSEND, CTL_RESET};
enum {CTL_XOFF = 0, CTL_XON};

typedef struct {
	char pkgType;
	char valType;
	char value[6];
}SComm_CtlVal_T;

const int CTL_PKT_SIZE	= sizeof(SComm_CtlVal_T);

static commResource_t CommConn[SCOMM_MAX_CONN];
static sem_T ScommSem;
static int Init = 0;


#if (OS == UCOS)
extern "C" unsigned int OSTimeGet(void);
static time_t timex(time_t *t)
{
	time_t tmp_t;
	tmp_t = (OSTimeGet() / CLOCKS_PER_SEC);
	if (t != NULL) *t = tmp_t;
	return(tmp_t);
}
#else
inline static time_t timex(time_t *t){return(time(t));}
#endif


//dj: debug stuff below:
static unsigned int PktDropped=0, PktDataRcv=0, PktDataSnd=0, PktCtlRcv=0, PktCtlSnd=0;
static unsigned int PktChkSumErr=0, PktBadRcv=0, PktUnkwnRcv=0;

extern "C" int memLog(char *format,...);
extern void llcStats(void);

/*-----------------------------------------------------------------------------
   Function: 	scommStats
   Description: Print statistical information -- for debugging
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void scommStats(void)
{
   SComm_Print("\nScommStats: PktChkSumErr=%d, PktDataSnd=%d, PktDataRcv=%d, PktCtlSnd=%d,\n\
	    PktCtlRcv=%d, PktUnkwnRcv=%d, PktDropped=%d\n",
	   PktChkSumErr, PktDataSnd, PktDataRcv, PktCtlSnd, PktCtlRcv, PktUnkwnRcv, PktDropped);

   llcStats();
   return;
}


//dj: end of debug stuff



/*-----------------------------------------------------------------------------
   Function: 	SComm_FindPort
   Description: Locate a connection port
   Returns:		An index to the associated connection or an invalid indicator
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_FindPort(int portNum)
{
   for(int i = 0; i < SCOMM_MAX_CONN; i++)
      if((CommConn[i].sConnId == i) && (CommConn[i].portNum == portNum)){
         return(i);
      }
   return(SCOMM_INVLDID);
}


/*-----------------------------------------------------------------------------
   Function: 	SComm_Ctl
   Description: Set and read comm device parameters
   Returns:		
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Ctl(int sConnId, int request, char buffType, void *buffer, unsigned int length, unsigned int timeout)
{
   int err;
   if((unsigned int)sConnId >= SCOMM_MAX_CONN)return(SCOMM_INVLDID);
   switch(request){
      case SCOMM_PUTREQ:
         switch(buffType){
            case SCOMM_RCVBUF: return(CommConn[sConnId].recvBuf->put((char *)buffer, length, timeout));
		 }
	  case SCOMM_GETREQ:
         switch(buffType){
            case SCOMM_RCVBUF: return(CommConn[sConnId].recvBuf->get((char *)buffer, length, timeout));
		 }
	  case SCOMM_DATAVAIL:
         switch(buffType){
            case SCOMM_RCVBUF: return(CommConn[sConnId].recvBuf->getLen());
		 }
	  case SCOMM_SPACEAVAIL:
         switch(buffType){
            case SCOMM_RCVBUF: return(CommConn[sConnId].recvBuf->spaceLeft());
		 }
	  case SCOMM_SETERR:
		  CommConn[sConnId].err = buffType;
          return(SCOMM_SUCCESS);
	  case SCOMM_GETERR:
		  // read and clear error
		  err = CommConn[sConnId].err;
		  CommConn[sConnId].err = SCOMM_SUCCESS;
		  return(err);
	  case SCOMM_RESET:
		  CommConn[sConnId].err = 0;
		  CommConn[sConnId].xmit = CTL_XON;
		  CommConn[sConnId].recvBuf->clear();
          return(SCOMM_SUCCESS);
	  case SCOMM_SETXOFF:
		  CommConn[sConnId].xmit = buffType;
          return(SCOMM_SUCCESS);
	  case SCOMM_GETXOFF:
		  return(CommConn[sConnId].xmit);
   }
   return(SCOMM_FAIL);
}



/*-----------------------------------------------------------------------------
   Function: 	sendCtlMsg
   Description: Send a control message to the destination port
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void sendCtlMsg(int devNum, int srcPort, int destPort, SComm_CtlVal_T *ctl)
{
   unsigned short checksum=0;
   unsigned int pktHdr, i;
   char *cPtr;
   iovec_t	iovec[3];

   pktHdr = ((unsigned int)srcPort & 0x7f) << 9;		// source port
   pktHdr |= ((unsigned int)destPort & 0x7f) << 2;		// destination port
   pktHdr |= PKT_CTRL & 0x3;							// type
   iovec[0].base = (char *)&pktHdr;
   iovec[0].len =  PKT_HDR_SIZE;

   iovec[1].base = (char *)ctl;
   iovec[1].len =  sizeof(SComm_CtlVal_T);

   // calculate the checksum for the header and the payload
   for(cPtr = (char *)&pktHdr, i = 0; i < PKT_HDR_SIZE; i++)
	   checksum = (short) Crc_Calc(checksum, *cPtr++);
   for(cPtr = (char *)ctl, i = 0; i < sizeof(SComm_CtlVal_T); i++)
	   checksum = (short) Crc_Calc(checksum, *cPtr++);
   iovec[2].base = (char *)&checksum;
   iovec[2].len =  CHKSUM_SIZE;

   ComDevices[devNum]->sendPacket(iovec, 3);
   PktCtlSnd++;

   return;
}


/*-----------------------------------------------------------------------------
   Function: 	sendError
   Description: Send an error to the destination port
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void sendError(int devNum, int destPort, int errNum)
{
   SComm_CtlVal_T ctl;

   ctl.pkgType = PKT_CTRL;
   ctl.valType = CTL_ERROR;
   ctl.value[0] = errNum;
   sendCtlMsg(devNum, SCOMM_INVLDPORT, destPort, &ctl);
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	sendFlowCtl
   Description: Send a flow control message to the destination port
   Returns:	
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void sendFlowCtl(int devNum, int destPort, int onOff)
{
   SComm_CtlVal_T ctl;

   ctl.pkgType = PKT_CTRL;
   ctl.valType = CTL_FLOW;
   ctl.value[0] = onOff;
   sendCtlMsg(devNum, SCOMM_INVLDPORT, destPort, &ctl);
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	reqToSendCtl
   Description: Send a control message requesting to data to destPort.  This
				is generally called if the xmit bit has been set, and the sender
				wants to determine if space is avaiable on the destination's buffer.
				If the requested is granted, then the xmit bit on the source port
				will be cleared.
   Returns:	
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void reqToSendCtl(int devNum, int srcPort, int destPort, int dataSize)
{
   SComm_CtlVal_T ctl;

   ctl.pkgType = PKT_CTRL;
   ctl.valType = CTL_REQTOSEND;
   memmove(&ctl.value, &dataSize, sizeof(int));
   sendCtlMsg(devNum, srcPort, destPort, &ctl);
   return;
}



/*-----------------------------------------------------------------------------
   Function: 	recvPacket
   Description: Receive packet and perform appropriate action based on
				the packet type.
   Returns:	
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void recvPacket(int devId, char *packet, int packetLen)
{
   unsigned short	checksum, calcChecksum;
   unsigned char	ctl, srcPort, destPort;
   int				sConnId, pktLen, i;

   // examine packet to determine its type and destination
   ctl = packet[0] & 0x3;


   // ignore packet if the size is wrong or we are not intialized
   if((packetLen < MIN_PKT_SIZE) || (packetLen >= MAX_PKT_SIZE) || (!Init)){
       PktBadRcv++;
	   PktDropped++;
	  return;
   }

   // perform checksum
   memmove(&checksum, &packet[packetLen - CHKSUM_SIZE], CHKSUM_SIZE);
   calcChecksum = (short)Crc_StrChecksum(packet, packetLen - CHKSUM_SIZE);
   if(calcChecksum != checksum){
 	   // just drop it -- can't trust the data to send back an error message
	  PktChkSumErr++;
	  PktDropped++;
	  return;
   }

   unsigned int pktHdr;
   memmove(&pktHdr, packet, PKT_HDR_SIZE);
   srcPort = (pktHdr >> 9) & 0x7f;		// source port
   destPort = (pktHdr >> 2) & 0x7f;		// destination port

   switch(ctl){
      case PKT_DATA:
         Sem_Wait(ScommSem, SEM_WAITFOREVER);
	     PktDataRcv++;
         if((sConnId = SComm_FindPort(destPort)) == SCOMM_INVLDID){
			// return error and drop packet
            sendError(devId, srcPort, SCOMM_INVLDPORT);
			PktDropped++;
		    Sem_Post(ScommSem);
			break;
		 }
		 pktLen = packetLen - MIN_PKT_SIZE;
         Sem_Post(ScommSem);

		 // place the entire packet on the receive buffer
		 for(i = 0; i < SCOMM_XMTRETRYS; i++){
			 if(CommConn[sConnId].recvBuf->lockIfSpace(pktLen, (1000/*ms*/ / DELAYMS_RES)) >= pktLen){
				CommConn[sConnId].recvBuf->putNoLock(&packet[PKT_HDR_SIZE], pktLen);
                CommConn[sConnId].recvBuf->unLockBuf();
			   break;
			 }
		 }
         if(i >= SCOMM_XMTRETRYS){
			PktDropped++;
		    sendFlowCtl(devId, srcPort, CTL_XOFF);		// send xmit off to src
            sendError(devId, srcPort, SCOMM_XMITFAIL);	// it failed
		 }
		 break;

      case PKT_CTRL:{
	     PktCtlRcv++;
		 SComm_CtlVal_T *ctlVal = (SComm_CtlVal_T *)&packet[PKT_HDR_SIZE];
		 if(!ctlVal) return;

		 switch(ctlVal->valType){
	        case CTL_ERROR:
               Sem_Wait(ScommSem, SEM_WAITFOREVER);
               if((sConnId = SComm_FindPort(destPort)) == SCOMM_INVLDID){
                  Sem_Post(ScommSem);
			      PktDropped++;
				  return;
			   }
		       SComm_Ctl(sConnId, SCOMM_SETERR, ctlVal->value[0], NULL, 0, 0);
               Sem_Post(ScommSem);
			break;
	        case CTL_RESET:
               Sem_Wait(ScommSem, SEM_WAITFOREVER);
               if((sConnId = SComm_FindPort(destPort)) == SCOMM_INVLDID){
                  Sem_Post(ScommSem);
			      PktDropped++;
				  return;
			   }
		       SComm_Ctl(sConnId, SCOMM_RESET, ctlVal->value[0], NULL, 0, 0);
               Sem_Post(ScommSem);
			break;
	        case CTL_FLOW:
               Sem_Wait(ScommSem, SEM_WAITFOREVER);
               if((sConnId = SComm_FindPort(destPort)) == SCOMM_INVLDID){
                  Sem_Post(ScommSem);
			      PktDropped++;
				  return;
			   }
		       SComm_Ctl(sConnId, SCOMM_SETXOFF, ctlVal->value[0], NULL, 0, 0);
               Sem_Post(ScommSem);
			break;
	        case CTL_REQTOSEND:
			   int dataSize;
               Sem_Wait(ScommSem, SEM_WAITFOREVER);
               if((sConnId = SComm_FindPort(destPort)) == SCOMM_INVLDID){
                  Sem_Post(ScommSem);
			      PktDropped++;
				  return;
			   }
			   memmove(&dataSize, ctlVal->value, 2);
		       if(dataSize <= SComm_Ctl(sConnId, SCOMM_SPACEAVAIL, SCOMM_RCVBUF, NULL, 0, 0))
		          sendFlowCtl(devId, srcPort, CTL_XON);		// send xmit on to src
               Sem_Post(ScommSem);
			break;

		 }
		 break;
		}
      case PKT_BCAST:
      case PKT_INFO:
		  break;
	  default:
		  PktUnkwnRcv++;
	      PktDropped++;
		  break;
   }
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	SComm_InstallDev
   Description: Install a serial communication device.
   Returns:	
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_InstallDev(char *DevName)
{
   int i, devNum = -1, status = SCOMM_SUCCESS;

   //printf("SComm_InstallDev: creating and waiting on sem\n"); fflush(stdout);
   if(Init == 0){
      ScommSem = Sem_Init(1);
      Sem_Wait(ScommSem, SEM_WAITFOREVER);
      for(i = 0; i < SCOMM_MAX_CONN; i++){
	     CommConn[i].devId = NULL_DEV;
         CommConn[i].sConnId = -1;
         CommConn[i].portNum = -1;
         CommConn[i].err = 0;
         CommConn[i].xmit = CTL_XON;
         CommConn[i].recvBuf = NULL;
	  }
	}
    else Sem_Wait(ScommSem, SEM_WAITFOREVER);


   // check is device is already installed
   utl_strLowcase(DevName);
   for(i = 0; i < MAX_COMDEVS; i++){
	  if(ComDevices[i]){
		 if(!strcmp(ComDevices[i]->devNam, DevName)){
            Sem_Post(ScommSem);
			return(SCOMM_SUCCESS);
		 }
	  }else devNum = i;
   }

   // no more device slot
   if(devNum == -1){
      Sem_Post(ScommSem);
	  return(SCOMM_FAIL);
   }

   //printf("SComm_InstallDev: device %s\n", DevName); fflush(stdout);
   ComDevices[devNum] = new ComDev(DevName, devNum, recvPacket);
   if(!ComDevices[devNum]->device) status = SCOMM_FAIL;
   Sem_Post(ScommSem);
   Init = 1;


   return(status);
}

/*-----------------------------------------------------------------------------
   Function: 	SComm_Open
   Description: Open a connection on the device and bind the port number
   Returns:		Connection ID
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Open(char *DevName, int portNum)
{
   if(!Init) return(SCOMM_UNAVAIL);
   Sem_Wait(ScommSem, SEM_WAITFOREVER);

   // check is device is installed
   int devNum;
   utl_strLowcase(DevName);

   for(devNum = 0; devNum < MAX_COMDEVS; devNum++)
	  if(ComDevices[devNum] && !strcmp(ComDevices[devNum]->devNam, DevName)) break;

   if(devNum >= MAX_COMDEVS){
      Sem_Post(ScommSem);
	  return(SCOMM_INVLDDEV);
   }

   // check if port number is unique
   if(SComm_FindPort(portNum) != SCOMM_INVLDID){
      Sem_Post(ScommSem);
	  return(SCOMM_DUPLPORT);
   }

   for(int i = 0; i < SCOMM_MAX_CONN; i++)
      if(CommConn[i].sConnId == -1){
         CommConn[i].sConnId = i;
         CommConn[i].devId = devNum;
	     CommConn[i].portNum = portNum;
		 CommConn[i].err = 0;
		 CommConn[i].xmit = CTL_XON;
	     CommConn[i].recvBuf = new Buffer(DEF_BUFSIZE);
         Sem_Post(ScommSem);
         return(i);
	  }
   Sem_Post(ScommSem);
   return(SCOMM_INVLDID);
}

/*-----------------------------------------------------------------------------
   Function: 	SComm_Close
   Description: Close the connection identified by the sConnId parameter.
   Returns:		SCOMM_SUCCESS
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Close(int sConnId)
{
   VERIFY_SCOMMID(sConnId);

   Sem_Wait(ScommSem, SEM_WAITFOREVER);
   CommConn[sConnId].sConnId = -1;
   CommConn[sConnId].devId = NULL_DEV;
   CommConn[sConnId].portNum = -1;
   CommConn[sConnId].xmit = CTL_XOFF;
   delete(CommConn[sConnId].recvBuf);
   Sem_Post(ScommSem);

   return(SCOMM_SUCCESS);
}


/*-----------------------------------------------------------------------------
   Function: 	SComm_Recv
   Description: Read a specified number char from the connection within the
				approx. number of seconds specified by the timeout parameter.
   Returns:		The number of char read
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Recv(int sConnId, void *buffer, unsigned int length, unsigned int timeout)
{
   VERIFY_SCOMMID(sConnId);
   return(SComm_Ctl(sConnId, SCOMM_GETREQ, SCOMM_RCVBUF, buffer, length, timeout));
}


/*-----------------------------------------------------------------------------
   Function: 	SComm_Send
   Description: Packet data and send to destination port within the
				approx. number of seconds specified by the timeout parameter.
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Send(int sConnId, int destPort, void *buffer, unsigned int bufLen, unsigned int timeout)
{
   int s = 0;
   unsigned int i;
   unsigned short checksum=0;
   iovec_t	iovec[3];
   char	*cPtr;
   time_t endTime, currentTime;

   VERIFY_SCOMMID(sConnId);

   // drop it if the comm link is closed
   if(CommConn[sConnId].devId == NULL_DEV) return(SCOMM_INVLDID);
   if(bufLen > MAX_PKT_SIZE) return(SCOMM_BADPARAM);

   // if xmit==off wait or return
   if(CommConn[sConnId].xmit == CTL_XOFF){
	  if(timeout == 0) return(0);

      //ask request to send bufLen bytes
	  reqToSendCtl(CommConn[sConnId].devId, CommConn[sConnId].portNum, destPort, bufLen);

      endTime = timex(&currentTime) + timeout;

      while((CommConn[sConnId].xmit == CTL_XOFF) && (endTime > timex(&currentTime))){
    	 delayTick(1);
    	 reqToSendCtl(CommConn[sConnId].devId, CommConn[sConnId].portNum, destPort, bufLen);
	  }
	  if(CommConn[sConnId].xmit == CTL_XOFF) return(0);		// timed out
   }


   // packet format:
   // checksum(2 bytes), payload(>= 0 bytes), ctl(srcPort<15:9>, destPort<8:2>, type<1:0>)

   unsigned int pktHdr;
   pktHdr = (CommConn[sConnId].portNum & 0x7f) << 9;	// source port
   pktHdr |= (destPort  & 0x7f) << 2;					// destination port
   pktHdr |= PKT_DATA & 0x3;							// type
   iovec[0].base = (char *)&pktHdr;
   iovec[0].len = PKT_HDR_SIZE;

   // copy the payload
   iovec[1].base = (char *)buffer;
   iovec[1].len = bufLen;
 

   // calculate the checksum for the header and the payload
   for(cPtr = (char *)&pktHdr, i = 0; i < PKT_HDR_SIZE; i++)
	   checksum = (short) Crc_Calc(checksum, *cPtr++);
   for(cPtr = (char *)buffer, i = 0; i < bufLen; i++)
	   checksum = (short) Crc_Calc(checksum, *cPtr++);

   iovec[2].base = (char *)&checksum;
   iovec[2].len = CHKSUM_SIZE;

   PktDataSnd++;

   if((s = ComDevices[CommConn[sConnId].devId]->sendPacket(iovec, 3)) != (int)(bufLen + MIN_PKT_SIZE))
	   SComm_Print("SComm_Send: Error writing %d bytes to comm device, wrote %d bytes\n", 
	           (bufLen + MIN_PKT_SIZE), s);
   
   return((s - MIN_PKT_SIZE));
}

/*-----------------------------------------------------------------------------
   Function: 	SComm_DataLen
   Description: Determine the number of bytes that are currently in the
				receive-buffer.
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_DataLen(int sConnId)
{
   return(SComm_Ctl(sConnId, SCOMM_DATAVAIL, SCOMM_RCVBUF, NULL, NULL, 0));
}



/*-----------------------------------------------------------------------------
   Function: 	SComm_GetError
   Description: Read and clear the associated connection error
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_GetError(int sConnId)
{
   return(SComm_Ctl(sConnId, SCOMM_GETERR, 0, NULL, NULL, 0));
}

/*-----------------------------------------------------------------------------
   Function: 	SComm_Reset
   Description: Reset the comm port
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_Reset(int sConnId)
{
   return(SComm_Ctl(sConnId, SCOMM_RESET, 0, NULL, NULL, 0));
}


/*-----------------------------------------------------------------------------
   Function: 	SComm_SendReset
   Description: Send a reset message to the destination port
   Returns:	
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int SComm_SendReset(int sConnId, int destPort)
{
   SComm_CtlVal_T ctl;
   int devNum;

   if((unsigned int)sConnId >= SCOMM_MAX_CONN)return(SCOMM_INVLDID);
   Sem_Wait(ScommSem, SEM_WAITFOREVER);
   if(CommConn[sConnId].devId == NULL_DEV){
      Sem_Post(ScommSem);
	  return(SCOMM_INVLDID);
   }
   devNum = CommConn[sConnId].devId;
   Sem_Post(ScommSem);

   ctl.pkgType = PKT_CTRL;
   ctl.valType = CTL_RESET;
   ctl.value[0] = 0;
   sendCtlMsg(devNum, SCOMM_INVLDPORT, destPort, &ctl);
   return(SCOMM_SUCCESS);
}
