/* scomm_buff.cpp
 * Serial Communication Library
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: SA1200
 * subsystem: Serial Communication Library
 * author: drj Mar 16, 99 Initial version
 * revisions:
 * $History: scomm_buff.cpp $
 * 
 * *****************  Version 18  *****************
 * User: Ewalker      Date: 9/20/00    Time: 9:58a
 * Updated in $/dev_1200/Share/scomm
 * Performance cleanup
 * 
 * *****************  Version 17  *****************
 * User: Drjohns1     Date: 9/08/00    Time: 5:10p
 * Updated in $/dev_1200/Share/scomm
 * Re-implemented timex
 * 
 * *****************  Version 16  *****************
 * User: Drjohns1     Date: 9/07/00    Time: 3:38p
 * Updated in $/dev_1200/Share/scomm
 * Corrected RTC_CNTR register address
 * 
 * *****************  Version 15  *****************
 * User: Drjohns1     Date: 4/17/00    Time: 12:24p
 * Updated in $/dev_1200/Share/scomm
 * Included ostype.h
 * 
 * ---------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "ostype.h"
#include "delay.h"
#include "scomm_buff.h"

static const char __copyright[] =
"Copyright Intel Corp, 1999. All rights reserved";

#define MaxModSemDly (1000/*ms*/ / DELAYMS_RES)   // max system-ticks to wait for semaphore
#define MaxMsgSemDly (1000/*ms*/ / DELAYMS_RES)   // max system-ticks to wait for message
#define MaxSpaceSemDly (1000/*ms*/ / DELAYMS_RES) // max system-ticks to wait for space


#if (OS == UCOS)
extern "C" unsigned int OSTimeGet(void);
static time_t timex(time_t *t)
{
	time_t tmp_t;
	tmp_t = (OSTimeGet() / CLOCKS_PER_SEC);
	if (t != NULL) *t = tmp_t;
	return(tmp_t);
}
#else
inline static time_t timex(time_t *t){return(time(t));}
#endif


/*-----------------------------------------------------------------------------
   Function: 	FastBuf::FastBuf
   Description: Buffer Class constructor
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
FastBuf::FastBuf(unsigned int bufSize)
{
   pBufSize = bufSize;
   pBufPtr = new char[pBufSize];
   pDataLen = pRdIndx = pWrIndx = 0;
   pWaitForSpaceSem = Sem_Init(1);
   pWaitForMsgSem = Sem_Init(1);
   Sem_Wait(pWaitForSpaceSem, SEM_WAITFOREVER);
   Sem_Wait(pWaitForMsgSem, SEM_WAITFOREVER);
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	Buffer::Buffer
   Description: Buffer Class constructor
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
Buffer::Buffer(unsigned int bufSize)
{
   size = bufSize;
   bufPtr = new char[size];
   len = head = tail = waitingForMsg = waitingForSpace = 0;
   modSem = Sem_Init(1);
   waitForSpaceSem = Sem_Init(1);
   waitForMsgSem = Sem_Init(1);
   Sem_Wait(waitForSpaceSem, SEM_WAITFOREVER);
   Sem_Wait(waitForMsgSem, SEM_WAITFOREVER);
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	Buffer::~Buffer
   Description: Buffer Class destructor
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
Buffer::~Buffer(){
   Sem_Destroy(modSem);
   Sem_Destroy(waitForSpaceSem);
   Sem_Destroy(waitForMsgSem);
   if(bufPtr){
	   delete(bufPtr);
	   bufPtr = NULL;
   }
}


/*-----------------------------------------------------------------------------
   Function: 	Buffer::lockIfSpace
   Description: Lock the buffer if the available space is at least the requested
				dataLen. 
   Returns:		Return the number of available space in the buffer, or -1 if
				failed to obtain the semaphore.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int Buffer::lockIfSpace(unsigned int dataLen, int timeout /* ticks */)
{
   unsigned int left;

   // wait no longer than MaxModSemDly ticks for the semaphore
   if(Sem_Wait(modSem, timeout) != SEM_SUCCESS) return(-1);

   // check if the sufficient space is available
   if((left = (size - len)) < dataLen){
	   Sem_Post(modSem);
   }
   return(left);
}


/*-----------------------------------------------------------------------------
   Function: 	Buffer::put
   Description: Attempt to write dataLen number of chars to the buffer.  A
				partial copy of the data may occur if the entire data cannot
				be copied, due to unavailable space, within approx. timeout
				seconds.
 
   Returns:		The number of chars copied
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int Buffer::put(char *dataPtr, unsigned int dataLen, unsigned int timeout)
{
   unsigned int bytesWritten = 0, bytesToWrite;
   time_t endTime, currentTime;

   endTime = timex(&currentTime) + timeout;

   while((bytesWritten < dataLen) && (endTime > timex(&currentTime))){      
	  // wait no longer than MaxModSemDly ticks for the semaphore
	  if(Sem_Wait(modSem, MaxModSemDly) != SEM_SUCCESS) continue;

	  // check if space is available
      if((size - len) == 0){
         waitingForSpace++;
		 Sem_Post(modSem);				// release modification semaphore

		 // wait until space is available
		 if(Sem_Wait(waitForSpaceSem, MaxSpaceSemDly) != SEM_SUCCESS){
		    waitingForSpace--;
			continue;
		 }

		 if(Sem_Wait(modSem, MaxModSemDly) != SEM_SUCCESS){
		    waitingForSpace--;
			continue;
		 }
		 waitingForSpace--;
	  }

	  // copy to available space
      bytesToWrite = (dataLen - bytesWritten);
      bytesWritten += putNoLock(&dataPtr[bytesWritten], bytesToWrite);
	  if(waitingForMsg) Sem_Post(waitForMsgSem);
      Sem_Post(modSem);
   }

   return(bytesWritten);
}



/*-----------------------------------------------------------------------------
   Function: 	Buffer::get
   Description: Attempt to read dataLen number of chars from the buffer.  A
				partial read of the data may occur if the entire data cannot
				be read, due to unavailable data, within the approx. timeout
				seconds. 
   Returns:		The number of chars read
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int Buffer::get(char *dataPtr, unsigned int dataLen, unsigned int timeout)
{
   unsigned int bytesRead = 0, bytesToRead;
   time_t endTime, currentTime;

   timex(&endTime);
   endTime += timeout;

   while((bytesRead < dataLen) && (endTime > timex(&currentTime))){      
	  if(Sem_Wait(modSem, MaxModSemDly) != SEM_SUCCESS) continue;

      // check if enough data is available
      if(len == 0){

		 // wait for available data
         waitingForMsg++;
		 Sem_Post(modSem);					// release modification semaphore

		 // wait until msg arrive
		 if(Sem_Wait(waitForMsgSem, MaxMsgSemDly) != SEM_SUCCESS){
		    waitingForMsg--;
			continue;
		 }
         if(Sem_Wait(modSem, MaxModSemDly) != SEM_SUCCESS){
		    waitingForMsg--;
			continue;
		 }
		 waitingForMsg--;
	  }

	  // get the available chars
      bytesToRead = (dataLen - bytesRead);
      bytesRead += getNoLock(&dataPtr[bytesRead], bytesToRead);

	  if(waitingForSpace) Sem_Post(waitForSpaceSem);
      Sem_Post(modSem);
   }

   return(bytesRead);
}


/*-----------------------------------------------------------------------------
   Function: 	Buffer::peep
   Description: A non-destructive read from the buffer
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int Buffer::peep(char *dataPtr, unsigned int dataLen)
{
   unsigned int i, h;
      
   Sem_Wait(modSem, SEM_WAITFOREVER);

   if (!len){
      Sem_Post(modSem);
      return(0);	// empty
   }
   h = head;
   if (dataLen > len) dataLen = len;
	       
   for (i = 0; i < dataLen; i++){
      dataPtr[i] = bufPtr[h];
      if(h+1 < size) h++;
      else h = 0;
   }
   Sem_Post(modSem);
   return(dataLen);
}
