/* scomm_buff.h
 * Serial communication buffer
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 *  system: SA1200
 *  subsystem: Serial Communication Library
 *  author: drj March 1, 99 Initial version
 *  revisions:
 *  $History: scomm_buff.h $
 * 
 * *****************  Version 14  *****************
 * User: Drjohns1     Date: 8/30/00    Time: 2:06p
 * Updated in $/dev_1200/Share/scomm
 * 
 * *****************  Version 12  *****************
 * User: Drjohns1     Date: 3/27/00    Time: 5:12p
 * Updated in $/dev_1200/Share/scomm
 * Fixed bug...missing time function on hardware (ARM lib)
 * 
 * *****************  Version 9  *****************
 * User: Drjohns1     Date: 12/03/99   Time: 1:14p
 * Updated in $/dev_1200/Share/scomm
 * 
 * *****************  Version 8  *****************
 * User: Drjohns1     Date: 10/29/99   Time: 10:04a
 * Updated in $/dev_1200/Share/scomm
 * Adjusted timeout/delay parmeter to system clock resolution, modified
 * ISR handle and frame parsing.
 * 
 * *****************  Version 7  *****************
 * User: Drjohns1     Date: 10/22/99   Time: 5:44p
 * Updated in $/dev_1200/Share/scomm
 * Fixed deadlock problem when running at 115k baud
 * 
 * *****************  Version 6  *****************
 * User: Drjohns1     Date: 8/06/99    Time: 5:24p
 * Updated in $/dev_1200/Share/scomm
 * Added timout param to Sem_Wait
 * 
 * *****************  Version 3  *****************
 * User: C-tclancy    Date: 3/23/99    Time: 4:17p
 * Updated in $/development/scomm
 * update COPYRIGHT
 * 
 * *****************  Version 2  *****************
 * User: Drjohns1     Date: 3/19/99    Time: 3:57p
 * Updated in $/users/Johnson/SA1200/scomm
 * 
 * *****************  Version 1  *****************
 * User: Drjohns1     Date: 3/05/99    Time: 9:06p
 * Created in $/users/Johnson/SA1200/scomm
 * 
 * ---------------------------------------------------------------------
 */

#ifndef __SCOMM_BUF_H
#define __SCOMM_BUF_H

#include <signal.h>
#include "sem.h"


class FastBuf {
public:
   FastBuf(unsigned int bufSize);
   ~FastBuf(){}

   inline int put(char *dataPtr, int dataLen){
      int availSpace = pBufSize - pDataLen;
	  if(dataLen > availSpace) dataLen = availSpace;

	  // copy data to buffer
	  for(int i = 0; i < dataLen; i++){
		 pBufPtr[pWrIndx] = dataPtr[i];
 		 pWrIndx = incIndx(pWrIndx);
	  }
      pDataLen = pDataLen + dataLen;
	  if(pDataLen) Sem_Post(pWaitForMsgSem);
	  return(dataLen);
   }

   inline int get(char *dataPtr, int dataLen){
	  if(!pDataLen) Sem_Wait(pWaitForMsgSem, SEM_WAITFOREVER);

	  if(dataLen > pDataLen) dataLen = pDataLen;

	  // copy data from buffer
	  for(int i = 0; i < dataLen; i++){
		 dataPtr[i] = pBufPtr[pRdIndx];
		 pRdIndx = incIndx(pRdIndx);
	  }
	  pDataLen = pDataLen - dataLen;
	  return(dataLen);
   }


   inline int getReadIndx(void){return(pRdIndx);}
   inline unsigned int getDataLen(void){return(pDataLen);}
   inline unsigned int spaceLeft(){return(pBufSize - pDataLen);}
   inline int isEmpty(void){return((pDataLen == 0) ? 1 : 0);}
   inline int dataReady(void){return((pDataLen) ? 1 : 0);}
   inline int isFull(void){return((pDataLen >= pBufSize) ? 1 : 0);}
   inline void clear(void){pRdIndx = pWrIndx = pDataLen = 0;}
   inline int incIndx(int indx){
		if(indx < (pBufSize-1))return((++indx));
		else return(0);
   }
 
   inline int incIndxByN(int indx, int n){return(((indx + n) % pBufSize));}
  
   inline char getChar(int indx){return(pBufPtr[indx]);}
   inline void reduceData(int nChars){pDataLen = pDataLen - nChars; Sem_Post(pWaitForSpaceSem);}
   inline void expandData(int nChars){pDataLen = pDataLen + nChars; Sem_Post(pWaitForMsgSem);}
   inline unsigned int waitForMsg(){Sem_Wait(pWaitForMsgSem, SEM_WAITFOREVER); return(pDataLen);}


private:
   sem_T  pWaitForMsgSem, pWaitForSpaceSem;
   char   *pBufPtr;
   int pBufSize;
   sig_atomic_t	pRdIndx, pWrIndx, pDataLen;
};


class Buffer {
public:
   Buffer(unsigned int bufSize);
   ~Buffer();
   int put(char *dataPtr, unsigned int dataLen, unsigned int delay);
   int get(char *dataPtr, unsigned int dataLen, unsigned int delay);
   int peep(char *dataPtr, unsigned int dataLen);
   int lockIfSpace(unsigned int dataLen, int delay = SEM_WAITFOREVER);

   inline void unLockBuf(void){Sem_Post(modSem);}
   inline int lockBuf(int delay = SEM_WAITFOREVER){if(Sem_Wait(modSem, delay) != SEM_SUCCESS) return(-1); return(0);}
   inline unsigned int getLen(){Sem_Wait(modSem,SEM_WAITFOREVER); int l = len; Sem_Post(modSem); return(l);}

   inline int putNoLock(char *dataPtr, unsigned int dataLen){
	  if(dataLen > (size - len)) dataLen = (size - len);
	  for(unsigned int i = 0; i < dataLen; i++){
		 bufPtr[tail] = dataPtr[i];
		 if(tail+1 < size) tail++;
		 else tail = 0;
	  }
	  len += dataLen;
	  if(waitingForMsg) Sem_Post(waitForMsgSem);
	  return(dataLen);
   }

   inline int getNoLock(char *dataPtr, unsigned int dataLen){
	  if(dataLen > len) dataLen = len;
	  for(unsigned int i = 0; i < dataLen; i++){
		 dataPtr[i] = bufPtr[head];
		 if(head+1 < size) head++;
		 else head = 0;
	  }
	  len -= dataLen;
	  if(waitingForSpace) Sem_Post(waitForSpaceSem);
	  return(dataLen);
   }

   inline unsigned int spaceLeft(){
      Sem_Wait(modSem,SEM_WAITFOREVER);
      int s = size - len;
      Sem_Post(modSem);
      return(s);
   }
   inline int isEmpty(){
      Sem_Wait(modSem,SEM_WAITFOREVER);
      int s = ((len == 0) ? 1 : 0);
      Sem_Post(modSem);
      return(s);
   }
   inline int isFull(){
      Sem_Wait(modSem,SEM_WAITFOREVER);
      int s = ((len >= size) ? 1 : 0);
      Sem_Post(modSem);
      return(s);
   }

   inline void clear(){
	  Sem_Wait(modSem, SEM_WAITFOREVER);
      tail = head = len = 0;
	  if(waitingForSpace) Sem_Post(waitForSpaceSem);
	  Sem_Post(modSem);
   }

private:
   unsigned int size, len;
   sem_T  modSem, waitForMsgSem, waitForSpaceSem;
   char   *bufPtr;
   unsigned int	tail, head;
   unsigned int waitingForMsg, waitingForSpace;
};

#endif				// __SCOMM_BUF_H
