/*----------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
//**************************************************************************
//
// File = $Id: $
//
//
// Edit History
//
// $Log: /development/uca/poolmgr.cpp $
 * 
 * 3     3/23/99 4:20p C-tclancy
 * update COPYRIGHT
 * 
 * 2     5/20/98 1:59p Jguilford
 * Change copyright notice to Intel
 * 
 * 1     3/25/98 12:53p Jguilford
 * Utilities for managing object allocation
//
*/

#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include "poolmgr.h"

/***********************************************************************/
// create pool for objects of size n, allocating alloc_gran items each time
PoolMgr::PoolMgr(size_t n, unsigned int alloc_gran) :
    m_pool_size(n),
    m_alloc_gran(alloc_gran)
{
    m_head = NULL;
    m_freelist = NULL;
    
#ifdef CHECK_FOR_LEAKS
    m_num_allocs = 0;
#endif 
}
/*---------------------------------------------------------------------*/
PoolMgr::~PoolMgr()          // deallocate all memory in pool
{
    flush();
}
/*---------------------------------------------------------------------*/
void
PoolMgr::flush()          // deallocate all memory in pool
{
    chunk* c;
    chunk* next_c;

    for (c = m_head; c; c = next_c) {
        next_c = c->next;
        delete c;
    } // end for c
    m_head = NULL;
    m_freelist = NULL;

#ifdef CHECK_FOR_LEAKS
    if (m_num_allocs != 0) {
        fprintf(stderr,"Memory leak in PoolMgr(%d): %d outstanding allocs\n",
                m_pool_size, m_num_allocs);
    }
    m_num_allocs = 0;
#endif 
}
/*---------------------------------------------------------------------*/
void*
PoolMgr::alloc(size_t n)
{
    obj* o;
    chunk* c;
    unsigned int i;
    char *cptr;

#ifdef CHECK_FOR_LEAKS
    m_num_allocs++;
#endif 

    if (n != m_pool_size)
        return ::operator new(n);

    o = m_freelist;
    if (o != NULL) {
        m_freelist = o->next;
    } else {
        // need to allocate a new chuck
        c = (chunk*) ::operator new(sizeof(chunk) + m_alloc_gran * m_pool_size);
        if (c == NULL)
            return NULL;
        c->next = m_head;
        m_head = c;
        // return first block
        o = (obj*) (c+1);
        // link in remaining blocks
        cptr = (char*)o;
        cptr += m_pool_size;
        m_freelist = (obj*) cptr;
        for (i=1; i < m_alloc_gran-1; i++) {
            ((obj*)cptr)->next = (obj*)(cptr + m_pool_size);
            cptr += m_pool_size;
        }
        ((obj*)cptr)->next = NULL;
    }
    return o;
}
/*---------------------------------------------------------------------*/
void
PoolMgr::free(void* p, size_t n)
{
    obj* o;

    if (p == NULL)
        return;
#ifdef CHECK_FOR_LEAKS
    m_num_allocs--;
#endif 
    if (n != m_pool_size) {
        delete (p);
        return;
    }
    
#ifdef _DEBUG
    memset(p, 0xdc, m_pool_size);
#endif

    o = (obj*)p;
    o->next = m_freelist;
    m_freelist = o;
}
/*---------------------------------------------------------------------*/

#ifdef DO_TEST

class test
{
private:
    static PoolMgr m_pool;
public:
    int m_int;

    static void* operator new(size_t size) {return m_pool.alloc(size);}
    static void operator delete(void *p, size_t size) {m_pool.free(p, size);}

    static void flush() {m_pool.flush();}
};

PoolMgr test::m_pool(sizeof(test));

main()
{
    test* array[100];
    int i;

    for (i=0; i<100; i++) {
        array[i] = new test;
    }
    for (i=0; i<100; i++) {
        delete array[i];
    }
    for (i=0; i<100; i++) {
        array[i] = new test;
    }
    for (i=0; i<100; i++) {
        delete array[i];
    }
    test::flush();
    for (i=0; i<100; i++) {
        array[i] = new test;
    }
    for (i=0; i<50; i++) {
        delete array[i];
    }

    return 0;
}

#endif // ifdef DO_TEST
