/*----------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
//**************************************************************************
//
//
// File = $Id: $
//
//
// Edit History
//
// $Log: /development/uca/poolmgr.h $
 * 
 * 3     3/23/99 4:20p C-tclancy
 * update COPYRIGHT
 * 
 * 2     5/20/98 1:59p Jguilford
 * Change copyright notice to Intel
 * 
 * 1     3/25/98 12:53p Jguilford
 * Utilities for managing object allocation
*/

#ifndef POOLMGR_H
#define POOLMGR_H

// PoolMgr is a class that allows one to easily create a pool of fixed-sized
// objects for use in new and delete, so that these operations are much faster.
// Use of PoolMgr is illustrated by the following example:
//   class test
//   {
//   private:
//       static PoolMgr m_pool;
//   public:
//       ...
//   
//       static void* operator new(size_t size) {return m_pool.alloc(size);}
//       static void operator delete(void *p, size_t size) {m_pool.free(p,size);}
//   };
//   
//   PoolMgr test::m_pool(sizeof(test), 20);
//
// The object (in this example "test") declares a pool as static. This is then
// initialized at process start with the size of the objects it will contain,
// in this case "sizeof(test)". Optionally, a second argument (in this case 20)
// tells how many objects should be allocated at once. The default is 10.
// Then the operator new and operator delete just call alloc and free from the
// pool. 
//
// Note that the pool will be freed automatically at process termination
// (when the static object's destructor is called). If you want to free the
// memory pool earlier, you can call the pool "flush" method (not shown in
// the above example).
//
//   If the pool is destroyed (or flush is called) while the application
//   has any outstanding pointers to the objects, then use of these objects
//   is undefined and will probably cause the program to crash. For example,
//   in the above example if the application were to create an object of type
//   "test", then cause the pool to be flushed, then if the "test" object were
//   to be destroyed, bad things would happen. This problem can be avoided
//   by not using the "flush" method and allowing the pool to be destroyed
//   at process destroy time.
//
// Caveats: If "test" is a base class, then the derived sub-classes will only
// inherit use of the pool if they are the same size as the base class 
// (i.e. if they do not add any class variables). If they do, then they would
// need to define their own pool, new, and delete (and the destructor for the 
// base class would need to be virtual).
//
// If _DEBUG is define (i.e. if Win32 debugging is enabled), then the free
// method will scribble over the object when it is destroyed.
//
// If CHECK_FOR_LEAKS is defined, then the PoolMgr object will track how many
// outstanding allocations there are. If it is destroyed or flushed with the
// outstanding allocations not zero, it will print a message to stderr.
//

class PoolMgr {
public:
    // create pool for objects of size n, allocating alloc_gran items each time
    PoolMgr(size_t n, unsigned int alloc_gran = 10);   
    ~PoolMgr();          // deallocate all memory in pool
    void* alloc(size_t n);
    void free(void* p, size_t n);
    void flush();
private:
    struct chunk {
        chunk* next;
    };
    struct obj {
        obj* next;
    };

    const size_t m_pool_size;
    const unsigned int m_alloc_gran;
    chunk* m_head;
    obj* m_freelist;

#ifdef CHECK_FOR_LEAKS
    unsigned int m_num_allocs;
#endif 
};

#endif // ifdef POOLMGR_H
