/* utl_bits.c
 * utilities to manipulate bits
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: SA1200
 * subsystem: Ucode Linker/Loader
 * author: drj Dec 31, 97 Initial version
 * revisions:
 * $History: utl_bits.c $
 * 
 * *****************  Version 17  *****************
 * User: Drjohns1     Date: 4/04/00    Time: 6:18p
 * Updated in $/dev_1200/Utils
 * OS type check and cleanup
 * 
 *
 * ---------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "utl_bits.h"

static const char __copyright[]=
"Copyright Intel Corp, 1997-1998. All rights reserved";


#define BITS_IN_WORD 32


typedef struct{
   unsigned int b0 : 1;
   unsigned int b1 : 1;
   unsigned int b2 : 1;
   unsigned int b3 : 1;
   unsigned int b4 : 1;
   unsigned int b5 : 1;
   unsigned int b6 : 1;
   unsigned int b7 : 1;
}bits_t;   

typedef union{
   char b;
   bits_t bb;
}byte_t;


/*-----------------------------------------------------------------------------
   Function: 	prnByteBin
   Description: Prints the byte value in binary format to std IO
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void prnByteBin(unsigned char byte)
{
   byte_t bval;
   bval.b = byte;

   printf("%d%d%d%d%d%d%d%d",
       bval.bb.b7,bval.bb.b6,bval.bb.b5,
       bval.bb.b4,bval.bb.b3,bval.bb.b2,
       bval.bb.b1,bval.bb.b0); 
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	prnWordBin
   Description: Prints the value of the word in binary format
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
void prnWordBin(unsigned int word)
{
   char *w = (char *)&word;
   int i;
   for(i=3; i >= 0; i--) prnByteBin((unsigned char)w[i]);
   return;
}


/*-----------------------------------------------------------------------------
   Function: 	setField
   Description: Sets a specified range of bits of 'word' equal to the bits of
                'value'--starting from the zero bit-position of 'value'.
   Returns:	The result of the operation
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int setField(unsigned int word, int startBit, int fieldLen,
                      unsigned int value)
{
   unsigned int bit, stopBit, mask = 0;

   /* check that the field is within bound of the word */
   if((stopBit = (startBit + fieldLen)) > BITS_IN_WORD){
      printf("setField: field exceeds the size of the word\n");
      return(word);
   }

   /* determine the mask */
   for(bit = startBit; bit < stopBit; bit++) mask |= (1 << bit);

   /* Assign the first 'numBit' bits of 'value' to the field of 'fieldLen'
      starting at the 'startBit' of 'word'. */
   word &= ~mask;
   return((word |= mask & (value << startBit)));
}


/*-----------------------------------------------------------------------------
   Function: 	getField
   Description: Gets a specified range of bits of 'word' equal to the bits of
                'value'--starting from the zero bit-position of 'value'.
   Returns:	The result of the operation
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int getField(unsigned int word, int startBit, int fieldLen)
{
   unsigned int bit, stopBit, mask = 0;

   /* check that the field is within bound of the word */
   if((stopBit = (startBit + fieldLen)) > BITS_IN_WORD){
      printf("getField: field exceeds the size of the word\n");
      return(word);
   }

   /* determine the mask */
   for(bit = startBit; bit < stopBit; bit++) mask |= (1 << bit);
  
   return((word & mask) >> startBit);
}

/*-----------------------------------------------------------------------------
   Function: 	numBitsSet
   Description: Determine the number of bits that are set in a long-word.
   Returns:
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int numBitsSet(unsigned int word)
{
   int n = 0;

   if(word)
      do{
	     n++;
      } while ((word &= word - 1));

   return (n);
}


/*-----------------------------------------------------------------------------
   Function: 	bitParity
   Description: Determine parity of the number of bits.
   Returns:    0 if even; 1 if odd.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int bitParity(unsigned int word)
{
   word ^= word >> 1;
   word ^= word >> 2;
   word ^= word >> 4;
   word ^= word >> 8;
   word ^= word >> 16;
   return(word & 1);
}


/*-----------------------------------------------------------------------------
   Function: 	powOfTwo
   Description: Determine if a long-word value is the power-of-two.
   Returns:    1 is a power-of-two; 0 not a power-of-two.
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
int powOfTwo(unsigned int word)
{
   unsigned wm = word - 1;
   return((word ^ wm) >> 1 == wm);
}


/*-----------------------------------------------------------------------------
   Function: 	bitSwap
   Description: Swaps bits at position pos1, pos2 in word.
   Returns:    
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int bitSwap(unsigned int word, unsigned int pos1, unsigned int pos2)
{
   unsigned int m1 = 1 << pos1;
   unsigned int m2 = 1 << pos2;
   if (!(word & m1) != !(word & m2))  word ^= m1 | m2;
   return(word);
}


/*-----------------------------------------------------------------------------
   Function: 	rightRotate
   Description: Rotate the bit in 'word' 's' positions to the right.
   Returns:    
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int rightRotate(unsigned int word, unsigned int s)
{
   s = s & 31;
   return((word << (32 - s)) | (word >> s));
}


/*-----------------------------------------------------------------------------
   Function: 	leftRotate
   Description: Rotate the bit in 'word' 's' positions to the left.
   Returns:    
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int leftRotate(unsigned int word, unsigned int s)
{
   s = (32 - s) & (31);
   return((word << (32 - s)) | (word >> s));
}



#if DEBUG_UTL_BITS
int main()
{
   unsigned int w = 0xaaaaaaaa, val;

   printf("       10987654321098765432109876543210\n");
   w = toggleAllBits(w);
   printf("word = "); prnWordBin(w); printf("\tToggled all bits\n");

   w = toggleAllBits(w);
   printf("word = "); prnWordBin(w); printf("\tToggled all bits\n");

   w = SET_ALL_BITS();
   printf("word = "); prnWordBin(w); printf("\tbits <13:0> = 0xffffffff\n");

   w = CLR_ALL_BITS();
   printf("word = "); prnWordBin(w); printf("\tbits <13:0> = 0x00000000\n");

/*   printf("       10987654321098765432109876543210\n"); */
   w = setField(w, 0, 32, 0xaaaaaaaa);
   printf("word = "); prnWordBin(w); printf("\tbits <31:0> = 0xaaaaaaaa\n");

   w = setField(w, 0, 32, 0xffffffff);
   printf("word = "); prnWordBin(w); printf("\tbits <31:0> = 0xffffffff\n");

   w = setField(w, 0, 16, 0x0000);
   printf("word = "); prnWordBin(w); printf("\tbits <15:0> = 0x0000\n");

   w = setField(w, 16, 16, 0x0000);
   printf("word = "); prnWordBin(w); printf("\tbits <31:16> = 0x0000\n");

   w = setField(w, 16, 16, 0xffff);
   printf("word = "); prnWordBin(w); printf("\tbits <31:16> = 0xffff\n");

   w = setField(w, 0, 16, 0xffff);
   printf("word = "); prnWordBin(w); printf("\tbits <15:0> = 0xffff\n");

   w = setField(w, 0, 32, 0xf0f0f0f0);
   printf("word = "); prnWordBin(w); printf("\tbits <31:0> = 0xf0f0f0f0\n");

   w = setField(w, 14, 8, 0x55);
   printf("word = "); prnWordBin(w); printf("\tbits <21:14> = 0x55\n");

   w = 0xf0f00f0f;
   printf("\n       10987654321098765432109876543210\n");
   printf("word = "); prnWordBin(w); printf("\n");

   val = getField(w, 0, 16);
   printf("value= "); prnWordBin(val); printf("\tgetField bits <15:0> = 0x0000\n");

   val = getField(w, 16, 16);
   printf("value= "); prnWordBin(val); printf("\tgetField bits <31:16> = 0xffff\n");

   return(1);
}

#endif
