/* utl_crc.c
 * CRC checksum calculation
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: SA1200
 * subsystem: Ucode Linker/Loader
 * author: drj Dec 31, 97 Initial version
 * revisions:
 * $History: utl_crc.c $
 * 
 * *****************  Version 8  *****************
 * User: Drjohns1     Date: 4/04/00    Time: 6:18p
 * Updated in $/dev_1200/Utils
 * OS type check and cleanup
 * 
 *
 * ---------------------------------------------------------------------
 */

#include <stdio.h>
#include <errno.h>
#include "utl_crc.h"

#define X25_STANDARD	0x1021		/* CRC-CCITT, ADCCP, SDLC/HDLC */
#define X25_REVERSED	0x0811
#define CRC16_STARNARD	0x8005
#define CRC16_REVERSED	0x4003		/* LHA */
#define CRC32		0x04c11db7	/* PKZIP, AUTODIN II, Ethernet, FDDI */

/* define CRC_POLY for one of the above standard */
#define CRC_POLY	X25_STANDARD

#define CRC_WIDTH 16
#define CRC_BITMASK(X) (1L << (X))
#define CRC_MAXLINE 256
#define CRC_WIDTHMASK(width) ((((1L<<(width-1))-1L)<<1)|1L)

#define CRC_FILE_BUFSIZE 1024


/*-----------------------------------------------------------------------------
   Function: 	Crc_Calc
   Description: Calculate the CRC checksum
   Returns:	The calculated checksum value for the string
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int Crc_Calc(unsigned int reg, int ch)
{
 int  i;
 unsigned int topbit = CRC_BITMASK(CRC_WIDTH - 1);
 unsigned int inbyte = (unsigned int) ((reg >> 24) ^ ch);

 reg ^= inbyte << (CRC_WIDTH - 8);
 for (i = 0; i < 8; i++)
    if (reg & topbit) reg = (reg << 1) ^ CRC_POLY;
      else reg <<= 1;

  return(reg & CRC_WIDTHMASK(CRC_WIDTH));
}


/*-----------------------------------------------------------------------------
   Function: 	Crc_StrChecksum
   Description: Calculate the CRC checksum of a string
   Returns:	The calculated checksum value for the string
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int Crc_StrChecksum(char *charPtr, int numChar)
{
   unsigned int chksum = 0;

   if(charPtr)
      while (numChar--) chksum = Crc_Calc(chksum, *charPtr++);
   return(chksum);
}



/*-----------------------------------------------------------------------------
   Function: 	Crc_FileChecksum
   Description: Calculate the CRC checksum of a file
   Returns:	The calculated checksum value for the string
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int Crc_FileChecksum(FILE *fp)
{
   unsigned int chksum = 0, len = 0;
   char buf[CRC_FILE_BUFSIZE], *p;

   while((len = fread(buf, 1, CRC_FILE_BUFSIZE, fp))){
      p = buf;
      if(len)
         while (len--) chksum = Crc_Calc(chksum, *p++);
   }

   return(chksum);
}


/*-----------------------------------------------------------------------------
   Function: 	Crc_FposChecksum
   Description: Calculate the CRC checksum of a file
   Returns:	The calculated checksum value for the string
   Uses:
   Modifies:
-----------------------------------------------------------------------------*/
unsigned int Crc_FposChecksum(FILE *fp, unsigned int nBytes)
{
   unsigned int chksum = 0, bytes, len = 0;
   char buf[CRC_FILE_BUFSIZE], *p;

   /* read nBytes and calculate crc */
   while(nBytes){
	  if(nBytes > CRC_FILE_BUFSIZE) bytes = CRC_FILE_BUFSIZE;
	  else bytes = nBytes;

	  nBytes -= bytes;
	  if((len = fread(buf, 1, bytes, fp))){
         p = buf;
         while (len--) chksum = Crc_Calc(chksum, *p++);
	  }
   }

   return(chksum);
}
