/* NetworkTraffic.cpp
 * 
 * ---------------------------------------------------------------------	
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ---------------------------------------------------------------------
 *
    This file represents an example of a network traffic simulation DLL
    which can be assigned to a IX bus device's port when debugging using 
    the Developer Workbench.

    When you Start Debugging, the Workbench loads the DLL and calls the
    Initialize() function immediately after the transactor initializes 
    the model.  This is the only time that this function is called. In
    the Initialize() function you can register console functions with the
    transactor.  You can then call these functions from a script in order
    configure your traffic simulation.  To call the transactor you must
    include the file xact_dll.h and link against xact_dll.lib.

    When the Workbench is activated to start receiving packets, it calls 
    the DLL to initialize each port.  If you have checked the option for
    packets to be received at the first clock cycle, then the initialization
    will occur at the first clock cycle.  Otherwise, it will occur when
    the start_packet_receive() function is invoked from your startup script.
    In either case, the Workbench iterates through all the ports on every 
    IX bus device.  If the DLL is connected to the receive side of the 
    port, then the function InitializeRxPort() is  called.  If it is 
    connected to the transmit side, then the function InitializeTxPort() 
    is called.  See the file NetTrafficSimInterface.h for the contents 
    of the port configuration data structures that are passed to these 
    initialization functions.

    As the simulation progresses, the Workbench calls function 
    GetNextByte() whenever a byte of data is required on a receive port.
    In addition to the byte of data, the DLL must also return a flag 
    indicating whether the byte is the last byte (EOP) in the frame/cell.  
    When the Workbench receives the last byte, it will call the 
    GetInterpacketTime() function in order to determine how long it waits 
    before asking for the first byte of the next frame/cell.  This allows 
    the DLL to randomize the arrival of frames/cells.
    
    On the transmit side, the Workbench calls the function SendNextByte() 
    when it takes a byte of data is out of the transmit buffer to be sent 
    over the network.  An EOP flag is asserted along with the last byte
    in the frame/cell.

    When the user Stops Debugging, the functions CloseRxPort() and 
    CloseTxPort() are called for each connected receive and transmit
    port, respectively.

    When the user closes the project or exits the Workbench, the function
    Close() is called just before the DLL is freed.
 *
 *
 */

#include "stdafx.h"
#include "NetTrafficSimInterface.h"
#include "xact_dll.h"

typedef struct
{
    unsigned int m_FrameIndex;
    unsigned int m_ByteIndex;
} PORT_HANDLE;

// These frames are sample ethernet IP traffic.
unsigned char Frame_1[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xb4, 0xc9, 0x00, 0x00, 0x00, 0x00, 0x01, 0x01, 0x01, 0x01, 0x00, 0x01, 
                            0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 
                            0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_2[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xb2, 0xc7, 0x00, 0x00, 0x00, 0x00, 0x02, 0x02, 0x02, 0x02, 0x1a, 0x1b, 
                            0x1c, 0x1d, 0x1e, 0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 
                            0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_3[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xb0, 0xc5, 0x00, 0x00, 0x00, 0x00, 0x03, 0x03, 0x03, 0x03, 0x34, 0x35, 
                            0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 
                            0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_4[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xae, 0xc3, 0x00, 0x00, 0x00, 0x00, 0x04, 0x04, 0x04, 0x04, 0x4e, 0x4f, 
                            0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 
                            0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_5[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xac, 0xc1, 0x00, 0x00, 0x00, 0x00, 0x05, 0x05, 0x05, 0x05, 0x68, 0x69, 
                            0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 
                            0x76, 0x77, 0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_6[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xaa, 0xbf, 0x00, 0x00, 0x00, 0x00, 0x06, 0x06, 0x06, 0x06, 0x82, 0x83, 
                            0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 
                            0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_7[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xa8, 0xbd, 0x00, 0x00, 0x00, 0x00, 0x07, 0x07, 0x07, 0x07, 0x9c, 0x9d, 
                            0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 
                            0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf, 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_8[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xa6, 0xbb, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08, 0x08, 0x08, 0xb6, 0xb7, 
                            0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf, 0xc0, 0xc1, 0xc2, 0xc3, 
                            0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_9[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                            0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                            0xa4, 0xb9, 0x00, 0x00, 0x00, 0x00, 0x09, 0x09, 0x09, 0x09, 0xd0, 0xd1, 
                            0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 
                            0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 
                            0x00, 0x00, 0x00, 0x00};

unsigned char Frame_10[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0xa2, 0xb7, 0x00, 0x00, 0x00, 0x00, 0x0a, 0x0a, 0x0a, 0x0a, 0xea, 0xeb, 
                             0xec, 0xed, 0xee, 0xef, 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 
                             0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff, 0x00, 0x01, 0x02, 0x03, 
                             0x00, 0x00, 0x00, 0x00};

unsigned char Frame_11[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0xa0, 0xb5, 0x00, 0x00, 0x00, 0x00, 0x0b, 0x0b, 0x0b, 0x0b, 0x04, 0x05, 
                             0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 
                             0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 
                             0x00, 0x00, 0x00, 0x00};

unsigned char Frame_12[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0x9e, 0xb3, 0x00, 0x00, 0x00, 0x00, 0x0c, 0x0c, 0x0c, 0x0c, 0x1e, 0x1f, 
                             0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 
                             0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 
                             0x00, 0x00, 0x00, 0x00};


unsigned char Frame_13[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0x9c, 0xb1, 0x00, 0x00, 0x00, 0x00, 0x0d, 0x0d, 0x0d, 0x0d, 0x38, 0x39, 
                             0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 
                             0x46, 0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f, 0x50, 0x51, 
                             0x00, 0x00, 0x00, 0x00 };

unsigned char Frame_14[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0x9a, 0xaf, 0x00, 0x00, 0x00, 0x00, 0x0e, 0x0e, 0x0e, 0x0e, 0x52, 0x53, 
                             0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 
                             0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x6b, 
                             0x00, 0x00, 0x00, 0x00};

unsigned char Frame_15[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0x98, 0xad, 0x00, 0x00, 0x00, 0x00, 0x0f, 0x0f, 0x0f, 0x0f, 0x6c, 0x6d, 
                             0x6e, 0x6f, 0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 
                             0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f, 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 
                             0x00, 0x00, 0x00, 0x00};

unsigned char Frame_16[] = { 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 
                             0x05, 0xdd, 0x45, 0x00, 0x00, 0x2e, 0x00, 0x00, 0x00, 0x00, 0x04, 0x06, 
                             0x96, 0xab, 0x00, 0x00, 0x00, 0x00, 0x10, 0x10, 0x10, 0x10, 0x86, 0x87, 
                             0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91, 0x92, 0x93, 
                             0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f, 
                             0x00, 0x00, 0x00, 0x00};

typedef struct
{
    unsigned int m_FrameSize;
    unsigned char *m_pFrame;
} FRAME_T;

FRAME_T Frames[] = { 
    { sizeof(Frame_1)/ sizeof(Frame_1[0]), Frame_1}, 
    { sizeof(Frame_2)/ sizeof(Frame_2[0]), Frame_2}, 
    { sizeof(Frame_3)/ sizeof(Frame_3[0]), Frame_3}, 
    { sizeof(Frame_4)/ sizeof(Frame_4[0]), Frame_4}, 
    { sizeof(Frame_5)/ sizeof(Frame_5[0]), Frame_5}, 
    { sizeof(Frame_6)/ sizeof(Frame_6[0]), Frame_6}, 
    { sizeof(Frame_7)/ sizeof(Frame_7[0]), Frame_7}, 
    { sizeof(Frame_8)/ sizeof(Frame_8[0]), Frame_8}, 
    { sizeof(Frame_9)/ sizeof(Frame_9[0]), Frame_9},
    { sizeof(Frame_10)/ sizeof(Frame_10[0]), Frame_10}, 
    { sizeof(Frame_11)/ sizeof(Frame_11[0]), Frame_11}, 
    { sizeof(Frame_12)/ sizeof(Frame_12[0]), Frame_12},
    { sizeof(Frame_13)/ sizeof(Frame_13[0]), Frame_13},
    { sizeof(Frame_14)/ sizeof(Frame_14[0]), Frame_14},
    { sizeof(Frame_15)/ sizeof(Frame_15[0]), Frame_15},
    { sizeof(Frame_16)/ sizeof(Frame_16[0]), Frame_16}};

static unsigned int NumFrames = sizeof(Frames) / sizeof(Frames[0]);

/*-----------------------------------------------------------------
    net_traffic_init

    This function is an example of a function that can be registered
    with the transactor and called from a script or the command line.

*/
#if defined(__cplusplus)
extern "C"
#endif
int
net_traffic_init()
{
    return 1;
}

/*-----------------------------------------------------------------
    Initialize

    This function is called once right after the DLL is loaded.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
Initialize()
{
	if (!XACT_register_console_function("net_traffic_init", net_traffic_init, 0, 0) )
            fprintf(stderr,"XACT_register_console_function() net_traffic_init failed\n");
}

/*-----------------------------------------------------------------
    Close

    This function is called once just before the DLL is freed.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
Close()
{
}

/*-----------------------------------------------------------------
    InitializeRxPort

    This function is called when debugging is started but before any
    simulation cycles are executed.  It is called once for each IX bus receive 
    port that is being simulated.  A unique handle is returned by the DLL to the
    Workbench.  The Workbench passes this handle along with other parameters
    when it requires data from the DLL for the specific port.
    
*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
InitializeRxPort(
    unsigned int DeviceNumber,             // IX bus device number
    unsigned int PortNumber,               // # of port to be initialized
    RX_PORT_CONFIG_DATA *pPortConfigData,  // pointer to port configuration data
    void **ppRxPortHandle)                 // pointer to where unique handle is to be returned
{
    // Allocate a PORT_HANDLE structure for each port to keep track of which byte in which frame
    // is to be processed next.
    PORT_HANDLE *pHandle = new PORT_HANDLE;
    pHandle->m_FrameIndex = DeviceNumber * 8 + PortNumber;
    pHandle->m_ByteIndex = 0;
    *ppRxPortHandle = (void *)pHandle;
}

/*-----------------------------------------------------------------
    GetNextByte

    This function returns the next byte of data coming from the network.
    If there is no data, the function returns false. Else, it returns true.
*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) bool __cdecl
GetNextByte(
    void *pPortHandle,          // pointer to unique port handle
    unsigned int Cycle,         // IX bus cycle count
    unsigned char *pDataByte,   // pointer to where byte is returned
    bool *pbEop)                // pointer to where boolean flag is returned to indicate if the byte
                                // is the last byte in packet
{
    PORT_HANDLE *pHandle = (PORT_HANDLE *)pPortHandle;
    unsigned char *pFrame = Frames[pHandle->m_FrameIndex].m_pFrame;
    *pDataByte = pFrame[pHandle->m_ByteIndex++];
    if (pHandle->m_ByteIndex >= Frames[pHandle->m_FrameIndex].m_FrameSize)
    {
        // reached eop, move on to next frame.
        *pbEop = true;
        pHandle->m_ByteIndex = 0;
        pHandle->m_FrameIndex++;
        // if just processed last frame, loop back to the first.
        if (pHandle->m_FrameIndex >= NumFrames)
            pHandle->m_FrameIndex = 0;
    }
    else
        // haven't reached eop yet.
        *pbEop = false;
    return true;
}

/*-----------------------------------------------------------------
    GetInterpacketTime

    This function returns the interpacket gap to be used before the next
    packet is requested from the DLL. The value is specified in # of nanoseconds. 

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) unsigned int __cdecl
GetInterpacketTime(
    void *pPortHandle)          // pointer to unique port handle
{
    return 1200;
}

/*-----------------------------------------------------------------
    CloseRxPort

    This function is called when the user stops debugging in the Workbench.
    It is called once for each IX bus receive port that is being simulated.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
CloseRxPort(
    void *pPortHandle)          // pointer to unique port handle
{
    // free memory allocated for the port.
    delete (PORT_HANDLE *)pPortHandle;
}

/*-----------------------------------------------------------------
    InitializeTxPort

    This function is called when debugging is started but before any
    simulation cycles are executed.  It is called once for each IX bus receive 
    port that is being simulated.  A unique handle is returned by the DLL to the
    Workbench.  The Workbench passes this handle along with other parameters
    when it requires data from the DLL for the specific port.
    
*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
InitializeTxPort(
    unsigned int DeviceNumber,             // IX bus device number
    unsigned int PortNumber,               // # of port to be initialized
    TX_PORT_CONFIG_DATA *pPortConfigData,  // pointer to port configuration data
    void **ppTxPortHandle)                 // pointer to where unique handle is to be returned
{
    *ppTxPortHandle = NULL;
}

/*-----------------------------------------------------------------
    SendNextByte

    This function is called when a byte of data is transmitted 
    out of the IX bus device onto the network.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
SendNextByte(
    void *pTxPortHandle,        // pointer to unique port handle
    unsigned int Cycle,         // IX bus cycle count
    unsigned char DataByte,     // byte of frame/cell data
    bool bEop)                  // true if byte represents the last byte in a frame/cell
{
    // this example just ignores transmitted bytes.
}

/*-----------------------------------------------------------------
    CloseRxPort

    This function is called when the user stops debugging in the Workbench.
    It is called once for each IX bus transmit port that is being simulated.

*/
#if defined(__cplusplus)
extern "C"
#endif
__declspec(dllexport) void __cdecl
CloseTxPort(
    void *pPortHandle)          // pointer to unique port handle
{
}

