/* sem.c
 * Common Utility Library
 *
 *---------------------------------------------------------------------------
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  1998-1999 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 *---------------------------------------------------------------------------
 *
 * system: Common utility
 * subsystem: Common
 * author: drj Dec 31, 97 Initial version
 * revisions:
 * $History: sem.c $
 * 
 * *****************  Version 2  *****************
 * User: Ewalker      Date: 8/11/00    Time: 10:03
 * Updated in $/dev_1200/osApi
 * Minor comment changes for OS == xxx
 * 
 * *****************  Version 1  *****************
 * User: Drjohns1     Date: 7/24/00    Time: 5:42p
 * Created in $/dev_1200/osApi
 * 
 * *****************  Version 24  *****************
 * User: Dmwhite      Date: 4/05/00    Time: 2:55p
 * Updated in $/dev_1200/Utils
 * no more  U_COS macro, changed to (OS==UCOS)
 * 
 * *****************  Version 23  *****************
 * User: Drjohns1     Date: 4/04/00    Time: 6:18p
 * Updated in $/dev_1200/Utils
 * OS type check and cleanup
 * 
 * ---------------------------------------------------------------------
 */
#include <stdlib.h>
#include "ostype.h"
#include "sem.h"

#if (OS == WIN_32)
#include<wtypes.h>
#include<winbase.H>


sem_T Sem_Init(unsigned int initCount)
{
   return((sem_T)CreateSemaphore(NULL, initCount, 10, NULL));
}


int Sem_Destroy(sem_T sem)
{
   CloseHandle((HANDLE)sem);
   return(SEM_SUCCESS);
}

int Sem_Wait(sem_T sem, unsigned int timeout)
{
   unsigned int status;
   if((status = WaitForSingleObject((HANDLE)sem, timeout)) == WAIT_FAILED)
	   return(SEM_FAILED);
   if(status == WAIT_TIMEOUT) return(SEM_TIMEDOUT);
   return (SEM_SUCCESS);
}

int Sem_Post(sem_T sem)
{
   if(ReleaseSemaphore((HANDLE)sem, 1, NULL)) return(SEM_FAILED);
   return(SEM_SUCCESS);
}
#endif /* OS == WIN_32 */


#if (OS == VXWORKS)
#include <semLib.h>
sem_T Sem_Init(unsigned int initCount)
{
   return((sem_T)semCCreate(SEM_Q_FIFO, initCount));
}

int Sem_Destroy(sem_T sem)
{
   semDelete((SEM_ID)sem);
   return(SEM_SUCCESS);
}

int Sem_Wait(sem_T sem, unsigned int timeout)
{
   return(semTake((SEM_ID)sem, timeout));
}

int Sem_Post(sem_T sem)
{
   return(semGive((SEM_ID)sem));
}
#endif /* OS == VXWORKS */


#if (OS == UCOS)
#include "os.h"

sem_T Sem_Init(unsigned int initCount)
{
   return((sem_T)OSSemCreate(initCount));
}

int Sem_Destroy(sem_T sem)
{
   ;/*dj: Henry we need ucos function to remove semaphores! */
   return(SEM_SUCCESS);
}

int Sem_Wait(sem_T sem, unsigned int timeout)
{
   unsigned int err;

   OSSemPend((OS_EVENT *)sem, timeout, &err);
   if(err == OS_NO_ERR) return(SEM_SUCCESS);
   if(err == OS_TIMEOUT) return(SEM_TIMEDOUT);
   return(SEM_FAILED);
}

int Sem_Post(sem_T sem)
{
   if(OSSemPost((OS_EVENT *) sem) == OS_NO_ERR) return(SEM_SUCCESS);
   return(SEM_FAILED);
}
#endif /* OS == UCOS */



#if (OS == UNIX)
#include <pthread.h>
typedef struct semx_S{
   unsigned int		count;			/* number of SemPost called */
   unsigned int		waiting;		/* number of Sem_Wait called */  
   pthread_mutex_t	mutex;			/* protect member variables */
   pthread_cond_t	condCount;		/* condition variable to block on */
}semx_T;


sem_T Sem_Init(unsigned int initCount)
{
   semx_T *sem;

   if(!(sem = (semx_T *)malloc(sizeof(semx_T)))) return(NULL);
   pthread_mutex_init(&sem->mutex, NULL);
   pthread_cond_init(&sem->condCount, NULL);
   sem->count = initCount;
   sem->waiting = 0;
   return((sem_T) sem);
}

int Sem_Wait(sem_T sem, unsigned int timeout)
{
   semx_T *semx = (semx_T *)sem;
   struct timespec abstime, delta;
   int status;
   
   if(timeout >= 1000) delta.tv_sec = timeout/1000;
   else delta.tv_sec = 0;
   delta.tv_nsec = (timeout - (delta.tv_sec * 1000)) * 1000000;

   pthread_get_expiration_np(&delta, &abstime);
					    
   pthread_mutex_lock(&semx->mutex);			/* obtain the semaphore mutex */
   semx->waiting++;   					/* indicate that we are waiting */

   /* wait for condition count to be > 0 and to be signaled or timed out */
   while(semx->count == 0)
	  if((status = pthread_cond_timedwait(&semx->condCount, &semx->mutex, &abstime))){
         semx->waiting--;
         pthread_mutex_unlock(&semx->mutex);
         if(status == ETIMEDOUT) return(SEM_TIMEDOUT);
         return(SEM_FAILED);
	   }

   /* We've gotten the mutex, so... decrement counters and release mutex */
   semx->waiting--;
   semx->count--;
   pthread_mutex_unlock(&semx->mutex);
   return(SEM_SUCCESS);
}


int Sem_Post(sem_T sem)
{
   semx_T *semx = (semx_T *)sem;

   pthread_mutex_lock(&semx->mutex);

   /* if threads are waiting, then signal one to continue */
   if(semx->waiting > 0) pthread_cond_signal(&semx->condCount);

   /* increment the semaphore's count and released the mutex */
   semx->count++;

   pthread_mutex_unlock(&semx->mutex);

   return(SEM_SUCCESS);
}

int Sem_Destroy(sem_T sem)
{
   semx_T *semx = (semx_T *)sem;

   pthread_mutex_destroy(&semx->mutex);
   pthread_cond_destroy(&semx->condCount);
   free(semx);
   return(SEM_SUCCESS);
}

#endif /* OS == UNIX */
