/* fmpcit.cpp

    This file contains sample code demonstrating one way to write a
    PCI transactor as a foreign model. This example links against a
    static transactor library (xact_main.lib), although with minor
    modifications, it could be linked against the DLL version.

    Note that the state machines used here are "mostly correct" in that
    in most normal cases they do the right thing, but they are far from
    complete and should not be used without modification in a real design.

    The intent of this file is a "proof of concept" that an external
    model of a PCI device can interface with the transactor. As such
    this example contains several severe limitations. Among these are:
    1) The master transactions are compiled into the code
    2) There is no slave memory, a variation of the address is returned 
       on slave reads.
 * 
 * ---------------------------------------------------------------------	
 *                                                                      
 *                  I N T E L   P R O P R I E T A R Y                   
 *                                                                      
 *     COPYRIGHT (c)  2000 BY  INTEL  CORPORATION.  ALL RIGHTS          
 *     RESERVED.   NO  PART  OF THIS PROGRAM  OR  PUBLICATION  MAY      
 *     BE  REPRODUCED,   TRANSMITTED,   TRANSCRIBED,   STORED  IN  A    
 *     RETRIEVAL SYSTEM, OR TRANSLATED INTO ANY LANGUAGE OR COMPUTER    
 *     LANGUAGE IN ANY FORM OR BY ANY MEANS, ELECTRONIC, MECHANICAL,    
 *     MAGNETIC,  OPTICAL,  CHEMICAL, MANUAL, OR OTHERWISE,  WITHOUT    
 *     THE PRIOR WRITTEN PERMISSION OF :                                
 *                                                                      
 *                        INTEL  CORPORATION                            
 *                                                                     
 *                     2200 MISSION COLLEGE BLVD                        
 *                                                                      
 *               SANTA  CLARA,  CALIFORNIA  95052-8119                  
 *                                                                      
 * ---------------------------------------------------------------------
 *
 */

static const char __copyright[]=
        "Copyright (c) Intel Corporation, 2000. All rights reserved";

static const char __fmpcit_version__ []=
        "\nFILEINFO: $Workfile: fmpcit.cpp $ $Date: 3/24/00 2:52p $ $Revision: 1 $\n";

#include <stdio.h>
#include <stdlib.h>
#define ext_model_port
#include "xact.h"

//
// Foreign_model interface:
//   Calls from transactor to us:
//     int foreign_model_initialize();
//     int foreign_model_pre_sim();
//     int foreign_model_post_sim();
//     int foreign_model_exit();
//     int foreign_model_reset();
//
//   Calls that we may want to make:
//     XACT_register_console_function

void do_cycle();
void m_state_init();
void s_state_init();

typedef unsigned int uint;

enum PCI_CMD {NONE = 0,
              IRD = 0x2,
              IWR = 0x3,
              MRD = 0x6,
              MWR = 0x7,
              CRD = 0xA,
              CWR = 0xB,
              MRM = 0xC};

struct m_cmd {
    PCI_CMD cmd;
    uint addr;
    uint cbe;
    uint delay;        // between 0 and 100
    uint num_data;
    uint data;
    uint *pdata;
};

#define IS_WRITE_CMD(cmd) ((cmd) & 1)
#define IS_WRITE(m_cmd) ((m_cmd).cmd & 1)

struct State {
    // bus signals
    XACT_HANDLE PCI__AD;
    XACT_HANDLE PCI__CBE;
    XACT_HANDLE PCI__PAR;
    XACT_HANDLE PCI__FRAME;
    XACT_HANDLE PCI__TRDY;
    XACT_HANDLE PCI__IRDY;
    XACT_HANDLE PCI__STOP;
    XACT_HANDLE PCI__DEVSEL;
    XACT_HANDLE PCI__PERR;
    XACT_HANDLE PCI__SERR;
    XACT_HANDLE PCI__REQ0;
    XACT_HANDLE PCI__REQ1;
    XACT_HANDLE PCI__GNT0;
    XACT_HANDLE PCI__GNT1;
    XACT_HANDLE PCI__RST;
    XACT_HANDLE PCI__IRQ;

    // normal signals
    XACT_HANDLE PCI_CFN;
    XACT_HANDLE PCI__IDSEL;
    XACT_HANDLE RESET_OUT;

    // bus clients
    XACT_HANDLE PCI__AD_client;
    XACT_HANDLE PCI__CBE_client;
    XACT_HANDLE PCI__PAR_client;
    XACT_HANDLE PCI__FRAME_client;
    XACT_HANDLE PCI__TRDY_client;
    XACT_HANDLE PCI__IRDY_client;
    XACT_HANDLE PCI__STOP_client;
    XACT_HANDLE PCI__DEVSEL_client;
    XACT_HANDLE PCI__PERR_client;
    XACT_HANDLE PCI__SERR_client;
    XACT_HANDLE PCI__REQ0_client;
    XACT_HANDLE PCI__REQ1_client;
    XACT_HANDLE PCI__GNT0_client;
    XACT_HANDLE PCI__GNT1_client;
    XACT_HANDLE PCI__RST_client;
    XACT_HANDLE PCI__IRQ_client;

    XACT_HANDLE pci_clk_cycle;
    XACT_HANDLE sim_halt;

    unsigned int AD;
    unsigned int CBE;
    unsigned int PAR;
    unsigned int FRAME;
    unsigned int TRDY;
    unsigned int IRDY;
    unsigned int STOP;
    unsigned int DEVSEL;
    unsigned int PERR;
    unsigned int SERR;
    unsigned int REQ0;
    unsigned int REQ1;
    unsigned int GNT0;
    unsigned int GNT1;
    unsigned int RST;
    unsigned int IRQ;

    // normal signals
    unsigned int CFN;
    unsigned int IDSEL;
    unsigned int Reset_out;

    // master state
    int m_state;
    m_cmd *p_m_cmd;
    m_cmd m_curr_cmd;
    int m_devsel_cnt;

    // slave state
    int s_state;
    uint s_addr_lo;
    uint s_addr_hi;
    uint s_addr;

    // debug state
    XACT_HANDLE core_clk_cycle_h;
    uint core_clk_cycle;
};

State state;

int 
foreign_model_initialize()
{
    // bus signals
    state.PCI__AD = XACT_get_handle( "PCI__AD", -1);
    state.PCI__CBE = XACT_get_handle( "PCI__CBE", -1);
    state.PCI__PAR = XACT_get_handle( "PCI__PAR", -1);
    state.PCI__FRAME = XACT_get_handle( "PCI__FRAME", -1);
    state.PCI__TRDY = XACT_get_handle( "PCI__TRDY", -1);
    state.PCI__IRDY = XACT_get_handle( "PCI__IRDY", -1);
    state.PCI__STOP = XACT_get_handle( "PCI__STOP", -1);
    state.PCI__DEVSEL = XACT_get_handle( "PCI__DEVSEL", -1);
    state.PCI__PERR = XACT_get_handle( "PCI__PERR", -1);
    state.PCI__SERR = XACT_get_handle( "PCI__SERR", -1);
    state.PCI__REQ0 = XACT_get_handle( "PCI__REQ0", -1);
    state.PCI__REQ1 = XACT_get_handle( "PCI__REQ1", -1);
    state.PCI__GNT0 = XACT_get_handle( "PCI__GNT0", -1);
    state.PCI__GNT1 = XACT_get_handle( "PCI__GNT1", -1);
    state.PCI__RST = XACT_get_handle( "PCI__RST", -1);
    state.PCI__IRQ = XACT_get_handle( "PCI__IRQ", -1);

    // normal signals
    state.PCI_CFN = XACT_get_handle( "PCI_CFN", -1);
    state.PCI__IDSEL = XACT_get_handle( "PCI__IDSEL", -1);
    state.RESET_OUT = XACT_get_handle( "RESET_OUT", -1);
    
    // bus clients
    XACT_add_bus_client( "PCI__AD", "fmpcit_PCI__AD_client" );
    state.PCI__AD_client = XACT_get_handle( "fmpcit_PCI__AD_client", -1);

    XACT_add_bus_client( "PCI__CBE", "fmpcit_PCI__CBE_client" );
    state.PCI__CBE_client = XACT_get_handle( "fmpcit_PCI__CBE_client", -1);

    XACT_add_bus_client( "PCI__PAR", "fmpcit_PCI__PAR_client" );
    state.PCI__PAR_client = XACT_get_handle( "fmpcit_PCI__PAR_client", -1);

    XACT_add_bus_client( "PCI__FRAME", "fmpcit_PCI__FRAME_client" );
    state.PCI__FRAME_client = XACT_get_handle( "fmpcit_PCI__FRAME_client", -1);

    XACT_add_bus_client( "PCI__TRDY", "fmpcit_PCI__TRDY_client" );
    state.PCI__TRDY_client = XACT_get_handle( "fmpcit_PCI__TRDY_client", -1);

    XACT_add_bus_client( "PCI__IRDY", "fmpcit_PCI__IRDY_client" );
    state.PCI__IRDY_client = XACT_get_handle( "fmpcit_PCI__IRDY_client", -1);

    XACT_add_bus_client( "PCI__STOP", "fmpcit_PCI__STOP_client" );
    state.PCI__STOP_client = XACT_get_handle( "fmpcit_PCI__STOP_client", -1);

    XACT_add_bus_client( "PCI__DEVSEL", "fmpcit_PCI__DEVSEL_client" );
    state.PCI__DEVSEL_client = XACT_get_handle( "fmpcit_PCI__DEVSEL_client", -1);

    XACT_add_bus_client( "PCI__PERR", "fmpcit_PCI__PERR_client" );
    state.PCI__PERR_client = XACT_get_handle( "fmpcit_PCI__PERR_client", -1);

    XACT_add_bus_client( "PCI__SERR", "fmpcit_PCI__SERR_client" );
    state.PCI__SERR_client = XACT_get_handle( "fmpcit_PCI__SERR_client", -1);

    XACT_add_bus_client( "PCI__REQ0", "fmpcit_PCI__REQ0_client" );
    state.PCI__REQ0_client = XACT_get_handle( "fmpcit_PCI__REQ0_client", -1);

    XACT_add_bus_client( "PCI__REQ1", "fmpcit_PCI__REQ1_client" );
    state.PCI__REQ1_client = XACT_get_handle( "fmpcit_PCI__REQ1_client", -1);

    XACT_add_bus_client( "PCI__GNT0", "fmpcit_PCI__GNT0_client" );
    state.PCI__GNT0_client = XACT_get_handle( "fmpcit_PCI__GNT0_client", -1);

    XACT_add_bus_client( "PCI__GNT1", "fmpcit_PCI__GNT1_client" );
    state.PCI__GNT1_client = XACT_get_handle( "fmpcit_PCI__GNT1_client", -1);

    XACT_add_bus_client( "PCI__RST", "fmpcit_PCI__RST_client" );
    state.PCI__RST_client = XACT_get_handle( "fmpcit_PCI__RST_client", -1);

    XACT_add_bus_client( "PCI__IRQ", "fmpcit_PCI__IRQ_client" );
    state.PCI__IRQ_client = XACT_get_handle( "fmpcit_PCI__IRQ_client", -1);

    state.pci_clk_cycle = XACT_get_handle( "sim.pci_clk_cycle", -1);
    state.sim_halt = XACT_get_handle( "sim.halt", -1);

    state.core_clk_cycle_h = XACT_get_handle( "sim.core_clk_cycle", -1);

    XACT_import_sim_state( "fmpcit_m_state", (uint*)&state.m_state, 1);
    XACT_import_sim_state( "fmpcit_m_cmd", (uint*)&state.m_curr_cmd.cmd, 1);
    XACT_import_sim_state( "fmpcit_m_addr", (uint*)&state.m_curr_cmd.addr, 1);
    XACT_import_sim_state( "fmpcit_m_cbe", (uint*)&state.m_curr_cmd.cbe, 1);
    XACT_import_sim_state( "fmpcit_m_delay", (uint*)&state.m_curr_cmd.delay, 1);
    XACT_import_sim_state( "fmpcit_m_num_data", (uint*)&state.m_curr_cmd.num_data, 1);
    XACT_import_sim_state( "fmpcit_m_data", (uint*)&state.m_curr_cmd.data, 1);
    XACT_import_sim_state( "fmpcit_m_devsel_cnt", (uint*)&state.m_devsel_cnt, 1);

    XACT_import_sim_state( "fmpcit_s_state", (uint*)&state.s_state, 1);
    XACT_import_sim_state( "fmpcit_s_addr_lo", (uint*)&state.s_addr_lo, 1);
    XACT_import_sim_state( "fmpcit_s_addr_hi", (uint*)&state.s_addr_hi, 1);
    XACT_import_sim_state( "fmpcit_s_addr", (uint*)&state.s_addr, 1);

    m_state_init();
    s_state_init();

    return 1;
}

int 
foreign_model_pre_sim()
{
    if (! XACT_start_of_cycle(CLK_PCI))
        return 1;

    XACT_get_state_value(state.PCI__AD, &state.AD);
    XACT_get_state_value(state.PCI__CBE, &state.CBE);
    XACT_get_state_value(state.PCI__PAR, &state.PAR);
    XACT_get_state_value(state.PCI__FRAME, &state.FRAME);
    XACT_get_state_value(state.PCI__TRDY, &state.TRDY);
    XACT_get_state_value(state.PCI__IRDY, &state.IRDY);
    XACT_get_state_value(state.PCI__STOP, &state.STOP);
    XACT_get_state_value(state.PCI__DEVSEL, &state.DEVSEL);
    XACT_get_state_value(state.PCI__PERR, &state.PERR);
    XACT_get_state_value(state.PCI__SERR, &state.SERR);
    XACT_get_state_value(state.PCI__REQ0, &state.REQ0);
    XACT_get_state_value(state.PCI__REQ1, &state.REQ1);
    XACT_get_state_value(state.PCI__GNT0, &state.GNT0);
    XACT_get_state_value(state.PCI__GNT1, &state.GNT1);
    XACT_get_state_value(state.PCI__RST, &state.RST);
    XACT_get_state_value(state.PCI__IRQ, &state.IRQ);
    
    XACT_get_state_value(state.PCI_CFN, &state.CFN);
    XACT_get_state_value(state.PCI__IDSEL, &state.IDSEL);
    XACT_get_state_value(state.RESET_OUT, &state.Reset_out);

    XACT_get_state_value(state.core_clk_cycle_h, &state.core_clk_cycle);

    do_cycle();

    return 1;
}

int 
foreign_model_post_sim()
{
    return 1;
}

int 
foreign_model_exit()
{
    return 1;
}

int 
foreign_model_reset()
{
    return 1;
}

////////////////////////////////////////////////
// real code follows
////////////////////////////////////////////////

// Master state machine
// states are upper case
// IDLE:
//   if (!gnt# & frame# & irdy#) state = ADDR
// ADDR:
//   state = M_DATA
// M_DATA:
//   

void
drive(XACT_HANDLE handle, unsigned int value)
{
    XACT_set_state_value(handle, &value);
}

uint data0[] = {0x23242526,
                0x34353637,
                0x45464748};

uint data1[] = {0x66660000, 0x66660001, 0x66660002, 0x66660003,
                0x66660004, 0x66660005, 0x66660006, 0x66660007,
                0x66660008, 0x66660009, 0x6666000a, 0x6666000b,
                0x6666000c, 0x6666000d, 0x6666000e, 0x6666000f,
                0x66660010, 0x66660011, 0x66660012, 0x66660013,
                0x66660014, 0x66660015, 0x66660016, 0x66660017,
                0x66660018, 0x66660019, 0x6666001a, 0x6666001b,
                0x6666001c, 0x6666001d, 0x6666001e, 0x6666001f};
uint data2[] = {0x11110000};
uint data3[] = {0x22220000, 0x22220001};
uint data4[] = {0x33330000, 0x33330001, 0x33330002};
uint data5[] = {0x44440000, 0x44440001, 0x44440002, 0x44440003};

m_cmd m_cmds[] = {
//   cmd  addr        cbe delay n  data        pdata
    {CWR, 0x80000010, 0x0, 100, 1, 0x02000000, NULL},
    {CWR, 0x80000014, 0x0,   0, 1, 0x03000000, NULL},
    {CWR, 0x80000018, 0x0,   0, 1, 0x10000000, NULL},
    {CWR, 0x80000004, 0xC,   0, 1, 0x00000017, NULL},

    {MWR, 0x02000050, 0x0,  20, 1, 0x12131415, NULL},
    {MWR, 0x02000054, 0x0,   0, 3, 0,          data0},
    {IRD, 0x03000050, 0x0,   0, 1, 0,          NULL},
    {IRD, 0x03000054, 0x0,   0, 3, 0,          NULL},

    {MWR, 0x10000000, 0x0,   0, 1, 0,          data2},
    {MWR, 0x10000020, 0x0,   0, 2, 0,          data3},
    {MWR, 0x10000040, 0x0,   0, 3, 0,          data4},
    {MWR, 0x10000060, 0x0,   0, 4, 0,          data5},

    {MWR, 0x10000FF0, 0x0,   0,32, 0,          data1},
    {MRM, 0x10000FF0, 0x0,   0,32, 0,          NULL},

    {NONE, 0,           0,   0, 0, 0,          NULL}};

void
m_state_next_cmd()
{
    static char *names[] = {"NONE", "UNK", "IRD", "IWR",
                            "UNK",  "UNK", "MRD", "MWR",
                            "UNK",  "UNK", "CRD", "CWR",
                            "MRM",  "UNK", "UNK", "UNK"};
    state.p_m_cmd++;
    state.m_curr_cmd = *state.p_m_cmd;
    fprintf(stderr,"Starting cmd = %s, addr = %08X, cbe = %X, delay = %d, num_data = %d\n",
            names[state.m_curr_cmd.cmd & 0xF],
            state.m_curr_cmd.addr,
            state.m_curr_cmd.cbe,
            state.m_curr_cmd.delay,
            state.m_curr_cmd.num_data);
            
    state.m_state = 0 - state.m_curr_cmd.delay;
    if (state.m_curr_cmd.pdata == NULL) {
        state.m_curr_cmd.pdata = &state.m_curr_cmd.data;
        if (IS_WRITE(state.m_curr_cmd) && state.m_curr_cmd.num_data != 1)
            XACT_printf_error("fmpcit: No pdata for write with num_data>1\n");
    }
    if (state.m_curr_cmd.cmd == NONE)
        state.m_state = 100;
}

void
m_state_init()
{
    state.p_m_cmd = &m_cmds[-1];
    m_state_next_cmd();
}

void
m_state_cycle()
{
//    if ((state.m_state % 10) == 0)
//        fprintf(stderr,"m_state = %d\n",state.m_state);

    switch (state.m_state) {
    case 0:
        drive(state.PCI__REQ0_client, 0);
        state.m_state++;
        break;
    case 1:
        if (!state.GNT0 && state.FRAME && state.IRDY) {
            drive(state.PCI__FRAME_client, 0);
            drive(state.PCI__AD_client, state.m_curr_cmd.addr);
            drive(state.PCI__CBE_client, state.m_curr_cmd.cmd);
            state.m_state++;
            state.m_devsel_cnt = 0;
        } else {
            drive(state.PCI__REQ0_client, 0);
        }
        break;
    case 2:
        drive(state.PCI__IRDY_client, 0);
        drive(state.PCI__CBE_client, state.m_curr_cmd.cbe);
        if (state.DEVSEL)
            state.m_devsel_cnt++;
        if (state.m_devsel_cnt > 10)
            XACT_printf_error("No devices responding to address %08X\n",state.m_curr_cmd.addr);
        if (IS_WRITE(state.m_curr_cmd)) {
            if (state.STOP) {
                // stop not asserted
                if (state.TRDY) {
                    // not STOP, not TRDY : do nothing, stay in same state
                } else {
                    // not STOP, TRDY     : transfer data
                    state.m_curr_cmd.pdata++;
                    state.m_curr_cmd.num_data--;
                    state.m_curr_cmd.addr += 4;
                    if (state.m_curr_cmd.num_data <= 0)
                        m_state_next_cmd();
                }
            } else {
                // stop asserted
                if (state.TRDY) {
                    // STOP, not TRDY : retry
                    state.m_state = -5;
                } else {
                    // STOP, TRDY : disconnect with data
                    state.m_curr_cmd.pdata++;
                    state.m_curr_cmd.num_data--;
                    state.m_curr_cmd.addr += 4;
                    if (state.m_curr_cmd.num_data <= 0)
                        m_state_next_cmd();
                    else {
                        state.m_state = -5;
                    }
                }
            }
            drive(state.PCI__AD_client, state.m_curr_cmd.pdata[0]);
        } else {
            // is READ
            if (state.STOP) {
                // stop not asserted
                if (state.TRDY) {
                    // not STOP, not TRDY : do nothing, stay in same state
                } else {
                    // not STOP, TRDY     : transfer data
                    fprintf(stderr,"FMPCIT: READ(%08X) = %08X\n",
                            state.m_curr_cmd.addr,state.AD);
                    state.m_curr_cmd.num_data--;
                    state.m_curr_cmd.addr += 4;
                    if (state.m_curr_cmd.num_data == 0)
                        m_state_next_cmd();
                }
            } else {
                // stop asserted
                if (state.TRDY) {
                    // STOP, not TRDY : retry
                    state.m_state = -5;
                } else {
                    // STOP, TRDY : disconnect with data
                    fprintf(stderr,"FMPCIT: READ(%08X) = %08X\n",
                            state.m_curr_cmd.addr,state.AD);
                    state.m_curr_cmd.num_data--;
                    state.m_curr_cmd.addr += 4;
                    if (state.m_curr_cmd.num_data == 0)
                        m_state_next_cmd();
                    else {
                        state.m_state = -5;
                    }
                }
            }
        } // end if WRITE, ELSE
        if (state.m_curr_cmd.num_data > 1)
            drive(state.PCI__FRAME_client, 0);
        break;
    case 120:
//        drive(state.sim_halt, 1);
        fprintf(stderr,"fmpcit: done--halting simulation\n");
        state.m_state++;
        break;
    case 121:
        break;
    default:
        state.m_state++;
    } // end switch
}

void
s_state_init()
{
    state.s_state = 0;
    state.s_addr_lo = 0x08000000;
    state.s_addr_hi = 0x09000000;
}

void
s_state_cycle()
{
    switch (state.s_state) {
    case 0:
        if (!state.FRAME &&
            (state.AD >= state.s_addr_lo) &&
            (state.AD <  state.s_addr_hi)) {
            state.s_addr = state.AD;
            if (IS_WRITE_CMD(state.CBE)) {
                drive(state.PCI__DEVSEL_client, 0);
                drive(state.PCI__TRDY_client, 0);
                state.s_state=3;
            } else {
                // need turn around
                state.s_state+=1;
            }
        } else if (!state.FRAME) {
            // not for us
            state.s_state=4;
        }
        break;
    case 1: // turn around
        drive(state.PCI__DEVSEL_client, 0);
        drive(state.PCI__TRDY_client, 0);
        drive(state.PCI__AD_client, 0xDA7A | ((0xFFFF & state.s_addr)<<16));
        if (!state.IRDY) {
            fprintf(stderr,"fmpcit: Read from fmpcit(%08X): %08X\n",state.s_addr,
                    0xDA7A | ((0xFFFF & state.s_addr)<<16));
            state.s_addr+=4;
        }
        if (state.FRAME)
            state.s_state = 0;
        break;
    case 3: // data for write
        drive(state.PCI__DEVSEL_client, 0);
        drive(state.PCI__TRDY_client, 0);
        if (!state.IRDY) {
            fprintf(stderr,"fmpcit: Write to fmpcit(%08X): %08X\n",state.s_addr,state.AD);
            state.s_addr+=4;
            if (state.FRAME)
                state.s_state = 0;
        }
        break;
    case 4: // transaction not for us
        if (state.FRAME) {
            state.s_state = 0;
        }
        break;
    } // end switch s_state
}

void do_cycle()
{
    m_state_cycle();
    s_state_cycle();
}
